@@ -28,7 +28,6 @@ QUIET=''
 # link jemalloc (needed for tokudb). For now pxc doesn't build tokudb
 # if in future pxc supports tokudb we will need it.
 WITH_JEMALLOC=''
-DEBUG_EXTRA=''
 
 # Galera can be build locally as part of the script if the directory is placed
 # at said location. Alternatively, skip building galera instead just copy
@@ -36,6 +35,11 @@ DEBUG_EXTRA=''
 # script will install it at said location while making file tar.gz.
 COPYGALERA=0
 
+# suffix to add in case of debug build to tar.gz.
+DEBUG_EXTNAME=''
+
+WITH_MECAB_OPTION=''
+
 # build with ssl. configuration related to ssl.
 WITH_SSL='/usr'
 WITH_SSL_TYPE='system'
@@ -45,17 +49,18 @@ CRYPTO_LIBRARY=''
 GALERA_SSL=''
 SSL_OPT=''
 STAG=${STAG:-}
+DEBUG_EXTRA=''
+
 # tag build with this name.
-TAG=''
+TAG='1'
 
 # cmake build option to preset. based on configuration default options are used.
 CMAKE_BUILD_TYPE='RelWithDebInfo'
 COMMON_FLAGS=''
-
-# tokudb backup version [IGNORED]
-TOKUDB_BACKUP_VERSION='@@TOKUDB_BACKUP_VERSION@@'
-
-# build related utilities
+#
+TOKUDB_BACKUP_VERSION=''
+#
+# Some programs that may be overriden
 TAR=${TAR:-tar}
 SCONS_ARGS=${SCONS_ARGS:-""}
 
@@ -74,8 +79,8 @@ ENABLE_ASAN=0
 # parse input option and configure build enviornment acccordingly.
 if ! getopt --test
 then
-    go_out="$(getopt --options=iqGadvjt: \
-        --longoptions=i686,verbose,copygalera,asan,debug,valgrind,with-jemalloc:,with-yassl,keep-build,with-ssl:,tag: \
+    go_out="$(getopt --options=iqGadvjmt: \
+        --longoptions=i686,verbose,copygalera,asan,debug,valgrind,with-jemalloc:,with-mecab:,with-yassl,keep-build,with-ssl:,tag: \
         --name="$(basename "$0")" -- "$@")"
     test $? -eq 0 || exit 1
     eval set -- $go_out
@@ -94,8 +99,8 @@ do
         shift
         CMAKE_BUILD_TYPE='Debug'
         BUILD_COMMENT="${BUILD_COMMENT:-}-debug"
-        DEBUG_EXTRA="-DDEBUG_EXTNAME=ON"
-        SCONS_ARGS+=' debug=0'
+        DEBUG_EXTNAME="-DDEBUG_EXTNAME=OFF"
+        SCONS_ARGS+=' debug=1'
         ;;
     -G | --copygalera )
         shift
@@ -119,6 +124,11 @@ do
         WITH_JEMALLOC="$1"
         shift
         ;;
+    -m | --with-mecab )
+        shift
+        WITH_MECAB_OPTION="-DWITH_MECAB=$1"
+        shift
+        ;;
     --with-yassl )
         shift
         WITH_SSL_TYPE="bundled"
@@ -242,11 +252,13 @@ PS_VERSION="$MYSQL_VERSION-$PERCONA_SERVER_EXTENSION"
 
 # Note: WSREP_INTERFACE_VERSION act as compatibility check between wsrep-plugin
 # and galera plugin so we include it as part of our component.
-WSREP_VERSION="$(grep WSREP_INTERFACE_VERSION wsrep/src/wsrep_api.h | cut -d '"' -f2).$(grep 'SET(WSREP_PATCH_VERSION'  "cmake/wsrep.cmake" | cut -d '"' -f2)"
+WSREP_VERSION="$(grep WSREP_INTERFACE_VERSION wsrep/wsrep_api.h | cut -d '"' -f2).$(grep 'SET(WSREP_PATCH_VERSION'  "cmake/wsrep.cmake" | cut -d '"' -f2)"
 
 if [[ $COPYGALERA -eq 0 ]];then
-    GALERA_REVISION="$(cd "$SOURCEDIR/percona-xtradb-cluster-galera"; test -r GALERA-REVISION && cat GALERA-REVISION)"
+    #GALERA_REVISION="$(cd "$SOURCEDIR/percona-xtradb-cluster-galera"; test -r GALERA-REVISION && cat GALERA-REVISION)"
+    GALERA_REVISION="$(cd "$SOURCEDIR/percona-xtradb-cluster-galera"; git log --pretty=oneline | head -1 | cut --delimiter=' ' -f 1 | cut -b 1-7)"
 fi
+TOKUDB_BACKUP_VERSION="${MYSQL_VERSION}${MYSQL_VERSION_EXTRA}"
 
 RELEASE_TAG=''
 PRODUCT_NAME="Percona-XtraDB-Cluster-$MYSQL_VERSION-$PERCONA_SERVER_EXTENSION"
@@ -286,6 +298,25 @@ if [[ $ENABLE_ASAN -eq 1 ]]; then
     fi
 fi
 
+#
+# TokuDB cmake flags [IGNORED FOR PXC]
+if test -d "$SOURCEDIR/storage/tokudb"
+then
+    CMAKE_OPTS="${CMAKE_OPTS:-} -DBUILD_TESTING=OFF -DUSE_GTAGS=OFF -DUSE_CTAGS=OFF -DUSE_ETAGS=OFF -DUSE_CSCOPE=OFF -DTOKUDB_BACKUP_PLUGIN_VERSION=${TOKUDB_BACKUP_VERSION}"
+
+    if test "x$CMAKE_BUILD_TYPE" != "xDebug"
+    then
+        CMAKE_OPTS="${CMAKE_OPTS:-} -DTOKU_DEBUG_PARANOID=OFF"
+    else
+        CMAKE_OPTS="${CMAKE_OPTS:-} -DTOKU_DEBUG_PARANOID=ON"
+    fi
+
+    if [[ $CMAKE_OPTS == *WITH_VALGRIND=ON* ]]
+    then
+        CMAKE_OPTS="${CMAKE_OPTS:-} -DUSE_VALGRIND=ON"
+    fi
+fi
+
 #
 
 COMMON_FLAGS="-DPERCONA_INNODB_VERSION=$PERCONA_SERVER_EXTENSION"
@@ -293,13 +324,6 @@ export CFLAGS=" $COMMON_FLAGS -static-libgcc $MACHINE_SPECS_CFLAGS ${CFLAGS:-}"
 export CXXFLAGS=" $COMMON_FLAGS $MACHINE_SPECS_CFLAGS ${CXXFLAGS:-}"
 export MAKE_JFLAG="${MAKE_JFLAG:--j$PROCESSORS}"
 
-export DEBIAN_VERSION="$(lsb_release -sc)"
-echo $DEBIAN_VERSION
-if [[ "$CMAKE_BUILD_TYPE" == "Debug" ]] && [[ "$DEBIAN_VERSION" == "yakkety" ]]; then
-    export CFLAGS=" $CFLAGS -fno-strict-aliasing -Wno-unused-parameter -Wno-sign-compare -Wno-error=deprecated-declarations -Wno-error=nonnull-compare -Wno-error=shift-negative-value"
-    export CXXFLAGS=" $CFLAGS -fno-strict-aliasing -Wno-unused-parameter -Wno-sign-compare -Wno-error=deprecated-declarations -Wno-error=nonnull-compare -Wno-error=shift-negative-value"
-fi
-
 #
 # Test jemalloc directory
 if test "x$WITH_JEMALLOC" != "x"
@@ -331,15 +355,21 @@ fi
         cd "percona-xtradb-cluster-galera"
         if grep builtin <<< "$STAG";then
             # No builtin SSL in galera yet.
+            scons $MAKE_JFLAG psi=1 --config=force ssl=0 revno="$GALERA_REVISION" ${SCONS_ARGS} boost_pool=0 \
+                libgalera_smm.so
             scons $MAKE_JFLAG --config=force ssl=0 revno="$GALERA_REVISION" ${SCONS_ARGS} boost_pool=0 \
-                garb/garbd libgalera_smm.so
+                garb/garbd
         elif grep static <<< "$STAG";then
             # Disable SSL in galera for now
+            scons $MAKE_JFLAG psi=1 --config=force static_ssl=1 with_ssl=$GALERA_SSL \
+                revno="$GALERA_REVISION" ${SCONS_ARGS} boost_pool=0 libgalera_smm.so
             scons $MAKE_JFLAG --config=force static_ssl=1 with_ssl=$GALERA_SSL \
-            revno="$GALERA_REVISION" ${SCONS_ARGS} boost_pool=0 garb/garbd libgalera_smm.so
+                revno="$GALERA_REVISION" ${SCONS_ARGS} boost_pool=0 garb/garbd
         else
+            scons $MAKE_JFLAG psi=1 --config=force revno="$GALERA_REVISION" ${SCONS_ARGS} \
+                libgalera_smm.so
             scons $MAKE_JFLAG --config=force revno="$GALERA_REVISION" ${SCONS_ARGS} \
-                garb/garbd libgalera_smm.so
+                garb/garbd
         fi
         mkdir -p "$TARGETDIR/usr/local/$PRODUCT_FULL_NAME/bin" \
              "$TARGETDIR/usr/local/$PRODUCT_FULL_NAME/lib"
@@ -368,6 +398,7 @@ fi
         cmake ../../ ${CMAKE_OPTS:-} -DBUILD_CONFIG=mysql_release \
             -DCMAKE_BUILD_TYPE=Debug \
             $DEBUG_EXTRA \
+            -DDEBUG_EXTNAME=ON \
             -DWITH_EMBEDDED_SERVER=OFF \
             -DFEATURE_SET=community \
             -DENABLE_DTRACE=OFF \
@@ -384,16 +415,18 @@ fi
             -DCOMPILATION_COMMENT="$COMMENT - UNIV_DEBUG ON" \
             -DWITH_PAM=ON \
             -DWITH_INNODB_MEMCACHED=ON \
-            -DWITH_SCALABILITY_METRICS=ON \
-            $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
+            -DDOWNLOAD_BOOST=1 \
+            -DWITH_BOOST="$TARGETDIR/libboost" \
+            $WITH_MECAB_OPTION $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
 
-        make $MAKE_JFLAG $QUIET || exit 1
-        make install || exit 1
+        (make $MAKE_JFLAG $QUIET) || exit 1
+        (make install) || exit 1
         cp -v sql/mysqld-debug $TARGETDIR/usr/local/$PRODUCT_FULL_NAME/bin/mysqld
         echo "mysqld in build in debug mode"
     else
         cmake ../../ ${CMAKE_OPTS:-} -DBUILD_CONFIG=mysql_release \
             -DCMAKE_BUILD_TYPE=${CMAKE_BUILD_TYPE:-RelWithDebInfo} \
+            $DEBUG_EXTNAME \
             -DWITH_EMBEDDED_SERVER=OFF \
             -DFEATURE_SET=community \
             -DENABLE_DTRACE=OFF \
@@ -409,11 +442,12 @@ fi
             -DCOMPILATION_COMMENT="$COMMENT" \
             -DWITH_PAM=ON \
             -DWITH_INNODB_MEMCACHED=ON \
-            -DWITH_SCALABILITY_METRICS=ON \
-            $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
+            -DDOWNLOAD_BOOST=1 \
+            -DWITH_BOOST="$TARGETDIR/libboost" \
+            $WITH_MECAB_OPTION $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
 
-        make $MAKE_JFLAG $QUIET || exit 1
-        make install || exit 1
+        (make $MAKE_JFLAG $QUIET) || exit 1
+        (make install) || exit 1
         echo "mysqld in build in release mode"
     fi
 
@@ -450,6 +484,7 @@ fi
 
 if [[ $KEEP_BUILD -eq 0 ]]
 then
+    rm -rf $TARGETDIR/libboost
     rm -rf $TARGETDIR/bld
     rm -rf $TARGETDIR/usr
 fi
