@@ -1,20 +1,13 @@
--- Copyright (c) 2008, 2020, Oracle and/or its affiliates. All rights reserved.
+-- Copyright (c) 2008, 2015, Oracle and/or its affiliates. All rights reserved.
 --
 -- This program is free software; you can redistribute it and/or modify
--- it under the terms of the GNU General Public License, version 2.0,
--- as published by the Free Software Foundation.
---
--- This program is also distributed with certain software (including
--- but not limited to OpenSSL) that is licensed under separate terms,
--- as designated in a particular file or component or in included license
--- documentation.  The authors of MySQL hereby grant you an additional
--- permission to link the program and your derivative works with the
--- separately licensed software that they have included with MySQL.
+-- it under the terms of the GNU General Public License as published by
+-- the Free Software Foundation; version 2 of the License.
 --
 -- This program is distributed in the hope that it will be useful,
 -- but WITHOUT ANY WARRANTY; without even the implied warranty of
 -- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--- GNU General Public License, version 2.0, for more details.
+-- GNU General Public License for more details.
 --
 -- You should have received a copy of the GNU General Public License
 -- along with this program; if not, write to the Free Software Foundation,
@@ -106,7 +99,8 @@ INSERT INTO global_suppressions VALUES
  ("Forcing shutdown of [0-9]* plugins"),
  ("Forcing close of thread"),
 
- ("Percona Server cannot operate under OpenSSL FIPS mode"),
+ ("innodb-page-size has been changed"),
+
  /*
    Due to timing issues, it might be that this warning
    is printed when the server shuts down and the
@@ -183,6 +177,8 @@ INSERT INTO global_suppressions VALUES
  ("No existing UUID has been found, so we assume that this is the first time that this server has been started.*"),
  /*It will print a warning if server is run without --explicit_defaults_for_timestamp.*/
  ("TIMESTAMP with implicit DEFAULT value is deprecated. Please use --explicit_defaults_for_timestamp server option (see documentation for more details)*"),
+ /*It will print a warning if a server is run without NO_AUTO_CREATE_USER sql mode.*/
+ ("'NO_AUTO_CREATE_USER' sql mode was not set."),
 
  /* Added 2009-08-XX after fixing Bug #42408 */
 
@@ -205,43 +201,24 @@ INSERT INTO global_suppressions VALUES
  ("==[0-9]*== Warning: set address range perms: large range"),
  /* valgrind-3.5.0 dumps this */
  ("==[0-9]*== Command: "),
- /* Messages from valgrind tools */
- ("==[0-9]*== Callgrind"),
- ("==[0-9]*== For interactive control, run 'callgrind_control -h'"),
- ("==[0-9]*== Events    :"),
- ("==[0-9]*== Collected : [0-9]+"),
- ("==[0-9]*== I   refs:      [0-9]+"),
- ("==[0-9]*== Massif"),
- ("==[0-9]*== Helgrind"),
 
  /* valgrind warnings: invalid file descriptor -1 in syscall
     write()/read(). Bug #50414 */
  ("==[0-9]*== Warning: invalid file descriptor -1 in syscall write()"),
  ("==[0-9]*== Warning: invalid file descriptor -1 in syscall read()"),
 
- /* Suppress warnings caused by foreign clients, see Bug#31893901 */
-
- ("IP address .* could not be resolved.*"),
-
  /*
    Transient network failures that cause warnings on reconnect.
    BUG#47743 and BUG#47983.
  */
- ("Slave I/O: Get master SERVER_UUID failed with error:.*"),
- ("Slave I/O: Get master SERVER_ID failed with error:.*"),
- ("Slave I/O: Get master clock failed with error:.*"),
- ("Slave I/O: Get master COLLATION_SERVER failed with error:.*"),
- ("Slave I/O: Get master TIME_ZONE failed with error:.*"),
- ("Slave I/O: The slave I/O thread stops because a fatal error is encountered when it tried to SET @master_binlog_checksum on master.*"),
- ("Slave I/O: Get master BINLOG_CHECKSUM failed with error.*"),
- ("Slave I/O: Notifying master by SET @master_binlog_checksum= @@global.binlog_checksum failed with error.*"),
- /*
-   BUG#42147 - Concurrent DML and LOCK TABLE ... READ for InnoDB 
-   table cause warnings in errlog
-   Note: This is a temporary suppression until Bug#42147 can be 
-   fixed properly. See bug page for more information.
-  */
- ("Found lock of type 6 that is write and read locked"),
+ ("Slave I/O.*: Get master SERVER_UUID failed with error:.*"),
+ ("Slave I/O.*: Get master SERVER_ID failed with error:.*"),
+ ("Slave I/O.*: Get master clock failed with error:.*"),
+ ("Slave I/O.*: Get master COLLATION_SERVER failed with error:.*"),
+ ("Slave I/O.*: Get master TIME_ZONE failed with error:.*"),
+ ("Slave I/O.*: The slave I/O thread stops because a fatal error is encountered when it tried to SET @master_binlog_checksum on master.*"),
+ ("Slave I/O.*: Get master BINLOG_CHECKSUM failed with error.*"),
+ ("Slave I/O.*: Notifying master by SET @master_binlog_checksum= @@global.binlog_checksum failed with error.*"),
 
  /*
    Warning message is printed out whenever a slave is started with
@@ -259,7 +236,7 @@ INSERT INTO global_suppressions VALUES
   In MTS if the user issues a stop slave sql while it is scheduling a group
   of events, this warning is emitted.
   */
- ("Slave SQL: Coordinator thread of multi-threaded slave is being stopped in the middle of assigning a group of events.*"),
+ ("Slave SQL.*: Coordinator thread of multi-threaded slave is being stopped in the middle of assigning a group of events.*"),
  
  ("Changed limits: max_open_files: *"),
  ("Changed limits: max_connections: *"),
@@ -267,25 +244,24 @@ INSERT INTO global_suppressions VALUES
  ("Could not increase number of max_open_files to more than *"),
 
  /*
-  Warnings related to --secure-file-priv
+   Warning message if xb-doublewrite buffer file is located on tmpfs
+   Introduced by the parallel doublewrite buffer feature introduced in
+   PS-5.7.11-4
+ */
+ ("InnoDB: Failed to set O_DIRECT on "),
+
+ /*
+   Warning message introduced by wl#7706
+ */
+ ("CA certificate .* is self signed"),
+
+ /*
+   Warnings related to --secure-file-priv
  */
  ("Insecure configuration for --secure-file-priv:*"),
 
  /*
-  Bug#26585560, warning related to --pid-file
- */
- ("Insecure configuration for --pid-file:*"),
- ("Few location(s) are inaccessible while checking PID filepath"),
- /*
-   Following WL#12670, this warning is expected.
- */
- ("Setting named_pipe_full_access_group='\\*everyone\\*' is insecure"),
-
- /* ASan memory allocation warnings */
- ("==[0-9]*== WARNING: AddressSanitizer failed to allocate 0x[0-9a-f]+ bytes"),
-
- /*
-   Galera suppressions
+   Galera suppressions 
  */
  ("WSREP:*down context*"),
  ("WSREP: Failed to send state UUID:*"),
@@ -318,22 +294,26 @@ INSERT INTO global_suppressions VALUES
  ("WSREP: no nodes coming from prim view, prim not possible"),
  ("WSREP: Failed to prepare for incremental state transfer: Local state seqno is undefined:"),
  ("WSREP: gcs_caused\\(\\) returned -107 \\(Transport endpoint is not connected\\)"),
- ("WSREP: gcs_caused\\(\\) returned -57 \\(Socket is not connected\\)"),
  ("WSREP: gcs_caused\\(\\) returned -1 \\(Operation not permitted\\)"),
+
  ("Action message in non-primary configuration from member"),
- ("SYNC message from member"),
  ("InnoDB: Resizing redo log from"),
  ("InnoDB: Starting to delete and rewrite log files"),
  ("InnoDB: New log files created, LSN="),
 -- WSREP: Send action {0x7f86280147f0, 73, STATE_REQUEST} returned -107 (Transport endpoint is not connected)
  ("Transport endpoint is not connected"),
- ("Socket is not connected"),
 -- "WSREP: Protocol violation. JOIN message sender 1.0 (host-91-221-67-96) is not in state transfer (SYNCED). Message ignored.
  ("is not in state transfer"),
- ("JOIN message from member .* in non-primary configuration"),
  ("install timer expired"),
  ("Last Applied Action message in non-primary configuration from member"),
- ("IP address '127.0.0.2' could not be resolved: Name or service not known"),
+
+ ("IP address \'127.0.0.2\' could not be resolved: Name or service not known"),
+ ("JOIN message from member .* in non-primary configuration"),
+ ("SYNC message from member"),
+ ("WSREP: Percona-XtraDB-Cluster prohibits setting binlog_format to STATEMENT or MIXED at global level"),
+ ("WSREP: Table without explict primary key \\(not-recommended\\) and certification of nonPK table is OFF too"),
+ ("WSREP: Node is not a cluster node. Disabling pxc_strict_mode"),
+ ("WSREP: pxc_strict_mode can be changed only if node is cluster-node"),
 
  ("THE_LAST_SUPPRESSION")||
 
