@@ -1,21 +1,15 @@
 # -*- cperl -*-
-# Copyright (c) 2007, 2017, Oracle and/or its affiliates. All rights reserved.
+# Copyright (c) 2007, 2016, Oracle and/or its affiliates. All rights reserved.
 #
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License, version 2.0,
-# as published by the Free Software Foundation.
-#
-# This program is also distributed with certain software (including
-# but not limited to OpenSSL) that is licensed under separate terms,
-# as designated in a particular file or component or in included license
-# documentation.  The authors of MySQL hereby grant you an additional
-# permission to link the program and your derivative works with the
-# separately licensed software that they have included with MySQL.
+# This program is free software; you can redistribute it and/or
+# modify it under the terms of the GNU Library General Public
+# License as published by the Free Software Foundation; version 2
+# of the License.
 #
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License, version 2.0, for more details.
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# Library General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
@@ -33,6 +27,7 @@ use My::Platform;
 
 use File::Basename;
 
+
 #
 # Rules to run first of all
 #
@@ -41,9 +36,8 @@ sub add_opt_values {
   my ($self, $config)= @_;
 
   # add auto-options
-  $config->insert('OPT', 'port'   => fix_port($self, $config) );
-  $config->insert('OPT', 'vardir' => sub { $self->{ARGS}->{vardir} });
-  $config->insert('mysqld', "loose-skip-$_" => undef) for (@::optional_plugins);
+  $config->insert('OPT', 'port'   => sub { fix_port($self, $config) });
+  $config->insert('mysqld', "loose-skip-plugin-$_" => undef) for (@::optional_plugins);
 }
 
 my @pre_rules=
@@ -150,7 +144,17 @@ sub fix_socket {
   my ($self, $config, $group_name, $group)= @_;
   # Put socket file in tmpdir
   my $dir= $self->{ARGS}->{tmpdir};
-  return "$dir/$group_name.sock";
+  my $socket = "$dir/$group_name.sock";
+ 
+  # Make sure the socket path does not become longer then the path
+  # which mtr uses to test if a new tmpdir should be created
+  if (length($socket) > length("$dir/mysql_testsocket.sock"))
+  {
+    # Too long socket path, generate shorter based on port
+    my $port = $group->value('port');
+    $socket = "$dir/mysqld-$port.sock"; 
+  }
+  return $socket;
 }
 
 sub fix_tmpdir {
@@ -192,11 +196,12 @@ sub ssl_supported {
   return $self->{ARGS}->{ssl};
 }
 
-sub fix_skip_ssl {
+sub fix_ssl_disabled {
   return if !ssl_supported(@_);
-  # Add skip-ssl if ssl is supported to avoid
+  return if $::opt_ssl;
+  # Add ssl-mode=DISABLED to avoid
   # that mysqltest connects with SSL by default
-  return 1;
+  return "DISABLED";
 }
 
 sub fix_ssl_ca {
@@ -205,6 +210,12 @@ sub fix_ssl_ca {
   return "$std_data/cacert.pem"
 }
 
+sub fix_client_ssl_ca {
+  return if !$::opt_ssl;
+  my $std_data= fix_std_data(@_); 
+  return "$std_data/cacert.pem"
+}
+
 sub fix_ssl_server_cert {
   return if !ssl_supported(@_);
   my $std_data= fix_std_data(@_);
@@ -264,6 +275,7 @@ my @mysqld_rules=
  { 'ssl-ca' => \&fix_ssl_ca },
  { 'ssl-cert' => \&fix_ssl_server_cert },
  { 'ssl-key' => \&fix_ssl_server_key },
+ { 'loose-sha256_password_auto_generate_rsa_keys' => "0"},
   );
 
 if (IS_WINDOWS)
@@ -318,6 +330,8 @@ my @ndbd_rules=
  { 'BackupDataDir' => \&fix_cluster_backup_dir },
 );
 
+
+#
 # Rules to run for each memcached in the config
 #  - will be run in order listed here
 #
@@ -355,10 +369,10 @@ my @client_rules=
 #
 my @mysqltest_rules=
 (
- { 'ssl-ca' => \&fix_ssl_ca },
+ { 'ssl-ca' => \&fix_client_ssl_ca },
  { 'ssl-cert' => \&fix_ssl_client_cert },
  { 'ssl-key' => \&fix_ssl_client_key },
- { 'skip-ssl' => \&fix_skip_ssl },
+ { 'ssl-mode' => \&fix_ssl_disabled },
 );
 
 
@@ -369,7 +383,6 @@ my @mysqltest_rules=
 my @mysqlbinlog_rules=
 (
  { 'character-sets-dir' => \&fix_charset_dir },
- { 'local-load' => sub { return shift->{ARGS}->{tmpdir}; } },
 );
 
 
@@ -379,7 +392,6 @@ my @mysqlbinlog_rules=
 #
 my @mysql_upgrade_rules=
 (
- { 'tmpdir' => sub { return shift->{ARGS}->{tmpdir}; } },
 );
 
 
@@ -655,20 +667,8 @@ sub new_config {
     croak "you must pass '$required'" unless defined $args->{$required};
   }
 
-  # Fill in hosts/port hash
-  my $hosts= {};
-  my $baseport= $args->{baseport};
-  $args->{hosts}= [ 'localhost' ] unless exists($args->{hosts});
-  foreach my $host ( @{$args->{hosts}} ) {
-     $hosts->{$host}= $baseport;
-  }
-
   # Open the config template
   my $config= My::Config->new($args->{'template_path'});
-  my $extra_template_path= $args->{'extra_template_path'};
-  if ($extra_template_path){
-    $config->append(My::Config->new($extra_template_path));
-  }
   my $self= bless {
 		   CONFIG       => $config,
 		   ARGS         => $args,
