@@ -1,21 +1,14 @@
 # -*- cperl -*-
-# Copyright (c) 2005, 2017, Oracle and/or its affiliates. All rights reserved.
+# Copyright (c) 2005, 2016, Oracle and/or its affiliates. All rights reserved.
 # 
 # This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License, version 2.0,
-# as published by the Free Software Foundation.
-#
-# This program is also distributed with certain software (including
-# but not limited to OpenSSL) that is licensed under separate terms,
-# as designated in a particular file or component or in included license
-# documentation.  The authors of MySQL hereby grant you an additional
-# permission to link the program and your derivative works with the
-# separately licensed software that they have included with MySQL.
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; version 2 of the License.
 #
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License, version 2.0, for more details.
+# GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
@@ -33,7 +26,6 @@ our @EXPORT= qw(collect_option collect_test_cases);
 
 use mtr_report;
 use mtr_match;
-use My::Constants;
 
 # Options used for the collect phase
 our $start_from;
@@ -53,7 +45,7 @@ our $quick_collect;
 # default storage engine settings, and use MyISAM
 # as default.  (temporary option used in connection
 # with the change of default storage engine to InnoDB)
-our $default_myisam= 1;
+our $default_myisam= 0;
  
 
 sub collect_option {
@@ -151,13 +143,7 @@ sub collect_test_cases ($$$$) {
 
   if ( @$opt_cases )
   {
-    # A list of tests was specified on the command line.
-    # Among those, the tests which are not already collected will be
-    # collected and stored temporarily in an array of hashes pointed
-    # by the below reference. This array is eventually appeneded to
-    # the one having all collected test cases.
-    my $cmdline_cases;
-
+    # A list of tests was specified on the command line
     # Check that the tests specified was found
     # in at least one suite
     foreach my $test_name_spec ( @$opt_cases )
@@ -176,56 +162,20 @@ sub collect_test_cases ($$$$) {
       }
       if ( not $found )
       {
-        if ( $sname )
+	$sname= "main" if !$opt_reorder and !$sname;
+	mtr_error("Could not find '$tname' in '$suites' suite(s)") unless $sname;
+	# If suite was part of name, find it there, may come with combinations
+	my @this_case = collect_one_suite($sname, [ $tname ]);
+	if (@this_case)
         {
-	  # If suite was part of name, find it there, may come with combinations
-	  my @this_case = collect_one_suite($sname, [ $tname ]);
-
-          # If a test is specified multiple times on the command line, all
-          # instances of the test need to be picked. Hence, such tests are
-          # stored in the temporary array instead of adding them to $cases
-          # directly so that repeated tests are not run only once
-	  if (@this_case)
-          {
-	    push (@$cmdline_cases, @this_case);
-	  }
-	  else
-	  {
-	    mtr_error("Could not find '$tname' in '$sname' suite");
-          }
-        }
-        else
-        {
-          if ( !$opt_reorder )
-          {
-            # If --no-reorder is passed and if suite was not part of name,
-            # search in all the suites
-            foreach my $suite (split(",", $suites))
-            {
-              my @this_case = collect_one_suite($suite, [ $tname ]);
-              if ( @this_case )
-              {
-                push (@$cmdline_cases, @this_case);
-                $found= 1;
-              }
-              @this_case= collect_one_suite("i_".$suite, [ $tname ]);
-              if ( @this_case )
-              {
-                push (@$cmdline_cases, @this_case);
-                $found= 1;
-              }
-            }
-          }
-          if ( !$found )
-          {
-            mtr_error("Could not find '$tname' in '$suites' suite(s)");
-          }
+	  push (@$cases, @this_case);
+	}
+	else
+	{
+	  mtr_error("Could not find '$tname' in '$sname' suite");
         }
       }
     }
-    # Add test cases collected in the temporary array to the one
-    # containing all previously collected test cases
-    push (@$cases, @$cmdline_cases) if $cmdline_cases;
   }
 
   if ( $opt_reorder && !$quick_collect)
@@ -347,6 +297,8 @@ sub collect_one_suite($)
 			      # Look in plugin specific suite dir
 			      "plugin/$suite/tests",
 			      "internal/plugin/$suite/tests",
+			      "rapid/plugin/$suite/tests",
+			      "rapid/mysql-test/suite",
 			     ],
 			     [$suite, "mtr"], ($suite =~ /^i_/));
       return unless $suitedir;
@@ -431,6 +383,12 @@ sub collect_one_suite($)
 
   # Read suite.opt file
   my $suite_opt_file=  "$testdir/suite.opt";
+
+  if ( $::opt_suite_opt )
+  {
+    $suite_opt_file= "$testdir/$::opt_suite_opt";
+  }
+
   my $suite_opts= [];
   if ( -f $suite_opt_file )
   {
@@ -656,7 +614,6 @@ sub optimize_cases {
 	if ( !$supported )
 	{
 	  $tinfo->{'skip'}= 1;
-	  $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
 	  $tinfo->{'comment'}=
 	    "Doesn't support --binlog-format='$binlog_format'";
 	}
@@ -683,7 +640,6 @@ sub optimize_cases {
 	if ( !$supported )
 	{
 	  $tinfo->{'skip'}= 1;
-	  $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
 	  $tinfo->{'comment'}=
 	    "Doesn't support --binlog-format='$test_binlog_format'";
 	  next;
@@ -719,15 +675,14 @@ sub optimize_cases {
 	     ! exists $builtin_engines{$default_engine} )
 	{
 	  $tinfo->{'skip'}= 1;
-	  $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
 	  $tinfo->{'comment'}=
 	    "'$default_engine' not supported";
 	}
 
 	$tinfo->{'ndb_test'}= 1
 	  if ( $default_engine =~ /^ndb/i );
-	$tinfo->{'innodb_test'}= 1
-	  if ( $default_engine =~ /^innodb/i );
+	$tinfo->{'myisam_test'}= 1
+	  if ( $default_engine =~ /^myisam/i );
       }
       if (defined $default_tmp_engine){
 
@@ -741,15 +696,14 @@ sub optimize_cases {
 	     ! exists $builtin_engines{$default_tmp_engine} )
 	{
 	  $tinfo->{'skip'}= 1;
-	  $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
 	  $tinfo->{'comment'}=
 	    "'$default_tmp_engine' not supported";
 	}
 
 	$tinfo->{'ndb_test'}= 1
 	  if ( $default_tmp_engine =~ /^ndb/i );
-	$tinfo->{'innodb_test'}= 1
-	  if ( $default_tmp_engine =~ /^innodb/i );
+	$tinfo->{'myisam_test'}= 1
+	  if ( $default_tmp_engine =~ /^myisam/i );
       }
     }
 
@@ -911,6 +865,14 @@ sub collect_one_test_case {
     return $tinfo;
   }
 
+  # ----------------------------------------------------------------------
+  # Check for replicaton tests
+  # ----------------------------------------------------------------------
+  if ( $suitedir =~ 'rpl' )
+  {
+    $tinfo->{'rpl_test'}= 1;
+  }
+
   # ----------------------------------------------------------------------
   # Check for disabled tests
   # ----------------------------------------------------------------------
@@ -980,7 +942,6 @@ sub collect_one_test_case {
     if ( IS_WIN32PERL )
     {
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "No tests with sh scripts on Windows";
       return $tinfo;
     }
@@ -999,7 +960,6 @@ sub collect_one_test_case {
     if ( IS_WIN32PERL )
     {
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "No tests with sh scripts on Windows";
       return $tinfo;
     }
@@ -1025,15 +985,14 @@ sub collect_one_test_case {
     $tinfo->{'ndb_test'}= 1
       if ( $default_storage_engine =~ /^ndb/i );
 
-    $tinfo->{'innodb_test'}= 1
-      if ( $default_storage_engine =~ /^innodb/i );
+    $tinfo->{'mysiam_test'}= 1
+      if ( $default_storage_engine =~ /^mysiam/i );
 
   }
 
   if ( $tinfo->{'big_test'} and ! $::opt_big_test )
   {
     $tinfo->{'skip'}= 1;
-    $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
     $tinfo->{'comment'}= "Test needs 'big-test' option";
     return $tinfo
   }
@@ -1041,7 +1000,6 @@ sub collect_one_test_case {
   if ( $tinfo->{'need_debug'} && ! $::debug_compiled_binaries )
   {
     $tinfo->{'skip'}= 1;
-    $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
     $tinfo->{'comment'}= "Test needs debug binaries";
     return $tinfo
   }
@@ -1053,7 +1011,6 @@ sub collect_one_test_case {
     {
       # ndbcluster is disabled
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "ndbcluster disabled";
       return $tinfo;
     }
@@ -1065,7 +1022,6 @@ sub collect_one_test_case {
     {
       # Only the ndb test should be run, all other should be skipped
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "Only ndbcluster tests";
       return $tinfo;
     }
@@ -1077,24 +1033,7 @@ sub collect_one_test_case {
     push(@{$tinfo->{'master_opt'}}, "--loose-federated");
     push(@{$tinfo->{'slave_opt'}}, "--loose-federated");
   }
-
-  if ( $tinfo->{'innodb_test'} )
-  {
-    # This is a test that needs innodb
-    if ( $::mysqld_variables{'innodb'} eq "OFF" ||
-         ! exists $::mysqld_variables{'innodb'} )
-    {
-      # innodb is not supported, skip it
-      $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
-      # This comment is checked for running with innodb plugin (see above),
-      # please keep that in mind if changing the text.
-      $tinfo->{'comment'}= "No innodb support";
-      # But continue processing if we may run it with innodb plugin
-      return $tinfo unless $do_innodb_plugin;
-    }
-  }
-  elsif ($default_myisam)
+  if ( $tinfo->{'myisam_test'})
   {
     # This is a temporary fix to allow non-innodb tests to run even if
     # the default storage engine is innodb.
@@ -1103,13 +1042,11 @@ sub collect_one_test_case {
     push(@{$tinfo->{'master_opt'}}, "--default-tmp-storage-engine=MyISAM");
     push(@{$tinfo->{'slave_opt'}}, "--default-tmp-storage-engine=MyISAM");
   }
-
   if ( $tinfo->{'need_binlog'} )
   {
     if (grep(/^--skip-log-bin/,  @::opt_extra_mysqld_opt) )
     {
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "Test needs binlog";
       return $tinfo;
     }
@@ -1127,7 +1064,6 @@ sub collect_one_test_case {
     if ( $skip_rpl )
     {
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "No replication tests(--skip-rpl)";
       return $tinfo;
     }
@@ -1138,10 +1074,17 @@ sub collect_one_test_case {
     if ( $tinfo->{'not_embedded'} )
     {
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "Not run for embedded server";
       return $tinfo;
     }
+#Setting the default storage engine to InnoDB for embedded tests as the default
+#storage engine for mysqld in embedded mode is still MyISAM.
+#To be removed after completion of WL #6911.
+    if ( !$tinfo->{'myisam_test'} && !defined $default_storage_engine)
+    {
+      push(@{$tinfo->{'master_opt'}}, "--default-storage-engine=InnoDB");
+      push(@{$tinfo->{'master_opt'}}, "--default-tmp-storage-engine=InnoDB");
+    }
   }
 
   if ( $tinfo->{'need_ssl'} )
@@ -1150,7 +1093,6 @@ sub collect_one_test_case {
     if ( ! $::opt_ssl_supported ) {
       # SSL is not supported, skip it
       $tinfo->{'skip'}= 1;
-      $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
       $tinfo->{'comment'}= "No SSL support";
       return $tinfo;
     }
@@ -1159,7 +1101,6 @@ sub collect_one_test_case {
   if ( $tinfo->{'not_windows'} && IS_WINDOWS )
   {
     $tinfo->{'skip'}= 1;
-    $tinfo->{'skip_reason'}= MTR_SKIP_BY_FRAMEWORK;
     $tinfo->{'comment'}= "Test not supported on Windows";
     return $tinfo;
   }
@@ -1233,8 +1174,8 @@ my @tags=
   "binlog_formats", ["row", "statement"]],
 
  ["include/have_log_bin.inc", "need_binlog", 1],
-
- ["include/have_innodb.inc", "innodb_test", 1],
+# an empty file to use test that needs myisam engine.
+ ["include/force_myisam_default.inc", "myisam_test", 1],
  ["include/big_test.inc", "big_test", 1],
  ["include/have_debug.inc", "need_debug", 1],
  ["include/have_ndb.inc", "ndb_test", 1],
