@@ -1,4 +1,4 @@
-#!@SHELL_PATH@
+#!/bin/sh
 # Copyright Abandoned 1996 TCX DataKonsult AB & Monty Program KB & Detron HB
 # This file is public domain and comes with NO WARRANTY of any kind
 #
@@ -34,9 +34,11 @@ pid_file=
 err_log=
 wsrep_data_home_dir=""
 grastate_loc=""
+timestamp_format=UTC
 
 syslog_tag_mysqld=mysqld
 syslog_tag_mysqld_safe=mysqld_safe
+syslog_facility=daemon
 
 trap '' 1 2 3 15			# we shouldn't let anyone kill us
 trap '' 13                              # not even SIGPIPE
@@ -82,12 +84,9 @@ EOF
         fi
         cat <<EOF
 Usage: $0 [OPTIONS]
- The following options may be given as the first argument:
   --no-defaults              Don't read the system defaults file
   --defaults-file=FILE       Use the specified defaults file
   --defaults-extra-file=FILE Also use defaults from the specified file
-
- Other options:
   --ledir=DIRECTORY          Look for mysqld in the specified directory
   --open-files-limit=LIMIT   Limit the number of open files
   --core-file-size=LIMIT     Limit core files to the specified size
@@ -107,6 +106,10 @@ Usage: $0 [OPTIONS]
   --numa-interleave          Run mysqld with its memory interleaved
                              on all NUMA nodes
 ${thp_usage}
+  --mysqld-safe-log-         TYPE must be one of UTC (ISO 8601 UTC),
+    timestamps=TYPE          system (ISO 8601 local time), hyphen
+                             (hyphenated date a la mysqld 5.6), legacy
+                             (legacy non-ISO 8601 mysqld_safe timestamps)
 
 All other options are passed to the mysqld program.
 
@@ -147,18 +150,13 @@ log_generic () {
   priority="$1"
   shift
 
-  msg="`date +'%y%m%d %H:%M:%S'` mysqld_safe $*"
+  msg="`eval $DATE` mysqld_safe $*"
   echo "$msg"
   case $logging in
     init) ;;  # Just echo the message, don't save it anywhere
-    file)
-      if [ -w / -o "$USER" = "root" ]; then
-        true
-      else
-        echo "$msg" >> "$err_log"
-      fi
-      ;;
+    file) echo "$msg" >> "$err_log" ;;
     syslog) logger -t "$syslog_tag_mysqld_safe" -p "$priority" "$*" ;;
+    both) echo "$msg" >> "$err_log"; logger -t "$syslog_tag_mysqld_safe" -p "$priority" "$*" ;;
     *)
       echo "Internal program error (non-fatal):" \
            " unknown logging method '$logging'" >&2
@@ -167,32 +165,22 @@ log_generic () {
 }
 
 log_error () {
-  log_generic daemon.error "$@" >&2
+  log_generic ${syslog_facility}.error "$@" >&2
 }
 
 log_notice () {
-  log_generic daemon.notice "$@"
+  log_generic ${syslog_facility}.notice "$@"
 }
 
 eval_log_error () {
   local cmd="$1"
   case $logging in
-    file)
-      if [ -w / -o "$USER" = "root" ]; then
-        cmd="$cmd > /dev/null 2>&1"
-      else
-        cmd="$cmd >> "`shell_quote_string "$err_log"`" 2>&1"
-      fi
-      ;;
+    file) cmd="$cmd >> "`shell_quote_string "$err_log"`" 2>&1" ;;
     syslog)
-      # mysqld often prefixes its messages with a timestamp, which is
-      # redundant when logging to syslog (which adds its own timestamp)
-      # However, we don't strip the timestamp with sed here, because
-      # sed buffers output (only GNU sed supports a -u (unbuffered) option)
-      # which means that messages may not get sent to syslog until the
-      # mysqld process quits.
-      cmd="$cmd 2>&1 | logger -t '$syslog_tag_mysqld' -p daemon.error"
+      cmd="$cmd --log-syslog=1 --log-syslog-facility=$syslog_facility '--log-syslog-tag=$syslog_tag' > /dev/null 2>&1"
       ;;
+    both)
+      cmd="$cmd --log-syslog=1 --log-syslog-facility=$syslog_facility '--log-syslog-tag=$syslog_tag' >> "`shell_quote_string "$err_log"`" 2>&1" ;;
     *)
       echo "Internal program error (non-fatal):" \
            " unknown logging method '$logging'" >&2
@@ -254,7 +242,7 @@ wsrep_recover_position() {
   # If sequence number is not equal to -1, wsrep-recover co-ordinates aren't used.
   # lp:1112724
   # So, directly pass whatever is obtained from grastate.dat
-  if [ ! -z $seqno ] && [ $seqno -ne -1 ]; then
+  if [ $seqno -ne -1 ];then 
     log_notice "Skipping wsrep-recover for $uuid:$seqno pair"
     log_notice "Assigning $uuid:$seqno to wsrep_start_position"
     wsrep_start_position_opt="--wsrep_start_position=$uuid:$seqno"
@@ -321,19 +309,14 @@ parse_arguments() {
     # the parameter after "=", or the whole $arg if no match
     val=`echo "$arg" | sed -e 's;^--[^=]*=;;'`
     # what's before "=", or the whole $arg if no match
-    optname=`echo "$arg" | sed -e 's;^\(--[^=]*\)=.*$;\1;'`
+    optname=`echo "$arg" | sed -e 's/^\(--[^=]*\)=.*$/\1/'`
     # replace "_" by "-" ; mysqld_safe must accept "_" like mysqld does.
     optname_subst=`echo "$optname" | sed 's/_/-/g'`
-    arg=`echo $arg | sed "s;^$optname;$optname_subst;"`
+    arg=`echo $arg | sed "s/^$optname/$optname_subst/"`
     case "$arg" in
       # these get passed explicitly to mysqld
       --basedir=*) MY_BASEDIR_VERSION="$val" ;;
-      --datadir=*)
-        case $val in
-          /) DATADIR=$val ;;
-          *) DATADIR="`echo $val | sed 's;/*$;;'`" ;;
-        esac
-        ;;
+      --datadir=*) DATADIR="$val" ;;
       --pid-file=*) pid_file="$val" ;;
       --plugin-dir=*) PLUGIN_DIR="$val" ;;
       --user=*) user="$val"; SET_USER=1 ;;
@@ -346,28 +329,13 @@ parse_arguments() {
 
       # mysqld_safe-specific options - must be set in my.cnf ([mysqld_safe])!
       --core-file-size=*) core_file_size="$val" ;;
-      --ledir=*)
-        if [ -z "$pick_args" ]; then
-          log_error "--ledir option can only be used as command line option, found in config file"
-          exit 1
-        fi
-        ledir="$val"
-        ;;
-     --malloc-lib=*)
+      --ledir=*) ledir="$val" ;;
+      --malloc-lib=*)
 	set_malloc_lib "$val"
 	load_jemalloc=0
 	;;
-      --mysqld=*)
-        if [ -z "$pick_args" ]; then
-          log_error "--mysqld option can only be used as command line option, found in config file"
-          exit 1
-        fi
-        MYSQLD="$val" ;;
+      --mysqld=*) MYSQLD="$val" ;;
       --mysqld-version=*)
-        if [ -z "$pick_args" ]; then
-          log_error "--mysqld-version option can only be used as command line option, found in config file"
-          exit 1
-        fi
         if test -n "$val"
         then
           MYSQLD="mysqld-$val"
@@ -382,6 +350,7 @@ parse_arguments() {
       --skip-kill-mysqld*) KILL_MYSQLD=0 ;;
       --thp-setting=*) thp_setting="$val" ;;
       --preload-hotbackup) load_hotbackup=1 ;;
+      --mysqld-safe-log-timestamps=*) timestamp_format="$val" ;;
       --syslog) want_syslog=1 ;;
       --skip-syslog) want_syslog=0 ;;
       --syslog-tag=*) syslog_tag="$val" ;;
@@ -389,7 +358,6 @@ parse_arguments() {
       --wsrep[-_]urls=*) wsrep_urls="$val"; ;;
       --wsrep-data-home-dir=*) wsrep_data_home_dir="$val"; ;;
       --flush-caches=*) flush_caches="$val" ;;
-      --numa-interleave=*) numa_interleave="$val" ;;
       --exit-on-recover-fail) resume_on_fail=0 ;;
       --wsrep[-_]provider=*)
         if test -n "$val" && test "$val" != "none"
@@ -398,6 +366,7 @@ parse_arguments() {
         fi
         append_arg_to_args "$arg"
         ;;
+
       --help) usage ;;
 
       *)
@@ -421,11 +390,10 @@ add_mysqld_ld_preload() {
   lib_to_add="$1"
   lib_to_add=$(readlink -f $lib_to_add)
   log_notice "Adding '$lib_to_add' to LD_PRELOAD for mysqld"
-  real_basedir=$(readlink -f ${MY_BASEDIR_VERSION})
 
   # Check if the library is in the reduced number of standard system directories
   case "$lib_to_add" in
-    /usr/lib64/* | /usr/lib/* | ${MY_BASEDIR_VERSION}/lib/* | ${real_basedir}/lib/*)
+    /usr/lib64/* | /usr/lib/* | ${MY_BASEDIR_VERSION}/lib/*)
       ;;
     *)
       log_error "ld_preload libraries can only be loaded from system directories (/usr/lib64, /usr/lib, ${MY_BASEDIR_VERSION}/lib)"
@@ -480,41 +448,66 @@ mysqld_ld_preload_text() {
   echo "$text"
 }
 
+
+mysql_config=
+get_mysql_config() {
+  if [ -z "$mysql_config" ]; then
+    mysql_config=`echo "$0" | sed 's,/[^/][^/]*$,/mysql_config,'`
+    if [ ! -x "$mysql_config" ]; then
+      log_error "Can not run mysql_config $@ from '$mysql_config'"
+      exit 1
+    fi
+  fi
+
+  "$mysql_config" "$@"
+}
+
+
 # set_malloc_lib LIB
 # - If LIB is empty, do nothing and return
+# - If LIB is 'tcmalloc', look for tcmalloc shared library in /usr/lib
+#   then pkglibdir.  tcmalloc is part of the Google perftools project.
 # - If LIB is an absolute path, assume it is a malloc shared library
 #
 # Put LIB in mysqld_ld_preload, which will be added to LD_PRELOAD when
 # running mysqld.  See ld.so for details.
 set_malloc_lib() {
-  # This list is kept intentionally simple.
-  malloc_dirs="/usr/lib /usr/lib64 /usr/lib/i386-linux-gnu /usr/lib/x86_64-linux-gnu"
   malloc_lib="$1"
 
+  if [ "$malloc_lib" = tcmalloc ]; then
+    pkglibdir=`get_mysql_config --variable=pkglibdir`
+    malloc_lib=
+    # This list is kept intentionally simple.  Simply set --malloc-lib
+    # to a full path if another location is desired.
+    for libdir in /usr/lib "$pkglibdir" "$pkglibdir/mysql"; do
+      for flavor in _minimal '' _and_profiler _debug; do
+        tmp="$libdir/libtcmalloc$flavor.so"
+        #log_notice "DEBUG: Checking for malloc lib '$tmp'"
+        [ -r "$tmp" ] || continue
+        malloc_lib="$tmp"
+        break 2
+      done
+    done
+
+    if [ -z "$malloc_lib" ]; then
+      log_error "no shared library for --malloc-lib=tcmalloc found in /usr/lib or $pkglibdir"
+      exit 1
+    fi
+  fi
+
   # Allow --malloc-lib='' to override other settings
   [ -z  "$malloc_lib" ] && return
 
   case "$malloc_lib" in
     /*)
       if [ ! -r "$malloc_lib" ]; then
-        log_error "--malloc-lib can not be read and will not be used"
+        log_error "--malloc-lib '$malloc_lib' can not be read and will not be used"
         exit 1
       fi
-
-      # Restrict to a the list in $malloc_dirs above
-      case "`dirname "$malloc_lib"`" in
-        /usr/lib) ;;
-        /usr/lib64) ;;
-        /usr/lib/i386-linux-gnu) ;;
-        /usr/lib/x86_64-linux-gnu) ;;
-        *)
-          log_error "--malloc-lib must be located in one of the directories: $malloc_dirs"
-          exit 1
-          ;;
-      esac
       ;;
     *)
-      log_error "--malloc-lib must be an absolute path ignoring value '$malloc_lib'"
+      log_error "--malloc-lib must be an absolute path or 'tcmalloc'; " \
+        "ignoring value '$malloc_lib'"
       exit 1
       ;;
   esac
@@ -535,15 +528,7 @@ else
   relpkgdata='@pkgdatadir@'
 fi
 
-case "$0" in
-  /*)
-  MY_PWD='@prefix@'
-  ;;
-  *)
-  MY_PWD=`dirname $0`
-  MY_PWD=`dirname $MY_PWD`
-  ;;
-esac
+MY_PWD=`pwd`
 # Check for the directories we would expect from a binary release install
 if test -n "$MY_BASEDIR_VERSION" -a -d "$MY_BASEDIR_VERSION"
 then
@@ -587,10 +572,6 @@ fi
 if test -d $MY_BASEDIR_VERSION/data/mysql
 then
   DATADIR=$MY_BASEDIR_VERSION/data
-  if test -z "$defaults" -a -r "$DATADIR/my.cnf"
-  then
-    defaults="--defaults-extra-file=$DATADIR/my.cnf"
-  fi
 # Next try where the source installs put it
 elif test -d $MY_BASEDIR_VERSION/var/mysql
 then
@@ -602,23 +583,7 @@ fi
 
 if test -z "$MYSQL_HOME"
 then 
-  if test -r "$MY_BASEDIR_VERSION/my.cnf" && test -r "$DATADIR/my.cnf"
-  then
-    log_error "WARNING: Found two instances of my.cnf -
-$MY_BASEDIR_VERSION/my.cnf and
-$DATADIR/my.cnf
-IGNORING $DATADIR/my.cnf"
-
-    MYSQL_HOME=$MY_BASEDIR_VERSION
-  elif test -r "$DATADIR/my.cnf"
-  then
-    log_error "WARNING: Found $DATADIR/my.cnf
-The data directory is a deprecated location for my.cnf, please move it to
-$MY_BASEDIR_VERSION/my.cnf"
-    MYSQL_HOME=$DATADIR
-  else
-    MYSQL_HOME=$MY_BASEDIR_VERSION
-  fi
+  MYSQL_HOME=$MY_BASEDIR_VERSION
 fi
 export MYSQL_HOME
 
@@ -651,7 +616,7 @@ append_arg_to_args () {
 args=
 
 SET_USER=2
-parse_arguments `$print_defaults $defaults --loose-verbose mysqld server | sed 's/\s//g'`
+parse_arguments `$print_defaults $defaults --loose-verbose mysqld server`
 if test $SET_USER -eq 2
 then
   SET_USER=0
@@ -678,7 +643,7 @@ fi
 #
 if test $load_hotbackup -eq 1
 then
-  for libhb in "${MY_BASEDIR_VERSION}/lib" "/usr/lib64" "/usr/lib/x86_64-linux-gnu" "/usr/lib"; do
+  for libhb in "${MY_BASEDIR_VERSION}/lib" "/usr/lib64" "/usr/lib/x86_64-linux-gnu" "/usr/lib" "${MY_BASEDIR_VERSION}/lib/mysql" "/usr/lib64/mysql" "/usr/lib/x86_64-linux-gnu/mysql" "/usr/lib/mysql"; do
     if [ -r "$libhb/libHotBackup.so" ]; then
       add_mysqld_ld_preload "$libhb/libHotBackup.so"
       break
@@ -686,6 +651,19 @@ then
   done
 fi
 
+#
+# Sort out date command from $timestamp_format early so we'll start off
+# with correct log messages.
+#
+case "$timestamp_format" in
+    UTC|utc)       DATE="date -u +%Y-%m-%dT%H:%M:%S.%06NZ";;
+    SYSTEM|system) DATE="date +%Y-%m-%dT%H:%M:%S.%06N%:z";;
+    HYPHEN|hyphen) DATE="date +'%Y-%m-%d %H:%M:%S'";;
+    LEGACY|legacy) DATE="date +'%y%m%d %H:%M:%S'";;
+    *)             DATE="date -u +%Y-%m-%dT%H:%M:%S.%06NZ";
+                   log_error "unknown data format $timestamp_format, using UTC";;
+esac
+
 #
 # Try to find the plugin directory
 #
@@ -709,6 +687,32 @@ else
 fi
 plugin_dir="${plugin_dir}${PLUGIN_VARIANT}"
 
+# A pid file is created for the mysqld_safe process. This file protects the
+# server instance resources during race conditions.
+safe_pid="$DATADIR/mysqld_safe.pid"
+if test -f $safe_pid
+then
+  PID=`cat "$safe_pid"`
+  if @CHECK_PID@
+  then
+    if @FIND_PROC@
+    then
+      log_error "A mysqld_safe process already exists"
+      exit 1
+    fi
+  fi
+  rm -f "$safe_pid"
+  if test -f "$safe_pid"
+  then
+    log_error "Fatal error: Can't remove the mysqld_safe pid file"
+    exit 1
+  fi
+fi
+
+# Insert pid proerply into the pid file.
+ps -e | grep  [m]ysqld_safe | awk '{print $1}' | sed -n 1p > $safe_pid
+# End of mysqld_safe pid(safe_pid) check.
+
 # Determine what logging facility to use
 
 # Ensure that 'logger' exists, if it's requested
@@ -718,10 +722,24 @@ then
   if [ $? -ne 0 ]
   then
     log_error "--syslog requested, but no 'logger' program found.  Please ensure that 'logger' is in your PATH, or do not specify the --syslog option to mysqld_safe."
+    rm -f "$safe_pid"                 # Clean Up of mysqld_safe.pid file.
     exit 1
   fi
 fi
 
+if [ $want_syslog -eq 1 ]
+then
+  if [ -n "$syslog_tag" ]
+  then
+    # Sanitize the syslog tag
+    syslog_tag=`echo "$syslog_tag" | sed -e 's/[^a-zA-Z0-9_-]/_/g'`
+    syslog_tag_mysqld_safe="${syslog_tag_mysqld_safe}-$syslog_tag"
+    syslog_tag_mysqld="${syslog_tag_mysqld}-$syslog_tag"
+  fi
+  log_notice "Logging to syslog."
+  logging=syslog
+fi
+
 if [ -n "$err_log" -o $want_syslog -eq 0 ]
 then
   if [ -n "$err_log" ]
@@ -750,66 +768,20 @@ then
 
   append_arg_to_args "--log-error=$err_log"
 
+  # Log to err_log file
+  log_notice "Logging to '$err_log'."
   if [ $want_syslog -eq 1 ]
   then
-    # User explicitly asked for syslog, so warn that it isn't used
-    log_error "Can't log to error log and syslog at the same time.  Remove all --log-error configuration options for --syslog to take effect."
-  fi
-
-  # Log to err_log file
-  logging=file
-else
-  if [ -n "$syslog_tag" ]
-  then
-    # Sanitize the syslog tag
-    syslog_tag=`echo "$syslog_tag" | sed -e 's/[^a-zA-Z0-9_-]/_/g'`
-    syslog_tag_mysqld_safe="${syslog_tag_mysqld_safe}-$syslog_tag"
-    syslog_tag_mysqld="${syslog_tag_mysqld}-$syslog_tag"
-  fi
-  log_notice "Logging to syslog."
-  logging=syslog
-fi
-
-logdir=`dirname "$err_log"`
-# Change the err log to the right user, if possible and it is in use
-if [ $logging = "file" -o $logging = "both" ]; then
-  if [ ! -e "$err_log" -a ! -h "$err_log" ]; then
-    if test -w / -o "$USER" = "root"; then
-      case $logdir in
-        /var/log)
-          (
-            umask 0137
-            set -o noclobber
-            > "$err_log" && chown $user "$err_log"
-          ) ;;
-        *) ;;
-      esac
-    else
-      (
-        umask 0137
-        set -o noclobber
-        > "$err_log"
-      )
-    fi
-  fi
-
-  if [ -f "$err_log" -o -p "$err_log" ]; then        # Log to err_log file
-    log_notice "Logging to '$err_log'."
-  elif [ "x$user" = "xroot" ]; then # running as root, mysqld can create log file; continue
-    echo "Logging to '$err_log'." >&2
+    logging=both
   else
-    case $logdir in
-      # We can't create $err_log, however mysqld can; continue
-      /tmp|/var/tmp|/var/log/mysql|$DATADIR)
-        echo "Logging to '$err_log'." >&2
-        ;;
-      # We can't create $err_log and don't know if mysqld can; error out
-      *)
-        log_error "error: log-error set to '$err_log', however file does not exist. Create writable for user '$user'."
-        exit 1
-        ;;
-    esac
+    logging=file
   fi
+
+  if [ ! -f "$err_log" -a ! -h "$err_log" ]; then  # if error log already exists,
+    touch "$err_log"                               # we just append. otherwise,
+    chmod "$fmode" "$err_log"                      # fix the permissions here!
+  fi
+
 fi
 
 USER_OPTION=""
@@ -819,6 +791,11 @@ then
   then
     USER_OPTION="--user=$user"
   fi
+  # Change the err log to the right user, if it is in use
+  if [ $want_syslog -eq 0 -a ! -h "$err_log" ]; then
+    touch "$err_log"
+    chown $user "$err_log"
+  fi
   if test -n "$open_files"
   then
     ulimit -n $open_files
@@ -831,16 +808,14 @@ then
 fi
 
 safe_mysql_unix_port=${mysql_unix_port:-${MYSQL_UNIX_PORT:-@MYSQL_UNIX_ADDR@}}
-# Check that directory for $safe_mysql_unix_port exists
+# Make sure that directory for $safe_mysql_unix_port exists
 mysql_unix_port_dir=`dirname $safe_mysql_unix_port`
 if [ ! -d $mysql_unix_port_dir ]
 then
-  if [ ! -h $mysql_unix_port_dir ];
-  then
-    install -d -m 0755 -o $user $mysql_unix_port_dir
-  else
-    log_error "Directory '$mysql_unix_port_dir' for UNIX socket file does not exist."
-    exit 1
+  if [ ! -h $mysql_unix_port_dir ]; then
+    mkdir $mysql_unix_port_dir
+    chown $user $mysql_unix_port_dir
+    chmod 755 $mysql_unix_port_dir
   fi
 fi
 
@@ -857,6 +832,7 @@ does not exist or is not executable. Please cd to the mysql installation
 directory and restart this script from there as follows:
 ./bin/mysqld_safe&
 See http://dev.mysql.com/doc/mysql/en/mysqld-safe.html for more information"
+  rm -f "$safe_pid"                 # Clean Up of mysqld_safe.pid file.
   exit 1
 fi
 
@@ -955,18 +931,22 @@ then
     if @FIND_PROC@
     then    # The pid contains a mysqld process
       log_error "A mysqld process already exists"
+      rm -f "$safe_pid"                 # Clean Up of mysqld_safe.pid file.
       exit 1
     fi
   fi
+
   if [ ! -h "$pid_file" ]; then
-      rm -f "$pid_file"
+    rm -f "$pid_file"
   fi
+
   if test -f "$pid_file"
   then
     log_error "Fatal error: Can't remove the pid file:
 $pid_file
 Please remove it manually and start $0 again;
 mysqld daemon not started"
+    rm -f "$safe_pid"                 # Clean Up of mysqld_safe.pid file.
     exit 1
   fi
 fi
@@ -1017,10 +997,7 @@ fi
 # Change transparent huge pages setting if thp-setting option specified
 if [ -n "$thp_setting" ]
 then
-  if [ $(id -u) -ne 0 ]; then
-    log_error "mysqld_safe must be run as root for setting transparent huge pages!"
-    exit 1
-  elif [ $thp_setting != "always" -a $thp_setting != "madvise" -a $thp_setting != "never" ]; then
+  if [ $thp_setting != "always" -a $thp_setting != "madvise" -a $thp_setting != "never" ]; then
     log_error "Invalid value for thp-setting=$thp_setting in config file. Valid values are: always, madvise or never"
     exit 1
   else
@@ -1033,6 +1010,9 @@ then
     fi
     if [ $STATUS_THP -eq 0 ]; then
       log_notice "Transparent huge pages are already set to: ${thp_setting}."
+    elif [ $(id -u) -ne 0 ]; then
+      log_error "mysqld_safe must be run as root for setting transparent huge pages!"
+      exit 1
     else
       if [ -f /sys/kernel/mm/transparent_hugepage/defrag ]; then
         echo $thp_setting > /sys/kernel/mm/transparent_hugepage/defrag
@@ -1087,7 +1067,7 @@ log_notice "Starting $MYSQLD daemon with databases from $DATADIR"
 
 # variable to track the current number of "fast" (a.k.a. subsecond) restarts
 fast_restart=0
-# maximum number of restarts before throttling kicks in
+# maximum number of restarts before trottling kicks in
 max_fast_restarts=5
 # flag whether a usable sleep command exists
 have_sleep=1
@@ -1110,7 +1090,9 @@ do
   if [ ! -h "$pid_file" ]; then
     rm -f "$pid_file"
   fi
-
+  if [ ! -h "$pid_file.shutdown" ]; then
+    rm -f "$pid_file.shutdown"
+  fi
   start_time=`date +%M%S`
 
   # This file is checked for empty directory because 
@@ -1150,31 +1132,11 @@ do
     eval_log_error "$cmd $wsrep_start_position_opt --wsrep_cluster_address=$url $nohup_redir"
   fi
 
-  # hypothetical: log was renamed but not
-  # flushed yet. we'd recreate it with
-  # wrong owner next time we log, so set
-  # it up correctly while we can!
-
   if [ $want_syslog -eq 0 -a ! -f "$err_log" -a ! -h "$err_log" ]; then
-    if test -w / -o "$USER" = "root"; then
-      logdir=`dirname "$err_log"`
-      case $logdir in
-        /var/log)
-          (
-            umask 0137
-            set -o noclobber
-            > "$err_log" && chown $user "$err_log"
-          ) ;;
-        *) ;;
-      esac
-    else
-      (
-        umask 0137
-        set -o noclobber
-        > "$err_log"
-      )
-    fi
-  fi
+    touch "$err_log"                    # hypothetical: log was renamed but not
+    chown $user "$err_log"              # flushed yet. we'd recreate it with
+    chmod "$fmode" "$err_log"           # wrong owner next time we log, so set
+  fi                                    # it up correctly while we can!
 
   end_time=`date +%M%S`
 
@@ -1183,6 +1145,12 @@ do
     break
   fi
 
+  if test -f "$pid_file.shutdown"	# created to signal that it must stop
+  then
+    log_notice "$pid_file.shutdown present. The server will not restart."
+    break
+  fi
+
 
   # sanity check if time reading is sane and there's sleep
   if test $end_time -gt 0 -a $have_sleep -gt 0
@@ -1198,7 +1166,7 @@ do
         sleep_state=$?
         if test $sleep_state -gt 0
         then
-          log_notice "The server is respawning too fast and no working sleep command. Turning off throttling."
+          log_notice "The server is respawning too fast and no working sleep command. Turning off trottling."
           have_sleep=0
         fi
 
@@ -1255,5 +1223,9 @@ do
   log_notice "mysqld restarted"
 done
 
+rm -f "$pid_file.shutdown"
+
 log_notice "mysqld from pid file $pid_file ended"
 
+rm -f "$safe_pid"                       # Some Extra Safety. File is deleted
+                                        # once the mysqld process ends.
