@@ -1,4 +1,4 @@
-#!/usr/bin/env bash
+#!/bin/bash -ue
 
 # Copyright (C) 2010-2014 Codership Oy
 #
@@ -18,10 +18,10 @@
 
 # This is a reference script for rsync-based state snapshot tansfer
 
-set -o nounset -o errexit
-
 RSYNC_PID=
 RSYNC_CONF=
+keyring=""
+KEYRING_FILE=
 OS=$(uname)
 [ "$OS" == "Darwin" ] && export -n LD_LIBRARY_PATH
 
@@ -32,6 +32,11 @@ export PATH="/usr/sbin:/sbin:$PATH"
 
 wsrep_check_programs rsync
 
+keyring=$(parse_cnf mysqld keyring-file-data "")
+if [[ -z $keyring ]]; then
+    keyring=$(parse_cnf sst keyring-file-data "")
+fi
+
 cleanup_joiner()
 {
     local PID=$(cat "$RSYNC_PID" 2>/dev/null || echo 0)
@@ -41,6 +46,7 @@ cleanup_joiner()
     rm -rf "$RSYNC_CONF"
     rm -rf "$MAGIC_FILE"
     rm -rf "$RSYNC_PID"
+    rm -rf "$KEYRING_FILE"
     wsrep_log_info "Joiner cleanup done."
     if [ "${WSREP_SST_OPT_ROLE}" = "joiner" ];then
         wsrep_cleanup_progress_file
@@ -59,25 +65,10 @@ check_pid_and_port()
     local rsync_pid=$2
     local rsync_port=$3
 
-    case $OS in
-    FreeBSD)
-        local port_info="$(sockstat -46lp ${rsync_port} 2>/dev/null | \
-            grep ":${rsync_port}")"
-        local is_rsync="$(echo $port_info | \
-            grep -E '[[:space:]]+(rsync|stunnel)[[:space:]]+'"$rsync_pid" 2>/dev/null)"
-        ;;
-    *)
-        if ! which lsof > /dev/null; then
-            wsrep_log_error "lsof tool not found in PATH! Make sure you have it installed."
-            exit 2 # ENOENT
-        fi
-
-        local port_info="$(lsof -i :$rsync_port -Pn 2>/dev/null | \
-            grep "(LISTEN)")"
-        local is_rsync="$(echo $port_info | \
-            grep -E '^(rsync|stunnel)[[:space:]]+'"$rsync_pid" 2>/dev/null)"
-        ;;
-    esac
+    local port_info=$(lsof -i :$rsync_port -Pn 2>/dev/null | \
+        grep "(LISTEN)")
+    local is_rsync=$(echo $port_info | \
+        grep -w '^rsync[[:space:]]\+'"$rsync_pid" 2>/dev/null)
 
     if [ -n "$port_info" -a -z "$is_rsync" ]; then
         wsrep_log_error "rsync daemon port '$rsync_port' has been taken"
@@ -101,6 +92,10 @@ then
     BINLOG_FILENAME=$(basename $WSREP_SST_OPT_BINLOG)
 fi
 
+KEYRING_FILE="$WSREP_SST_OPT_DATA/keyring-sst"
+rm -rf "$KEYRING_FILE"
+
+
 WSREP_LOG_DIR=${WSREP_LOG_DIR:-""}
 # if WSREP_LOG_DIR env. variable is not set, try to get it from my.cnf
 if [ -z "$WSREP_LOG_DIR" ]; then
@@ -128,16 +123,8 @@ fi
 #         --exclude '*.[0-9][0-9][0-9][0-9][0-9][0-9]' --exclude '*.index')
 
 # New filter - exclude everything except dirs (schemas) and innodb files
-FILTER=(-f '- /lost+found'
-        -f '- /.fseventsd'
-        -f '- /.Trashes'
-        -f '+ /wsrep_sst_binlog.tar'
-        -f '+ /ib_lru_dump'
-        -f '+ /ibdata*'
-        -f '+ /undo*'
-        -f '+ /*.ibd'
-        -f '+ /*/'
-        -f '- /*')
+FILTER=(-f '- /lost+found' -f '- /.fseventsd' -f '- /.Trashes'
+        -f '+ /wsrep_sst_binlog.tar' -f '+ /ib_lru_dump' -f '+ /ibdata*' -f '+ /*/' -f '- /*')
 
 if [ "$WSREP_SST_OPT_ROLE" = "donor" ]
 then
@@ -219,6 +206,19 @@ then
             exit 255 # unknown error
         fi
 
+        # third, transfer the keyring file (this requires SSL, check for encryption)
+        if [[ -r $keyring ]]; then
+            rsync --owner --group --perms --links --specials \
+                  --ignore-times --inplace --quiet \
+                  $WHOLE_FILE_OPT "$keyring" \
+                  rsync://$WSREP_SST_OPT_ADDR/keyring-sst >&2 || RC=$?
+
+            if [ $RC -ne 0 ]; then
+                wsrep_log_error "rsync keyring returned code $RC:"
+                exit 255 # unknown error
+            fi
+        fi
+
         # then, we parallelize the transfer of database directories, use . so that pathconcatenation works
         pushd "$WSREP_SST_OPT_DATA" >/dev/null
 
@@ -226,8 +226,7 @@ then
         [ "$OS" == "Linux" ] && count=$(grep -c processor /proc/cpuinfo)
         [ "$OS" == "Darwin" -o "$OS" == "FreeBSD" ] && count=$(sysctl -n hw.ncpu)
 
-        find . -maxdepth 1 -mindepth 1 -type d -not -name "lost+found" \
-             -print0 | xargs -I{} -0 -P $count \
+        find . -maxdepth 1 -mindepth 1 -type d -print0 | xargs -I{} -0 -P $count \
              rsync --owner --group --perms --links --specials \
              --ignore-times --inplace --recursive --delete --quiet \
              $WHOLE_FILE_OPT --exclude '*/ib_logfile*' "$WSREP_SST_OPT_DATA"/{}/ \
@@ -254,6 +253,7 @@ then
 
 elif [ "$WSREP_SST_OPT_ROLE" = "joiner" ]
 then
+    wsrep_check_programs lsof
 
     touch $SST_PROGRESS_FILE
     MYSQLD_PID=$WSREP_SST_OPT_PARENT
@@ -269,24 +269,25 @@ then
     fi
     rm -rf "$RSYNC_PID"
 
+    ADDR=$WSREP_SST_OPT_ADDR
+    RSYNC_PORT=$(echo $ADDR | awk -F ':' '{ print $2 }')
+    if [ -z "$RSYNC_PORT" ]
+    then
+        RSYNC_PORT=4444
+        ADDR="$(echo $ADDR | awk -F ':' '{ print $1 }'):$RSYNC_PORT"
+    fi
+
     trap "exit 32" HUP PIPE
     trap "exit 3"  INT TERM ABRT
     trap cleanup_joiner EXIT
 
     RSYNC_CONF="$WSREP_SST_OPT_DATA/$MODULE.conf"
 
-    if [ -n "${MYSQL_TMP_DIR:-}" ] ; then
-        SILENT="log file = $MYSQL_TMP_DIR/rsyncd.log"
-    else
-        SILENT=""
-    fi
-
 cat << EOF > "$RSYNC_CONF"
 pid file = $RSYNC_PID
 use chroot = no
 read only = no
 timeout = 300
-$SILENT
 [$MODULE]
     path = $WSREP_SST_OPT_DATA
 [$MODULE-log_dir]
@@ -296,7 +297,6 @@ EOF
 #    rm -rf "$DATA"/ib_logfile* # we don't want old logs around
 
     # listen at all interfaces (for firewalled setups)
-    readonly RSYNC_PORT=${WSREP_SST_OPT_PORT:-4444}
     rsync --daemon --no-detach --port $RSYNC_PORT --config "$RSYNC_CONF" &
     RSYNC_REAL_PID=$!
 
@@ -305,7 +305,7 @@ EOF
         sleep 0.2
     done
 
-    echo "ready $WSREP_SST_OPT_HOST:$RSYNC_PORT/$MODULE"
+    echo "ready $ADDR/$MODULE"
 
     # wait for SST to complete by monitoring magic file
     while [ ! -r "$MAGIC_FILE" ] && check_pid "$RSYNC_PID" && \
@@ -338,6 +338,31 @@ EOF
         fi
         popd &> /dev/null
     fi
+
+    if [[ -n $keyring ]]; then
+        if [[ -r $KEYRING_FILE ]]; then
+            wsrep_log_info "Moving sst keyring into place: moving $KEYRING_FILE to $keyring"
+            mv $KEYRING_FILE $keyring
+        else
+            # error, missing file
+            wsrep_log_error "FATAL: rsync could not find '${KEYRING_FILE}'"
+            wsrep_log_error "The joiner is using a keyring file but the donor has not sent"
+            wsrep_log_error "a keyring file.  Please check your configuration to ensure that"
+            wsrep_log_error "both sides are using a keyring file"
+            exit 32
+        fi
+    else
+        if [[ -r $KEYRING_FILE ]]; then
+            # error, file should not be here
+            wsrep_log_error "FATAL: rsync found '${KEYRING_FILE}'"
+            wsrep_log_error "The joiner is not using a keyring file but the donor has sent"
+            wsrep_log_error "a keyring file.  Please check your configuration to ensure that"
+            wsrep_log_error "both sides are using a keyring file"
+            rm -rf $KEYRING_FILE
+            exit 32
+        fi
+    fi
+
     if [ -r "$MAGIC_FILE" ]
     then
         cat "$MAGIC_FILE" # output UUID:seqno
