@@ -1,21 +1,14 @@
 #ifndef BINLOG_H_INCLUDED
-/* Copyright (c) 2010, 2019, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2010, 2016, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
@@ -23,14 +16,68 @@
 
 #define BINLOG_H_INCLUDED
 
-#include "mysqld.h"                             /* opt_relay_logname */
-#include "log_event.h"
-#include "log.h"
+#include "sql_class.h"
+#include "my_global.h"
+#include "m_string.h"                  // llstr
+#include "binlog_event.h"              // enum_binlog_checksum_alg
+#include "mysqld.h"                    // opt_relay_logname
+#include "tc_log.h"                    // TC_LOG
+#include "atomic_class.h"
+#include "rpl_gtid.h"                  // Gtid_set, Sid_map
 
 class Relay_log_info;
 class Master_info;
-
+class Slave_worker;
 class Format_description_log_event;
+class Transaction_boundary_parser;
+class Rows_log_event;
+class Rows_query_log_event;
+class Incident_log_event;
+class Log_event;
+class Gtid_set;
+struct Gtid;
+
+typedef int64 query_id_t;
+
+/**
+  Logical timestamp generator for logical timestamping binlog transactions.
+  A transaction is associated with two sequence numbers see
+  @c Transaction_ctx::last_committed and @c Transaction_ctx::sequence_number.
+  The class provides necessary interfaces including that of
+  generating a next consecutive value for the latter.
+*/
+class  Logical_clock
+{
+private:
+  int64 state;
+  /*
+    Offset is subtracted from the actual "absolute time" value at
+    logging a replication event. That is the event holds logical
+    timestamps in the "relative" format. They are meaningful only in
+    the context of the current binlog.
+    The member is updated (incremented) per binary log rotation.
+  */
+  int64 offset;
+public:
+  Logical_clock();
+  int64 step();
+  int64 set_if_greater(int64 new_val);
+  int64 get_timestamp();
+  int64 get_offset() { return offset; }
+  /*
+    Updates the offset.
+    This operation is invoked when binlog rotates and at that time
+    there can't any concurrent step() callers so no need to guard
+    the assignement.
+  */
+  void update_offset(int64 new_offset)
+  {
+    DBUG_ASSERT(offset <= new_offset);
+
+    offset= new_offset;
+  }
+  ~Logical_clock() { }
+};
 
 /**
   Class for maintaining the commit stages for binary log group commit.
@@ -41,7 +88,7 @@ public:
     friend class Stage_manager;
   public:
     Mutex_queue()
-      : m_first(NULL), m_last(&m_first)
+      : m_first(NULL), m_last(&m_first), m_size(0)
     {
     }
 
@@ -73,6 +120,11 @@ public:
 
     std::pair<bool,THD*> pop_front();
 
+    inline int32 get_size()
+    {
+      return my_atomic_load32(&m_size);
+    }
+
   private:
     void lock() { mysql_mutex_lock(&m_lock); }
     void unlock() { mysql_mutex_unlock(&m_lock); }
@@ -91,6 +143,9 @@ public:
     */
     THD **m_last;
 
+    /** size of the queue */
+    int32 m_size;
+
     /** Lock for protecting the queue. */
     mysql_mutex_t m_lock;
   } MY_ATTRIBUTE((aligned(CPU_LEVEL1_DCACHE_LINESIZE)));
@@ -125,10 +180,10 @@ public:
             )
   {
     mysql_mutex_init(key_LOCK_done, &m_lock_done, MY_MUTEX_INIT_FAST);
-    mysql_cond_init(key_COND_done, &m_cond_done, NULL);
+    mysql_cond_init(key_COND_done, &m_cond_done);
 #ifndef DBUG_OFF
-    /* reuse key_COND_done 'cos a new PSI object would be wasteful in DBUG_ON */
-    mysql_cond_init(key_COND_done, &m_cond_preempt, NULL);
+    /* reuse key_COND_done 'cos a new PSI object would be wasteful in !DBUG_OFF */
+    mysql_cond_init(key_COND_done, &m_cond_preempt);
 #endif
     m_queue[FLUSH_STAGE].init(
 #ifdef HAVE_PSI_INTERFACE
@@ -206,13 +261,22 @@ public:
     return m_queue[stage].fetch_and_empty();
   }
 
-  void signal_done(THD *queue) {
-    mysql_mutex_lock(&m_lock_done);
-    for (THD *thd= queue ; thd ; thd = thd->next_to_commit)
-      thd->transaction.flags.pending= false;
-    mysql_mutex_unlock(&m_lock_done);
-    mysql_cond_broadcast(&m_cond_done);
-  }
+  /**
+    Introduces a wait operation on the executing thread.  The
+    waiting is done until the timeout elapses or count is
+    reached (whichever comes first).
+
+    If count == 0, then the session will wait until the timeout
+    elapses. If timeout == 0, then there is no waiting.
+
+    @param usec     the number of microseconds to wait.
+    @param count    wait for as many as count to join the queue the
+                    session is waiting on
+    @param stage    which stage queue size to compare count against.
+   */
+  time_t wait_count_or_timeout(ulong count, time_t usec, StageID stage);
+
+  void signal_done(THD *queue);
 
 private:
   /**
@@ -238,13 +302,68 @@ private:
 #endif
 };
 
+/* log info errors */
+#define LOG_INFO_EOF -1
+#define LOG_INFO_IO  -2
+#define LOG_INFO_INVALID -3
+#define LOG_INFO_SEEK -4
+#define LOG_INFO_MEM -6
+#define LOG_INFO_FATAL -7
+#define LOG_INFO_IN_USE -8
+#define LOG_INFO_EMFILE -9
 
-class MYSQL_BIN_LOG: public TC_LOG, private MYSQL_LOG
+/* bitmap to MYSQL_BIN_LOG::close() */
+#define LOG_CLOSE_INDEX		1
+#define LOG_CLOSE_TO_BE_OPENED	2
+#define LOG_CLOSE_STOP_EVENT	4
+
+
+/*
+  Note that we destroy the lock mutex in the destructor here.
+  This means that object instances cannot be destroyed/go out of scope
+  until we have reset thd->current_linfo to NULL;
+ */
+typedef struct st_log_info
 {
- private:
+  char log_file_name[FN_REFLEN];
+  my_off_t index_file_offset, index_file_start_offset;
+  my_off_t pos;
+  bool fatal; // if the purge happens to give us a negative offset
+  int entry_index; //used in purge_logs(), calculatd in find_log_pos().
+  st_log_info()
+    : index_file_offset(0), index_file_start_offset(0),
+      pos(0), fatal(0), entry_index(0)
+    {
+      memset(log_file_name, 0, FN_REFLEN);
+    }
+} LOG_INFO;
+
+/*
+  TODO use mmap instead of IO_CACHE for binlog
+  (mmap+fsync is two times faster than write+fsync)
+*/
+
+class MYSQL_BIN_LOG: public TC_LOG
+{
+  enum enum_log_state { LOG_OPENED, LOG_CLOSED, LOG_TO_BE_OPENED };
+
+  /* LOCK_log is inited by init_pthread_objects() */
+  mysql_mutex_t LOCK_log;
+  char *name;
+  char log_file_name[FN_REFLEN];
+  char db[NAME_LEN + 1];
+  bool write_error, inited;
+  IO_CACHE log_file;
+  const enum cache_type io_cache_type;
 #ifdef HAVE_PSI_INTERFACE
+  /** Instrumentation key to use for file io in @c log_file */
+  PSI_file_key m_log_file_key;
+  /** The instrumentation key to use for @ LOCK_log. */
+  PSI_mutex_key m_key_LOCK_log;
   /** The instrumentation key to use for @ LOCK_index. */
   PSI_mutex_key m_key_LOCK_index;
+  /** The instrumentation key to use for @ LOCK_binlog_end_pos. */
+  PSI_mutex_key m_key_LOCK_binlog_end_pos;
 
   PSI_mutex_key m_key_COND_done;
 
@@ -266,13 +385,20 @@ class MYSQL_BIN_LOG: public TC_LOG, private MYSQL_LOG
   PSI_file_key m_key_file_log;
   /** The instrumentation key to use for opening the log index file. */
   PSI_file_key m_key_file_log_index;
+  /** The instrumentation key to use for opening a log cache file. */
+  PSI_file_key m_key_file_log_cache;
+  /** The instrumentation key to use for opening a log index cache file. */
+  PSI_file_key m_key_file_log_index_cache;
 #endif
   /* POSIX thread objects are inited by init_pthread_objects() */
   mysql_mutex_t LOCK_index;
   mysql_mutex_t LOCK_commit;
   mysql_mutex_t LOCK_sync;
+  mysql_mutex_t LOCK_binlog_end_pos;
   mysql_mutex_t LOCK_xids;
   mysql_cond_t update_cond;
+
+  my_off_t binlog_end_pos;
   ulonglong bytes_written;
   IO_CACHE index_file;
   char index_file_name[FN_REFLEN];
@@ -314,53 +440,23 @@ class MYSQL_BIN_LOG: public TC_LOG, private MYSQL_LOG
   uint *sync_period_ptr;
   uint sync_counter;
 
-  my_atomic_rwlock_t m_prep_xids_lock;
   mysql_cond_t m_prep_xids_cond;
-  volatile int32 m_prep_xids;
+  Atomic_int32 m_prep_xids;
 
   /**
     Increment the prepared XID counter.
    */
-  void inc_prep_xids(THD *thd) {
-    DBUG_ENTER("MYSQL_BIN_LOG::inc_prep_xids");
-    my_atomic_rwlock_wrlock(&m_prep_xids_lock);
-#ifndef DBUG_OFF
-    int result= my_atomic_add32(&m_prep_xids, 1);
-#else
-    (void) my_atomic_add32(&m_prep_xids, 1);
-#endif
-    DBUG_PRINT("debug", ("m_prep_xids: %d", result + 1));
-    my_atomic_rwlock_wrunlock(&m_prep_xids_lock);
-    thd->transaction.flags.xid_written= true;
-    DBUG_VOID_RETURN;
-  }
+  void inc_prep_xids(THD *thd);
 
   /**
     Decrement the prepared XID counter.
 
     Signal m_prep_xids_cond if the counter reaches zero.
    */
-  void dec_prep_xids(THD *thd) {
-    DBUG_ENTER("MYSQL_BIN_LOG::dec_prep_xids");
-    my_atomic_rwlock_wrlock(&m_prep_xids_lock);
-    int32 result= my_atomic_add32(&m_prep_xids, -1);
-    DBUG_PRINT("debug", ("m_prep_xids: %d", result - 1));
-    my_atomic_rwlock_wrunlock(&m_prep_xids_lock);
-    thd->transaction.flags.xid_written= false;
-    /* If the old value was 1, it is zero now. */
-    if (result == 1)
-    {
-      mysql_mutex_lock(&LOCK_xids);
-      mysql_cond_signal(&m_prep_xids_cond);
-      mysql_mutex_unlock(&LOCK_xids);
-    }
-    DBUG_VOID_RETURN;
-  }
+  void dec_prep_xids(THD *thd);
 
   int32 get_prep_xids() {
-    my_atomic_rwlock_rdlock(&m_prep_xids_lock);
-    int32 result= my_atomic_load32(&m_prep_xids);
-    my_atomic_rwlock_rdunlock(&m_prep_xids_lock);
+    int32 result= m_prep_xids.atomic_get();
     return result;
   }
 
@@ -382,9 +478,18 @@ class MYSQL_BIN_LOG: public TC_LOG, private MYSQL_LOG
   Stage_manager stage_manager;
   void do_flush(THD *thd);
 
+  bool open(
+#ifdef HAVE_PSI_INTERFACE
+            PSI_file_key log_file_key,
+#endif
+            const char *log_name,
+            const char *new_name);
+  bool init_and_set_log_file_name(const char *log_name,
+                                  const char *new_name);
 public:
-  using MYSQL_LOG::generate_name;
-  using MYSQL_LOG::is_open;
+  const char *generate_name(const char *log_name, const char *suffix,
+                            char *buff);
+  bool is_open() { return log_state.atomic_get() != LOG_CLOSED; }
 
   /* This is relay log */
   bool is_relay_log;
@@ -423,9 +528,10 @@ public:
     (A)    - checksum algorithm descriptor value
     FD.(A) - the value of (A) in FD
   */
-  uint8 relay_log_checksum_alg;
+  binary_log::enum_binlog_checksum_alg relay_log_checksum_alg;
 
-  MYSQL_BIN_LOG(uint *sync_period);
+  MYSQL_BIN_LOG(uint *sync_period,
+                enum cache_type io_cache_type_arg);
   /*
     note that there's no destructor ~MYSQL_BIN_LOG() !
     The reason is that we don't want it to be automatically called
@@ -439,6 +545,7 @@ public:
                     PSI_mutex_key key_LOCK_done,
                     PSI_mutex_key key_LOCK_flush_queue,
                     PSI_mutex_key key_LOCK_log,
+                    PSI_mutex_key key_LOCK_binlog_end_pos,
                     PSI_mutex_key key_LOCK_sync,
                     PSI_mutex_key key_LOCK_sync_queue,
                     PSI_mutex_key key_LOCK_xids,
@@ -446,7 +553,9 @@ public:
                     PSI_cond_key key_update_cond,
                     PSI_cond_key key_prep_xids_cond,
                     PSI_file_key key_file_log,
-                    PSI_file_key key_file_log_index)
+                    PSI_file_key key_file_log_index,
+                    PSI_file_key key_file_log_cache,
+                    PSI_file_key key_file_log_index_cache)
   {
     m_key_COND_done= key_COND_done;
 
@@ -457,6 +566,7 @@ public:
 
     m_key_LOCK_index= key_LOCK_index;
     m_key_LOCK_log= key_LOCK_log;
+    m_key_LOCK_binlog_end_pos= key_LOCK_binlog_end_pos;
     m_key_LOCK_commit= key_LOCK_commit;
     m_key_LOCK_sync= key_LOCK_sync;
     m_key_LOCK_xids= key_LOCK_xids;
@@ -464,8 +574,18 @@ public:
     m_key_prep_xids_cond= key_prep_xids_cond;
     m_key_file_log= key_file_log;
     m_key_file_log_index= key_file_log_index;
+    m_key_file_log_cache= key_file_log_cache;
+    m_key_file_log_index_cache= key_file_log_index_cache;
   }
 #endif
+
+public:
+  /* Committed transactions timestamp */
+   Logical_clock max_committed_transaction;
+  /* "Prepared" transactions timestamp */
+   Logical_clock transaction_counter;
+  void update_max_committed(THD *thd);
+
   /**
     Find the oldest binary log that contains any GTID that
     is not in the given gtid set.
@@ -484,33 +604,82 @@ public:
                                       const char **errmsg);
 
   /**
-    Reads the set of all GTIDs in the binary log, and the set of all
-    lost GTIDs in the binary log, and stores each set in respective
-    argument.
+    Reads the set of all GTIDs in the binary/relay log, and the set
+    of all lost GTIDs in the binary log, and stores each set in
+    respective argument.
 
-    @param gtid_set Will be filled with all GTIDs in this binary log.
+    @param gtid_set Will be filled with all GTIDs in this binary/relay
+    log.
     @param lost_groups Will be filled with all GTIDs in the
     Previous_gtids_log_event of the first binary log that has a
-    Previous_gtids_log_event.
-    @param last_gtid Will be filled with the last availble GTID information
-    in the binary/relay log files.
+    Previous_gtids_log_event. This is requested to binary logs but not
+    to relay logs.
     @param verify_checksum If true, checksums will be checked.
     @param need_lock If true, LOCK_log, LOCK_index, and
     global_sid_lock->wrlock are acquired; otherwise they are asserted
     to be taken already.
+    @param trx_parser [out] This will be used to return the actual
+    relaylog transaction parser state because of the possibility
+    of partial transactions.
+    @param [out] gtid_partial_trx If a transaction was left incomplete
+    on the relaylog, it's GTID should be returned to be used in the
+    case of the rest of the transaction be added to the relaylog.
     @param is_server_starting True if the server is starting.
     @return false on success, true on error.
   */
   bool init_gtid_sets(Gtid_set *gtid_set, Gtid_set *lost_groups,
-                      Gtid *last_gtid, bool verify_checksum,
-                      bool need_lock, bool is_server_starting= false);
+                      bool verify_checksum,
+                      bool need_lock,
+                      Transaction_boundary_parser *trx_parser,
+                      Gtid *gtid_partial_trx,
+                      bool is_server_starting= false);
 
-  void set_previous_gtid_set(Gtid_set *previous_gtid_set_param)
+  void set_previous_gtid_set_relaylog(Gtid_set *previous_gtid_set_param)
   {
-    previous_gtid_set= previous_gtid_set_param;
+    DBUG_ASSERT(is_relay_log);
+    previous_gtid_set_relaylog= previous_gtid_set_param;
   }
+  /**
+    If the thread owns a GTID, this function generates an empty
+    transaction and releases ownership of the GTID.
+
+    - If the binary log is disabled for this thread, the GTID is
+      inserted directly into the mysql.gtid_executed table and the
+      GTID is included in @@global.gtid_executed.  (This only happens
+      for DDL, since DML will save the GTID into table and release
+      ownership inside ha_commit_trans.)
+
+    - If the binary log is enabled for this thread, an empty
+      transaction consisting of GTID, BEGIN, COMMIT is written to the
+      binary log, the GTID is included in @@global.gtid_executed, and
+      the GTID is added to the mysql.gtid_executed table on the next
+      binlog rotation.
+
+    This function must be called by any committing statement (COMMIT,
+    implicitly committing statements, or Xid_log_event), after the
+    statement has completed execution, regardless of whether the
+    statement updated the database.
+
+    This logic ensures that an empty transaction is generated for the
+    following cases:
+
+    - Explicit empty transaction:
+      SET GTID_NEXT = 'UUID:NUMBER'; BEGIN; COMMIT;
+
+    - Transaction or DDL that gets completely filtered out in the
+      slave thread.
+
+    @param thd The committing thread
+
+    @retval 0 Success
+    @retval nonzero Error
+  */
+  int gtid_end_transaction(THD *thd);
 private:
-  Gtid_set* previous_gtid_set;
+  Atomic_int32 log_state; /* atomic enum_log_state */
+
+  /* The previous gtid set in relay log. */
+  Gtid_set* previous_gtid_set_relaylog;
 
   bool snapshot_lock_acquired;
 
@@ -560,11 +729,56 @@ public:
   {
     bytes_written = 0;
   }
-  void harvest_bytes_written(Relay_log_info *rli, bool need_log_space_lock);
+  void harvest_bytes_written(ulonglong* counter)
+  {
+#ifndef DBUG_OFF
+    char buf1[22],buf2[22];
+#endif
+    DBUG_ENTER("harvest_bytes_written");
+    (*counter)+=bytes_written;
+    DBUG_PRINT("info",("counter: %s  bytes_written: %s", llstr(*counter,buf1),
+		       llstr(bytes_written,buf2)));
+    bytes_written=0;
+    DBUG_VOID_RETURN;
+  }
   void set_max_size(ulong max_size_arg);
-  void signal_update();
+  void signal_update()
+  {
+    DBUG_ENTER("MYSQL_BIN_LOG::signal_update");
+    signal_cnt++;
+    mysql_cond_broadcast(&update_cond);
+    DBUG_VOID_RETURN;
+  }
+
+  void update_binlog_end_pos()
+  {
+    /*
+      binlog_end_pos is used only on master's binlog right now. It is possible
+      to use it on relay log.
+    */
+    if (is_relay_log)
+      signal_update();
+    else
+    {
+      lock_binlog_end_pos();
+      binlog_end_pos= my_b_tell(&log_file);
+      signal_update();
+      unlock_binlog_end_pos();
+    }
+  }
+
+  void update_binlog_end_pos(my_off_t pos)
+  {
+    lock_binlog_end_pos();
+    if (pos > binlog_end_pos)
+      binlog_end_pos= pos;
+    signal_update();
+    unlock_binlog_end_pos();
+  }
+
   int wait_for_update_relay_log(THD* thd, const struct timespec * timeout);
-  int  wait_for_update_bin_log(THD* thd, const struct timespec * timeout);
+  int wait_for_update_bin_log(THD* thd, const struct timespec * timeout);
+  bool do_write_cache(IO_CACHE *cache, class Binlog_event_writer *writer);
 public:
   void init_pthread_objects();
   void cleanup();
@@ -573,8 +787,6 @@ public:
     @param log_name Name of binlog
     @param new_name Name of binlog, too. todo: what's the difference
     between new_name and log_name?
-    @param io_cache_type_arg Specifies how the IO cache is opened:
-    read-only or read-write.
     @param max_size The size at which this binlog will be rotated.
     @param null_created If false, and a Format_description_log_event
     is written, then the Format_description_log_event will have the
@@ -588,10 +800,8 @@ public:
   */
   bool open_binlog(const char *log_name,
                    const char *new_name,
-                   enum cache_type io_cache_type_arg,
                    ulong max_size,
                    bool null_created,
-                   bool need_lock_log,
                    bool need_lock_index, bool need_sid_lock,
                    Format_description_log_event *extra_description_event);
   bool open_index_file(const char *index_file_name_arg,
@@ -600,14 +810,34 @@ public:
   int new_file(Format_description_log_event *extra_description_event);
 
   bool write_event(Log_event* event_info);
-  bool write_cache(THD *thd, class binlog_cache_data *binlog_cache_data);
-  int  do_write_cache(THD *thd, IO_CACHE *cache);
+  bool write_cache(THD *thd, class binlog_cache_data *binlog_cache_data,
+                   class Binlog_event_writer *writer);
+  bool write_gtid(THD *thd, binlog_cache_data *cache_data,
+                  class Binlog_event_writer *writer);
+
+  /**
+     Write a dml into statement cache and then flush it into binlog. It writes
+     Gtid_log_event and BEGIN, COMMIT automatically.
+
+     It is aimed to handle cases of "background" logging where a statement is
+     logged indirectly, like "DELETE FROM a_memory_table". So don't use it on any
+     normal statement.
+
+     @param[IN] thd  the THD object of current thread.
+     @param[IN] stmt the DELETE statement.
+     @param[IN] stmt_len the length of DELETE statement.
+
+     @return Returns false if succeeds, otherwise true is returned.
+  */
+  bool write_dml_directly(THD* thd, const char *stmt, size_t stmt_len);
 
   void set_write_error(THD *thd, bool is_transactional);
   bool check_write_error(THD *thd);
   bool write_incident(THD *thd, bool need_lock_log,
+                      const char* err_msg,
                       bool do_flush_and_sync= true);
   bool write_incident(Incident_log_event *ev, bool need_lock_log,
+                      const char* err_msg,
                       bool do_flush_and_sync= true);
 
   void start_union_events(THD *thd, query_id_t query_id_param);
@@ -652,7 +882,7 @@ public:
   int set_crash_safe_index_file_name(const char *base_file_name);
   int open_crash_safe_index_file();
   int close_crash_safe_index_file();
-  int add_log_to_index(uchar* log_file_name, int name_len,
+  int add_log_to_index(uchar* log_file_name, size_t name_len,
                        bool need_lock_index);
   int move_crash_safe_index_file_to_index_file(bool need_lock_index);
   int set_purge_index_file_name(const char *base_file_name);
@@ -665,13 +895,14 @@ public:
   int register_create_index_entry(const char* entry);
   int purge_index_entry(THD *thd, ulonglong *decrease_log_space,
                         bool need_lock_index);
-  bool reset_logs(THD* thd);
-  void close(uint exiting, bool need_lock_log, bool need_lock_index);
+  bool reset_logs(THD* thd, bool delete_only= false);
+  void close(uint exiting);
 
   // iterating through the log index file
   int find_log_pos(LOG_INFO* linfo, const char* log_name,
                    bool need_lock_index);
   int find_next_log(LOG_INFO* linfo, bool need_lock_index);
+  int find_next_relay_log(char log_name[FN_REFLEN+1]);
   int get_current_log(LOG_INFO* linfo, bool need_lock_log= true);
   int raw_get_current_log(LOG_INFO* linfo);
   uint next_file_id();
@@ -687,57 +918,37 @@ public:
   inline IO_CACHE *get_index_file() { return &index_file;}
   inline uint32 get_open_count() { return open_count; }
 
-  /**
-    Function to report the missing GTIDs.
-
-    This function logs the missing transactions on master to its error log
-    as a warning. If the missing GTIDs are too long to print in a message,
-    it suggests the steps to extract the missing transactions.
-
-    This function also informs slave about the GTID set sent by the slave,
-    transactions missing on the master and few suggestions to recover from
-    the error. This message shall be wrapped by
-    ER_MASTER_FATAL_ERROR_READING_BINLOG on slave and will be logged as an
-    error.
-
-    This function will be called from mysql_binlog_send() function.
-
-    @param slave_executed_gtid_set     GTID set executed by slave
-    @param errmsg                      Pointer to the error message
-
-    @return void
+  /*
+    It is called by the threads(e.g. dump thread) which want to read
+    hot log without LOCK_log protection.
   */
-  void report_missing_purged_gtids(const Gtid_set* slave_executed_gtid_set,
-                                   const char** errmsg);
+  my_off_t get_binlog_end_pos() const
+  {
+    mysql_mutex_assert_not_owner(&LOCK_log);
+    mysql_mutex_assert_owner(&LOCK_binlog_end_pos);
+    return binlog_end_pos;
+  }
+  mysql_mutex_t* get_binlog_end_pos_lock() { return &LOCK_binlog_end_pos; }
+  void lock_binlog_end_pos() { mysql_mutex_lock(&LOCK_binlog_end_pos); }
+  void unlock_binlog_end_pos() { mysql_mutex_unlock(&LOCK_binlog_end_pos); }
+  void set_status_variables(THD *thd);
 
   /**
-    Function to report the missing GTIDs.
+    Deep copy global_sid_map to @param sid_map and
+    gtid_state->get_executed_gtids() to @param gtid_set
+    Both operations are done under LOCK_commit and global_sid_lock
+    protection.
 
-    This function logs the missing transactions on master to its error log
-    as a warning. If the missing GTIDs are too long to print in a message,
-    it suggests the steps to extract the missing transactions.
+    @param[out] sid_map  The Sid_map to which global_sid_map will
+                         be copied.
+    @param[out] gtid_set The Gtid_set to which gtid_executed will
+                         be copied.
 
-    This function also informs slave about the GTID set sent by the slave,
-    transactions missing on the master and few suggestions to recover from
-    the error. This message shall be wrapped by
-    ER_MASTER_FATAL_ERROR_READING_BINLOG on slave and will be logged as an
-    error.
-
-    This function will be called from find_first_log_not_in_gtid_set()
-    function.
-
-    @param previous_gtid_set           Previous GTID set found
-    @param slave_executed_gtid_set     GTID set executed by slave
-    @param errmsg                      Pointer to the error message
-
-    @return void
+    @return the operation status
+      @retval 0      OK
+      @retval !=0    Error
   */
-  void report_missing_gtids(const Gtid_set* previous_gtid_set,
-                            const Gtid_set* slave_executed_gtid_set,
-                            const char** errmsg);
-  static const int MAX_RETRIES_FOR_DELETE_RENAME_FAILURE = 5;
-private:
-  void publish_coordinates_for_global_status(void) const;
+  int get_gtid_executed(Sid_map *sid_map, Gtid_set *gtid_set);
 };
 
 typedef struct st_load_file_info
@@ -772,15 +983,27 @@ void check_binlog_cache_size(THD *thd);
 void check_binlog_stmt_cache_size(THD *thd);
 bool binlog_enabled();
 void register_binlog_handler(THD *thd, bool trx);
-int gtid_empty_group_log_and_cleanup(THD *thd);
+int query_error_code(THD *thd, bool not_killed);
+
+bool generate_new_log_name(char *new_name, ulong *new_ext,
+                           const char *log_name, bool is_binlog);
 
 extern const char *log_bin_index;
 extern const char *log_bin_basename;
 extern bool opt_binlog_order_commits;
 
+/*
+  Maximum unique log filename extension.
+  Note: setting to 0x7FFFFFFF due to atol windows
+  overflow/truncate.
+*/
+#define MAX_LOG_UNIQUE_FN_EXT 0x7FFFFFFF
+
 /**
   Turns a relative log binary log path into a full path, based on the
-  opt_bin_logname or opt_relay_logname.
+  opt_bin_logname or opt_relay_logname. Also trims the cr-lf at the
+  end of the full_path before return to avoid any server startup
+  problem on windows.
 
   @param from         The log name we want to make into an absolute path.
   @param to           The buffer where to put the results of the 
@@ -829,11 +1052,42 @@ inline bool normalize_binlog_name(char *to, const char *from, bool is_relay_log)
   }
 
   DBUG_ASSERT(ptr);
-
   if (ptr)
-    strmake(to, ptr, strlen(ptr));
+  {
+    size_t length= strlen(ptr);
 
+    // Strips the CR+LF at the end of log name and \0-terminates it.
+    if (length && ptr[length-1] == '\n')
+    {
+      ptr[length-1]= 0;
+      length--;
+      if (length && ptr[length-1] == '\r')
+      {
+        ptr[length-1]= 0;
+        length--;
+      }
+    }
+    if (!length)
+    {
+      error= true;
+      goto end;
+    }
+    strmake(to, ptr, length);
+  }
 end:
   DBUG_RETURN(error);
 }
+
+
+#ifdef WITH_WSREP
+
+IO_CACHE* wsrep_get_trans_log(THD * thd);
+bool wsrep_trans_cache_is_empty(THD *thd);
+void wsrep_thd_binlog_flush_pending_rows_event(THD *thd, bool stmt_end);
+void wsrep_thd_binlog_trx_reset(THD * thd);
+TC_LOG::enum_result wsrep_thd_binlog_commit(THD* thd, bool all);
+int wsrep_thd_binlog_rollback(THD * thd, bool all);
+
+#endif /* WITH_WSREP */
+
 #endif /* BINLOG_H_INCLUDED */
