@@ -1,20 +1,13 @@
-/* Copyright (c) 2000, 2015, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2000, 2016, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
@@ -80,13 +73,12 @@
   we are forced to use mysql_lock_merge.
 */
 
-#include "sql_priv.h"
 #include "debug_sync.h"
-#include "unireg.h"                    // REQUIRED: for other includes
 #include "lock.h"
 #include "sql_base.h"                       // close_tables_for_reopen
 #include "sql_parse.h"                     // is_log_table_write_query
-#include "sql_acl.h"                       // SUPER_ACL
+#include "auth_common.h"                   // SUPER_ACL
+#include "session_tracker.h"
 #include <hash.h>
 #include <assert.h>
 #include "my_atomic.h"
@@ -104,7 +96,7 @@ extern HASH open_cache;
 #define GET_LOCK_UNLOCK         1
 #define GET_LOCK_STORE_LOCKS    2
 
-static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
+static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, size_t count,
                                  uint flags);
 static int lock_external(THD *thd, TABLE **table,uint count);
 static int unlock_external(THD *thd, TABLE **table,uint count);
@@ -123,10 +115,9 @@ static int thr_lock_errno_to_mysql[]=
   @return 0 if all the check passed, non zero if a check failed.
 */
 static int
-lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
+lock_tables_check(THD *thd, TABLE **tables, size_t count, uint flags)
 {
   uint system_count= 0, i= 0;
-  bool enforce_ro= true;
   /*
     Identifies if the executed sql command can updated either a log
     or rpl info table.
@@ -135,8 +126,6 @@ lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
 
   DBUG_ENTER("lock_tables_check");
 
-  if (!opt_super_readonly)
-    enforce_ro= !(thd->security_ctx->master_access & SUPER_ACL);
   log_table_write_query=
      is_log_table_write_query(thd->lex->sql_command);
 
@@ -188,18 +177,12 @@ lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
       write we must own metadata lock of MDL_SHARED_WRITE or stronger
       type. For table to be locked for read we must own metadata lock
       of MDL_SHARED_READ or stronger type).
-      The only exception are HANDLER statements which are allowed to
-      lock table for read while having only MDL_SHARED lock on it.
     */
     DBUG_ASSERT(t->s->tmp_table ||
-                thd->mdl_context.is_lock_owner(MDL_key::TABLE,
-                                 t->s->db.str, t->s->table_name.str,
-                                 t->reginfo.lock_type >= TL_WRITE_ALLOW_WRITE ?
-                                 MDL_SHARED_WRITE : MDL_SHARED_READ) ||
-                (t->open_by_handler &&
-                 thd->mdl_context.is_lock_owner(MDL_key::TABLE,
-                                  t->s->db.str, t->s->table_name.str,
-                                  MDL_SHARED)));
+                thd->mdl_context.owns_equal_or_stronger_lock(MDL_key::TABLE,
+                                   t->s->db.str, t->s->table_name.str,
+                                   t->reginfo.lock_type >= TL_WRITE_ALLOW_WRITE ?
+                                   MDL_SHARED_WRITE : MDL_SHARED_READ));
 
     /*
       Prevent modifications to base tables if READ_ONLY is activated.
@@ -208,18 +191,8 @@ lock_tables_check(THD *thd, TABLE **tables, uint count, uint flags)
     if (!(flags & MYSQL_LOCK_IGNORE_GLOBAL_READ_ONLY) && !t->s->tmp_table)
     {
       if (t->reginfo.lock_type >= TL_WRITE_ALLOW_WRITE &&
-#ifdef WITH_WSREP
-          enforce_ro && opt_readonly && !thd->slave_thread
-          && !thd->wsrep_applier
-         )
-#else
-          enforce_ro && opt_readonly && !thd->slave_thread)
-#endif /* WITH_WSREP */
-      {
-        my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0),
-                 opt_super_readonly ? "--read-only (super)" : "--read-only");
-        DBUG_RETURN(1);
-      }
+        check_readonly(thd, true))
+          DBUG_RETURN(1);
     }
   }
 
@@ -275,6 +248,35 @@ static void reset_lock_data(MYSQL_LOCK *sql_lock)
 }
 
 
+/**
+  Scan array of tables for access types; update transaction tracker
+  accordingly.
+
+   @param thd          The current thread.
+   @param tables       An array of pointers to the tables to lock.
+   @param count        The number of tables to lock.
+*/
+
+static void track_table_access(THD *thd, TABLE **tables, size_t count)
+{
+  Transaction_state_tracker *tst= (Transaction_state_tracker *)
+    thd->session_tracker.get_tracker(TRANSACTION_INFO_TRACKER);
+  enum enum_tx_state         s;
+
+  while (count--)
+  {
+    TABLE *t= tables[count];
+
+    if (t)
+    {
+      s= tst->calc_trx_state(thd, t->reginfo.lock_type,
+                             t->file->has_transactions());
+      tst->add_trx_state(thd, s);
+    }
+  }
+}
+
+
 /**
   Reset lock type in lock data and free.
 
@@ -304,7 +306,7 @@ static void reset_lock_data_and_free(MYSQL_LOCK **mysql_lock)
    @retval  NULL if an error or if wait on a lock was killed.
 */
 
-MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, uint count, uint flags)
+MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, size_t count, uint flags)
 {
   int rc;
   MYSQL_LOCK *sql_lock;
@@ -335,12 +337,16 @@ MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, uint count, uint flags)
 #ifdef WITH_WSREP
   //thd->main_lock_id.info->in_lock_tables= thd->in_lock_tables;
     thd->lock_info.in_lock_tables= thd->in_lock_tables;
-#endif    /* Lock on the copied half of the lock data array. */
+#endif /* WITH_WSREP */
   /* Lock on the copied half of the lock data array. */
   rc= thr_lock_errno_to_mysql[(int) thr_multi_lock(sql_lock->locks +
                                                    sql_lock->lock_count,
                                                    sql_lock->lock_count,
                                                    &thd->lock_info, timeout)];
+
+  DBUG_EXECUTE_IF("mysql_lock_tables_kill_query",
+                  thd->killed= THD::KILL_QUERY;);
+
   if (rc)
   {
     if (sql_lock->table_count)
@@ -349,6 +355,7 @@ MYSQL_LOCK *mysql_lock_tables(THD *thd, TABLE **tables, uint count, uint flags)
     if (! thd->killed)
       my_error(rc, MYF(0));
   }
+
 end:
   if (!(flags & MYSQL_OPEN_IGNORE_KILLED) && thd->killed)
   {
@@ -360,6 +367,9 @@ end:
     }
   }
 
+  if (thd->variables.session_track_transaction_info > TX_TRACK_NONE)
+    track_table_access(thd, tables, count);
+
   thd->set_time_after_lock();
   DBUG_RETURN(sql_lock);
 }
@@ -523,16 +533,16 @@ void mysql_lock_remove(THD *thd, MYSQL_LOCK *locked,TABLE *table)
         removed_locks= table->lock_count;
 
         /* Move down all table pointers above 'i'. */
-	bmove((char*) (locked->table+i),
-	      (char*) (locked->table+i+1),
-	      (old_tables - i) * sizeof(TABLE*));
+	memmove(reinterpret_cast<char*> (locked->table + i),
+                reinterpret_cast<char*> (locked->table + i + 1),
+                (old_tables - i) * sizeof(TABLE*));
 
         lock_data_end= table->lock_data_start + table->lock_count;
         /* Move down all lock data pointers above 'table->lock_data_end-1' */
-        bmove((char*) (locked->locks + table->lock_data_start),
-              (char*) (locked->locks + lock_data_end),
-              (locked->lock_count - lock_data_end) *
-              sizeof(THR_LOCK_DATA*));
+        memmove(reinterpret_cast<char*> (locked->locks + table->lock_data_start),
+                reinterpret_cast<char*> (locked->locks + lock_data_end),
+                (locked->lock_count - lock_data_end) *
+                sizeof(THR_LOCK_DATA*));
 
         /*
           Fix moved table elements.
@@ -581,30 +591,23 @@ void mysql_lock_abort(THD *thd, TABLE *table, bool upgrade_lock)
 
   @param thd	   Thread handler
   @param table	   Table that should be removed from lock queue
-
-  @retval
-    0  Table was not locked by another thread
-  @retval
-    1  Table was locked by at least one other thread
 */
 
-bool mysql_lock_abort_for_thread(THD *thd, TABLE *table)
+void mysql_lock_abort_for_thread(THD *thd, TABLE *table)
 {
   MYSQL_LOCK *locked;
-  bool result= FALSE;
   DBUG_ENTER("mysql_lock_abort_for_thread");
 
   if ((locked= get_lock_data(thd, &table, 1, GET_LOCK_UNLOCK)))
   {
     for (uint i=0; i < locked->lock_count; i++)
     {
-      if (thr_abort_locks_for_thread(locked->locks[i]->lock,
-                                     table->in_use->thread_id))
-        result= TRUE;
+      thr_abort_locks_for_thread(locked->locks[i]->lock,
+                                 table->in_use->thread_id());
     }
     my_free(locked);
   }
-  DBUG_RETURN(result);
+  DBUG_VOID_RETURN;
 }
 
 
@@ -615,7 +618,8 @@ MYSQL_LOCK *mysql_lock_merge(MYSQL_LOCK *a,MYSQL_LOCK *b)
   DBUG_ENTER("mysql_lock_merge");
 
   if (!(sql_lock= (MYSQL_LOCK*)
-	my_malloc(sizeof(*sql_lock)+
+	my_malloc(key_memory_MYSQL_LOCK,
+                  sizeof(*sql_lock)+
 		  sizeof(THR_LOCK_DATA*)*(a->lock_count+b->lock_count)+
 		  sizeof(TABLE*)*(a->table_count+b->table_count),MYF(MY_WME))))
     DBUG_RETURN(0);				// Fatal error
@@ -687,7 +691,7 @@ static int unlock_external(THD *thd, TABLE **table,uint count)
            - GET_LOCK_STORE_LOCKS : Store lock info in TABLE
 */
 
-static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
+static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, size_t count,
                                  uint flags)
 {
   uint i,tables,lock_count;
@@ -697,7 +701,7 @@ static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
   DBUG_ENTER("get_lock_data");
 
   DBUG_ASSERT((flags == GET_LOCK_UNLOCK) || (flags == GET_LOCK_STORE_LOCKS));
-  DBUG_PRINT("info", ("count %d", count));
+  DBUG_PRINT("info", ("count %zu", count));
 
   for (i=tables=lock_count=0 ; i < count ; i++)
   {
@@ -717,7 +721,8 @@ static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
     from the first part immediately before calling thr_multi_lock().
   */
   if (!(sql_lock= (MYSQL_LOCK*)
-	my_malloc(sizeof(*sql_lock) +
+	my_malloc(key_memory_MYSQL_LOCK,
+                  sizeof(*sql_lock) +
 		  sizeof(THR_LOCK_DATA*) * tables * 2 +
                   sizeof(table_ptr) * lock_count,
 		  MYF(0))))
@@ -735,7 +740,8 @@ static MYSQL_LOCK *get_lock_data(THD *thd, TABLE **table_ptr, uint count,
     if ((table=table_ptr[i])->s->tmp_table == NON_TRANSACTIONAL_TMP_TABLE)
       continue;
     lock_type= table->reginfo.lock_type;
-    DBUG_ASSERT(lock_type != TL_WRITE_DEFAULT && lock_type != TL_READ_DEFAULT);
+    DBUG_ASSERT(lock_type != TL_WRITE_DEFAULT && lock_type != TL_READ_DEFAULT &&
+                lock_type != TL_WRITE_CONCURRENT_DEFAULT);
     locks_start= locks;
     locks= table->file->store_lock(thd, locks,
                                    (flags & GET_LOCK_UNLOCK) ? TL_IGNORE :
@@ -803,22 +809,23 @@ bool lock_schema_name(THD *thd, const char *db)
 
   if (thd->locked_tables_mode)
   {
-    my_message(ER_LOCK_OR_ACTIVE_TRANSACTION,
-               ER(ER_LOCK_OR_ACTIVE_TRANSACTION), MYF(0));
+    my_error(ER_LOCK_OR_ACTIVE_TRANSACTION, MYF(0));
     return TRUE;
   }
 
   if (thd->global_read_lock.can_acquire_protection())
     return TRUE;
-  global_request.init(MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
-                      MDL_STATEMENT);
+  MDL_REQUEST_INIT(&global_request,
+                   MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
+                   MDL_STATEMENT);
 
   if (thd->backup_tables_lock.abort_if_acquired())
-    return true;
+      return true;
   thd->backup_tables_lock.init_protection_request(&backup_request,
                                                   MDL_STATEMENT);
 
-  mdl_request.init(MDL_key::SCHEMA, db, "", MDL_EXCLUSIVE, MDL_TRANSACTION);
+  MDL_REQUEST_INIT(&mdl_request,
+                   MDL_key::SCHEMA, db, "", MDL_EXCLUSIVE, MDL_TRANSACTION);
 
   mdl_requests.push_front(&mdl_request);
   mdl_requests.push_front(&backup_request);
@@ -833,6 +840,113 @@ bool lock_schema_name(THD *thd, const char *db)
 }
 
 
+/**
+  Obtain an exclusive metadata lock on a tablespace name.
+
+  @param thd         Thread handle.
+  @param tablespace  The tablespace name.
+
+  This function cannot be called while holding the LOCK_open mutex.
+  To avoid deadlocks, we do not try to obtain exclusive metadata
+  locks in LOCK TABLES mode, since in this mode there may be
+  other metadata locks already taken by the current connection,
+  and we must not wait for MDL locks while holding locks.
+
+  @retval false  Success.
+  @retval true   Failure: we're in LOCK TABLES mode, out of memory,
+                 connection was killed, or numerous other reasons.
+*/
+
+bool lock_tablespace_name(THD *thd, const char *tablespace)
+{
+  MDL_request_list mdl_requests;
+  MDL_request global_request;
+  MDL_request mdl_request;
+
+  if (thd->locked_tables_mode)
+  {
+    my_error(ER_LOCK_OR_ACTIVE_TRANSACTION, MYF(0));
+    return true;
+  }
+
+  if (thd->global_read_lock.can_acquire_protection())
+    return true;
+
+  MDL_REQUEST_INIT(&global_request,
+                   MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
+                   MDL_STATEMENT);
+
+  MDL_REQUEST_INIT(&mdl_request,
+                   MDL_key::TABLESPACE, "", tablespace,
+                   MDL_EXCLUSIVE, MDL_TRANSACTION);
+
+  mdl_requests.push_front(&mdl_request);
+  mdl_requests.push_front(&global_request);
+
+  if (thd->mdl_context.acquire_locks(&mdl_requests,
+                                     thd->variables.lock_wait_timeout))
+    return true;
+
+  DEBUG_SYNC(thd, "after_wait_locked_tablespace_name");
+  return false;
+}
+
+// Function generating hash key for Tablespace_hash_set.
+extern "C" uchar *tablespace_set_get_key(const uchar *record,
+                                         size_t *length,
+                                         my_bool not_used MY_ATTRIBUTE((unused)))
+{
+  const char *tblspace_name= reinterpret_cast<const char *>(record);
+  *length= strlen(tblspace_name);
+  return reinterpret_cast<uchar*>(const_cast<char*>(tblspace_name));
+}
+
+/**
+  Acquire IX MDL lock each tablespace name from the given set.
+
+  @param thd               - Thread invoking this function.
+  @param tablespace_set    - Set of tablespace names to be lock.
+  @param lock_wait_timeout - Lock timeout.
+
+  @return true - On failure
+  @return false - On Success.
+*/
+bool lock_tablespace_names(
+       THD *thd,
+       Tablespace_hash_set *tablespace_set,
+       ulong lock_wait_timeout)
+{
+  // Stop if we have nothing to lock
+  if (tablespace_set->is_empty())
+    return false;
+
+  // Prepare MDL_request's for all tablespace names.
+  MDL_request_list mdl_tablespace_requests;
+  Tablespace_hash_set::Iterator it(*tablespace_set);
+  char *tablespace= NULL;
+  while ((tablespace= it++))
+  {
+    DBUG_ASSERT(strlen(tablespace));
+
+    MDL_request *tablespace_request= new (thd->mem_root) MDL_request;
+    if (tablespace_request == NULL)
+      return true;
+    MDL_REQUEST_INIT(tablespace_request, MDL_key::TABLESPACE,
+                     "", tablespace, MDL_INTENTION_EXCLUSIVE,
+                     MDL_TRANSACTION);
+    mdl_tablespace_requests.push_front(tablespace_request);
+  }
+
+  // Finally, acquire IX MDL locks.
+  if (thd->mdl_context.acquire_locks(&mdl_tablespace_requests,
+                                     lock_wait_timeout))
+    return true;
+
+  DEBUG_SYNC(thd, "after_wait_locked_tablespace_name_for_table");
+
+  return false;
+}
+
 /**
   Obtain an exclusive metadata lock on an object name.
 
@@ -842,10 +956,8 @@ bool lock_schema_name(THD *thd, const char *db)
   @param db          The schema the object belongs to.
   @param name        Object name in the schema.
 
-  This function assumes that no metadata locks were acquired
-  before calling it. Additionally, it cannot be called while
-  holding LOCK_open mutex. Both these invariants are enforced by
-  asserts in MDL_context::acquire_locks().
+  This function cannot be called while holding LOCK_open_mutex.
+  This invariant is enforced by asserts in MDL_context::acquire_locks.
   To avoid deadlocks, we do not try to obtain exclusive metadata
   locks in LOCK TABLES mode, since in this mode there may be
   other metadata locks already taken by the current connection,
@@ -866,8 +978,7 @@ bool lock_object_name(THD *thd, MDL_key::enum_mdl_namespace mdl_type,
 
   if (thd->locked_tables_mode)
   {
-    my_message(ER_LOCK_OR_ACTIVE_TRANSACTION,
-               ER(ER_LOCK_OR_ACTIVE_TRANSACTION), MYF(0));
+    my_error(ER_LOCK_OR_ACTIVE_TRANSACTION, MYF(0));
     return TRUE;
   }
 
@@ -876,11 +987,14 @@ bool lock_object_name(THD *thd, MDL_key::enum_mdl_namespace mdl_type,
 
   if (thd->global_read_lock.can_acquire_protection())
     return TRUE;
-  global_request.init(MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
-                      MDL_STATEMENT);
-  schema_request.init(MDL_key::SCHEMA, db, "", MDL_INTENTION_EXCLUSIVE,
-                      MDL_TRANSACTION);
-  mdl_request.init(mdl_type, db, name, MDL_EXCLUSIVE, MDL_TRANSACTION);
+  MDL_REQUEST_INIT(&global_request,
+                   MDL_key::GLOBAL, "", "", MDL_INTENTION_EXCLUSIVE,
+                   MDL_STATEMENT);
+  MDL_REQUEST_INIT(&schema_request,
+                   MDL_key::SCHEMA, db, "", MDL_INTENTION_EXCLUSIVE,
+                   MDL_TRANSACTION);
+  MDL_REQUEST_INIT(&mdl_request,
+                   mdl_type, db, name, MDL_EXCLUSIVE, MDL_TRANSACTION);
 
   mdl_requests.push_front(&mdl_request);
   mdl_requests.push_front(&schema_request);
@@ -901,22 +1015,22 @@ static void print_lock_error(int error, const char *table)
 
   switch (error) {
   case HA_ERR_LOCK_WAIT_TIMEOUT:
-    my_error(ER_LOCK_WAIT_TIMEOUT, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), error);
+    my_error(ER_LOCK_WAIT_TIMEOUT, MYF(0), error);
     break;
   case HA_ERR_READ_ONLY_TRANSACTION:
-    my_error(ER_READ_ONLY_TRANSACTION, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG),
+    my_error(ER_READ_ONLY_TRANSACTION, MYF(0),
              error);
     break;
   case HA_ERR_LOCK_DEADLOCK:
-    my_error(ER_LOCK_DEADLOCK, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), error);
+    my_error(ER_LOCK_DEADLOCK, MYF(0), error);
     break;
   case HA_ERR_WRONG_COMMAND:
-    my_error(ER_ILLEGAL_HA, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG), table);
+    my_error(ER_ILLEGAL_HA, MYF(0), table);
     break;
   default:
     {
       char errbuf[MYSYS_STRERROR_SIZE];
-      my_error(ER_CANT_LOCK, MYF(ME_BELL+ME_OLDWIN+ME_WAITTANG),
+      my_error(ER_CANT_LOCK, MYF(0),
                error, my_strerror(errbuf, sizeof(errbuf), error));
     }
     break;
@@ -994,35 +1108,31 @@ volatile int32 Global_read_lock::m_active_requests;
   @retval True   Failure, thread was killed.
 */
 
-#ifdef WITH_WSREP
 bool Global_read_lock::lock_global_read_lock(THD *thd, bool *own_lock)
-#else
-bool Global_read_lock::lock_global_read_lock(THD *thd)
-#endif
 {
   DBUG_ENTER("lock_global_read_lock");
 
-#ifdef WITH_WSREP
   *own_lock= FALSE;
-#endif
 
   if (!m_state)
   {
     MDL_request mdl_request;
 
-    DBUG_ASSERT(! thd->mdl_context.is_lock_owner(MDL_key::GLOBAL, "", "",
-                                                 MDL_SHARED));
+    DBUG_ASSERT(! thd->mdl_context.owns_equal_or_stronger_lock(MDL_key::GLOBAL,
+                                                               "", "",
+                                                               MDL_SHARED));
 
     /*
-       Do not allow upgrading backup locks to FTWRL. Otherwise we can end up
-       with deadlocks being reported for queries previously blocked on a backup
-       lock.
+      Do not allow upgrading backup locks to FTWRL. Otherwise we can end up
+      with deadlocks being reported for queries previously blocked on a backup
+      lock.
     */
     if (thd->backup_tables_lock.abort_if_acquired() ||
         thd->backup_binlog_lock.abort_if_acquired())
-      DBUG_RETURN(1);
+        DBUG_RETURN(true);
 
-    mdl_request.init(MDL_key::GLOBAL, "", "", MDL_SHARED, MDL_EXPLICIT);
+    MDL_REQUEST_INIT(&mdl_request,
+                     MDL_key::GLOBAL, "", "", MDL_SHARED, MDL_EXPLICIT);
 
     /* Increment static variable first to signal innodb memcached server
        to release mdl locks held by it */
@@ -1037,9 +1147,7 @@ bool Global_read_lock::lock_global_read_lock(THD *thd)
     m_mdl_global_shared_lock= mdl_request.ticket;
     m_state= GRL_ACQUIRED;
 
-#ifdef WITH_WSREP
     *own_lock= TRUE;
-#endif
   }
   /*
     We DON'T set global_read_lock_blocks_commit now, it will be set after
@@ -1091,7 +1199,7 @@ void Global_read_lock::unlock_global_read_lock(THD *thd)
       if (ret != WSREP_OK)
       {
         WSREP_WARN("resync failed %d for FTWRL: db: %s, query: %s", ret,
-                   (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+                   (thd->db().length ? thd->db().str : "(null)"), WSREP_QUERY(thd));
         DBUG_VOID_RETURN;
       }
       provider_desynced_paused= false;
@@ -1103,6 +1211,7 @@ void Global_read_lock::unlock_global_read_lock(THD *thd)
   my_atomic_add32(&Global_read_lock::m_active_requests, -1);
   m_mdl_global_shared_lock= NULL;
   m_state= GRL_NONE;
+
   DBUG_VOID_RETURN;
 }
 
@@ -1143,7 +1252,8 @@ bool Global_read_lock::make_global_read_lock_block_commit(THD *thd)
   if (m_state != GRL_ACQUIRED)
     DBUG_RETURN(0);
 
-  mdl_request.init(MDL_key::COMMIT, "", "", MDL_SHARED, MDL_EXPLICIT);
+  MDL_REQUEST_INIT(&mdl_request,
+                   MDL_key::COMMIT, "", "", MDL_SHARED, MDL_EXPLICIT);
 
   if (thd->mdl_context.acquire_lock(&mdl_request,
                                     thd->variables.lock_wait_timeout))
@@ -1165,14 +1275,10 @@ bool Global_read_lock::make_global_read_lock_block_commit(THD *thd)
     int rcode;
     WSREP_DEBUG("running implicit desync for node");
     rcode = wsrep->desync(wsrep);
-    if (rcode == WSREP_TRX_FAIL && strcmp(wsrep_cluster_status, "Primary") != 0)
-    {
-      WSREP_DEBUG("desync failed while non-Primary, ignoring failure");
-    }
-    else if (rcode != WSREP_OK)
+    if (rcode != WSREP_OK)
     {
       WSREP_WARN("FTWRL desync failed %d for schema: %s, query: %s",
-                 rcode, (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+                 rcode, (thd->db().length ? thd->db().str : "(null)"), WSREP_QUERY(thd));
       my_message(ER_LOCK_DEADLOCK, "wsrep desync failed for FTWRL", MYF(0));
       DBUG_RETURN(TRUE);
     }
@@ -1296,11 +1402,12 @@ bool Global_backup_lock::acquire(THD *thd)
 
   DBUG_ENTER("Global_backup_lock::acquire");
 
-  DBUG_ASSERT(m_lock == NULL);
-  DBUG_ASSERT(!thd->mdl_context.is_lock_owner(m_namespace, "", "",
-                                              MDL_SHARED));
+  DBUG_ASSERT(m_lock == NULL &&
+              !thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "",
+                                                            "", MDL_SHARED));
 
-  mdl_request.init(m_namespace, "", "", MDL_SHARED, MDL_EXPLICIT);
+  MDL_REQUEST_INIT(&mdl_request, m_namespace, "", "", MDL_SHARED,
+                   MDL_EXPLICIT);
 
   if (thd->mdl_context.acquire_lock(&mdl_request,
                                     thd->variables.lock_wait_timeout))
@@ -1321,9 +1428,9 @@ void Global_backup_lock::release(THD *thd)
 {
   DBUG_ENTER("Global_backup_lock::release");
 
-  DBUG_ASSERT(m_lock != NULL);
-  DBUG_ASSERT(thd->mdl_context.is_lock_owner(m_namespace, "", "",
-                                             MDL_SHARED));
+  DBUG_ASSERT(m_lock != NULL &&
+              thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
+                                                           MDL_SHARED));
 
   thd->mdl_context.release_lock(m_lock);
 
@@ -1340,21 +1447,37 @@ void Global_backup_lock::release(THD *thd)
 
 void Global_backup_lock::set_explicit_locks_duration(THD *thd)
 {
+  bool should_own;
+
   DBUG_ENTER("Global_backup_lock::set_explicit_lock_duration");
 
   if (m_lock)
+  {
+    should_own= true;
     thd->mdl_context.set_lock_duration(m_lock, MDL_EXPLICIT);
+  }
+  else
+  {
+    should_own= false;
+  }
 
-  DBUG_ASSERT((m_lock != NULL) ==
-              thd->mdl_context.is_lock_owner(m_namespace, "", "",
-                                             MDL_SHARED));
+  DBUG_ASSERT(should_own ==
+              thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
+                                                           MDL_SHARED));
 
   if (m_prot_lock)
+  {
+    should_own= true;
     thd->mdl_context.set_lock_duration(m_prot_lock, MDL_EXPLICIT);
+  }
+  else
+  {
+    should_own= false;
+  }
 
-  DBUG_ASSERT((m_prot_lock != NULL) ==
-              thd->mdl_context.is_lock_owner(m_namespace, "", "",
-                                             MDL_INTENTION_EXCLUSIVE));
+  DBUG_ASSERT(should_own ==
+              thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
+                                                     MDL_INTENTION_EXCLUSIVE));
 
   DBUG_VOID_RETURN;
 }
@@ -1380,8 +1503,8 @@ bool Global_backup_lock::acquire_protection(THD *thd,
   DBUG_ENTER("Global_backup_lock::acquire_protection");
 
   DBUG_ASSERT(duration != MDL_EXPLICIT ||
-              !thd->mdl_context.is_lock_owner(m_namespace, "", "",
-                                              MDL_INTENTION_EXCLUSIVE));
+              !thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "",
+                                                "", MDL_INTENTION_EXCLUSIVE));
 
   init_protection_request(&mdl_request, duration);
 
@@ -1407,8 +1530,8 @@ Global_backup_lock::init_protection_request(MDL_request *mdl_request,
 {
   DBUG_ENTER("Global_backup_lock::init_protection_request");
 
-  mdl_request->init(m_namespace, "", "", MDL_INTENTION_EXCLUSIVE,
-                    duration);
+  MDL_REQUEST_INIT(mdl_request, m_namespace, "", "", MDL_INTENTION_EXCLUSIVE,
+                   duration);
 
   DBUG_VOID_RETURN;
 }
@@ -1423,8 +1546,8 @@ void Global_backup_lock::release_protection(THD *thd)
 {
   DBUG_ENTER("Global_backup_lock::release_protection");
 
-  DBUG_ASSERT(m_prot_lock != NULL);
-  DBUG_ASSERT(thd->mdl_context.is_lock_owner(m_namespace, "", "",
+  DBUG_ASSERT(m_prot_lock != NULL &&
+              thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
                                              MDL_INTENTION_EXCLUSIVE));
 
   thd->mdl_context.release_lock(m_prot_lock);
