@@ -14,39 +14,47 @@
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
 #include "my_global.h"    // NO_EMBEDDED_ACCESS_CHECKS
-#include "sql_priv.h"
 #include "sp_instr.h"
 #include "item.h"         // Item_splocal
+#include "log.h"          // query_logger
 #include "opt_trace.h"    // opt_trace_disable_etc
 #include "probes_mysql.h" // MYSQL_QUERY_EXEC_START
 #include "sp_head.h"      // sp_head
 #include "sp.h"           // sp_get_item_value
 #include "sp_rcontext.h"  // sp_rcontext
-#include "sql_acl.h"      // SELECT_ACL
+#include "auth_common.h"  // SELECT_ACL
 #include "sql_base.h"     // open_temporary_tables
 #include "sql_parse.h"    // check_table_access
 #include "sql_prepare.h"  // reinit_stmt_before_use
 #include "transaction.h"  // trans_commit_stmt
+#include "prealloced_array.h"
 #include "sql_audit.h"
+#include "binlog.h"
+#include "item_cmpfunc.h" // Item_func_eq
 
 #include <algorithm>
+#include <functional>
 
-///////////////////////////////////////////////////////////////////////////
-// Static function implementation.
-///////////////////////////////////////////////////////////////////////////
+#include "trigger.h"                  // Trigger
+#include "table_trigger_dispatcher.h" // Table_trigger_dispatcher
 
 
-static int cmp_splocal_locations(Item_splocal * const *a,
-                                 Item_splocal * const *b)
+class Cmp_splocal_locations :
+  public std::binary_function<const Item_splocal*, const Item_splocal*, bool>
 {
-  return (int)((*a)->pos_in_query - (*b)->pos_in_query);
-}
-
+public:
+  bool operator()(const Item_splocal *a, const Item_splocal *b)
+  {
+    DBUG_ASSERT(a->pos_in_query != b->pos_in_query);
+    return a->pos_in_query < b->pos_in_query;
+  }
+};
 
 /*
   StoredRoutinesBinlogging
   This paragraph applies only to statement-based binlogging. Row-based
-  binlogging does not need anything special like this.
+  binlogging does not need anything special like this except for a special
+  case that is mentioned below in section 2.1
 
   Top-down overview:
 
@@ -73,6 +81,14 @@ static int cmp_splocal_locations(Item_splocal * const *a,
     of SP local variables with NAME_CONST('spvar_name', <spvar-value>) calls.
     This substitution is done in subst_spvars().
 
+  2.1 Miscellaneous case: DDLs (Eg: ALTER EVENT) in StoredProcedure(SP) uses
+      its local variables
+
+  * Irrespective of binlog format, DDLs are always binlogged in statement mode.
+    Hence if there are any DDLs, in stored procedure, that uses SP local
+    variables,  those should be replaced with NAME_CONST('spvar_name', <spvar-value>)
+    even if binlog format is 'row'.
+
   3. FUNCTION calls
 
   In sp_head::execute_function(), we check
@@ -138,10 +154,8 @@ static int cmp_splocal_locations(Item_splocal * const *a,
 */
 static bool subst_spvars(THD *thd, sp_instr *instr, LEX_STRING *query_str)
 {
-  Dynamic_array<Item_splocal*> sp_vars_uses;
-  char *pbuf, *cur, buffer[512];
-  String qbuf(buffer, sizeof(buffer), &my_charset_bin);
-  int prev_pos, res, buf_len;
+  // Stack-local array, does not need instrumentation.
+  Prealloced_array<Item_splocal*, 16> sp_vars_uses(PSI_NOT_INSTRUMENTED);
 
   /* Find all instances of Item_splocal used in this statement */
   for (Item *item= instr->free_list; item; item= item->next)
@@ -150,27 +164,30 @@ static bool subst_spvars(THD *thd, sp_instr *instr, LEX_STRING *query_str)
     {
       Item_splocal *item_spl= (Item_splocal*)item;
       if (item_spl->pos_in_query)
-        sp_vars_uses.append(item_spl);
+        sp_vars_uses.push_back(item_spl);
     }
   }
 
-  if (!sp_vars_uses.elements())
+  if (sp_vars_uses.empty())
     return false;
 
   /* Sort SP var refs by their occurrences in the query */
-  sp_vars_uses.sort(cmp_splocal_locations);
+  std::sort(sp_vars_uses.begin(), sp_vars_uses.end(), Cmp_splocal_locations());
 
   /*
     Construct a statement string where SP local var refs are replaced
     with "NAME_CONST(name, value)"
   */
+  char buffer[512];
+  String qbuf(buffer, sizeof(buffer), &my_charset_bin);
   qbuf.length(0);
-  cur= query_str->str;
-  prev_pos= res= 0;
+  char *cur= query_str->str;
+  int prev_pos= 0;
+  int res= 0;
   thd->query_name_consts= 0;
 
-  for (Item_splocal **splocal= sp_vars_uses.front(); 
-       splocal <= sp_vars_uses.back(); splocal++)
+  for (Item_splocal **splocal= sp_vars_uses.begin(); 
+       splocal != sp_vars_uses.end(); splocal++)
   {
     Item *val;
 
@@ -219,25 +236,11 @@ static bool subst_spvars(THD *thd, sp_instr *instr, LEX_STRING *query_str)
       qbuf.append(cur + prev_pos, query_str->length - prev_pos))
     return true;
 
-  /*
-    Allocate additional space at the end of the new query string for the
-    query_cache_send_result_to_client function.
-
-    The query buffer layout is:
-       buffer :==
-            <statement>   The input statement(s)
-            '\0'          Terminating null char
-            <length>      Length of following current database name (size_t)
-            <db_name>     Name of current database
-            <flags>       Flags struct
-  */
-  buf_len= qbuf.length() + 1 + sizeof(size_t) + thd->db_length + 
-           QUERY_CACHE_FLAGS_SIZE + 1;
-  if ((pbuf= (char *) alloc_root(thd->mem_root, buf_len)))
+  char *pbuf;
+  if ((pbuf= static_cast<char*>(thd->alloc(qbuf.length() + 1))))
   {
     memcpy(pbuf, qbuf.ptr(), qbuf.length());
     pbuf[qbuf.length()]= 0;
-    memcpy(pbuf+qbuf.length()+1, (char *) &thd->db_length, sizeof(size_t));
   }
   else
     return true;
@@ -263,8 +266,6 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
                                            uint *nextp,
                                            bool open_tables)
 {
-  bool rc= false;
-
   /*
     The flag is saved at the entry to the following substatement.
     It's reset further in the common code part.
@@ -272,8 +273,8 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
   */
 
   unsigned int parent_unsafe_rollback_flags=
-    thd->transaction.stmt.get_unsafe_rollback_flags();
-  thd->transaction.stmt.reset_unsafe_rollback_flags();
+    thd->get_transaction()->get_unsafe_rollback_flags(Transaction_ctx::STMT);
+  thd->get_transaction()->reset_unsafe_rollback_flags(Transaction_ctx::STMT);
 
   /* Check pre-conditions. */
 
@@ -315,81 +316,102 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
     }
   }
 
-  /* Reset LEX-object before re-use. */
+  bool error= reinit_stmt_before_use(thd, m_lex);
 
-  reinit_stmt_before_use(thd, m_lex);
+  /*
+    In case a session state exists do not cache the SELECT stmt. If we
+    cache SELECT statment when session state information exists, then
+    the result sets of this SELECT are cached which contains changed
+    session information. Next time when same query is executed when there
+    is no change in session state, then result sets are picked from cache
+    which is wrong as the result sets picked from cache have changed
+    state information.
+    In case of embedded server since session state information is not
+    sent there is no need to turn off cache.
+  */
+
+#ifndef EMBEDDED_LIBRARY
+  if (thd->get_protocol()->has_client_capability(CLIENT_SESSION_TRACK) &&
+      thd->session_tracker.enabled_any() &&
+      thd->session_tracker.changed_any())
+    thd->lex->safe_to_cache_query= 0;
+#endif
 
   /* Open tables if needed. */
 
-  if (open_tables)
+  if (!error)
   {
-    /*
-      IF, CASE, DECLARE, SET, RETURN, have 'open_tables' true; they may
-      have a subquery in parameter and are worth tracing. They don't
-      correspond to a SQL command so we pretend that they are SQLCOM_SELECT.
-    */
-    Opt_trace_start ots(thd, m_lex->query_tables, SQLCOM_SELECT,
-                        &m_lex->var_list, NULL, 0, this,
-                        thd->variables.character_set_client);
-    Opt_trace_object trace_command(&thd->opt_trace);
-    Opt_trace_array trace_command_steps(&thd->opt_trace, "steps");
-
-    /*
-      Check whenever we have access to tables for this statement
-      and open and lock them before executing instructions core function.
-      If we are not opening any tables, we don't need to check permissions
-      either.
-    */
-    if (m_lex->query_tables)
-      rc= (open_temporary_tables(thd, m_lex->query_tables) ||
-            check_table_access(thd, SELECT_ACL, m_lex->query_tables, false,
-                               UINT_MAX, false));
-
-    if (!rc)
-      rc= open_and_lock_tables(thd, m_lex->query_tables, true, 0);
-
-    if (!rc)
+    if (open_tables)
     {
-      rc= exec_core(thd, nextp);
-      DBUG_PRINT("info",("exec_core returned: %d", rc));
-    }
+      // todo: break this block out into a separate function.
+      /*
+        IF, CASE, DECLARE, SET, RETURN, have 'open_tables' true; they may
+        have a subquery in parameter and are worth tracing. They don't
+        correspond to a SQL command so we pretend that they are SQLCOM_SELECT.
+      */
+      Opt_trace_start ots(thd, m_lex->query_tables, SQLCOM_SELECT,
+                          &m_lex->var_list, NULL, 0, this,
+                          thd->variables.character_set_client);
+      Opt_trace_object trace_command(&thd->opt_trace);
+      Opt_trace_array trace_command_steps(&thd->opt_trace, "steps");
 
-    /*
-      Call after unit->cleanup() to close open table
-      key read.
-    */
+      /*
+        Check whenever we have access to tables for this statement
+        and open and lock them before executing instructions core function.
+        If we are not opening any tables, we don't need to check permissions
+        either.
+      */
+      if (m_lex->query_tables)
+        error= (open_temporary_tables(thd, m_lex->query_tables) ||
+                check_table_access(thd, SELECT_ACL, m_lex->query_tables, false,
+                                   UINT_MAX, false));
 
-    m_lex->unit.cleanup();
+      if (!error)
+        error= open_and_lock_tables(thd, m_lex->query_tables, 0);
 
-    /* Here we also commit or rollback the current statement. */
-
-    if (! thd->in_sub_stmt)
-    {
-      thd->get_stmt_da()->set_overwrite_status(true);
-      thd->is_error() ? trans_rollback_stmt(thd) : trans_commit_stmt(thd);
-      thd->get_stmt_da()->set_overwrite_status(false);
-    }
-    thd_proc_info(thd, "closing tables");
-    close_thread_tables(thd);
-    thd_proc_info(thd, 0);
-
-    if (! thd->in_sub_stmt)
-    {
-      if (thd->transaction_rollback_request)
+      if (!error)
       {
-        trans_rollback_implicit(thd);
-        thd->mdl_context.release_transactional_locks();
+        error= exec_core(thd, nextp);
+        DBUG_PRINT("info",("exec_core returned: %d", error));
+      }
+
+      /*
+        Call after unit->cleanup() to close open table
+        key read.
+      */
+
+      m_lex->unit->cleanup(true);
+
+      /* Here we also commit or rollback the current statement. */
+
+      if (! thd->in_sub_stmt)
+      {
+        thd->get_stmt_da()->set_overwrite_status(true);
+        thd->is_error() ? trans_rollback_stmt(thd) : trans_commit_stmt(thd);
+        thd->get_stmt_da()->set_overwrite_status(false);
+      }
+      thd_proc_info(thd, "closing tables");
+      close_thread_tables(thd);
+      thd_proc_info(thd, 0);
+
+      if (! thd->in_sub_stmt)
+      {
+        if (thd->transaction_rollback_request)
+        {
+          trans_rollback_implicit(thd);
+          thd->mdl_context.release_transactional_locks();
+        }
+        else if (! thd->in_multi_stmt_transaction_mode())
+          thd->mdl_context.release_transactional_locks();
+        else
+          thd->mdl_context.release_statement_locks();
       }
-      else if (! thd->in_multi_stmt_transaction_mode())
-        thd->mdl_context.release_transactional_locks();
-      else
-        thd->mdl_context.release_statement_locks();
     }
-  }
-  else
-  {
-    rc= exec_core(thd, nextp);
-    DBUG_PRINT("info",("exec_core returned: %d", rc));
+    else
+    {
+      error= exec_core(thd, nextp);
+      DBUG_PRINT("info",("exec_core returned: %d", error));
+    }
   }
 
   if (m_lex->query_tables_own_last)
@@ -418,10 +440,10 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
     open_tables stage.
   */
 
-  if (!rc || !thd->is_error() ||
-      (thd->get_stmt_da()->sql_errno() != ER_CANT_REOPEN_TABLE &&
-       thd->get_stmt_da()->sql_errno() != ER_NO_SUCH_TABLE &&
-       thd->get_stmt_da()->sql_errno() != ER_UPDATE_TABLE_USED))
+  if (!error || !thd->is_error() ||
+      (thd->get_stmt_da()->mysql_errno() != ER_CANT_REOPEN_TABLE &&
+       thd->get_stmt_da()->mysql_errno() != ER_NO_SUCH_TABLE &&
+       thd->get_stmt_da()->mysql_errno() != ER_UPDATE_TABLE_USED))
     thd->stmt_arena->state= Query_arena::STMT_EXECUTED;
 
   /*
@@ -429,7 +451,16 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
     what is needed from the substatement gained
   */
 
-  thd->transaction.stmt.add_unsafe_rollback_flags(parent_unsafe_rollback_flags);
+  thd->get_transaction()->add_unsafe_rollback_flags(
+    Transaction_ctx::STMT,
+    parent_unsafe_rollback_flags);
+
+  if (thd->variables.session_track_transaction_info > TX_TRACK_NONE)
+  {
+    ((Transaction_state_tracker *)
+     thd->session_tracker.get_tracker(TRANSACTION_INFO_TRACKER))
+      ->add_trx_state_from_thd(thd);
+  }
 
   /* Restore original lex. */
 
@@ -445,7 +476,7 @@ bool sp_lex_instr::reset_lex_and_exec_core(THD *thd,
     cleanup_items() is called in sp_head::execute()
   */
 
-  return rc || thd->is_error();
+  return error || thd->is_error();
 }
 
 
@@ -476,7 +507,8 @@ LEX *sp_lex_instr::parse_expr(THD *thd, sp_head *sp)
 
   // Cleanup and re-init the lex mem_root for re-parse.
   free_root(&m_lex_mem_root, MYF(0));
-  init_sql_alloc(&m_lex_mem_root, MEM_ROOT_BLOCK_SIZE, MEM_ROOT_PREALLOC);
+  init_sql_alloc(PSI_NOT_INSTRUMENTED, &m_lex_mem_root,
+                 MEM_ROOT_BLOCK_SIZE, MEM_ROOT_PREALLOC);
 
   /*
     Switch mem-roots. We store the new LEX and its Items in the
@@ -541,19 +573,21 @@ LEX *sp_lex_instr::parse_expr(THD *thd, sp_head *sp)
         execution.
       */
 
-      Table_triggers_list *ttl= sp->m_trg_list;
-      int event= sp->m_trg_chistics.event;
-      int action_time= sp->m_trg_chistics.action_time;
-      GRANT_INFO *grant_table= &ttl->subject_table_grants[event][action_time];
+      Trigger *t= sp->m_trg_list->find_trigger(thd->lex->sphead->m_name);
 
-      for (Item_trigger_field *trg_field=
-             sp->m_cur_instr_trig_field_items.first;
-           trg_field;
-           trg_field= trg_field->next_trg_field)
+      DBUG_ASSERT(t);
+
+      if (!t)
+        return NULL; // Don't take chances in production.
+
+      for (Item_trigger_field *trg_fld= sp->m_cur_instr_trig_field_items.first;
+           trg_fld;
+           trg_fld= trg_fld->next_trg_field)
       {
-        trg_field->setup_field(thd, ttl->trigger_table, grant_table);
+        trg_fld->setup_field(thd, sp->m_trg_list->get_trigger_field_support(),
+                             t->get_subject_table_grant());
       }
-     
+
       /**
         Move Item_trigger_field's list to instruction's Item_trigger_field
         list.
@@ -598,6 +632,7 @@ bool sp_lex_instr::validate_lex_and_execute_core(THD *thd,
 
   while (true)
   {
+    DBUG_EXECUTE_IF("simulate_bug18831513", { invalidate(); });
     if (is_invalid())
     {
       LEX *lex= parse_expr(thd, thd->sp_runtime_ctx->sp);
@@ -663,7 +698,7 @@ bool sp_lex_instr::validate_lex_and_execute_core(THD *thd,
     if (stmt_reprepare_observer &&
         !thd->is_fatal_error &&
         !thd->killed &&
-        thd->get_stmt_da()->sql_errno() == ER_NEED_REPREPARE &&
+        thd->get_stmt_da()->mysql_errno() == ER_NEED_REPREPARE &&
         reprepare_attempt++ < 3)
     {
       DBUG_ASSERT(stmt_reprepare_observer->is_invalidated());
@@ -742,6 +777,10 @@ void sp_lex_instr::get_query(String *sql_query) const
 // sp_instr_stmt implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_stmt::psi_info=                                    
+{ 0, "stmt", 0};
+#endif
 
 bool sp_instr_stmt::execute(THD *thd, uint *nextp)
 {
@@ -751,7 +790,9 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
 
   DBUG_PRINT("info", ("query: '%.*s'", (int) m_query.length, m_query.str));
 
-  const CSET_STRING query_backup= thd->query_string;
+  MYSQL_SET_STATEMENT_TEXT(thd->m_statement_psi, m_query.str, m_query.length);
+
+  const LEX_CSTRING query_backup= thd->query();
 
 #if defined(ENABLED_PROFILING)
   /* This SP-instr is profilable and will be captured. */
@@ -783,7 +824,13 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
 
     - general log is off
 
-    - binary logging is off, or not in statement mode
+    - binary logging is off
+
+    - if the query generates row events in binlog row format mode
+    (DDLs are always written in statement format irrespective of binlog_format
+    and they can have SP variables in it. For example, 'ALTER EVENT' is allowed
+    inside a procedure and can contain SP variables in it. Those too need to be
+    substituted with NAME_CONST(...))
 
     We don't have to substitute on behalf of the query cache as
     queries with SP vars are not cached, anyway.
@@ -800,10 +847,11 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
     substitution immediately before writing to the log.
   */
 
-  need_subst= ((thd->variables.option_bits & OPTION_LOG_OFF) &&
+  need_subst= !((thd->variables.option_bits & OPTION_LOG_OFF) &&
                (!(thd->variables.option_bits & OPTION_BIN_LOG) ||
                 !mysql_bin_log.is_open() ||
-                thd->is_current_stmt_binlog_format_row())) ? FALSE : TRUE;
+                (thd->is_current_stmt_binlog_format_row() &&
+                 sqlcom_can_generate_row_events(m_lex->sql_command))));
 
   /*
     If we need to do a substitution but can't (OOM), give up.
@@ -817,10 +865,10 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
     queries with SP vars can't be cached)
   */
   if (unlikely((thd->variables.option_bits & OPTION_LOG_OFF)==0))
-    general_log_write(thd, COM_QUERY, thd->query(), thd->query_length());
+    query_logger.general_log_write(thd, COM_QUERY, thd->query().str,
+                                   thd->query().length);
 
-  if (query_cache_send_result_to_client(thd, thd->query(),
-                                        thd->query_length()) <= 0)
+  if (query_cache.send_result_to_client(thd, thd->query()) <= 0)
   {
     rc= validate_lex_and_execute_core(thd, nextp, false);
 
@@ -839,15 +887,19 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
     if (thd->get_stmt_da()->is_eof())
     {
       /* Finalize server status flags after executing a statement. */
-      thd->protocol->end_statement();
+      thd->update_server_status();
+
+      thd->send_statement_status();
     }
 
-    query_cache_end_of_result(thd);
+    query_cache.end_of_result(thd);
 
+#ifndef EMBEDDED_LIBRARY
     mysql_audit_general(thd, MYSQL_AUDIT_GENERAL_STATUS,
                         thd->get_stmt_da()->is_error() ?
-                            thd->get_stmt_da()->sql_errno() : 0,
+                            thd->get_stmt_da()->mysql_errno() : 0,
                         command_name[COM_QUERY].str);
+#endif
 
     if (!rc && unlikely(log_slow_applicable(thd)))
     {
@@ -878,9 +930,6 @@ bool sp_instr_stmt::execute(THD *thd, uint *nextp)
   thd->set_query(query_backup);
   thd->query_name_consts= 0;
 
-  if (!thd->is_error())
-    thd->get_stmt_da()->reset_diagnostics_area();
-
   /* Restore the original query start time */
   if (thd->enable_slow_log)
     thd->set_time(&time_info);
@@ -901,12 +950,12 @@ void sp_instr_stmt::print(String *str)
     Print the query string (but not too much of it), just to indicate which
     statement it is.
   */
-  uint len= m_query.length;
+  size_t len= m_query.length;
   if (len > SP_STMT_PRINT_MAXLEN)
     len= SP_STMT_PRINT_MAXLEN-3;
 
   /* Copy the query string and replace '\n' with ' ' in the process */
-  for (uint i= 0 ; i < len ; i++)
+  for (size_t i= 0 ; i < len ; i++)
   {
     char c= m_query.str[i];
     if (c == '\n')
@@ -915,24 +964,23 @@ void sp_instr_stmt::print(String *str)
   }
   if (m_query.length > SP_STMT_PRINT_MAXLEN)
     str->qs_append(STRING_WITH_LEN("...")); /* Indicate truncated string */
-  str->qs_append('"');
+  str->qs_append(STRING_WITH_LEN("\""));
 }
 
 
 bool sp_instr_stmt::exec_core(THD *thd, uint *nextp)
 {
-  MYSQL_QUERY_EXEC_START(thd->query(),
-                         thd->thread_id,
-                         (char *) (thd->db ? thd->db : ""),
-                         &thd->security_ctx->priv_user[0],
-                         (char *)thd->security_ctx->host_or_ip,
+  MYSQL_QUERY_EXEC_START(const_cast<char*>(thd->query().str),
+                         thd->thread_id(),
+                         (char *) (thd->db().str ? thd->db().str : ""),
+                         (char *) thd->security_context()->priv_user().str,
+                         (char *) thd->security_context()->host_or_ip().str,
                          3);
 
   thd->lex->set_sp_current_parsing_ctx(get_parsing_ctx());
   thd->lex->sphead= thd->sp_runtime_ctx->sp;
 
   PSI_statement_locker *statement_psi_saved= thd->m_statement_psi;
-  thd->m_statement_psi= NULL;
 
   bool rc= mysql_execute_command(thd);
 
@@ -952,6 +1000,10 @@ bool sp_instr_stmt::exec_core(THD *thd, uint *nextp)
 // sp_instr_set implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_set::psi_info=                                     
+{ 0, "set", 0};
+#endif
 
 bool sp_instr_set::exec_core(THD *thd, uint *nextp)
 {
@@ -975,7 +1027,7 @@ bool sp_instr_set::exec_core(THD *thd, uint *nextp)
 void sp_instr_set::print(String *str)
 {
   /* set name@offset ... */
-  int rsrv = SP_INSTR_UINT_MAXLEN+6;
+  size_t rsrv = SP_INSTR_UINT_MAXLEN+6;
   sp_variable *var = m_parsing_ctx->find_variable(m_offset);
 
   /* 'var' should always be non-null, but just in case... */
@@ -991,7 +1043,7 @@ void sp_instr_set::print(String *str)
   }
   str->qs_append(m_offset);
   str->qs_append(' ');
-  m_value_item->print(str, QT_ORDINARY);
+  m_value_item->print(str, QT_TO_ARGUMENT_CHARSET);
 }
 
 
@@ -999,12 +1051,29 @@ void sp_instr_set::print(String *str)
 // sp_instr_set_trigger_field implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_set_trigger_field::psi_info=                       
+{ 0, "set_trigger_field", 0};
+#endif
 
 bool sp_instr_set_trigger_field::exec_core(THD *thd, uint *nextp)
 {
   *nextp= get_ip() + 1;
   thd->count_cuted_fields= CHECK_FIELD_ERROR_FOR_NULL;
-  return m_trigger_field->set_value(thd, &m_value_item);
+  Strict_error_handler strict_handler(Strict_error_handler::
+                                      ENABLE_SET_SELECT_STRICT_ERROR_HANDLER);
+  /*
+    Before Triggers are executed after the 'data' is assigned
+    to the Field objects. If triggers wants to SET invalid value
+    to the Field objects (NEW.<variable_name>= <Invalid value>),
+    it should not be allowed.
+  */
+  if (thd->is_strict_mode() && !thd->lex->is_ignore())
+    thd->push_internal_handler(&strict_handler);
+  bool error= m_trigger_field->set_value(thd, &m_value_item);
+  if (thd->is_strict_mode() && !thd->lex->is_ignore())
+    thd->pop_internal_handler();
+  return error;
 }
 
 
@@ -1013,21 +1082,21 @@ void sp_instr_set_trigger_field::print(String *str)
   str->append(STRING_WITH_LEN("set_trigger_field "));
   m_trigger_field->print(str, QT_ORDINARY);
   str->append(STRING_WITH_LEN(":="));
-  m_value_item->print(str, QT_ORDINARY);
+  m_value_item->print(str, QT_TO_ARGUMENT_CHARSET);
 }
 
 
 bool sp_instr_set_trigger_field::on_after_expr_parsing(THD *thd)
 {
-  DBUG_ASSERT(thd->lex->select_lex.item_list.elements == 1);
+  DBUG_ASSERT(thd->lex->select_lex->item_list.elements == 1);
 
-  m_value_item= thd->lex->select_lex.item_list.head();
+  m_value_item= thd->lex->select_lex->item_list.head();
 
   DBUG_ASSERT(!m_trigger_field);
 
   m_trigger_field=
     new (thd->mem_root) Item_trigger_field(thd->lex->current_context(),
-                                           Item_trigger_field::NEW_ROW,
+                                           TRG_NEW_ROW,
                                            m_trigger_field_name.str,
                                            UPDATE_ACL,
                                            false);
@@ -1056,6 +1125,10 @@ void sp_instr_set_trigger_field::cleanup_before_parsing(THD *thd)
 // sp_instr_jump implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_jump::psi_info=                                    
+{ 0, "jump", 0};
+#endif
 
 void sp_instr_jump::print(String *str)
 {
@@ -1111,6 +1184,10 @@ void sp_instr_jump::opt_move(uint dst, List<sp_branch_instr> *bp)
 // sp_instr_jump_if_not class implementation
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_jump_if_not::psi_info=                             
+{ 0, "jump_if_not", 0};
+#endif
 
 bool sp_instr_jump_if_not::exec_core(THD *thd, uint *nextp)
 {
@@ -1203,6 +1280,10 @@ void sp_lex_branch_instr::opt_move(uint dst, List<sp_branch_instr> *bp)
 // sp_instr_jump_case_when implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_jump_case_when::psi_info=                          
+{ 0, "jump_case_when", 0};
+#endif
 
 bool sp_instr_jump_case_when::exec_core(THD *thd, uint *nextp)
 {
@@ -1260,9 +1341,9 @@ bool sp_instr_jump_case_when::build_expr_items(THD *thd)
 
   if (!m_expr_item)
   {
-    DBUG_ASSERT(thd->lex->select_lex.item_list.elements == 1);
+    DBUG_ASSERT(thd->lex->select_lex->item_list.elements == 1);
 
-    m_expr_item= thd->lex->select_lex.item_list.head();
+    m_expr_item= thd->lex->select_lex->item_list.head();
   }
 
   // Setup main expression item (m_expr_item).
@@ -1280,17 +1361,13 @@ bool sp_instr_jump_case_when::build_expr_items(THD *thd)
 // sp_instr_freturn implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_freturn::psi_info=                                 
+{ 0, "freturn", 0};
+#endif
 
 bool sp_instr_freturn::exec_core(THD *thd, uint *nextp)
 {
-  /*
-    RETURN is a "procedure statement" (in terms of the SQL standard).
-    That means, Diagnostics Area should be clean before its execution.
-  */
-
-  Diagnostics_area *da= thd->get_stmt_da();
-  da->clear_warning_info(da->warning_info_id());
-
   /*
     Change <next instruction pointer>, so that this will be the last
     instruction in the stored function.
@@ -1326,6 +1403,10 @@ void sp_instr_freturn::print(String *str)
 // sp_instr_hpush_jump implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_hpush_jump::psi_info=                              
+{ 0, "hpush_jump", 0};
+#endif
 
 bool sp_instr_hpush_jump::execute(THD *thd, uint *nextp)
 {
@@ -1346,17 +1427,7 @@ void sp_instr_hpush_jump::print(String *str)
   str->qs_append(' ');
   str->qs_append(m_frame);
 
-  switch (m_handler->type) {
-  case sp_handler::EXIT:
-    str->qs_append(STRING_WITH_LEN(" EXIT"));
-    break;
-  case sp_handler::CONTINUE:
-    str->qs_append(STRING_WITH_LEN(" CONTINUE"));
-    break;
-  default:
-    // The handler type must be either CONTINUE or EXIT.
-    DBUG_ASSERT(0);
-  }
+  m_handler->print(str);
 }
 
 
@@ -1398,6 +1469,10 @@ uint sp_instr_hpush_jump::opt_mark(sp_head *sp, List<sp_instr> *leads)
 // sp_instr_hpop implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_hpop::psi_info=                                    
+{ 0, "hpop", 0};
+#endif
 
 bool sp_instr_hpop::execute(THD *thd, uint *nextp)
 {
@@ -1411,16 +1486,13 @@ bool sp_instr_hpop::execute(THD *thd, uint *nextp)
 // sp_instr_hreturn implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_hreturn::psi_info=                                 
+{ 0, "hreturn", 0};
+#endif
 
 bool sp_instr_hreturn::execute(THD *thd, uint *nextp)
 {
-  /*
-    Remove the SQL conditions that were present in DA when the
-    handler was activated.
-  */
-
-  thd->get_stmt_da()->remove_marked_sql_conditions();
-
   /*
     Obtain next instruction pointer (m_dest is set for EXIT handlers, retrieve
     the instruction pointer from runtime context for CONTINUE handlers).
@@ -1436,7 +1508,7 @@ bool sp_instr_hreturn::execute(THD *thd, uint *nextp)
   */
 
   sp_instr *next_instr= rctx->sp->get_instr(*nextp);
-  rctx->exit_handler(next_instr->get_parsing_ctx());
+  rctx->exit_handler(thd, next_instr->get_parsing_ctx());
 
   return false;
 }
@@ -1486,6 +1558,10 @@ uint sp_instr_hreturn::opt_mark(sp_head *sp, List<sp_instr> *leads)
 // sp_instr_cpush implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_cpush::psi_info=                                   
+{ 0, "cpush", 0};
+#endif
 
 bool sp_instr_cpush::execute(THD *thd, uint *nextp)
 {
@@ -1512,7 +1588,7 @@ void sp_instr_cpush::print(String *str)
 {
   const LEX_STRING *cursor_name= m_parsing_ctx->find_cursor(m_cursor_idx);
 
-  uint rsrv= SP_INSTR_UINT_MAXLEN + 7 + m_cursor_query.length + 1;
+  size_t rsrv= SP_INSTR_UINT_MAXLEN + 7 + m_cursor_query.length + 1;
 
   if (cursor_name)
     rsrv+= cursor_name->length;
@@ -1535,6 +1611,10 @@ void sp_instr_cpush::print(String *str)
 // sp_instr_cpop implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_cpop::psi_info=                                    
+{ 0, "cpop", 0};
+#endif
 
 bool sp_instr_cpop::execute(THD *thd, uint *nextp)
 {
@@ -1559,9 +1639,16 @@ void sp_instr_cpop::print(String *str)
 // sp_instr_copen implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_copen::psi_info=                                   
+{ 0, "copen", 0};
+#endif
 
 bool sp_instr_copen::execute(THD *thd, uint *nextp)
 {
+  // Manipulating a CURSOR with an expression should clear DA.
+  clear_da(thd);
+
   *nextp= get_ip() + 1;
 
   // Get the cursor pointer.
@@ -1607,7 +1694,7 @@ void sp_instr_copen::print(String *str)
   const LEX_STRING *cursor_name= m_parsing_ctx->find_cursor(m_cursor_idx);
 
   /* copen name@offset */
-  uint rsrv= SP_INSTR_UINT_MAXLEN+7;
+  size_t rsrv= SP_INSTR_UINT_MAXLEN+7;
 
   if (cursor_name)
     rsrv+= cursor_name->length;
@@ -1627,9 +1714,16 @@ void sp_instr_copen::print(String *str)
 // sp_instr_cclose implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_cclose::psi_info=                                  
+{ 0, "cclose", 0};
+#endif
 
 bool sp_instr_cclose::execute(THD *thd, uint *nextp)
 {
+  // Manipulating a CURSOR with an expression should clear DA.
+  clear_da(thd);
+
   *nextp= get_ip() + 1;
 
   sp_cursor *c= thd->sp_runtime_ctx->get_cursor(m_cursor_idx);
@@ -1643,7 +1737,7 @@ void sp_instr_cclose::print(String *str)
   const LEX_STRING *cursor_name= m_parsing_ctx->find_cursor(m_cursor_idx);
 
   /* cclose name@offset */
-  uint rsrv= SP_INSTR_UINT_MAXLEN+8;
+  size_t rsrv= SP_INSTR_UINT_MAXLEN+8;
 
   if (cursor_name)
     rsrv+= cursor_name->length;
@@ -1663,9 +1757,16 @@ void sp_instr_cclose::print(String *str)
 // sp_instr_cfetch implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_cfetch::psi_info=                                  
+{ 0, "cfetch", 0};
+#endif
 
 bool sp_instr_cfetch::execute(THD *thd, uint *nextp)
 {
+  // Manipulating a CURSOR with an expression should clear DA.
+  clear_da(thd);
+
   *nextp= get_ip() + 1;
 
   sp_cursor *c= thd->sp_runtime_ctx->get_cursor(m_cursor_idx);
@@ -1681,7 +1782,7 @@ void sp_instr_cfetch::print(String *str)
   const LEX_STRING *cursor_name= m_parsing_ctx->find_cursor(m_cursor_idx);
 
   /* cfetch name@offset vars... */
-  uint rsrv= SP_INSTR_UINT_MAXLEN+8;
+  size_t rsrv= SP_INSTR_UINT_MAXLEN+8;
 
   if (cursor_name)
     rsrv+= cursor_name->length;
@@ -1710,6 +1811,10 @@ void sp_instr_cfetch::print(String *str)
 // sp_instr_error implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_error::psi_info=                                   
+{ 0, "error", 0};
+#endif
 
 void sp_instr_error::print(String *str)
 {
@@ -1725,6 +1830,10 @@ void sp_instr_error::print(String *str)
 // sp_instr_set_case_expr implementation.
 ///////////////////////////////////////////////////////////////////////////
 
+#ifdef HAVE_PSI_INTERFACE
+PSI_statement_info sp_instr_set_case_expr::psi_info=                           
+{ 0, "set_case_expr", 0};
+#endif
 
 bool sp_instr_set_case_expr::exec_core(THD *thd, uint *nextp)
 {
