@@ -1,40 +1,36 @@
-/* Copyright (c) 2010, 2018, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2010, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
-#include "sql_parse.h"                       // check_access
+#include "sql_alter.h"
+
+#include "auth_common.h"                     // check_access
 #include "sql_table.h"                       // mysql_alter_table,
                                              // mysql_exchange_partition
 #include "sql_base.h"                        // open_temporary_tables
-#include "sql_alter.h"
 #ifdef WITH_WSREP
 #include "wsrep_mysqld.h"
+#include "sql_parse.h"
 #endif /* WITH_WSREP */
+#include "log.h"
 
-bool has_external_data_or_index_dir(partition_info &pi);
 
 Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
   :drop_list(rhs.drop_list, mem_root),
   alter_list(rhs.alter_list, mem_root),
   key_list(rhs.key_list, mem_root),
+  alter_rename_key_list(rhs.alter_rename_key_list, mem_root),
   create_list(rhs.create_list, mem_root),
   delayed_key_list(rhs.delayed_key_list, mem_root),
   delayed_key_info(rhs.delayed_key_info),
@@ -44,7 +40,8 @@ Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
   partition_names(rhs.partition_names, mem_root),
   num_parts(rhs.num_parts),
   requested_algorithm(rhs.requested_algorithm),
-  requested_lock(rhs.requested_lock)
+  requested_lock(rhs.requested_lock),
+  with_validation(rhs.with_validation)
 {
   /*
     Make deep copies of used objects.
@@ -58,6 +55,7 @@ Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
   list_copy_and_replace_each_value(drop_list, mem_root);
   list_copy_and_replace_each_value(alter_list, mem_root);
   list_copy_and_replace_each_value(key_list, mem_root);
+  list_copy_and_replace_each_value(alter_rename_key_list, mem_root);
   list_copy_and_replace_each_value(create_list, mem_root);
   list_copy_and_replace_each_value(delayed_key_list, mem_root);
   /* partition_names are not deeply copied currently */
@@ -95,32 +93,12 @@ bool Alter_info::set_requested_lock(const LEX_STRING *str)
   return false;
 }
 
-/**
-  Checks if there are any columns with COLUMN_FORMAT COMRPESSED
-  attribute among field definitions in create_list.
-
-  @retval false there are no compressed columns
-  @retval true there is at least one compressed column
-*/
-bool Alter_info::has_compressed_columns() const
-{
-  List_iterator<Create_field> it(
-    const_cast<List<Create_field>&>(create_list));
-  const Create_field* sql_field;
-  while ((sql_field= it++))
-    if (sql_field->column_format() == COLUMN_FORMAT_TYPE_COMPRESSED)
-      return true;
-  return false;
-}
-
 
 Alter_table_ctx::Alter_table_ctx()
   : datetime_field(NULL), error_if_not_empty(false),
     tables_opened(0),
     db(NULL), table_name(NULL), alias(NULL),
-    new_db(NULL), new_name(NULL), new_alias(NULL),
-    fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL)
+    new_db(NULL), new_name(NULL), new_alias(NULL)
 #ifndef DBUG_OFF
     , tmp_table(false)
 #endif
@@ -130,12 +108,11 @@ Alter_table_ctx::Alter_table_ctx()
 
 Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
                                  uint tables_opened_arg,
-                                 char *new_db_arg, char *new_name_arg)
+                                 const char *new_db_arg,
+                                 const char *new_name_arg)
   : datetime_field(NULL), error_if_not_empty(false),
     tables_opened(tables_opened_arg),
-    new_db(new_db_arg), new_name(new_name_arg),
-    fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL)
+    new_db(new_db_arg), new_name(new_name_arg)
 #ifndef DBUG_OFF
     , tmp_table(false)
 #endif
@@ -158,13 +135,14 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
 
     if (lower_case_table_names == 1) // Convert new_name/new_alias to lower case
     {
-      my_casedn_str(files_charset_info, new_name);
+      my_casedn_str(files_charset_info, const_cast<char*>(new_name));
       new_alias= new_name;
     }
     else if (lower_case_table_names == 2) // Convert new_name to lower case
     {
-      strmov(new_alias= new_alias_buff, new_name);
-      my_casedn_str(files_charset_info, new_name);
+      my_stpcpy(new_alias_buff, new_name);
+      new_alias= (const char*)new_alias_buff;
+      my_casedn_str(files_charset_info, const_cast<char*>(new_name));
     }
     else
       new_alias= new_name; // LCTN=0 => case sensitive + case preserving
@@ -186,8 +164,8 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
     new_name= table_name;
   }
 
-  my_snprintf(tmp_name, sizeof(tmp_name), "%s-%lx_%lx", tmp_file_prefix,
-              current_pid, thd->thread_id);
+  my_snprintf(tmp_name, sizeof(tmp_name), "%s-%lx_%x", tmp_file_prefix,
+              current_pid, thd->thread_id());
   /* Safety fix for InnoDB */
   if (lower_case_table_names)
     my_casedn_str(files_charset_info, tmp_name);
@@ -223,9 +201,9 @@ bool Sql_cmd_alter_table::execute(THD *thd)
 {
   LEX *lex= thd->lex;
   /* first SELECT_LEX (have special meaning for many of non-SELECTcommands) */
-  SELECT_LEX *select_lex= &lex->select_lex;
+  SELECT_LEX *select_lex= lex->select_lex;
   /* first table of first SELECT_LEX */
-  TABLE_LIST *first_table= (TABLE_LIST*) select_lex->table_list.first;
+  TABLE_LIST *first_table= select_lex->get_table_list();
   /*
     Code in mysql_alter_table() may modify its HA_CREATE_INFO argument,
     so we have to use a copy of this structure to make execution
@@ -243,16 +221,6 @@ bool Sql_cmd_alter_table::execute(THD *thd)
 
   if (thd->is_fatal_error) /* out of memory creating a copy of alter_info */
     DBUG_RETURN(TRUE);
-
-#ifdef WITH_PARTITION_STORAGE_ENGINE
-  {
-    partition_info *part_info= thd->lex->part_info;
-    if (part_info != NULL && has_external_data_or_index_dir(*part_info) &&
-        check_access(thd, FILE_ACL, any_db, NULL, NULL, FALSE, FALSE))
-
-      DBUG_RETURN(TRUE);
-  }
-#endif
   /*
     We also require DROP priv for ALTER TABLE ... DROP PARTITION, as well
     as for RENAME TO, as being done by SQLCOM_RENAME_TABLE
@@ -325,7 +293,7 @@ bool Sql_cmd_alter_table::execute(THD *thd)
   {
     // Rename of table
     TABLE_LIST tmp_table;
-    memset(static_cast<void*>(&tmp_table), 0, sizeof(tmp_table));
+    memset(&tmp_table, 0, sizeof(tmp_table));
     tmp_table.table_name= lex->name.str;
     tmp_table.db= select_lex->db;
     tmp_table.grant.privilege= priv;
@@ -336,60 +304,198 @@ bool Sql_cmd_alter_table::execute(THD *thd)
 
   /* Don't yet allow changing of symlinks with ALTER TABLE */
   if (create_info.data_file_name)
-    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+    push_warning_printf(thd, Sql_condition::SL_WARNING,
                         WARN_OPTION_IGNORED, ER(WARN_OPTION_IGNORED),
                         "DATA DIRECTORY");
   if (create_info.index_file_name)
-    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+    push_warning_printf(thd, Sql_condition::SL_WARNING,
                         WARN_OPTION_IGNORED, ER(WARN_OPTION_IGNORED),
                         "INDEX DIRECTORY");
   create_info.data_file_name= create_info.index_file_name= NULL;
 
-  thd->set_slow_log_for_admin_command();
+  thd->enable_slow_log= opt_log_slow_admin_statements;
+
 
 #ifdef WITH_WSREP
-  /* Check if foreign keys are accessible.
-  1. Transaction is replicated first, then is done locally.
-  2. Replicated node applies write sets in context
-  of root user.
-  Above two conditions may cause that even if we have no access to FKs,
-  transactions will replicate with success, but fail locally.
+  /* PXC doesn't recommend/allow ALTER operation on table created using
+  non-transactional storage engine (like MyISAM, HEAP/MEMORY, etc....)
+  except ALTER operation to change storage engine to transactional storage
+  engine. (of-course if that is feasible with other constraints)
+  Application only for non-temporary (persistent) tables.
+  Temporarily tables are generally created by some internal workload
+  so we continue to allow them for now as avoid breaking the application.
+  Note: Temporary table are never replicated. */
 
-  Here we check only for FK access, not if the engine supports FKs.
-  That's enough, because right now we need only to know if transaction
-  should be rolled back because of lack of access and not replicated,
-  or we can replicate it and let replicated node do the rest of the job. */
-  if (alter_info.flags & Alter_info::ADD_FOREIGN_KEY)
+  char path[FN_REFLEN + 1];
+  enum legacy_db_type existing_db_type, new_db_type;
+  (void) build_table_filename(path, sizeof(path) - 1, first_table->db,
+                              first_table->table_name, reg_ext, 0);
+  dd_frm_type(thd, path, &existing_db_type);
+
+  bool is_system_db= (first_table &&
+                      ((strcmp(first_table->db, "mysql") == 0) ||
+                       (strcmp(first_table->db, "information_schema") == 0)));
+
+  if (!is_temporary_table(first_table) && !is_system_db)
   {
-    if (check_fk_parent_table_access(thd, select_lex->db, &create_info,
-                                     &alter_info, false))
+    bool safe_ops= false;
+
+    /* If create_info.db_type = NULL that means ALTER operation
+    is modifying the no changing the engine/dbtype. */
+    new_db_type = ((create_info.db_type != NULL)
+                  ? create_info.db_type->db_type : existing_db_type);
+
+    /* Existing table is created with non-transactional storage engine
+    and so switching it to transactional storage engine is allowed.
+    Currently transactional storage engine supported is InnoDB/XtraDB */
+    if ((existing_db_type != DB_TYPE_INNODB)                        &&
+        (alter_info.flags & Alter_info::ALTER_OPTIONS)              &&
+        (create_info.used_fields & HA_CREATE_USED_ENGINE)           &&
+        (new_db_type == DB_TYPE_INNODB ||
+         existing_db_type == new_db_type))
+      safe_ops= true;
+
+    /* Existing table is created with transactional storage engine
+    and switching it to non-transactional storage engine is blocked
+    other operations are allowed. */
+    if ((existing_db_type == DB_TYPE_INNODB) &&
+        (alter_info.flags & Alter_info::ALTER_OPTIONS)              &&
+        (create_info.used_fields & HA_CREATE_USED_ENGINE)           &&
+        (new_db_type != DB_TYPE_INNODB))
+      safe_ops= false;
+    else if (existing_db_type == DB_TYPE_INNODB ||
+             existing_db_type == DB_TYPE_PERFORMANCE_SCHEMA)
+      safe_ops= true;
+
+    if (!safe_ops &&
+        existing_db_type != DB_TYPE_INNODB &&
+        existing_db_type != DB_TYPE_UNKNOWN)
     {
-      DBUG_RETURN(TRUE);
+      bool block= false;
+
+      switch(pxc_strict_mode)
+      {
+      case PXC_STRICT_MODE_DISABLED:
+        break;
+      case PXC_STRICT_MODE_PERMISSIVE:
+        WSREP_WARN("Percona-XtraDB-Cluster doesn't recommend use of"
+                   " ALTER command on a table (%s.%s) that resides in"
+                   " non-transactional storage engine"
+                   " (except switching to transactional engine)"
+                   " with pxc_strict_mode = PERMISSIVE",
+                   first_table->db, first_table->table_name);
+        push_warning_printf(
+          thd, Sql_condition::SL_WARNING, ER_UNKNOWN_ERROR,
+          "Percona-XtraDB-Cluster doesn't recommend use of"
+          " ALTER command on a table (%s.%s) that resides in"
+          " non-transactional storage engine"
+          " (except switching to transactional engine)"
+          " with pxc_strict_mode = PERMISSIVE",
+          first_table->db, first_table->table_name);
+        break;
+      case PXC_STRICT_MODE_ENFORCING:
+      case PXC_STRICT_MODE_MASTER:
+      default:
+        block= true;
+        WSREP_ERROR("Percona-XtraDB-Cluster prohibits use of"
+                    " ALTER command on a table (%s.%s) that resides in"
+                    " non-transactional storage engine"
+                    " (except switching to transactional engine)"
+                    " with pxc_strict_mode = ENFORCING or MASTER",
+                    first_table->db, first_table->table_name);
+        char message[1024];
+        sprintf(message,
+                "Percona-XtraDB-Cluster prohibits use of"
+                " ALTER command on a table (%s.%s) that resides in"
+                " non-transactional storage engine"
+                " (except switching to transactional engine)"
+                " with pxc_strict_mode = ENFORCING or MASTER",
+                first_table->db, first_table->table_name);
+        my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+        break;
+      }
+
+      if (block)
+        DBUG_RETURN(TRUE);
+    }
+    else if (!safe_ops && existing_db_type == DB_TYPE_INNODB)
+    {
+      bool block= false;
+
+      switch(pxc_strict_mode)
+      {
+      case PXC_STRICT_MODE_DISABLED:
+        break;
+      case PXC_STRICT_MODE_PERMISSIVE:
+        WSREP_WARN("Percona-XtraDB-Cluster doesn't recommend changing"
+                   " storage engine of a table (%s.%s) from"
+                   " transactional to non-transactional"
+                   " with pxc_strict_mode = PERMISSIVE",
+                   first_table->db, first_table->table_name);
+        push_warning_printf(
+          thd, Sql_condition::SL_WARNING, ER_UNKNOWN_ERROR,
+          "Percona-XtraDB-Cluster doesn't recommend changing"
+          " storage engine of a table (%s.%s) from"
+          " transactional to non-transactional"
+          " with pxc_strict_mode = PERMISSIVE",
+          first_table->db, first_table->table_name);
+        break;
+      case PXC_STRICT_MODE_ENFORCING:
+      case PXC_STRICT_MODE_MASTER:
+      default:
+        block= true;
+        WSREP_ERROR("Percona-XtraDB-Cluster prohibits changing"
+                    " storage engine of a table (%s.%s) from"
+                    " transactional to non-transactional"
+                    " with pxc_strict_mode = ENFORCING or MASTER",
+                    first_table->db, first_table->table_name);
+        char message[1024];
+        sprintf(message,
+                "Percona-XtraDB-Cluster prohibits changing"
+                " storage engine of a table (%s.%s) from"
+                " transactional to non-transactional"
+                " with pxc_strict_mode = ENFORCING or MASTER",
+                first_table->db, first_table->table_name);
+        my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+        break;
+      }
+
+      if (block)
+        DBUG_RETURN(TRUE);
     }
   }
+#endif /* WITH_WSREP */
 
+  /* Push Strict_error_handler for alter table*/
+  Strict_error_handler strict_handler;
+  if (!thd->lex->is_ignore() && thd->is_strict_mode())
+    thd->push_internal_handler(&strict_handler);
+
+#ifdef WITH_WSREP
   TABLE *find_temporary_table(THD *thd, const TABLE_LIST *tl);
 
   if ((!thd->is_current_stmt_binlog_format_row() ||
        !find_temporary_table(thd, first_table)))
     {
-      WSREP_TO_ISOLATION_BEGIN_ALTER(((lex->name.str) ? select_lex->db : NULL),
-                                     ((lex->name.str) ? lex->name.str : NULL),
-                                     first_table,
-                                     &alter_info);
+      WSREP_TO_ISOLATION_BEGIN(((lex->name.str) ? select_lex->db : NULL),
+			       ((lex->name.str) ? lex->name.str : NULL),
+			       first_table);
     }
 #endif /* WITH_WSREP */
-  result= mysql_alter_table(thd, select_lex->db, lex->name.str,
-                            &create_info,
-                            first_table,
-                            &alter_info,
-                            select_lex->order_list.elements,
-                            select_lex->order_list.first,
-                            lex->ignore);
 
+  result= mysql_alter_table(thd, select_lex->db, lex->name.str,
+                            &create_info, first_table, &alter_info);
+
+  if (!thd->lex->is_ignore() && thd->is_strict_mode())
+    thd->pop_internal_handler();
   DBUG_RETURN(result);
+
 #ifdef WITH_WSREP
 error:
+
+  if (!thd->lex->is_ignore() && thd->is_strict_mode())
+    thd->pop_internal_handler();
+
   {
     WSREP_WARN("ALTER TABLE isolation failure");
     DBUG_RETURN(TRUE);
@@ -401,9 +507,9 @@ error:
 bool Sql_cmd_discard_import_tablespace::execute(THD *thd)
 {
   /* first SELECT_LEX (have special meaning for many of non-SELECTcommands) */
-  SELECT_LEX *select_lex= &thd->lex->select_lex;
+  SELECT_LEX *select_lex= thd->lex->select_lex;
   /* first table of first SELECT_LEX */
-  TABLE_LIST *table_list= (TABLE_LIST*) select_lex->table_list.first;
+  TABLE_LIST *table_list= select_lex->get_table_list();
 
   if (check_access(thd, ALTER_ACL, table_list->db,
                    &table_list->grant.privilege,
@@ -414,7 +520,7 @@ bool Sql_cmd_discard_import_tablespace::execute(THD *thd)
   if (check_grant(thd, ALTER_ACL, table_list, false, UINT_MAX, false))
     return true;
 
-  thd->set_slow_log_for_admin_command();
+  thd->enable_slow_log= opt_log_slow_admin_statements;
 
   /*
     Check if we attempt to alter mysql.slow_log or
@@ -422,14 +528,13 @@ bool Sql_cmd_discard_import_tablespace::execute(THD *thd)
     it is the case.
     TODO: this design is obsolete and will be removed.
   */
-  int table_kind= check_if_log_table(table_list->db_length, table_list->db,
-                                     table_list->table_name_length,
-                                     table_list->table_name, false);
+  enum_log_table_type table_kind=
+    query_logger.check_if_log_table(table_list, false);
 
-  if (table_kind)
+  if (table_kind != QUERY_LOG_NONE)
   {
-    /* Disable alter of enabled log tables */
-    if (logger.is_log_table_enabled(table_kind))
+    /* Disable alter of enabled query log tables */
+    if (query_logger.is_log_table_enabled(table_kind))
     {
       my_error(ER_BAD_LOG_STATEMENT, MYF(0), "ALTER");
       return true;
