@@ -1,25 +1,20 @@
 /* Copyright (c) 2010, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
+#include "sql_truncate.h"
+
 #include "debug_sync.h"  // DEBUG_SYNC
 #include "table.h"       // TABLE, FOREIGN_KEY_INFO
 #include "sql_class.h"   // THD
@@ -27,13 +22,13 @@
 #include "sql_table.h"   // write_bin_log
 #include "datadict.h"    // dd_recreate_table()
 #include "lock.h"        // MYSQL_OPEN_* flags
-#include "sql_acl.h"     // DROP_ACL
+#include "auth_common.h" // DROP_ACL
 #include "sql_parse.h"   // check_one_table_access()
-#include "sql_truncate.h"
 #include "sql_show.h"    //append_identifier()
 #ifdef WITH_WSREP
 #include "wsrep_mysqld.h"
 #endif /* WITH_WSREP */
+#include "sql_audit.h"
 
 
 /**
@@ -241,7 +236,7 @@ Sql_cmd_truncate_table::handler_truncate(THD *thd, TABLE_LIST *table_ref,
   }
 
   /* Open the table as it will handle some required preparations. */
-  if (open_and_lock_tables(thd, table_ref, FALSE, flags))
+  if (open_and_lock_tables(thd, table_ref, flags))
     DBUG_RETURN(TRUNCATE_FAILED_SKIP_BINLOG);
 
   /* Whether to truncate regardless of foreign keys. */
@@ -289,7 +284,7 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
   handlerton *table_type= table->s->db_type();
   DBUG_ENTER("recreate_temporary_table");
 
-  memset(static_cast<void*>(&create_info), 0, sizeof(create_info));
+  memset(&create_info, 0, sizeof(create_info));
 
   table->file->info(HA_STATUS_AUTO | HA_STATUS_NO_LOCK);
 
@@ -308,7 +303,7 @@ static bool recreate_temporary_table(THD *thd, TABLE *table)
     on table and schema names.
   */
   ha_create_table(thd, share->normalized_path.str, share->db.str,
-                  share->table_name.str, &create_info, 0, true, true);
+                  share->table_name.str, &create_info, true, true);
 
   if (open_table_uncached(thd, share->path.str, share->db.str,
                           share->table_name.str, true, true))
@@ -443,10 +438,9 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
   if (is_temporary_table(table_ref))
   {
     TABLE *tmp_table= table_ref->table;
-    /*    
-      bug 1313901 : THD::decide_logging_format has not yet been called and
-                    may not be called at all dependig on the engine, so call it
-                    here.
+    /*
+      bug 72475 : THD::decide_logging_format has not yet been called and may
+      not be called at all depending on the engine, so call it here.
     */
     if (thd->decide_logging_format(table_ref) != 0)
       DBUG_RETURN(TRUE);
@@ -460,7 +454,8 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
       if ((error= recreate_temporary_table(thd, tmp_table)))
         binlog_stmt= FALSE; /* No need to binlog failed truncate-by-recreate. */
 
-      DBUG_ASSERT(! thd->transaction.stmt.cannot_safely_rollback());
+      DBUG_ASSERT(! thd->get_transaction()->cannot_safely_rollback(
+        Transaction_ctx::STMT));
     }
     else
     {
@@ -485,16 +480,21 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
     bool hton_can_recreate;
 
 #ifdef WITH_WSREP
-    if (WSREP(thd) && wsrep_to_isolation_begin(thd, 
-                                               table_ref->db, 
-                                               table_ref->table_name, NULL))
-        DBUG_RETURN(TRUE);
+    error= true;
+    WSREP_TO_ISOLATION_BEGIN(table_ref->db, table_ref->table_name, NULL);
 #endif /* WITH_WSREP */
+
     if (lock_table(thd, table_ref, &hton_can_recreate))
       DBUG_RETURN(TRUE);
 
     if (hton_can_recreate)
     {
+#ifndef EMBEDDED_LIBRARY
+      if (mysql_audit_table_access_notify(thd, table_ref))
+      {
+        DBUG_RETURN(true);
+      }
+#endif /* !EMBEDDED_LIBRARY */
      /*
         The storage engine can truncate the table by creating an
         empty table with the same structure.
@@ -512,6 +512,8 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
       /*
         The engine does not support truncate-by-recreate.
         Attempt to use the handler truncate method.
+        MYSQL_AUDIT_TABLE_ACCESS_READ audit event is generated when opening
+        tables using open_tables function.
       */
       error= handler_truncate(thd, table_ref, FALSE);
 
@@ -534,12 +536,12 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
       query_cache_invalidate does not need a valid TABLE object.
     */
     table_ref->table= NULL;
-    query_cache_invalidate3(thd, table_ref, FALSE);
+    query_cache.invalidate(thd, table_ref, FALSE);
   }
 
   /* DDL is logged in statement format, regardless of binlog format. */
   if (binlog_stmt)
-    error|= write_bin_log(thd, !error, thd->query(), thd->query_length());
+    error|= write_bin_log(thd, !error, thd->query().str, thd->query().length);
 
   /*
     A locked table ticket was upgraded to a exclusive lock. After the
@@ -549,6 +551,9 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
   if (m_ticket_downgrade)
     m_ticket_downgrade->downgrade_lock(MDL_SHARED_NO_READ_WRITE);
 
+#ifdef WITH_WSREP
+error:
+#endif /* WITH_WSREP */
   DBUG_RETURN(error);
 }
 
@@ -563,14 +568,77 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
 bool Sql_cmd_truncate_table::execute(THD *thd)
 {
   bool res= TRUE;
-  TABLE_LIST *first_table= thd->lex->select_lex.table_list.first;
+  TABLE_LIST *first_table= thd->lex->select_lex->table_list.first;
   DBUG_ENTER("Sql_cmd_truncate_table::execute");
 
+#ifdef WITH_WSREP
+  char path[FN_REFLEN + 1];
+  enum legacy_db_type db_type= DB_TYPE_UNKNOWN;
+  (void) build_table_filename(path, sizeof(path) - 1, first_table->db,
+                              first_table->table_name, reg_ext, 0);
+  dd_frm_type(thd, path, &db_type);
+
+  bool is_system_db= (first_table &&
+                      ((strcmp(first_table->db, "mysql") == 0) ||
+                       (strcmp(first_table->db, "information_schema") == 0)));
+
+  if (db_type != DB_TYPE_INNODB             &&
+      db_type != DB_TYPE_UNKNOWN            &&
+      db_type != DB_TYPE_PERFORMANCE_SCHEMA &&
+      !is_system_db                         &&
+      !is_temporary_table(first_table))
+  {
+    bool block= false;
+
+    switch(pxc_strict_mode)
+    {
+    case PXC_STRICT_MODE_DISABLED:
+      break;
+    case PXC_STRICT_MODE_PERMISSIVE:
+      WSREP_WARN("Percona-XtraDB-Cluster doesn't recommend use of"
+                 " TRUNCATE command on a table (%s.%s) that resides in"
+                 " non-transactional storage engine"
+                 " with pxc_strict_mode = PERMISSIVE",
+                 first_table->db, first_table->table_name);
+      push_warning_printf(
+        thd, Sql_condition::SL_WARNING, ER_UNKNOWN_ERROR,
+        "Percona-XtraDB-Cluster doesn't recommend use of"
+        " TRUNCATE command on a table (%s.%s) that resides in"
+        " non-transactional storage engine"
+        " with pxc_strict_mode = PERMISSIVE",
+        first_table->db, first_table->table_name);
+      break;
+    case PXC_STRICT_MODE_ENFORCING:
+    case PXC_STRICT_MODE_MASTER:
+    default:
+      block= true;
+      WSREP_ERROR("Percona-XtraDB-Cluster prohibits use of"
+                  " TRUNCATE command on a table (%s.%s) that resides in"
+                  " non-transactional storage engine"
+                  " with pxc_strict_mode = ENFORCING or MASTER",
+                  first_table->db, first_table->table_name);
+      char message[1024];
+      sprintf(message,
+              "Percona-XtraDB-Cluster prohibits use of"
+              " TRUNCATE command on a table (%s.%s) that resides in"
+              " non-transactional storage engine"
+              " with pxc_strict_mode = ENFORCING or MASTER",
+              first_table->db, first_table->table_name);
+      my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+      break;
+    }
+ 
+    if (block)
+      DBUG_RETURN(res);
+  }
+#endif /* WITH_WSREP */
+
   if (check_one_table_access(thd, DROP_ACL, first_table))
     DBUG_RETURN(res);
 
   if (! (res= truncate_table(thd, first_table)))
     my_ok(thd);
+
   DBUG_RETURN(res);
 }
 
