@@ -15,19 +15,16 @@
 
 #include <my_global.h>
 #include <violite.h>
-#include <sql_priv.h>
 #include <sql_class.h>
-#include <my_pthread.h>
-#include <scheduler.h>
 #include <sql_connect.h>
 #include <mysqld.h>
 #include <debug_sync.h>
 #include <time.h>
 #include <sql_plist.h>
 #include <threadpool.h>
-#include <global_threads.h>
 #include <mysql/thread_pool_priv.h>             // thd_is_transaction_active()
 #include <time.h>
+#include <mysqld_thd_manager.h>
 #ifdef __linux__
 #include <sys/epoll.h>
 typedef struct epoll_event native_event;
@@ -41,6 +38,8 @@ typedef struct kevent native_event;
 typedef port_event_t native_event;
 #endif
 
+#include <conn_handler/connection_handler_impl.h>
+
 /** Maximum number of native events a listener can read in one go */
 #define MAX_EVENTS 1024
 
@@ -115,7 +114,6 @@ worker_list_t;
 
 struct connection_t
 {
-
   THD *thd;
   thread_group_t *thread_group;
   connection_t *next_in_queue;
@@ -188,6 +186,7 @@ static int  create_worker(thread_group_t *thread_group);
 static void *worker_main(void *param);
 static void check_stall(thread_group_t *thread_group);
 static void connection_abort(connection_t *connection);
+void tp_post_kill_notification(THD *thd);
 static void set_wait_timeout(connection_t *connection);
 static void set_next_timeout_check(ulonglong abstime);
 static void print_pool_blocked_message(bool);
@@ -249,7 +248,6 @@ static int io_poll_create()
   return epoll_create(1);
 }
 
-
 static int io_poll_associate_fd(int pollfd, int fd, void *data)
 {
   struct epoll_event ev;
@@ -259,15 +257,13 @@ static int io_poll_associate_fd(int pollfd, int fd, void *data)
   return epoll_ctl(pollfd, EPOLL_CTL_ADD,  fd, &ev);
 }
 
-
-
 static int io_poll_start_read(int pollfd, int fd, void *data)
 {
   struct epoll_event ev;
   ev.data.u64= 0; /* Keep valgrind happy */
   ev.data.ptr= data;
   ev.events=  EPOLLIN|EPOLLET|EPOLLERR|EPOLLRDHUP|EPOLLONESHOT;
-  return epoll_ctl(pollfd, EPOLL_CTL_MOD,  fd, &ev); 
+  return epoll_ctl(pollfd, EPOLL_CTL_MOD, fd, &ev);
 }
 
 static int io_poll_disassociate_fd(int pollfd, int fd)
@@ -276,7 +272,6 @@ static int io_poll_disassociate_fd(int pollfd, int fd)
   return epoll_ctl(pollfd, EPOLL_CTL_DEL,  fd, &ev);
 }
 
-
 /*
  Wrapper around epoll_wait.
  NOTE - in case of EINTR, it restarts with original timeout. Since we use
@@ -286,7 +281,7 @@ static int io_poll_wait(int pollfd, native_event *native_events, int maxevents,
                         int timeout_ms)
 {
   int ret;
-  do 
+  do
   {
     ret = epoll_wait(pollfd, native_events, maxevents, timeout_ms);
   }
@@ -294,7 +289,6 @@ static int io_poll_wait(int pollfd, native_event *native_events, int maxevents,
   return ret;
 }
 
-
 static void *native_event_get_userdata(native_event *event)
 {
   return event->data.ptr;
@@ -314,7 +308,6 @@ static int io_poll_start_read(int pollfd, int fd, void *data)
   return kevent(pollfd, &ke, 1, 0, 0, 0); 
 }
 
-
 static int io_poll_associate_fd(int pollfd, int fd, void *data)
 {
   struct kevent ke;
@@ -323,7 +316,6 @@ static int io_poll_associate_fd(int pollfd, int fd, void *data)
   return io_poll_start_read(pollfd,fd, data); 
 }
 
-
 static int io_poll_disassociate_fd(int pollfd, int fd)
 {
   struct kevent ke;
@@ -331,7 +323,6 @@ static int io_poll_disassociate_fd(int pollfd, int fd)
   return kevent(pollfd, &ke, 1, 0, 0, 0);
 }
 
-
 static int io_poll_wait(int pollfd, struct kevent *events, int maxevents,
                         int timeout_ms)
 {
@@ -436,23 +427,16 @@ inline bool too_many_busy_threads(thread_group_t *thread_group)
 /*
    Checks if a given connection is eligible to enter the high priority queue
    based on its current thread_pool_high_prio_mode value, available high
-   priority tickets and transactional state and whether any locks are held.
+   priority tickets and transactional state.
 */
 
-inline bool connection_is_high_prio(const connection_t *c)
+inline bool connection_is_high_prio(connection_t *c)
 {
   const ulong mode= c->thd->variables.threadpool_high_prio_mode;
 
   return (mode == TP_HIGH_PRIO_MODE_STATEMENTS) ||
-    (mode == TP_HIGH_PRIO_MODE_TRANSACTIONS && c->tickets > 0 &&
-     (thd_is_transaction_active(c->thd) ||
-      c->thd->variables.option_bits & OPTION_TABLE_LOCK ||
-      c->thd->locked_tables_mode != LTM_NONE ||
-      c->thd->mdl_context.has_locks() ||
-      c->thd->global_read_lock.is_acquired() ||
-      c->thd->backup_tables_lock.is_acquired() ||
-      c->thd->backup_binlog_lock.is_acquired() ||
-      c->thd->ull_hash.records > 0));
+    (mode == TP_HIGH_PRIO_MODE_TRANSACTIONS &&
+     c->tickets > 0 && thd_is_transaction_active(c->thd));
 }
 
 } // namespace
@@ -482,44 +466,33 @@ static connection_t *queue_get(thread_group_t *thread_group)
 }
 
 
-/* 
-  Handle wait timeout : 
-  Find connections that have been idle for too long and kill them.
-  Also, recalculate time when next timeout check should run.
-*/
-
-static void timeout_check(pool_timer_t *timer)
+class Thd_timeout_checker : public Do_THD_Impl
 {
-  std::set<THD*> global_thread_list_copy;
-  DBUG_ENTER("timeout_check");
+private:
+  pool_timer_t* m_timer;
 
-  mysql_mutex_lock(&LOCK_thd_remove);
-  copy_global_thread_list(&global_thread_list_copy);
+public:
+  Thd_timeout_checker(pool_timer_t *timer): m_timer(timer) { }
 
-  Thread_iterator it= global_thread_list_copy.begin();
-  Thread_iterator end= global_thread_list_copy.end();
+  virtual ~Thd_timeout_checker() { }
 
-  /* Reset next timeout check, it will be recalculated in the loop below */
-  my_atomic_fas64((volatile int64*)&timer->next_timeout_check, ULONGLONG_MAX);
-
-  THD *thd;
-  for ( ; it != end; ++it)
+  virtual void operator() (THD* thd)
   {
-    thd= (*it);
-    if (thd->net.reading_or_writing != 1)
-      continue;
- 
+    if (thd_get_net_read_write(thd) != 1)
+      return;
+
     connection_t *connection= (connection_t *)thd->event_scheduler.data;
     if (!connection)
     {
-      /* 
+      /*
         Connection does not have scheduler data. This happens for example
-        if THD belongs to a different scheduler, that is listening to extra_port.
+        if THD belongs to a different scheduler, that is listening to
+        extra_port.
       */
-      continue;
+      return;
     }
 
-    if(connection->abs_wait_timeout < timer->current_microtime)
+    if(connection->abs_wait_timeout < m_timer->current_microtime)
     {
       /* Wait timeout exceeded, kill connection. */
       mysql_mutex_lock(&thd->LOCK_thd_data);
@@ -527,12 +500,30 @@ static void timeout_check(pool_timer_t *timer)
       tp_post_kill_notification(thd);
       mysql_mutex_unlock(&thd->LOCK_thd_data);
     }
-    else 
+    else
     {
       set_next_timeout_check(connection->abs_wait_timeout);
     }
   }
-  mysql_mutex_unlock(&LOCK_thd_remove);
+};
+
+/*
+  Handle wait timeout :
+  Find connections that have been idle for too long and kill them.
+  Also, recalculate time when next timeout check should run.
+*/
+
+static void timeout_check(pool_timer_t *timer)
+{
+  DBUG_ENTER("timeout_check");
+
+  /* Reset next timeout check, it will be recalculated in the loop below */
+  my_atomic_fas64((volatile int64*)&timer->next_timeout_check, ULLONG_MAX);
+
+  Thd_timeout_checker thd_timeout_checker(timer);
+  Global_THD_manager::get_instance()
+    ->do_for_all_thd_copy(&thd_timeout_checker);
+
   DBUG_VOID_RETURN;
 }
 
@@ -561,7 +552,7 @@ static void* timer_thread(void *param)
 
   my_thread_init();
   DBUG_ENTER("timer_thread");
-  timer->next_timeout_check= ULONGLONG_MAX;
+  timer->next_timeout_check= ULLONG_MAX;
   timer->current_microtime= my_microsecond_getsystime();
 
   for(;;)
@@ -569,7 +560,7 @@ static void* timer_thread(void *param)
     struct timespec ts;
     int err;
 
-    set_timespec_nsec(ts,timer->tick_interval * 1000000ULL);
+    set_timespec_nsec(&ts, timer->tick_interval*1000000);
     mysql_mutex_lock(&timer->mutex);
     err= mysql_cond_timedwait(&timer->cond, &timer->mutex, &ts);
     if (timer->shutdown)
@@ -612,7 +603,7 @@ static bool queues_are_empty(thread_group_t *tg)
           (tg->queue.is_empty() || too_many_busy_threads(tg)));
 }
 
-void check_stall(thread_group_t *thread_group)
+static void check_stall(thread_group_t *thread_group)
 {
   if (mysql_mutex_trylock(&thread_group->mutex) != 0)
   {
@@ -677,10 +668,10 @@ void check_stall(thread_group_t *thread_group)
 
 static void start_timer(pool_timer_t* timer)
 {
-  pthread_t thread_id;
+  my_thread_handle thread_id;
   DBUG_ENTER("start_timer");
   mysql_mutex_init(key_timer_mutex,&timer->mutex, NULL);
-  mysql_cond_init(key_timer_cond, &timer->cond, NULL);
+  mysql_cond_init(key_timer_cond, &timer->cond);
   timer->shutdown = false;
   mysql_thread_create(key_timer_thread,&thread_id, NULL, timer_thread, timer);
   DBUG_VOID_RETURN;
@@ -871,7 +862,7 @@ static void add_thread_count(thread_group_t *thread_group, int32 count)
 
 static int create_worker(thread_group_t *thread_group)
 {
-  pthread_t thread_id;
+  my_thread_handle thread_id;
   bool max_threads_reached= false;
   int err;
   
@@ -890,12 +881,12 @@ static int create_worker(thread_group_t *thread_group)
   if (!err)
   {
     thread_group->last_thread_creation_time=my_microsecond_getsystime();
-    thread_created++;
+    Global_THD_manager::get_instance()->inc_thread_created();
     add_thread_count(thread_group, 1);
   }
   else
   {
-    my_errno= errno;
+    set_my_errno(errno);
   }
 
 end:
@@ -982,8 +973,6 @@ static int wake_or_create_thread(thread_group_t *thread_group)
   DBUG_RETURN(-1);
 }
 
-
-
 static int thread_group_init(thread_group_t *thread_group,
                              pthread_attr_t* thread_attr)
 {
@@ -996,7 +985,6 @@ static int thread_group_init(thread_group_t *thread_group,
   DBUG_RETURN(0);
 }
 
-
 static void thread_group_destroy(thread_group_t *thread_group)
 {
   mysql_mutex_destroy(&thread_group->mutex);
@@ -1059,7 +1047,6 @@ static void thread_group_close(thread_group_t *thread_group)
 
   if (pipe(thread_group->shutdown_pipe))
   {
-    mysql_mutex_unlock(&thread_group->mutex);
     DBUG_VOID_RETURN;
   }
   
@@ -1067,15 +1054,11 @@ static void thread_group_close(thread_group_t *thread_group)
   if (io_poll_associate_fd(thread_group->pollfd, 
       thread_group->shutdown_pipe[0], NULL))
   {
-    mysql_mutex_unlock(&thread_group->mutex);
     DBUG_VOID_RETURN;
   }
   char c= 0;
   if (write(thread_group->shutdown_pipe[1], &c, 1) < 0)
-  {
-    mysql_mutex_unlock(&thread_group->mutex);
     DBUG_VOID_RETURN;
-  }
 
   /* Wake all workers. */
   while(wake_thread(thread_group) == 0) 
@@ -1310,15 +1293,17 @@ static void wait_end(thread_group_t *thread_group)
 static connection_t *alloc_connection(THD *thd)
 {
   DBUG_ENTER("alloc_connection");
-  
-  connection_t* connection = (connection_t *)my_malloc(sizeof(connection_t),0);
+
+  connection_t* connection = (connection_t *)
+      my_malloc(key_memory_thread_pool_connection,
+                sizeof(connection_t),0);
   if (connection)
   {
     connection->thd = thd;
     connection->waiting= false;
     connection->logged_in= false;
     connection->bound_to_poll_descriptor= false;
-    connection->abs_wait_timeout= ULONGLONG_MAX;
+    connection->abs_wait_timeout= ULLONG_MAX;
     connection->tickets = 0;
   }
   DBUG_RETURN(connection);
@@ -1330,39 +1315,46 @@ static connection_t *alloc_connection(THD *thd)
   Add a new connection to thread pool..
 */
 
-void tp_add_connection(THD *thd)
+bool Thread_pool_connection_handler::add_connection(Channel_info *channel_info)
 {
-  DBUG_ENTER("tp_add_connection");
-  
-  add_global_thread(thd);
-  mysql_mutex_unlock(&LOCK_thread_count);
+  DBUG_ENTER("Thread_pool_connection_handler::add_connection");
+
+  THD* thd= channel_info->create_thd();
+  delete channel_info;
+
+  thd->set_new_thread_id();
+  thd->start_utime= thd->thr_create_utime= my_micro_time();
+
+  thd->scheduler= &tp_event_functions;
+
+  Global_THD_manager::get_instance()->add_thd(thd);
+
   connection_t *connection= alloc_connection(thd);
-  if (connection)
-  {
-    thd->event_scheduler.data= connection;
-      
-    /* Assign connection to a group. */
-    thread_group_t *group= 
-      &all_groups[thd->thread_id%group_count];
-    
-    connection->thread_group=group;
-      
-    mysql_mutex_lock(&group->mutex);
-    group->connection_count++;
-    mysql_mutex_unlock(&group->mutex);
-    
-    /*
-       Add connection to the work queue.Actual logon 
-       will be done by a worker thread.
-    */
-    queue_put(group, connection);
-  }
-  else
+
+  if (!connection)
   {
     /* Allocation failed */
     threadpool_remove_connection(thd);
-  } 
-  DBUG_VOID_RETURN;
+    DBUG_RETURN(true);
+  }
+
+  thd->event_scheduler.data= connection;
+
+  /* Assign connection to a group. */
+  thread_group_t *group= &all_groups[thd->thread_id() % group_count];
+
+  connection->thread_group=group;
+
+  mysql_mutex_lock(&group->mutex);
+  group->connection_count++;
+  mysql_mutex_unlock(&group->mutex);
+
+  /*
+    Add connection to the work queue. Actual login
+    will be done by a worker thread.
+  */
+  queue_put(group, connection);
+  DBUG_RETURN(false);
 }
 
 
@@ -1375,8 +1367,8 @@ static void connection_abort(connection_t *connection)
   DBUG_ENTER("connection_abort");
   thread_group_t *group= connection->thread_group;
 
-  threadpool_remove_connection(connection->thd); 
-  
+  threadpool_remove_connection(connection->thd);
+
   mysql_mutex_lock(&group->mutex);
   group->connection_count--;
   mysql_mutex_unlock(&group->mutex);
@@ -1395,9 +1387,10 @@ void tp_post_kill_notification(THD *thd)
   DBUG_ENTER("tp_post_kill_notification");
   if (current_thd == thd || thd->system_thread)
     DBUG_VOID_RETURN;
-  
-  if (thd->net.vio)
-    vio_cancel(thd->net.vio, SHUT_RD);
+
+  Vio* vio= thd->get_protocol_classic()->get_vio();
+  if (vio)
+    vio_cancel(vio, SHUT_RD);
   DBUG_VOID_RETURN;
 }
 
@@ -1470,7 +1463,7 @@ static void set_wait_timeout(connection_t *c)
 
   c->abs_wait_timeout= pool_timer.current_microtime +
     1000LL*pool_timer.tick_interval +
-    1000000LL*c->thd->get_wait_timeout();
+    1000000LL*c->thd->variables.net_wait_timeout;
 
   set_next_timeout_check(c->abs_wait_timeout);
   DBUG_VOID_RETURN;
@@ -1489,7 +1482,8 @@ static int change_group(connection_t *c,
  thread_group_t *new_group)
 { 
   int ret= 0;
-  int fd = mysql_socket_getfd(c->thd->net.vio->mysql_socket);
+  Vio* vio= c->thd->get_protocol_classic()->get_vio();
+  int fd = mysql_socket_getfd(vio->mysql_socket);
 
   DBUG_ASSERT(c->thread_group == old_group);
 
@@ -1516,8 +1510,9 @@ static int change_group(connection_t *c,
 
 
 static int start_io(connection_t *connection)
-{ 
-  int fd = mysql_socket_getfd(connection->thd->net.vio->mysql_socket);
+{
+  Vio* vio= connection->thd->get_protocol_classic()->get_vio();
+  int fd = mysql_socket_getfd(vio->mysql_socket);
 
   /*
     Usually, connection will stay in the same group for the entire
@@ -1530,7 +1525,7 @@ static int start_io(connection_t *connection)
     on thread_id and current group count, and migrate if necessary.
   */ 
   thread_group_t *group = 
-    &all_groups[connection->thd->thread_id%group_count];
+    &all_groups[connection->thd->thread_id() % group_count];
 
   if (group != connection->thread_group)
   {
@@ -1550,8 +1545,6 @@ static int start_io(connection_t *connection)
   return io_poll_start_read(group->pollfd, fd, connection);
 }
 
-
-
 static void handle_event(connection_t *connection)
 {
 
@@ -1563,7 +1556,7 @@ static void handle_event(connection_t *connection)
     err= threadpool_add_connection(connection->thd);
     connection->logged_in= true;
   }
-  else 
+  else
   {
     err= threadpool_process_request(connection->thd);
   }
@@ -1591,7 +1584,6 @@ static void *worker_main(void *param)
 {
   
   worker_thread_t this_thread;
-  pthread_detach_this_thread();
   my_thread_init();
   
   DBUG_ENTER("worker_main");
@@ -1599,12 +1591,12 @@ static void *worker_main(void *param)
   thread_group_t *thread_group = (thread_group_t *)param;
 
   /* Init per-thread structure */
-  mysql_cond_init(key_worker_cond, &this_thread.cond, NULL);
+  mysql_cond_init(key_worker_cond, &this_thread.cond);
   this_thread.thread_group= thread_group;
   this_thread.event_count=0;
 
 #ifdef HAVE_PSI_THREAD_INTERFACE
-    PSI_THREAD_CALL(set_thread_user_host)
+    PSI_THREAD_CALL(set_thread_account)
       (NULL, 0, NULL, 0);
 #endif
 
@@ -1613,7 +1605,7 @@ static void *worker_main(void *param)
   {
     connection_t *connection;
     struct timespec ts;
-    set_timespec(ts,threadpool_idle_timeout);
+    set_timespec(&ts, threadpool_idle_timeout);
     connection = get_event(&this_thread, thread_group, &ts);
     if (!connection)
       break;
@@ -1643,7 +1635,6 @@ bool tp_init()
 {
   DBUG_ENTER("tp_init");
   threadpool_started= true;
-  scheduler_init();
 
   for(uint i=0; i < array_elements(all_groups); i++)
   {
@@ -1777,29 +1768,26 @@ A likely cause of pool blocks are clients that lock resources for long time. \
 
 static void print_pool_blocked_message(bool max_threads_reached)
 {
-  ulonglong now= my_microsecond_getsystime();
+  ulonglong now;
   static bool msg_written;
-
+  
+  now= my_microsecond_getsystime();
   if (pool_block_start == 0)
   {
     pool_block_start= now;
-    msg_written= false;
+    msg_written = false;
+    return;
   }
-
-  if (!msg_written
-      && ((now > pool_block_start + BLOCK_MSG_DELAY)
-          || (now == pool_block_start)))
+  
+  if (now > pool_block_start + BLOCK_MSG_DELAY && !msg_written)
   {
     if (max_threads_reached)
       sql_print_error(MAX_THREADS_REACHED_MSG);
     else
-      sql_print_error(CREATE_THREAD_ERROR_MSG, my_errno);
-
-    if (now > pool_block_start)
-    {
-      sql_print_information("Threadpool has been blocked for %u seconds\n",
-                            (uint)((now - pool_block_start)/1000000));
-    }
+      sql_print_error(CREATE_THREAD_ERROR_MSG, my_errno());
+    
+    sql_print_information("Threadpool has been blocked for %u seconds\n",
+      (uint)((now- pool_block_start)/1000000));
     /* avoid reperated messages for the same blocking situation */
     msg_written= true;
   }
