@@ -1,31 +1,60 @@
-/* Copyright (c) 2000, 2013, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2000, 2016, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
    51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA */
 
 
-#include "sql_priv.h"
 #include "transaction.h"
 #include "rpl_handler.h"
 #include "debug_sync.h"         // DEBUG_SYNC
-#include "sql_acl.h"            // SUPER_ACL
+#include "auth_common.h"            // SUPER_ACL
+#include <pfs_transaction_provider.h>
+#include <mysql/psi/mysql_transaction.h>
+#include "rpl_context.h"
+#include "sql_class.h"
+#include "log.h"
+#include "binlog.h"
+
+/**
+  Helper: Tell tracker (if any) that transaction ended.
+*/
+void trans_track_end_trx(THD *thd)
+{
+  if (thd->variables.session_track_transaction_info > TX_TRACK_NONE)
+  {
+    ((Transaction_state_tracker *)
+     thd->session_tracker.get_tracker(TRANSACTION_INFO_TRACKER))->end_trx(thd);
+  }
+}
+
+/**
+  Helper: transaction ended, SET TRANSACTION one-shot variables
+  revert to session values. Let the transaction state tracker know.
+*/
+void trans_reset_one_shot_chistics(THD *thd)
+{
+  if (thd->variables.session_track_transaction_info > TX_TRACK_NONE)
+  {
+    Transaction_state_tracker *tst= (Transaction_state_tracker *)
+      thd->session_tracker.get_tracker(TRANSACTION_INFO_TRACKER);
+
+    tst->set_read_flags(thd, TX_READ_INHERIT);
+    tst->set_isol_level(thd, TX_ISOL_INHERIT);
+  }
+
+  thd->tx_isolation= (enum_tx_isolation) thd->variables.tx_isolation;
+  thd->tx_read_only= thd->variables.tx_read_only;
+}
 
 /**
   Check if we have a condition where the transaction state must
@@ -39,85 +68,24 @@
 
 bool trans_check_state(THD *thd)
 {
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
   DBUG_ENTER("trans_check");
 
   /*
     Always commit statement transaction before manipulating with
     the normal one.
   */
-  DBUG_ASSERT(thd->transaction.stmt.is_empty());
+  DBUG_ASSERT(thd->get_transaction()->is_empty(Transaction_ctx::STMT));
 
   if (unlikely(thd->in_sub_stmt))
+  {
     my_error(ER_COMMIT_NOT_ALLOWED_IN_SF_OR_TRG, MYF(0));
-  if (xa_state != XA_NOTR)
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-  else
-    DBUG_RETURN(FALSE);
-
-  DBUG_RETURN(TRUE);
-}
-
-
-/**
-  Mark a XA transaction as rollback-only if the RM unilaterally
-  rolled back the transaction branch.
-
-  @note If a rollback was requested by the RM, this function sets
-        the appropriate rollback error code and transits the state
-        to XA_ROLLBACK_ONLY.
-
-  @return TRUE if transaction was rolled back or if the transaction
-          state is XA_ROLLBACK_ONLY. FALSE otherwise.
-*/
-static bool xa_trans_rolled_back(XID_STATE *xid_state)
-{
-  if (xid_state->rm_error)
-  {
-    switch (xid_state->rm_error) {
-    case ER_LOCK_WAIT_TIMEOUT:
-      my_error(ER_XA_RBTIMEOUT, MYF(0));
-      break;
-    case ER_LOCK_DEADLOCK:
-      my_error(ER_XA_RBDEADLOCK, MYF(0));
-      break;
-    default:
-      my_error(ER_XA_RBROLLBACK, MYF(0));
-    }
-    xid_state->xa_state= XA_ROLLBACK_ONLY;
+    DBUG_RETURN(true);
   }
 
-  return (xid_state->xa_state == XA_ROLLBACK_ONLY);
-}
+  if (thd->get_transaction()->xid_state()->check_in_xa(true))
+    DBUG_RETURN(true);
 
-
-/**
-  Rollback the active XA transaction.
-
-  @note Resets rm_error before calling ha_rollback(), so
-        the thd->transaction.xid structure gets reset
-        by ha_rollback() / THD::transaction::cleanup().
-
-  @return TRUE if the rollback failed, FALSE otherwise.
-*/
-
-static bool xa_trans_force_rollback(THD *thd)
-{
-  /*
-    We must reset rm_error before calling ha_rollback(),
-    so thd->transaction.xid structure gets reset
-    by ha_rollback()/THD::transaction::cleanup().
-  */
-  thd->transaction.xid_state.rm_error= 0;
-#ifdef WITH_WSREP
-  wsrep_register_hton(thd, TRUE);
-#endif /* WITH_WSREP */
-  if (ha_rollback_trans(thd, true))
-  {
-    my_error(ER_XAER_RMERR, MYF(0));
-    return true;
-  }
-  return false;
+  DBUG_RETURN(false);
 }
 
 
@@ -137,11 +105,17 @@ static bool xa_trans_force_rollback(THD *thd)
 bool trans_begin(THD *thd, uint flags)
 {
   int res= FALSE;
+  Transaction_state_tracker *tst= NULL;
+
   DBUG_ENTER("trans_begin");
 
   if (trans_check_state(thd))
     DBUG_RETURN(TRUE);
 
+  if (thd->variables.session_track_transaction_info > TX_TRACK_NONE)
+    tst= (Transaction_state_tracker *)
+      thd->session_tracker.get_tracker(TRANSACTION_INFO_TRACKER);
+
   thd->locked_tables_list.unlock_locked_tables(thd);
 
   DBUG_ASSERT(!thd->locked_tables_mode);
@@ -163,7 +137,7 @@ bool trans_begin(THD *thd, uint flags)
   }
 
   thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
+  thd->get_transaction()->reset_unsafe_rollback_flags(Transaction_ctx::SESSION);
 
   if (res)
     DBUG_RETURN(TRUE);
@@ -178,7 +152,11 @@ bool trans_begin(THD *thd, uint flags)
   DBUG_ASSERT(!((flags & MYSQL_START_TRANS_OPT_READ_ONLY) &&
                 (flags & MYSQL_START_TRANS_OPT_READ_WRITE)));
   if (flags & MYSQL_START_TRANS_OPT_READ_ONLY)
+  {
     thd->tx_read_only= true;
+    if (tst)
+      tst->set_read_flags(thd, TX_READ_ONLY);
+  }
   else if (flags & MYSQL_START_TRANS_OPT_READ_WRITE)
   {
     /*
@@ -187,16 +165,15 @@ bool trans_begin(THD *thd, uint flags)
       Implicitly starting a RW transaction is allowed for backward
       compatibility.
     */
-    bool enforce_ro= true;
-    if (!opt_super_readonly)
-      enforce_ro= !(thd->security_ctx->master_access & SUPER_ACL);
-    if (opt_readonly && enforce_ro)
-    {
-      my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0),
-               opt_super_readonly ? "--read-only (super)" : "--read-only");
+    if (check_readonly(thd, true))
       DBUG_RETURN(true);
-    }
     thd->tx_read_only= false;
+    /*
+      This flags that tx_read_only was set explicitly, rather than
+      just from the session's default.
+    */
+    if (tst)
+      tst->set_read_flags(thd, TX_READ_WRITE);
   }
 
 #ifdef WITH_WSREP
@@ -204,6 +181,10 @@ bool trans_begin(THD *thd, uint flags)
   if (WSREP_CLIENT(thd) && wsrep_sync_wait(thd))
     DBUG_RETURN(TRUE);
 #endif /* WITH_WSREP */
+  DBUG_EXECUTE_IF("dbug_set_high_prio_trx", {
+    DBUG_ASSERT(thd->tx_priority==0);
+    thd->tx_priority= 1;
+  });
 
   thd->variables.option_bits|= OPTION_BEGIN;
   thd->server_status|= SERVER_STATUS_IN_TRANS;
@@ -211,9 +192,36 @@ bool trans_begin(THD *thd, uint flags)
     thd->server_status|= SERVER_STATUS_IN_TRANS_READONLY;
   DBUG_PRINT("info", ("setting SERVER_STATUS_IN_TRANS"));
 
+  if (tst)
+    tst->add_trx_state(thd, TX_EXPLICIT);
+
   /* ha_start_consistent_snapshot() relies on OPTION_BEGIN flag set. */
   if (flags & MYSQL_START_TRANS_OPT_WITH_CONS_SNAPSHOT)
+  {
+    if (tst)
+      tst->add_trx_state(thd, TX_WITH_SNAPSHOT);
     res= ha_start_consistent_snapshot(thd);
+  }
+
+  /*
+    Register transaction start in performance schema if not done already.
+    We handle explicitly started transactions here, implicitly started
+    transactions (and single-statement transactions in autocommit=1 mode)
+    are handled in trans_register_ha().
+    We can't handle explicit transactions in the same way as implicit
+    because we want to correctly attribute statements which follow
+    BEGIN but do not touch any transactional tables.
+  */
+#ifdef HAVE_PSI_TRANSACTION_INTERFACE
+  if (thd->m_transaction_psi == NULL)
+  {
+    thd->m_transaction_psi= MYSQL_START_TRANSACTION(&thd->m_transaction_state,
+                                                 NULL, NULL, thd->tx_isolation,
+                                                 thd->tx_read_only, false);
+    DEBUG_SYNC(thd, "after_set_transaction_psi_before_set_transaction_gtid");
+    gtid_set_performance_schema_values(thd);
+  }
+#endif
 
   DBUG_RETURN(MY_TEST(res));
 }
@@ -233,16 +241,6 @@ bool trans_commit(THD *thd)
   int res;
   DBUG_ENTER("trans_commit");
 
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
-
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
-
   if (trans_check_state(thd))
     DBUG_RETURN(TRUE);
 
@@ -253,13 +251,38 @@ bool trans_commit(THD *thd)
     ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
   DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
   res= ha_commit_trans(thd, TRUE);
+  if (res == FALSE)
+    if (thd->rpl_thd_ctx.session_gtids_ctx().
+        notify_after_transaction_commit(thd))
+      sql_print_warning("Failed to collect GTID to send in the response packet!");
+  /*
+    When gtid mode is enabled, a transaction may cause binlog
+    rotation, which inserts a record into the gtid system table
+    (which is probably a transactional table). Thence, the flag
+    SERVER_STATUS_IN_TRANS may be set again while calling
+    ha_commit_trans(...) Consequently, we need to reset it back,
+    much like we are doing before calling ha_commit_trans(...).
+
+    We would really only need to do this when gtid_mode=on.  However,
+    checking gtid_mode requires holding a lock, which is costly.  So
+    we clear the bit unconditionally.  This has no side effects since
+    if gtid_mode=off the bit is already cleared.
+  */
+  thd->server_status&= ~SERVER_STATUS_IN_TRANS;
   thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
+  thd->get_transaction()->reset_unsafe_rollback_flags(Transaction_ctx::SESSION);
 #ifdef WITH_WSREP
   wsrep_post_commit(thd, TRUE);
 #endif /* WITH_WSREP */
   thd->lex->start_transaction_opt= 0;
 
+  /* The transaction should be marked as complete in P_S. */
+  DBUG_ASSERT(thd->m_transaction_psi == NULL);
+
+  thd->tx_priority= 0;
+
+  trans_track_end_trx(thd);
+
   DBUG_RETURN(MY_TEST(res));
 }
 
@@ -280,24 +303,14 @@ bool trans_commit_implicit(THD *thd)
   bool res= FALSE;
   DBUG_ENTER("trans_commit_implicit");
 
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
-
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
-
   /*
     Ensure that trans_check_state() was called before trans_commit_implicit()
     by asserting that conditions that are checked in the former function are
     true.
   */
-  DBUG_ASSERT(thd->transaction.stmt.is_empty() &&
+  DBUG_ASSERT(thd->get_transaction()->is_empty(Transaction_ctx::STMT) &&
               !thd->in_sub_stmt &&
-              thd->transaction.xid_state.xa_state == XA_NOTR);
+              !thd->get_transaction()->xid_state()->check_in_xa(false));
 
   if (thd->in_multi_stmt_transaction_mode() ||
       (thd->variables.option_bits & OPTION_TABLE_LOCK))
@@ -319,8 +332,15 @@ bool trans_commit_implicit(THD *thd)
   else if (tc_log)
     tc_log->commit(thd, true);
 
+  if (res == FALSE)
+    if (thd->rpl_thd_ctx.session_gtids_ctx().
+        notify_after_transaction_commit(thd))
+      sql_print_warning("Failed to collect GTID to send in the response packet!");
   thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
+  thd->get_transaction()->reset_unsafe_rollback_flags(Transaction_ctx::SESSION);
+
+  /* The transaction should be marked as complete in P_S. */
+  DBUG_ASSERT(thd->m_transaction_psi == NULL);
 
   /*
     Upon implicit commit, reset the current transaction
@@ -328,8 +348,9 @@ bool trans_commit_implicit(THD *thd)
     @@session.completion_type since it's documented
     to not have any effect on implicit commit.
   */
-  thd->tx_isolation= (enum_tx_isolation) thd->variables.tx_isolation;
-  thd->tx_read_only= thd->variables.tx_read_only;
+  trans_reset_one_shot_chistics(thd);
+
+  trans_track_end_trx(thd);
 
   DBUG_RETURN(res);
 }
@@ -352,16 +373,6 @@ bool trans_rollback(THD *thd)
   thd->wsrep_PA_safe= true;
 #endif /* WITH_WSREP */
 
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
-
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
-
   if (trans_check_state(thd))
     DBUG_RETURN(TRUE);
 
@@ -373,9 +384,17 @@ bool trans_rollback(THD *thd)
   DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
   res= ha_rollback_trans(thd, TRUE);
   thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
+  thd->get_transaction()->reset_unsafe_rollback_flags(
+      Transaction_ctx::SESSION);
   thd->lex->start_transaction_opt= 0;
 
+  /* The transaction should be marked as complete in P_S. */
+  DBUG_ASSERT(thd->m_transaction_psi == NULL);
+
+  thd->tx_priority= 0;
+
+  trans_track_end_trx(thd);
+
   DBUG_RETURN(MY_TEST(res));
 }
 
@@ -406,7 +425,8 @@ bool trans_rollback_implicit(THD *thd)
     Don't perform rollback in the middle of sub-statement, wait till
     its end.
   */
-  DBUG_ASSERT(thd->transaction.stmt.is_empty() && !thd->in_sub_stmt);
+  DBUG_ASSERT(thd->get_transaction()->is_empty(Transaction_ctx::STMT) &&
+              !thd->in_sub_stmt);
 
 #ifdef WITH_WSREP
   wsrep_register_hton(thd, true);
@@ -415,15 +435,16 @@ bool trans_rollback_implicit(THD *thd)
     ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
   DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
   res= ha_rollback_trans(thd, true);
-  /*
-    We don't reset OPTION_BEGIN flag below to simulate implicit start
-    of new transacton in @@autocommit=1 mode. This is necessary to
-    preserve backward compatibility.
-  */
-  thd->transaction.all.reset_unsafe_rollback_flags();
+  thd->variables.option_bits&= ~OPTION_BEGIN;
+  thd->get_transaction()->reset_unsafe_rollback_flags(
+    Transaction_ctx::SESSION);
 
   /* Rollback should clear transaction_rollback_request flag. */
-  DBUG_ASSERT(! thd->transaction_rollback_request);
+  DBUG_ASSERT(!thd->transaction_rollback_request);
+  /* The transaction should be marked as complete in P_S. */
+  DBUG_ASSERT(thd->m_transaction_psi == NULL);
+
+  trans_track_end_trx(thd);
 
   DBUG_RETURN(MY_TEST(res));
 }
@@ -447,13 +468,6 @@ bool trans_rollback_implicit(THD *thd)
 bool trans_commit_stmt(THD *thd)
 {
   DBUG_ENTER("trans_commit_stmt");
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
-#endif
-
   int res= FALSE;
   /*
     We currently don't invoke commit/rollback at end of
@@ -463,36 +477,41 @@ bool trans_commit_stmt(THD *thd)
   */
   DBUG_ASSERT(! thd->in_sub_stmt);
 
-#ifndef DBUG_OFF
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
+  /*
+    Some code in MYSQL_BIN_LOG::commit and ha_commit_low() is not safe
+    for attachable transactions.
+  */
+  DBUG_ASSERT(!thd->is_attachable_ro_transaction_active());
 
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
+  thd->get_transaction()->merge_unsafe_rollback_flags();
 
-  thd->transaction.merge_unsafe_rollback_flags();
-
-  if (thd->transaction.stmt.ha_list)
+  if (thd->get_transaction()->is_active(Transaction_ctx::STMT))
   {
 #ifdef WITH_WSREP
     wsrep_register_hton(thd, FALSE);
 #endif /* WITH_WSREP */
     res= ha_commit_trans(thd, FALSE);
     if (! thd->in_active_multi_stmt_transaction())
+#ifdef WITH_WSREP
     {
-      thd->tx_isolation= (enum_tx_isolation) thd->variables.tx_isolation;
-      thd->tx_read_only= thd->variables.tx_read_only;
+#endif /* WITH_WSREP */
+      trans_reset_one_shot_chistics(thd);
 #ifdef WITH_WSREP
       wsrep_post_commit(thd, FALSE);
-#endif /* WITH_WSREP */
     }
+#endif /* WITH_WSREP */
   }
   else if (tc_log)
     tc_log->commit(thd, false);
+  if (res == FALSE && !thd->in_active_multi_stmt_transaction())
+    if (thd->rpl_thd_ctx.session_gtids_ctx().
+        notify_after_transaction_commit(thd))
+      sql_print_warning("Failed to collect GTID to send in the response packet!");
+  /* In autocommit=1 mode the transaction should be marked as complete in P_S */
+  DBUG_ASSERT(thd->in_active_multi_stmt_transaction() ||
+              thd->m_transaction_psi == NULL);
 
-  thd->transaction.stmt.reset();
+  thd->get_transaction()->reset(Transaction_ctx::STMT);
 
   DBUG_RETURN(MY_TEST(res));
 }
@@ -518,43 +537,107 @@ bool trans_rollback_stmt(THD *thd)
   */
   DBUG_ASSERT(! thd->in_sub_stmt);
 
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
+  /*
+    Some code in MYSQL_BIN_LOG::rollback and ha_rollback_low() is not safe
+    for attachable transactions.
+  */
+  DBUG_ASSERT(!thd->is_attachable_ro_transaction_active());
 
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
+  thd->get_transaction()->merge_unsafe_rollback_flags();
 
-  thd->transaction.merge_unsafe_rollback_flags();
-
-  if (thd->transaction.stmt.ha_list)
+  if (thd->get_transaction()->is_active(Transaction_ctx::STMT))
   {
 #ifdef WITH_WSREP
     wsrep_register_hton(thd, FALSE);
 #endif /* WITH_WSREP */
     ha_rollback_trans(thd, FALSE);
     if (! thd->in_active_multi_stmt_transaction())
-    {
-      thd->tx_isolation= (enum_tx_isolation) thd->variables.tx_isolation;
-      thd->tx_read_only= thd->variables.tx_read_only;
-    }
+      trans_reset_one_shot_chistics(thd);
   }
   else if (tc_log)
     tc_log->rollback(thd, false);
 
-  thd->transaction.stmt.reset();
+  if (!thd->owned_gtid.is_empty() &&
+      !thd->in_active_multi_stmt_transaction())
+  {
+    /*
+      To a failed single statement transaction on auto-commit mode,
+      we roll back its owned gtid if it does not modify
+      non-transational table or commit its owned gtid if it has modified
+      non-transactional table when rolling back it if binlog is disabled,
+      as we did when binlog is enabled.
+      We do not need to check if binlog is enabled here, since we already
+      released its owned gtid in MYSQL_BIN_LOG::rollback(...) right before
+      this if binlog is enabled.
+    */
+    if (thd->get_transaction()->has_modified_non_trans_table(
+          Transaction_ctx::STMT))
+      gtid_state->update_on_commit(thd);
+    else
+      gtid_state->update_on_rollback(thd);
+  }
+
+  /* In autocommit=1 mode the transaction should be marked as complete in P_S */
+  DBUG_ASSERT(thd->in_active_multi_stmt_transaction() ||
+              thd->m_transaction_psi == NULL ||
+              /* Todo: BUG#20488921 is in the way. */
+              DBUG_EVALUATE_IF("simulate_xa_commit_log_failure", true, false));
+
+  thd->get_transaction()->reset(Transaction_ctx::STMT);
 
   DBUG_RETURN(FALSE);
 }
 
+
+/**
+  Commit the attachable transaction.
+
+  @note This is slimmed down version of trans_commit_stmt() which commits
+        attachable transaction but skips code which is unnecessary and
+        unsafe for them (like dealing with GTIDs).
+
+  @param thd     Current thread
+
+  @retval False - Success
+  @retval True  - Failure
+*/
+bool trans_commit_attachable(THD *thd)
+{
+  DBUG_ENTER("trans_commit_attachable");
+  int res= 0;
+
+  /* This function only handles attachable transactions. */
+  DBUG_ASSERT(thd->is_attachable_ro_transaction_active());
+
+  /*
+    Since the attachable transaction is AUTOCOMMIT we only need to commit
+    statement transaction.
+  */
+  DBUG_ASSERT(! thd->get_transaction()->is_active(Transaction_ctx::SESSION));
+
+  /* Attachable transactions should not do anything unsafe. */
+  DBUG_ASSERT(!thd->get_transaction()->
+                 cannot_safely_rollback(Transaction_ctx::STMT));
+
+
+  if (thd->get_transaction()->is_active(Transaction_ctx::STMT))
+  {
+    res= ha_commit_attachable(thd);
+  }
+
+  DBUG_ASSERT(thd->m_transaction_psi == NULL);
+
+  thd->get_transaction()->reset(Transaction_ctx::STMT);
+
+  DBUG_RETURN(MY_TEST(res));
+}
+
+
 /* Find a named savepoint in the current transaction. */
 static SAVEPOINT **
 find_savepoint(THD *thd, LEX_STRING name)
 {
-  SAVEPOINT **sv= &thd->transaction.savepoints;
+  SAVEPOINT **sv= &thd->get_transaction()->m_savepoints;
 
   while (*sv)
   {
@@ -587,13 +670,18 @@ bool trans_savepoint(THD *thd, LEX_STRING name)
       !opt_using_transactions)
     DBUG_RETURN(FALSE);
 
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  if (xa_state != XA_NOTR && xa_state != XA_ACTIVE)
+  if (thd->variables.transaction_write_set_extraction != HASH_ALGORITHM_OFF)
   {
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-    DBUG_RETURN(TRUE);
+    // is_fatal_errror is needed to avoid stored procedures to skip the error.
+    thd->is_fatal_error= 1;
+    my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0),
+             "--transaction-write-set-extraction!=OFF");
+    DBUG_RETURN(true);
   }
 
+  if (thd->get_transaction()->xid_state()->check_has_uncommitted_xa())
+    DBUG_RETURN(true);
+
   sv= find_savepoint(thd, name);
 
   if (*sv) /* old savepoint of the same name exists */
@@ -602,14 +690,14 @@ bool trans_savepoint(THD *thd, LEX_STRING name)
     ha_release_savepoint(thd, *sv);
     *sv= (*sv)->prev;
   }
-  else if ((newsv= (SAVEPOINT *) alloc_root(&thd->transaction.mem_root,
-                                            savepoint_alloc_size)) == NULL)
+  else if ((newsv= (SAVEPOINT *) thd->get_transaction()->allocate_memory(
+      savepoint_alloc_size)) == NULL)
   {
     my_error(ER_OUT_OF_RESOURCES, MYF(0));
     DBUG_RETURN(TRUE);
   }
 
-  newsv->name= strmake_root(&thd->transaction.mem_root, name.str, name.length);
+  newsv->name= thd->get_transaction()->strmake(name.str, name.length);
   newsv->length= name.length;
 
   /*
@@ -620,8 +708,8 @@ bool trans_savepoint(THD *thd, LEX_STRING name)
   if (ha_savepoint(thd, newsv))
     DBUG_RETURN(TRUE);
 
-  newsv->prev= thd->transaction.savepoints;
-  thd->transaction.savepoints= newsv;
+  newsv->prev= thd->get_transaction()->m_savepoints;
+  thd->get_transaction()->m_savepoints= newsv;
 
   /*
     Remember locks acquired before the savepoint was set.
@@ -661,28 +749,14 @@ bool trans_rollback_to_savepoint(THD *thd, LEX_STRING name)
   SAVEPOINT *sv= *find_savepoint(thd, name);
   DBUG_ENTER("trans_rollback_to_savepoint");
 
-#ifndef DBUG_OFF
-  char buf1[256], buf2[256];
-  DBUG_PRINT("enter", ("stmt.ha_list: %s, all.ha_list: %s",
-                       ha_list_names(thd->transaction.stmt.ha_list, buf1),
-                       ha_list_names(thd->transaction.all.ha_list, buf2)));
-
-  thd->transaction.stmt.dbug_unsafe_rollback_flags("stmt");
-  thd->transaction.all.dbug_unsafe_rollback_flags("all");
-#endif
-
   if (sv == NULL)
   {
     my_error(ER_SP_DOES_NOT_EXIST, MYF(0), "SAVEPOINT", name.str);
     DBUG_RETURN(TRUE);
   }
 
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  if (xa_state != XA_NOTR && xa_state != XA_ACTIVE)
-  {
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-    DBUG_RETURN(TRUE);
-  }
+  if (thd->get_transaction()->xid_state()->check_has_uncommitted_xa())
+    DBUG_RETURN(true);
 
   /**
     Checking whether it is safe to release metadata locks acquired after
@@ -709,13 +783,9 @@ bool trans_rollback_to_savepoint(THD *thd, LEX_STRING name)
 #ifdef WITH_WSREP
   bool mdl_can_safely_rollback_to_savepoint=
                 (!((WSREP_EMULATE_BINLOG(thd) ||  mysql_bin_log.is_open()) 
-                   && thd->variables.sql_log_bin) ||
+		   && thd->variables.sql_log_bin) ||
                  ha_rollback_to_savepoint_can_release_mdl(thd));
-  /* No need to register wsrep plugin while rolling back to savepoint.
-  There is no action for wsrep plugin for savepoint rollback.
-  Action is associated with binlog to clearup/truncate till said point
-  and innodb.
-  wsrep_register_hton(thd, TRUE); */
+  wsrep_register_hton(thd, TRUE);
 #else
   bool mdl_can_safely_rollback_to_savepoint=
                 (!(mysql_bin_log.is_open() && thd->variables.sql_log_bin) ||
@@ -724,19 +794,19 @@ bool trans_rollback_to_savepoint(THD *thd, LEX_STRING name)
 
   if (ha_rollback_to_savepoint(thd, sv))
     res= TRUE;
-  else if (thd->transaction.all.cannot_safely_rollback() && !thd->slave_thread)
-    thd->transaction.push_unsafe_rollback_warnings(thd);
+  else if (thd->get_transaction()->cannot_safely_rollback(
+           Transaction_ctx::SESSION) &&
+           !thd->slave_thread)
+    thd->get_transaction()->push_unsafe_rollback_warnings(thd);
 
-  thd->transaction.savepoints= sv;
+  thd->get_transaction()->m_savepoints= sv;
 
-#ifdef WITH_WSREP
   /*
     Release metadata locks that were acquired during this savepoint unit
     unless binlogging is on. Releasing locks with binlogging on can break
     replication as it allows other connections to drop these tables before
     rollback to savepoint is written to the binlog.
   */
-#endif
   if (!res && mdl_can_safely_rollback_to_savepoint)
     thd->mdl_context.rollback_to_savepoint(sv->mdl_savepoint);
 
@@ -770,289 +840,13 @@ bool trans_release_savepoint(THD *thd, LEX_STRING name)
     DBUG_RETURN(TRUE);
   }
 
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  if (xa_state != XA_NOTR && xa_state != XA_ACTIVE)
-  {
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-    DBUG_RETURN(TRUE);
-  }
+  if (thd->get_transaction()->xid_state()->check_has_uncommitted_xa())
+    DBUG_RETURN(true);
 
   if (ha_release_savepoint(thd, sv))
     res= TRUE;
 
-  thd->transaction.savepoints= sv->prev;
+  thd->get_transaction()->m_savepoints= sv->prev;
 
   DBUG_RETURN(MY_TEST(res));
 }
-
-
-/**
-  Starts an XA transaction with the given xid value.
-
-  @param thd    Current thread
-
-  @retval FALSE  Success
-  @retval TRUE   Failure
-*/
-
-bool trans_xa_start(THD *thd)
-{
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  DBUG_ENTER("trans_xa_start");
-
-  if (xa_state == XA_IDLE && thd->lex->xa_opt == XA_RESUME)
-  {
-    bool not_equal= !thd->transaction.xid_state.xid.eq(thd->lex->xid);
-    if (not_equal)
-      my_error(ER_XAER_NOTA, MYF(0));
-    else
-      thd->transaction.xid_state.xa_state= XA_ACTIVE;
-    DBUG_RETURN(not_equal);
-  }
-
-  /* TODO: JOIN is not supported yet. */
-  if (thd->lex->xa_opt != XA_NONE)
-    my_error(ER_XAER_INVAL, MYF(0));
-  else if (xa_state != XA_NOTR)
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-  else if (thd->locked_tables_mode || thd->in_active_multi_stmt_transaction())
-    my_error(ER_XAER_OUTSIDE, MYF(0));
-  else if (!trans_begin(thd))
-  {
-    DBUG_ASSERT(thd->transaction.xid_state.xid.is_null());
-    thd->transaction.xid_state.xa_state= XA_ACTIVE;
-    thd->transaction.xid_state.rm_error= 0;
-    thd->transaction.xid_state.xid.set(thd->lex->xid);
-    if (xid_cache_insert(&thd->transaction.xid_state))
-    {
-      thd->transaction.xid_state.xa_state= XA_NOTR;
-      thd->transaction.xid_state.xid.null();
-      trans_rollback(thd);
-      DBUG_RETURN(true);
-    }
-    DBUG_RETURN(FALSE);
-  }
-
-  DBUG_RETURN(TRUE);
-}
-
-
-/**
-  Put a XA transaction in the IDLE state.
-
-  @param thd    Current thread
-
-  @retval FALSE  Success
-  @retval TRUE   Failure
-*/
-
-bool trans_xa_end(THD *thd)
-{
-  DBUG_ENTER("trans_xa_end");
-
-  /* TODO: SUSPEND and FOR MIGRATE are not supported yet. */
-  if (thd->lex->xa_opt != XA_NONE)
-    my_error(ER_XAER_INVAL, MYF(0));
-  else if (thd->transaction.xid_state.xa_state != XA_ACTIVE)
-    my_error(ER_XAER_RMFAIL, MYF(0),
-             xa_state_names[thd->transaction.xid_state.xa_state]);
-  else if (!thd->transaction.xid_state.xid.eq(thd->lex->xid))
-    my_error(ER_XAER_NOTA, MYF(0));
-  else if (!xa_trans_rolled_back(&thd->transaction.xid_state))
-    thd->transaction.xid_state.xa_state= XA_IDLE;
-
-  DBUG_RETURN(thd->is_error() ||
-              thd->transaction.xid_state.xa_state != XA_IDLE);
-}
-
-
-/**
-  Put a XA transaction in the PREPARED state.
-
-  @param thd    Current thread
-
-  @retval FALSE  Success
-  @retval TRUE   Failure
-*/
-
-bool trans_xa_prepare(THD *thd)
-{
-  DBUG_ENTER("trans_xa_prepare");
-
-  if (thd->transaction.xid_state.xa_state != XA_IDLE)
-    my_error(ER_XAER_RMFAIL, MYF(0),
-             xa_state_names[thd->transaction.xid_state.xa_state]);
-  else if (!thd->transaction.xid_state.xid.eq(thd->lex->xid))
-    my_error(ER_XAER_NOTA, MYF(0));
-  else if (ha_prepare(thd))
-  {
-    xid_cache_delete(&thd->transaction.xid_state);
-    thd->transaction.xid_state.xa_state= XA_NOTR;
-    my_error(ER_XA_RBROLLBACK, MYF(0));
-  }
-  else
-    thd->transaction.xid_state.xa_state= XA_PREPARED;
-
-  DBUG_RETURN(thd->is_error() ||
-              thd->transaction.xid_state.xa_state != XA_PREPARED);
-}
-
-
-/**
-  Commit and terminate the a XA transaction.
-
-  @param thd    Current thread
-
-  @retval FALSE  Success
-  @retval TRUE   Failure
-*/
-
-bool trans_xa_commit(THD *thd)
-{
-  bool res= TRUE;
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  DBUG_ENTER("trans_xa_commit");
-
-  if (!thd->transaction.xid_state.xid.eq(thd->lex->xid))
-  {
-    /*
-      xid_state.in_thd is always true beside of xa recovery procedure.
-      Note, that there is no race condition here between xid_cache_search
-      and xid_cache_delete, since we always delete our own XID
-      (thd->lex->xid == thd->transaction.xid_state.xid).
-      The only case when thd->lex->xid != thd->transaction.xid_state.xid
-      and xid_state->in_thd == 0 is in the function
-      xa_cache_insert(XID, xa_states), which is called before starting
-      client connections, and thus is always single-threaded.
-    */
-    XID_STATE *xs= xid_cache_search(thd->lex->xid);
-    res= !xs || xs->in_thd;
-    if (res)
-      my_error(ER_XAER_NOTA, MYF(0));
-    else
-    {
-      res= xa_trans_rolled_back(xs);
-      ha_commit_or_rollback_by_xid(thd, thd->lex->xid, !res);
-      xid_cache_delete(xs);
-    }
-    DBUG_RETURN(res);
-  }
-
-  if (xa_trans_rolled_back(&thd->transaction.xid_state))
-  {
-    xa_trans_force_rollback(thd);
-    res= thd->is_error();
-  }
-  else if (xa_state == XA_IDLE && thd->lex->xa_opt == XA_ONE_PHASE)
-  {
-#ifdef WITH_WSREP
-    wsrep_register_hton(thd, TRUE);
-#endif /* WITH_WSREP */
-    int r= ha_commit_trans(thd, TRUE);
-    if ((res= MY_TEST(r)))
-      my_error(r == 1 ? ER_XA_RBROLLBACK : ER_XAER_RMERR, MYF(0));
-#ifdef WITH_WSREP
-    wsrep_post_commit(thd, TRUE);
-#endif /* WITH_WSREP */
-  }
-  else if (xa_state == XA_PREPARED && thd->lex->xa_opt == XA_NONE)
-  {
-    MDL_request mdl_request;
-
-    /*
-      Acquire metadata lock which will ensure that COMMIT is blocked
-      by active FLUSH TABLES WITH READ LOCK (and vice versa COMMIT in
-      progress blocks FTWRL).
-
-      We allow FLUSHer to COMMIT; we assume FLUSHer knows what it does.
-    */
-    mdl_request.init(MDL_key::COMMIT, "", "", MDL_INTENTION_EXCLUSIVE,
-                     MDL_TRANSACTION);
-
-    if (thd->mdl_context.acquire_lock(&mdl_request,
-                                      thd->variables.lock_wait_timeout))
-    {
-#ifdef WITH_WSREP
-      wsrep_register_hton(thd, TRUE);
-#endif /* WITH_WSREP */
-      ha_rollback_trans(thd, TRUE);
-      my_error(ER_XAER_RMERR, MYF(0));
-    }
-    else
-    {
-      DEBUG_SYNC(thd, "trans_xa_commit_after_acquire_commit_lock");
-
-      if (tc_log)
-        res= MY_TEST(tc_log->commit(thd, /* all */ true));
-      else
-        res= MY_TEST(ha_commit_low(thd, /* all */ true));
-
-      if (res)
-        my_error(ER_XAER_RMERR, MYF(0));
-    }
-  }
-  else
-  {
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-    DBUG_RETURN(TRUE);
-  }
-
-  thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
-  thd->server_status&=
-    ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
-  DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
-  xid_cache_delete(&thd->transaction.xid_state);
-  thd->transaction.xid_state.xa_state= XA_NOTR;
-
-  DBUG_RETURN(res);
-}
-
-
-/**
-  Roll back and terminate a XA transaction.
-
-  @param thd    Current thread
-
-  @retval FALSE  Success
-  @retval TRUE   Failure
-*/
-
-bool trans_xa_rollback(THD *thd)
-{
-  bool res= TRUE;
-  enum xa_states xa_state= thd->transaction.xid_state.xa_state;
-  DBUG_ENTER("trans_xa_rollback");
-
-  if (!thd->transaction.xid_state.xid.eq(thd->lex->xid))
-  {
-    XID_STATE *xs= xid_cache_search(thd->lex->xid);
-    if (!xs || xs->in_thd)
-      my_error(ER_XAER_NOTA, MYF(0));
-    else
-    {
-      xa_trans_rolled_back(xs);
-      ha_commit_or_rollback_by_xid(thd, thd->lex->xid, 0);
-      xid_cache_delete(xs);
-    }
-    DBUG_RETURN(thd->get_stmt_da()->is_error());
-  }
-
-  if (xa_state != XA_IDLE && xa_state != XA_PREPARED && xa_state != XA_ROLLBACK_ONLY)
-  {
-    my_error(ER_XAER_RMFAIL, MYF(0), xa_state_names[xa_state]);
-    DBUG_RETURN(TRUE);
-  }
-
-  res= xa_trans_force_rollback(thd);
-
-  thd->variables.option_bits&= ~OPTION_BEGIN;
-  thd->transaction.all.reset_unsafe_rollback_flags();
-  thd->server_status&=
-    ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
-  DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
-  xid_cache_delete(&thd->transaction.xid_state);
-  thd->transaction.xid_state.xa_state= XA_NOTR;
-
-  DBUG_RETURN(res);
-}
