@@ -1,21 +1,14 @@
 /*
-   Copyright (c) 2000, 2017, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2000, 2016, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -31,13 +24,14 @@
     str is a (long) to record position where 0 is the first position.
 */
 
-#include "sql_priv.h"
 #include "unireg.h"
-#include "sql_partition.h"                    // struct partition_info
+#include "table.h"
+#include "sql_class.h"                        // THD, Internal_error_handler
+#include "partition_info.h"                   // partition_info
 #include "sql_table.h"                        // validate_comment_length   
-#include "sql_class.h"                  // THD, Internal_error_handler
-#include <m_ctype.h>
-#include <assert.h>
+
+#include "pfs_file_provider.h"
+#include "mysql/psi/mysql_file.h"
 
 #include <algorithm>
 
@@ -45,6 +39,9 @@ using std::min;
 using std::max;
 
 #define FCOMP			17		/* Bytes for a packed field */
+#define SC_INFO_LENGTH 4		/* Form format constant */
+#define TE_INFO_LENGTH 3
+#define MTYP_NOEMPTY_BIT 128
 
 static uchar * pack_screens(List<Create_field> &create_fields,
 			    uint *info_length, uint *screens, bool small_file);
@@ -69,33 +66,24 @@ static bool make_empty_rec(THD *thd, int file,
   XXX: what is a UNIREG  screen?
 */
 
-struct Pack_header_error_handler: public Internal_error_handler
+class Pack_header_error_handler: public Internal_error_handler
 {
+  bool m_is_handled;
+public:
   virtual bool handle_condition(THD *thd,
                                 uint sql_errno,
                                 const char* sqlstate,
-                                Sql_condition::enum_warning_level level,
-                                const char* msg,
-                                Sql_condition ** cond_hdl);
-  bool is_handled;
-  Pack_header_error_handler() :is_handled(FALSE) {}
+                                Sql_condition::enum_severity_level *level,
+                                const char* msg)
+  {
+    m_is_handled= (sql_errno == ER_TOO_MANY_FIELDS);
+    return m_is_handled;
+  }
+  Pack_header_error_handler() :m_is_handled(false) {}
+  bool is_handled() const { return m_is_handled; }
 };
 
 
-bool
-Pack_header_error_handler::
-handle_condition(THD *,
-                 uint sql_errno,
-                 const char*,
-                 Sql_condition::enum_warning_level,
-                 const char*,
-                 Sql_condition ** cond_hdl)
-{
-  *cond_hdl= NULL;
-  is_handled= (sql_errno == ER_TOO_MANY_FIELDS);
-  return is_handled;
-}
-
 /*
   Create a frm (table definition) file
 
@@ -131,14 +119,12 @@ bool mysql_create_frm(THD *thd, const char *file_name,
   uchar fileinfo[64],forminfo[288],*keybuff, *forminfo_p= forminfo;
   uchar *screen_buff= NULL;
   char buff[128];
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   partition_info *part_info= thd->work_part_info;
-#endif
   Pack_header_error_handler pack_header_error_handler;
   int error;
   const uint format_section_header_size= 8;
   uint format_section_length;
-  uint tablespace_length= 0;
+  size_t tablespace_length= 0;
   DBUG_ENTER("mysql_create_frm");
 
   DBUG_ASSERT(*fn_rext((char*)file_name)); // Check .frm extension
@@ -164,7 +150,7 @@ bool mysql_create_frm(THD *thd, const char *file_name,
   if (error)
   {
     my_free(screen_buff);
-    if (! pack_header_error_handler.is_handled)
+    if (! pack_header_error_handler.is_handled())
       DBUG_RETURN(1);
 
     // Try again without UNIREG screens (to get more columns)
@@ -194,12 +180,10 @@ bool mysql_create_frm(THD *thd, const char *file_name,
       => Total 6 byte
   */
   create_info->extra_size+= 6;
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   if (part_info)
   {
     create_info->extra_size+= part_info->part_info_len;
   }
-#endif
 
   for (i= 0; i < keys; i++)
   {
@@ -264,13 +248,30 @@ bool mysql_create_frm(THD *thd, const char *file_name,
     table and column properties
   */
   if (create_info->tablespace)
+  {
     tablespace_length= strlen(create_info->tablespace);
+    /*
+      Make sure we have at least an IX lock on the tablespace name,
+      unless this is a temporary table. For temporary tables, the
+      tablespace name is not IX locked.
+    */
+    if (tablespace_length > 0 &&
+        !(create_info->options & HA_LEX_CREATE_TMP_TABLE))
+      DBUG_ASSERT(thd->mdl_context.owns_equal_or_stronger_lock(
+                                     MDL_key::TABLESPACE, "",
+                                     create_info->tablespace,
+                                     MDL_INTENTION_EXCLUSIVE));
+  }
   format_section_length=
     format_section_header_size +
     tablespace_length + 1 +
     create_fields.elements;
   create_info->extra_size+= format_section_length;
 
+  create_info->extra_size+= 2 + create_info->compress.length;
+
+  create_info->extra_size+= 2 + create_info->encrypt_type.length;
+
   if ((file=create_frm(thd, file_name, db, table, reclength, fileinfo,
 		       create_info, keys, key_info)) < 0)
   {
@@ -279,28 +280,10 @@ bool mysql_create_frm(THD *thd, const char *file_name,
   }
 
   key_buff_length= uint4korr(fileinfo+47);
-  keybuff=(uchar*) my_malloc(key_buff_length, MYF(0));
+  keybuff=(uchar*) my_malloc(key_memory_frm,
+                             key_buff_length, MYF(0));
   key_info_length= pack_keys(keybuff, keys, key_info, data_offset);
 
-  /* key_info_length is currently stored in 2 bytes */
-  if (key_info_length > 65535U)
-  {
-    char *real_table_name= (char*) table;
-    List_iterator<Create_field> it(create_fields);
-    Create_field *field;
-    while ((field=it++))
-    {
-      if (field->field && field->field->table &&
-         (real_table_name= field->field->table->s->table_name.str))
-        break;
-    }
-    my_printf_error(ER_UNKNOWN_ERROR,
-                    "Index information size for the table %s.%s exceeds the "
-                    "maximum limit (Max: 2 bytes). Please recreate indexes "
-                    "accordingly.", MYF(0), db, real_table_name);
-    goto err;
-  }
-
   /*
     Ensure that there are no forms in this newly created form file.
     Even if the form file exists, create_frm must truncate it to
@@ -317,13 +300,11 @@ bool mysql_create_frm(THD *thd, const char *file_name,
                                 (create_info->min_rows == 1) && (keys == 0));
   int2store(fileinfo+28,key_info_length);
 
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   if (part_info)
   {
     fileinfo[61]= (uchar) ha_legacy_type(part_info->default_engine_type);
     DBUG_PRINT("info", ("part_db_type = %d", fileinfo[61]));
   }
-#endif
   int2store(fileinfo+59,db_file->extra_rec_buf_length());
 
   if (mysql_file_pwrite(file, fileinfo, 64, 0L, MYF_RW) ||
@@ -338,23 +319,22 @@ bool mysql_create_frm(THD *thd, const char *file_name,
 		     create_fields,reclength, data_offset, db_file))
     goto err;
 
-  int2store(buff, create_info->connect_string.length);
+  int2store(buff, static_cast<uint16>(create_info->connect_string.length));
   if (mysql_file_write(file, (const uchar*)buff, 2, MYF(MY_NABP)) ||
       mysql_file_write(file, (const uchar*)create_info->connect_string.str,
                create_info->connect_string.length, MYF(MY_NABP)))
       goto err;
 
-  int2store(buff, str_db_type.length);
+  int2store(buff, static_cast<uint16>(str_db_type.length));
   if (mysql_file_write(file, (const uchar*)buff, 2, MYF(MY_NABP)) ||
       mysql_file_write(file, (const uchar*)str_db_type.str,
                str_db_type.length, MYF(MY_NABP)))
     goto err;
 
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   if (part_info)
   {
     char auto_partitioned= part_info->is_auto_partitioned ? 1 : 0;
-    int4store(buff, part_info->part_info_len);
+    int4store(buff, static_cast<uint32>(part_info->part_info_len));
     if (mysql_file_write(file, (const uchar*)buff, 4, MYF_RW) ||
         mysql_file_write(file, (const uchar*)part_info->part_info_string,
                  part_info->part_info_len + 1, MYF_RW) ||
@@ -362,7 +342,6 @@ bool mysql_create_frm(THD *thd, const char *file_name,
       goto err;
   }
   else
-#endif
   {
     memset(buff, 0, 6);
     if (mysql_file_write(file, (uchar*) buff, 6, MYF_RW))
@@ -380,7 +359,7 @@ bool mysql_create_frm(THD *thd, const char *file_name,
   if (forminfo[46] == (uchar)255)
   {
     uchar comment_length_buff[2];
-    int2store(comment_length_buff,create_info->comment.length);
+    int2store(comment_length_buff, static_cast<uint16>(create_info->comment.length));
     if (mysql_file_write(file, comment_length_buff, 2, MYF(MY_NABP)) ||
         mysql_file_write(file, (uchar*) create_info->comment.str,
                          create_info->comment.length, MYF(MY_NABP)))
@@ -390,7 +369,8 @@ bool mysql_create_frm(THD *thd, const char *file_name,
   /* "Format section" with additional table and column properties */
   {
     uchar *ptr, *format_section_buff;
-    if (!(format_section_buff=(uchar*) my_malloc(format_section_length,
+    if (!(format_section_buff=(uchar*) my_malloc(key_memory_frm,
+                                                 format_section_length,
                                                  MYF(MY_WME))))
       goto err;
     ptr= format_section_buff;
@@ -436,34 +416,37 @@ bool mysql_create_frm(THD *thd, const char *file_name,
     my_free(format_section_buff);
   }
 
+  /* Write out the COMPRESS table attribute */
+  {
+    uchar length_buff[2];
+
+    int2store(length_buff, static_cast<uint16>(create_info->compress.length));
+
+    if (mysql_file_write(file, length_buff, 2, MYF(MY_NABP)) ||
+        mysql_file_write(file, (uchar*) create_info->compress.str,
+                         create_info->compress.length, MYF(MY_NABP)))
+      goto err;
+  }
+
+  /* Write out the ENCRYPT table attribute */
+  {
+    uchar length_buff[2];
+
+    int2store(length_buff,
+	      static_cast<uint16>(create_info->encrypt_type.length));
+
+    if (mysql_file_write(file, length_buff, 2, MYF(MY_NABP)) ||
+        mysql_file_write(file, (uchar*) create_info->encrypt_type.str,
+                         create_info->encrypt_type.length, MYF(MY_NABP)))
+      goto err;
+  }
+
   mysql_file_seek(file, filepos, MY_SEEK_SET, MYF(0));
   if (mysql_file_write(file, forminfo, 288, MYF_RW) ||
       mysql_file_write(file, screen_buff, info_length, MYF_RW) ||
       pack_fields(file, create_fields, data_offset))
     goto err;
 
-#ifdef HAVE_CRYPTED_FRM
-  if (create_info->password)
-  {
-    char tmp=2,*disk_buff=0;
-    SQL_CRYPT *crypted=new SQL_CRYPT(create_info->password);
-    if (!crypted || mysql_file_pwrite(file, &tmp, 1, 26, MYF_RW))// Mark crypted
-      goto err;
-    uint read_length=uint2korr(forminfo)-256;
-    mysql_file_seek(file, filepos+256, MY_SEEK_SET, MYF(0));
-    if (read_string(file,(uchar**) &disk_buff,read_length))
-      goto err;
-    crypted->encode(disk_buff,read_length);
-    delete crypted;
-    if (mysql_file_pwrite(file, disk_buff, read_length, filepos+256, MYF_RW))
-    {
-      my_free(disk_buff);
-      goto err;
-    }
-    my_free(disk_buff);
-  }
-#endif
-
   my_free(screen_buff);
   my_free(keybuff);
 
@@ -533,8 +516,9 @@ int rea_create_table(THD *thd, const char *path,
 {
   DBUG_ENTER("rea_create_table");
 
-  char frm_name[FN_REFLEN];
-  strxmov(frm_name, path, reg_ext, NullS);
+  char frm_name[FN_REFLEN + 1];
+  strxnmov(frm_name, sizeof(frm_name) - 1, path, reg_ext, NullS);
+
   if (mysql_create_frm(thd, frm_name, db, table_name, create_info,
                        create_fields, keys, key_info, file))
 
@@ -550,8 +534,7 @@ int rea_create_table(THD *thd, const char *path,
     goto err_handler_frm;
 
   if (!no_ha_table &&
-       ha_create_table(thd, path, db, table_name, create_info,
-                       &create_fields, 0))
+       ha_create_table(thd, path, db, table_name, create_info, 0))
     goto err_handler;
   DBUG_RETURN(0);
 
@@ -572,7 +555,8 @@ static uchar *pack_screens(List<Create_field> &create_fields,
 {
   uint i;
   uint row,start_row,end_row,fields_on_screen;
-  uint length,cols;
+  size_t length;
+  uint cols;
   uchar *info,*pos,*start_screen;
   uint fields=create_fields.elements;
   List_iterator<Create_field> it(create_fields);
@@ -585,9 +569,10 @@ static uchar *pack_screens(List<Create_field> &create_fields,
 
   Create_field *field;
   while ((field=it++))
-    length+=(uint) strlen(field->field_name)+1+TE_INFO_LENGTH+cols/2;
+    length+= strlen(field->field_name)+1+TE_INFO_LENGTH+cols/2;
 
-  if (!(info=(uchar*) my_malloc(length,MYF(MY_WME))))
+  if (!(info=(uchar*) my_malloc(key_memory_frm,
+                                length,MYF(MY_WME))))
     DBUG_RETURN(0);
 
   start_screen=0;
@@ -602,7 +587,7 @@ static uchar *pack_screens(List<Create_field> &create_fields,
       if (i)
       {
 	length=(uint) (pos-start_screen);
-	int2store(start_screen,length);
+	int2store(start_screen, static_cast<uint16>(length));
 	start_screen[2]=(uchar) (fields_on_screen+1);
 	start_screen[3]=(uchar) (fields_on_screen);
       }
@@ -615,7 +600,7 @@ static uchar *pack_screens(List<Create_field> &create_fields,
       strfill((char *) pos+3,(uint) (cols >> 1),' ');
       pos+=(cols >> 1)+4;
     }
-    length=(uint) strlen(cfield->field_name);
+    length= strlen(cfield->field_name);
     if (length > cols-3)
       length=cols-3;
 
@@ -631,7 +616,7 @@ static uchar *pack_screens(List<Create_field> &create_fields,
     cfield->sc_length= min<uint8>(cfield->length, cols - (length + 2));
   }
   length=(uint) (pos-start_screen);
-  int2store(start_screen,length);
+  int2store(start_screen, static_cast<uint16>(length));
   start_screen[2]=(uchar) (row-start_row+2);
   start_screen[3]=(uchar) (row-start_row+1);
 
@@ -657,7 +642,7 @@ static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
   {
     /* Replace HA_CLUSTERING with HA_SPATIAL | HA_FULLTEXT to allow storing
     TokuDB keys without changing the FRM format.  */
-    uint16 key_flags= (uint16)key->flags;
+    uint16 key_flags= static_cast<uint16>(key->flags);
     if (key->flags & HA_CLUSTERING)
     {
       key_flags|= HA_SPATIAL;
@@ -697,7 +682,7 @@ static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
   *pos++=(uchar) NAMES_SEP_CHAR;
   for (key=keyinfo ; key != end ; key++)
   {
-    uchar *tmp=(uchar*) strmov((char*) pos,key->name);
+    uchar *tmp=(uchar*) my_stpcpy((char*) pos,key->name);
     *tmp++= (uchar) NAMES_SEP_CHAR;
     *tmp=0;
     pos=tmp;
@@ -707,9 +692,9 @@ static uint pack_keys(uchar *keybuff, uint key_count, KEY *keyinfo,
   {
     if (key->flags & HA_USES_COMMENT)
     {
-      int2store(pos, key->comment.length);
-      uchar *tmp= (uchar*)strnmov((char*) pos+2,key->comment.str,
-                                  key->comment.length);
+      int2store(pos, static_cast<uint16>(key->comment.length));
+      uchar *tmp= (uchar*)my_stpnmov((char*) pos+2,key->comment.str,
+                                     key->comment.length);
       pos= tmp;
     }
   }
@@ -739,9 +724,10 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
                         uint info_length, uint screens, uint table_options,
                         ulong data_offset, handler *file)
 {
-  uint length,int_count,int_length,no_empty, int_parts;
+  size_t length;
+  uint int_count, int_length, no_empty, int_parts;
   uint time_stamp_pos,null_fields;
-  ulong reclength, totlength, n_length, com_length;
+  size_t reclength, totlength, n_length, com_length, gcol_info_length;
   DBUG_ENTER("pack_header");
 
   if (create_fields.elements > MAX_FIELDS)
@@ -752,8 +738,8 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
 
   totlength= 0L;
   reclength= data_offset;
-  no_empty=int_count=int_parts=int_length=time_stamp_pos=null_fields=
-    com_length=0;
+  no_empty=int_count=int_parts=int_length=time_stamp_pos=null_fields=0;
+  com_length=gcol_info_length=0;
   n_length=2L;
 
 	/* Check fields */
@@ -762,13 +748,58 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
   Create_field *field;
   while ((field=it++))
   {
-    if (validate_comment_length(current_thd,
+    THD *thd= current_thd;
+
+    if (validate_comment_length(thd,
                                 field->comment.str,
                                 &field->comment.length,
                                 COLUMN_COMMENT_MAXLEN,
                                 ER_TOO_LONG_FIELD_COMMENT,
                                 (char *) field->field_name))
       DBUG_RETURN(true);
+    if (field->gcol_info)
+    {
+      sql_mode_t sql_mode= thd->variables.sql_mode;
+      thd->variables.sql_mode&= ~MODE_ANSI_QUOTES;
+
+      /*
+        It is important to normalize the expression's text into the FRM, to
+        make it independent from sql_mode. For example, 'a||b' means 'a OR b'
+        or 'CONCAT(a,b)', depending on if PIPES_AS_CONCAT is on. Using
+        Item::print(), we get self-sufficient text containing 'OR' or
+        'CONCAT'. If sql_mode later changes, it will not affect the column.
+       */
+      String s;
+      // Printing db and table name is useless
+      field->gcol_info->expr_item->
+        print(&s, enum_query_type(QT_NO_DB | QT_NO_TABLE));
+
+      thd->variables.sql_mode= sql_mode;
+      /*
+        The new text must have exactly the same lifetime as the old text, it's
+        a replacement for it. So the same MEM_ROOT must be used: pass NULL.
+      */
+      field->gcol_info->dup_expr_str(NULL, s.ptr(), s.length());
+
+      uint tmp_len= system_charset_info->cset->charpos(system_charset_info,
+                                                       field->gcol_info->expr_str.str,
+                                                       field->gcol_info->expr_str.str +
+                                                       field->gcol_info->expr_str.length,
+                                                       GENERATED_COLUMN_EXPRESSION_MAXLEN);
+
+      if (tmp_len < field->gcol_info->expr_str.length)
+      {
+        my_error(ER_WRONG_STRING_LENGTH, MYF(0),
+                 field->gcol_info->expr_str.str,"GENERATED COLUMN EXPRESSION",
+                 (uint) GENERATED_COLUMN_EXPRESSION_MAXLEN);
+        DBUG_RETURN(1);
+      }
+      /*
+        Sum up the length of the expression string and mandatory header bytes
+        to the total length.
+      */
+      gcol_info_length+= field->gcol_info->expr_str.length+(uint)FRM_GCOL_HEADER_SIZE;
+    }
     totlength+= field->length;
     com_length+= field->comment.length;
     if (MTYP_TYPENR(field->unireg_check) == Field::NOEMPTY ||
@@ -788,10 +819,17 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
       time_stamp_pos= (uint) field->offset+ (uint) data_offset + 1;
     length=field->pack_length;
     /* Ensure we don't have any bugs when generating offsets */
-    DBUG_ASSERT(reclength == field->offset + data_offset);
-    if ((uint) field->offset+ (uint) data_offset+ length > reclength)
-      reclength=(uint) (field->offset+ data_offset + length);
-    n_length+= (ulong) strlen(field->field_name)+1;
+    /**
+      Because the virtual generated columns are not stored physically,
+      they are put at the tail of record. The details can be checked
+      in mysql_prepare_create_table. So the offset is messed up by
+      vitual generated columns. The original assert is not correct
+      any more.
+      DBUG_ASSERT(reclength == field->offset + data_offset);
+     */
+    if (field->offset + data_offset + length > reclength)
+      reclength= field->offset + data_offset + length;
+    n_length+= strlen(field->field_name) + 1;
     field->interval_id=0;
     field->save_interval= 0;
     if (field->interval)
@@ -822,7 +860,7 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
         {
           char *dst;
           const char *src= field->save_interval->type_names[pos];
-          uint hex_length;
+          size_t hex_length;
           length= field->save_interval->type_lengths[pos];
           hex_length= length * 2;
           field->interval->type_lengths[pos]= hex_length;
@@ -854,8 +892,9 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
   }
   /* Hack to avoid bugs with small static rows in MySQL */
   reclength= max<size_t>(file->min_record_length(table_options), reclength);
-  if (info_length+(ulong) create_fields.elements*FCOMP+288+
-      n_length+int_length+com_length > 65535L || int_count > 255)
+  if (info_length + (ulong) create_fields.elements * FCOMP + 288 +
+      n_length + int_length + com_length + gcol_info_length > 65535L ||
+      int_count > 255)
   {
     my_message(ER_TOO_MANY_FIELDS, ER(ER_TOO_MANY_FIELDS), MYF(0));
     DBUG_RETURN(1);
@@ -863,15 +902,15 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
 
   memset(forminfo, 0, 288);
   length=(info_length+create_fields.elements*FCOMP+288+n_length+int_length+
-	  com_length);
-  int2store(forminfo,length);
+	  com_length + gcol_info_length);
+  int2store(forminfo, static_cast<uint16>(length));
   forminfo[256] = (uint8) screens;
   int2store(forminfo+258,create_fields.elements);
   int2store(forminfo+260,info_length);
-  int2store(forminfo+262,totlength);
+  int2store(forminfo+262, static_cast<uint16>(totlength));
   int2store(forminfo+264,no_empty);
-  int2store(forminfo+266,reclength);
-  int2store(forminfo+268,n_length);
+  int2store(forminfo+266, static_cast<uint16>(reclength));
+  int2store(forminfo+268, static_cast<uint16>(n_length));
   int2store(forminfo+270,int_count);
   int2store(forminfo+272,int_parts);
   int2store(forminfo+274,int_length);
@@ -879,8 +918,9 @@ static bool pack_header(uchar *forminfo, enum legacy_db_type table_type,
   int2store(forminfo+278,80);			/* Columns needed */
   int2store(forminfo+280,22);			/* Rows needed */
   int2store(forminfo+282,null_fields);
-  int2store(forminfo+284,com_length);
-  /* Up to forminfo+288 is free to use for additional information */
+  int2store(forminfo+284, static_cast<uint16>(com_length));
+  int2store(forminfo+286,gcol_info_length);
+  /* forminfo+288 is free to use for additional information */
   DBUG_RETURN(0);
 } /* pack_header */
 
@@ -919,7 +959,8 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
                         ulong data_offset)
 {
   uint i;
-  uint int_count, comment_length=0;
+  uint int_count, gcol_info_length=0;
+  size_t comment_length= 0;
   uchar buff[MAX_FIELD_WIDTH];
   Create_field *field;
   DBUG_ENTER("pack_fields");
@@ -935,7 +976,7 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
     buff[0]= (uchar) field->row;
     buff[1]= (uchar) field->col;
     buff[2]= (uchar) field->sc_length;
-    int2store(buff+3, field->length);
+    int2store(buff+3, static_cast<uint16>(field->length));
     /* The +1 is here becasue the col offset in .frm file have offset 1 */
     recpos= field->offset+1 + (uint) data_offset;
     int3store(buff+5,recpos);
@@ -947,9 +988,6 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
     {
       buff[11]= 0;
       buff[14]= (uchar) field->geom_type;
-#ifndef HAVE_SPATIAL
-      DBUG_ASSERT(0);                           // Should newer happen
-#endif
     }
     else if (field->charset) 
     {
@@ -960,7 +998,12 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
     {
       buff[11]= buff[14]= 0;			// Numerical
     }
-    int2store(buff+15, field->comment.length);
+    if (field->gcol_info)
+    {
+      gcol_info_length+= field->gcol_info->expr_str.length;
+      buff[10]|= (uchar)Field::GENERATED_FIELD;
+    }
+    int2store(buff+15, static_cast<uint16>(field->comment.length));
     comment_length+= field->comment.length;
     set_if_bigger(int_count,field->interval_id);
     if (mysql_file_write(file, buff, FCOMP, MYF_RW))
@@ -975,7 +1018,7 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
   it.rewind();
   while ((field=it++))
   {
-    char *pos= strmov((char*) buff,field->field_name);
+    char *pos= my_stpcpy((char*) buff,field->field_name);
     *pos++=NAMES_SEP_CHAR;
     if (i == create_fields.elements-1)
       *pos++=0;
@@ -1054,10 +1097,142 @@ static bool pack_fields(File file, List<Create_field> &create_fields,
 	  DBUG_RETURN(1);
     }
   }
+  if (gcol_info_length)
+  {
+    it.rewind();
+    int_count=0;
+    while ((field=it++))
+    {
+      /*
+        Pack each virtual field as follows:
+        byte 1      = 1 (always 1 to allow for future extensions)
+        byte 2,3    = expression length
+        byte 4      = flags, as of now:
+                        0 - no flags
+                        1 - field is physically stored
+        byte 5-...  = generated column expression (text data)
+      */
+      if (field->gcol_info && field->gcol_info->expr_str.length)
+      {
+        buff[0]= (uchar)1;
+        int2store(buff + 1, field->gcol_info->expr_str.length);
+        buff[3]= (uchar) field->stored_in_db;
+        if (my_write(file, buff, FRM_GCOL_HEADER_SIZE, MYF_RW))
+          DBUG_RETURN(1);
+        if (my_write(file,
+                     (uchar*) field->gcol_info->expr_str.str,
+                     field->gcol_info->expr_str.length,
+                     MYF_RW))
+          DBUG_RETURN(1);
+      }
+    }
+  }
   DBUG_RETURN(0);
 }
 
 
+/**
+  Auxiliary function which stores field's explicit or implicit default
+  value in record buffer.
+
+  @param         thd         Connection's context.
+  @param         table       Table for which field default value to be stored.
+  @param         field       Field definition object.
+  @param         rec_pos     Pointer to the main part of record buffer where
+                             field values are stored (as opposed to record
+                             preamble).
+  @param         null_pos    Pointer to the preamble part of record buffer
+                             where null bits and leftover bits from BIT fields
+                             are stored.
+  @param[in/out] null_count  Index of bit in preamble to be used for storing
+                             NULL/leftover bits for the field if necessary. On
+                             return incremented by number of bits in preamble
+                             used for this field.
+
+  @retval true  An error occured.
+  @retval false Success.
+*/
+
+bool make_default_value(THD *thd, TABLE *table, Create_field *field,
+                        uchar *rec_pos, uchar *null_pos, uint *null_count)
+{
+  Field *regfield= make_field(table->s,
+                              rec_pos + field->offset,
+                              field->length,
+                              null_pos + *null_count / 8,
+                              *null_count & 7,
+                              field->pack_flag,
+                              field->sql_type,
+                              field->charset,
+                              field->geom_type,
+                              field->unireg_check,
+                              field->save_interval ? field->save_interval :
+                              field->interval,
+                              field->field_name);
+  if (!regfield)
+    return true;   // End of memory
+
+  /* save_in_field() will access regfield->table->in_use */
+  regfield->init(table);
+
+  if (!(field->flags & NOT_NULL_FLAG))
+  {
+    regfield->set_null();
+    (*null_count)++;
+  }
+
+  if (field->sql_type == MYSQL_TYPE_BIT && !f_bit_as_char(field->pack_flag))
+    (*null_count)+= field->length & 7;
+
+  Field::utype type= (Field::utype) MTYP_TYPENR(field->unireg_check);
+
+  if (field->def)
+  {
+    /*
+      Storing the value of a function is pointless as this function may not
+      be constant.
+    */
+    DBUG_ASSERT(field->def->type() != Item::FUNC_ITEM);
+    type_conversion_status res= field->def->save_in_field(regfield, true);
+    if (res != TYPE_OK && res != TYPE_NOTE_TIME_TRUNCATED &&
+        res != TYPE_NOTE_TRUNCATED)
+    {
+      /*
+        clear current error and report INVALID DEFAULT value error message
+      */
+      if (thd->is_error())
+        thd->clear_error();
+
+      my_error(ER_INVALID_DEFAULT, MYF(0), regfield->field_name);
+      /*
+        Delete to avoid memory leak for fields that allocate extra
+        memory (e.g Field_blob::value)
+      */
+      delete regfield;
+      return true;
+    }
+  }
+  else if (regfield->real_type() == MYSQL_TYPE_ENUM &&
+           (field->flags & NOT_NULL_FLAG))
+  {
+    regfield->set_notnull();
+    regfield->store((longlong) 1, TRUE);
+  }
+  else if (type == Field::YES)		// Old unireg type
+    regfield->store(ER(ER_YES), strlen(ER(ER_YES)),system_charset_info);
+  else if (type == Field::NO)			// Old unireg type
+    regfield->store(ER(ER_NO), strlen(ER(ER_NO)),system_charset_info);
+  else
+    regfield->reset();
+  /*
+    Delete to avoid memory leak for fields that allocate extra
+    memory (e.g Field_blob::value)
+  */
+  delete regfield;
+  return false;
+}
+
+
 /**
    Creates a record buffer consisting of default values for all columns and
    stores it in the formfile (.frm file.)
@@ -1091,21 +1266,22 @@ static bool make_empty_rec(THD *thd, File file,
                            handler *handler)
 {
   int error= 0;
-  Field::utype type;
   uint null_count;
   uchar *buff,*null_pos;
   TABLE table;
   TABLE_SHARE share;
   Create_field *field;
   enum_check_fields old_count_cuted_fields= thd->count_cuted_fields;
+  bool has_vgc= false;
   DBUG_ENTER("make_empty_rec");
 
   /* We need a table to generate columns for default values */
-  memset(static_cast<void*>(&table), 0, sizeof(table));
-  memset(static_cast<void*>(&share), 0, sizeof(share));
+  memset(&table, 0, sizeof(table));
+  memset(&share, 0, sizeof(share));
   table.s= &share;
 
-  if (!(buff=(uchar*) my_malloc((size_t) reclength,MYF(MY_WME | MY_ZEROFILL))))
+  if (!(buff=(uchar*) my_malloc(key_memory_frm,
+                                (size_t) reclength,MYF(MY_WME | MY_ZEROFILL))))
   {
     DBUG_RETURN(1);
   }
@@ -1125,83 +1301,38 @@ static bool make_empty_rec(THD *thd, File file,
   thd->count_cuted_fields= CHECK_FIELD_WARN;    // To find wrong default values
   while ((field=it++))
   {
-    Field *regfield= make_field(&share,
-                                buff+field->offset + data_offset,
-                                field->length,
-                                null_pos + null_count / 8,
-                                null_count & 7,
-                                field->pack_flag,
-                                field->sql_type,
-                                field->charset,
-                                field->geom_type,
-                                field->unireg_check,
-                                field->save_interval ? field->save_interval :
-                                field->interval, 
-                                field->field_name);
-    if (!regfield)
-    {
-      error= 1;
-      goto err;                                 // End of memory
-    }
-
-    /* save_in_field() will access regfield->table->in_use */
-    regfield->init(&table);
-
-    if (!(field->flags & NOT_NULL_FLAG))
-    {
-      regfield->set_null();
-      null_count++;
-    }
-
-    if (field->sql_type == MYSQL_TYPE_BIT && !f_bit_as_char(field->pack_flag))
-      null_count+= field->length & 7;
-
-    type= (Field::utype) MTYP_TYPENR(field->unireg_check);
-
-    if (field->def)
+    if (field->stored_in_db)
     {
       /*
-        Storing the value of a function is pointless as this function may not
-        be constant.
+        Even though Create_field::offset for virtual generated columns already
+        point at the end of record buffer we still need separate pass for them
+        in order to allocate null bits from preamble tail as well.
       */
-      DBUG_ASSERT(field->def->type() != Item::FUNC_ITEM);
-      type_conversion_status res= field->def->save_in_field(regfield, 1);
-      if (res != TYPE_OK && res != TYPE_NOTE_TIME_TRUNCATED &&
-          res != TYPE_NOTE_TRUNCATED)
+      if (make_default_value(thd, &table, field, buff + data_offset,
+                             null_pos, &null_count))
       {
-        /*
-          clear current error and report INVALID DEFAULT value error message
-          */
-        if (thd->is_error())
-          thd->clear_error();
-
-        my_error(ER_INVALID_DEFAULT, MYF(0), regfield->field_name);
         error= 1;
-        /*
-          Delete to avoid memory leak for fields that allocate extra
-          memory (e.g Field_blob::value)
-        */
-        delete regfield;
         goto err;
       }
     }
-    else if (regfield->real_type() == MYSQL_TYPE_ENUM &&
-	     (field->flags & NOT_NULL_FLAG))
-    {
-      regfield->set_notnull();
-      regfield->store((longlong) 1, TRUE);
-    }
-    else if (type == Field::YES)		// Old unireg type
-      regfield->store(ER(ER_YES),(uint) strlen(ER(ER_YES)),system_charset_info);
-    else if (type == Field::NO)			// Old unireg type
-      regfield->store(ER(ER_NO), (uint) strlen(ER(ER_NO)),system_charset_info);
     else
-      regfield->reset();
-    /*
-      Delete to avoid memory leak for fields that allocate extra
-      memory (e.g Field_blob::value)
-    */
-    delete regfield;
+      has_vgc= true;
+  }
+  if (has_vgc)
+  {
+    it.rewind();
+    while ((field=it++))
+    {
+      if (!field->stored_in_db)
+      {
+        if (make_default_value(thd, &table, field, buff + data_offset,
+                               null_pos, &null_count))
+        {
+          error= 1;
+          goto err;
+        }
+      }
+    }
   }
   DBUG_ASSERT(data_offset == ((null_count + 7) / 8));
 
