@@ -95,7 +95,7 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
   }
 
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-  wsrep_thd_set_query_state(thd, QUERY_EXEC);
+  thd->wsrep_query_state= QUERY_EXEC;
   if (thd->wsrep_conflict_state!= REPLAYING)
     thd->wsrep_conflict_state= NO_CONFLICT;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
@@ -118,10 +118,10 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
     }
 
     switch (ev->get_type_code()) {
-    case FORMAT_DESCRIPTION_EVENT:
+    case binary_log::FORMAT_DESCRIPTION_EVENT:
       wsrep_set_apply_format(thd, (Format_description_log_event*)ev);
       continue;
-    case GTID_LOG_EVENT:
+    case binary_log::GTID_LOG_EVENT:
     {
       Gtid_log_event* gev= (Gtid_log_event*)ev;
       if (gev->get_gno() == 0)
@@ -130,7 +130,6 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
         delete ev;
         continue;
       }
-      break;
     }
     default:
       break;
@@ -138,12 +137,14 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
 
     thd->server_id = ev->server_id; // use the original server id for logging
     thd->set_time();                // time the query
-    wsrep_xid_init(&thd->transaction.xid_state.xid,
+
+    Transaction_ctx *trn_ctx= thd->get_transaction();
+    wsrep_xid_init(trn_ctx->xid_state()->get_xid(),
                    thd->wsrep_trx_meta.gtid.uuid,
                    thd->wsrep_trx_meta.gtid.seqno);
-    thd->lex->current_select= 0;
-    if (!ev->when.tv_sec)
-      my_micro_time_to_timeval(my_micro_time(), &ev->when);
+    thd->lex->set_current_select(0);
+    if (!ev->common_header->when.tv_sec)
+      my_micro_time_to_timeval(my_micro_time(), &ev->common_header->when);
     ev->thd = thd;
     exec_res = ev->apply_event(thd->wsrep_rli);
     DBUG_PRINT("info", ("exec_event result: %d", exec_res));
@@ -178,29 +179,16 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
       DBUG_RETURN(WSREP_CB_FAILURE);
     }
 
-    switch (ev->get_type_code()) {
-    case ROWS_QUERY_LOG_EVENT:
-      /*
-        Setting binlog_rows_query_log_events to ON will generate
-        ROW_QUERY_LOG_EVENT. This event logs an extra information while logging
-        row information and so event should be kept infact till ROW_LOG_EVENT
-        is processed and should be freed once ROW_LOG_EVENT is done.
-
-        Keeping Rows_log event, it will be needed still, and will be deleted later
-        in rli->cleanup_context()
-        Also FORMAT_DESCRIPTION_EVENT is needed further, but it skipped from this loop
-        by 'continue' above, and thus  avoids the following 'delete ev'
-      */
-      continue;
-    default:
+    /* row-query-log-event is generated to log extra information when following
+    configuration is set "binlog_rows_query_log_events". The event is cleared
+    after processing of ROWS_LOG_EVENT. Avoid removing it here. */
+    if (ev->get_type_code() != binary_log::ROWS_QUERY_LOG_EVENT)
       delete ev;
-      break;
-    }
   }
 
  error:
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-  wsrep_thd_set_query_state(thd, QUERY_IDLE);
+  thd->wsrep_query_state= QUERY_IDLE;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
 
   assert(thd->wsrep_exec_mode== REPL_RECV);
@@ -220,29 +208,27 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
 {
   THD* const thd((THD*)ctx);
 
-  assert(thd->wsrep_apply_toi == false);
-
 // Allow tests to block the applier thread using the DBUG facilities
   DBUG_EXECUTE_IF("sync.wsrep_apply_cb",
                  {
                    const char act[]=
                      "now "
-                     "SIGNAL sync.wsrep_apply_cb_reached "
-                     "WAIT_FOR signal.wsrep_apply_cb";
+                     "wait_for signal.wsrep_apply_cb";
                    DBUG_ASSERT(!debug_sync_set_action(thd,
                                                       STRING_WITH_LEN(act)));
                  };);
 
   thd->wsrep_trx_meta = *meta;
 
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_applying_writeset);
+
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "applying write set %lld: %p, %zu",
-           (long long)wsrep_thd_trx_seqno(thd), buf, buf_len);
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "applying write set");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: applying write set (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
+
 
   /* tune FK and UK checking policy */
   if (wsrep_slave_UK_checks == FALSE) 
@@ -266,13 +252,18 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
   }
   wsrep_cb_status_t rcode(wsrep_apply_events(thd, buf, buf_len));
 
+  THD_STAGE_INFO_GUARD_LEAVE();
+
+  /* Provide this for some galera test cases that check the thread state */
+  DBUG_EXECUTE_IF("thd_proc_info.wsrep_apply_cb",
+                  { thd_proc_info(thd, "wsrep: applied write set"); };);
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "applied write set %lld", (long long)wsrep_thd_trx_seqno(thd));
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "applied write set");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: applied write set (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
+
 
   if (WSREP_CB_SUCCESS != rcode)
   {
@@ -282,8 +273,8 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
   TABLE *tmp;
   while ((tmp = thd->temporary_tables))
   {
-    WSREP_DEBUG("Applier %lu, has temporary tables: %s.%s",
-                  thd->thread_id, 
+    WSREP_DEBUG("Applier %u, has temporary tables: %s.%s",
+                  thd->thread_id(), 
                   (tmp->s) ? tmp->s->db.str : "void",
                   (tmp->s) ? tmp->s->table_name.str : "void");
       close_temporary_table(thd, tmp, 1, 1);    
@@ -294,24 +285,30 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
 
 static wsrep_cb_status_t wsrep_commit(THD* const thd)
 {
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_committing);
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "committing %lld", (long long)wsrep_thd_trx_seqno(thd));
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "committing");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: committing (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
+
 
   wsrep_cb_status_t const rcode(trans_commit(thd) ?
                                 WSREP_CB_FAILURE : WSREP_CB_SUCCESS);
 
+  THD_STAGE_INFO_GUARD_LEAVE();
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "committed %lld", (long long)wsrep_thd_trx_seqno(thd));
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "committed");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: committed (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
+
+
+  /* Provide this for some galera test cases that check the thread state */
+  DBUG_EXECUTE_IF("thd_proc_info.wsrep_commit",
+                  { thd_proc_info(thd, "wsrep: committed"); };);
 
   if (WSREP_CB_SUCCESS == rcode)
   {
@@ -329,13 +326,13 @@ static wsrep_cb_status_t wsrep_commit(THD* const thd)
 
 static wsrep_cb_status_t wsrep_rollback(THD* const thd)
 {
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_rolling_back);
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "rolling back %lld", (long long)wsrep_thd_trx_seqno(thd));
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "rolling back");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: rolling back (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
 
   /* Check for comments in Relay_log_info::cleanup_context */
   trans_rollback_stmt(thd);
@@ -343,13 +340,13 @@ static wsrep_cb_status_t wsrep_rollback(THD* const thd)
   wsrep_cb_status_t const rcode(trans_rollback(thd) ?
                                 WSREP_CB_FAILURE : WSREP_CB_SUCCESS);
 
+#ifdef WITH_WSREP
 #ifdef WSREP_PROC_INFO
-  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info) - 1,
-           "rolled back %lld", (long long)wsrep_thd_trx_seqno(thd));
-  thd_proc_info(thd, thd->wsrep_info);
-#else
-  thd_proc_info(thd, "rolled back");
+  if (WSREP(thd))
+    WSREP_DEBUG("wsrep: rolled back (%lld)", (long long) wsrep_thd_trx_seqno(thd));
 #endif /* WSREP_PROC_INFO */
+#endif /* WITH_WSREP */
+
   thd->wsrep_rli->cleanup_context(thd, 0);
   thd->variables.gtid_next.set_automatic();
 
@@ -375,7 +372,24 @@ wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
 
   wsrep_set_apply_format(thd, NULL);
   thd->mdl_context.release_transactional_locks();
-  free_root(thd->mem_root,MYF(MY_KEEP_PREALLOC));
+
+  if (thd->wsrep_applier)
+  {
+    /* This function is meant to initiate a commit after write-set has been
+    applied. This generally will be done by applier thread to replicate
+    action that some other node has executed.
+    There is an exception here: A local action may also follow this route
+    of applying write-set followed by commit if local transaction is forced
+    to rollback. Galera replay_transaction flow will try to use write-set
+    in that case instead of re-running the complete transaction.
+    In latter case it is not adviable to free the mem_root as follow-up
+    action associated with user thread will need it and so this action
+    is executed only when wsrep_applier = true which suggest true applier
+    (Note: this flags is set to flase for user thread even though it is using
+     write-set to complete the action) */
+    free_root(thd->mem_root,MYF(MY_KEEP_PREALLOC));
+  }
+
   thd->tx_isolation= (enum_tx_isolation) thd->variables.tx_isolation;
 
   if (wsrep_slave_count_change < 0 && commit && WSREP_CB_SUCCESS == rcode)
@@ -390,11 +404,15 @@ wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
     mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
   }
 
-  if (thd->wsrep_applier)
+  if (*exit == false && thd->wsrep_applier)
   {
+    /* This should be set as part of trans_begin and not explicitly.
+    Note sure why Galera thought of setting it here but it conflicts
+    as the first BEGIN statement is now intepreted wrongly as statement
+    inside an active transaction. */
     /* From trans_begin() */
-    thd->variables.option_bits|= OPTION_BEGIN;
-    thd->server_status|= SERVER_STATUS_IN_TRANS;
+    // thd->variables.option_bits|= OPTION_BEGIN;
+    //thd->server_status|= SERVER_STATUS_IN_TRANS;
     thd->wsrep_apply_toi= false;
   }
 
