@@ -15,6 +15,7 @@
 
 #include "wsrep_binlog.h"
 #include "wsrep_priv.h"
+#include "log_event.h"
 
 /*
   Write the contents of a cache to a memory buffer.
@@ -56,7 +57,8 @@ int wsrep_write_cache_buf(IO_CACHE *cache, uchar **buf, size_t *buf_len)
           goto error;
       }
 
-      uchar* tmp = (uchar *)my_realloc(*buf, total_length, MYF(0));
+      uchar* tmp = (uchar *)my_realloc(PSI_NOT_INSTRUMENTED,
+                                       *buf, total_length, MYF(0));
       if (!tmp)
       {
           WSREP_ERROR("could not (re)allocate buffer: %zu + %u",
@@ -159,6 +161,22 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
     uint length(my_b_bytes_in_cache(cache));
     if (unlikely(0 == length)) length = my_b_fill(cache);
 
+    /* Start 5.7 MySQL server will write GTID directly to binlog file.
+    This means GTID event is not present in the case if GTID is set explictly.
+    Let's construct the GTID event and prefix it to wsrep-binlog-buffer. */
+    if (thd->variables.gtid_next.type != AUTOMATIC_GROUP)
+    {
+      Gtid_log_event gtid_event(thd, true, 0, 0);
+      uchar gtid_buf[Gtid_log_event::MAX_EVENT_LENGTH];
+      uint32 gtid_len = gtid_event.write_to_memory(gtid_buf);
+
+      assert(Gtid_log_event::MAX_EVENT_LENGTH < STACK_SIZE);
+
+      memcpy(buf, gtid_buf, gtid_len);
+      used = gtid_len;
+      total_length = gtid_len;
+    }
+
     if (likely(length > 0)) do
     {
         total_length += length;
@@ -179,7 +197,8 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
         if (total_length > allocated)
         {
             size_t const new_size(heap_size(total_length));
-            uchar* tmp = (uchar *)my_realloc(heap_buf, new_size, MYF(0));
+            uchar* tmp = (uchar *)my_realloc(PSI_NOT_INSTRUMENTED,
+                                             heap_buf, new_size, MYF(0));
             if (!tmp)
             {
                 WSREP_ERROR("could not (re)allocate buffer: %zu + %u",
@@ -194,7 +213,6 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
 
             if (used <= STACK_SIZE && used > 0) // there's data in stack_buf
             {
-                DBUG_ASSERT(buf == stack_buf);
                 memcpy(heap_buf, stack_buf, used);
             }
         }
@@ -249,6 +267,22 @@ static int wsrep_write_cache_inc(wsrep_t*  const wsrep,
     uint length(my_b_bytes_in_cache(cache));
     if (unlikely(0 == length)) length = my_b_fill(cache);
 
+    /* Start 5.7 MySQL server will write GTID directly to binlog file.
+    This means GTID event is not present in the case if GTID is set explictly.
+    Let's construct the GTID event and prefix it to wsrep-binlog-buffer. */
+    if (thd->variables.gtid_next.type != AUTOMATIC_GROUP)
+    {
+      Gtid_log_event gtid_event(thd, true, 0, 0);
+      uchar gtid_buf[Gtid_log_event::MAX_EVENT_LENGTH];
+      uint32 gtid_len = gtid_event.write_to_memory(gtid_buf);
+
+      if (WSREP_OK != (err=wsrep_append_data(wsrep, &thd->wsrep_ws_handle,
+                                            gtid_buf, gtid_len)))
+        goto cleanup;
+
+      total_length = gtid_len;
+    }
+
     if (likely(length > 0)) do
     {
         total_length += length;
@@ -303,13 +337,11 @@ int wsrep_write_cache(wsrep_t*  const wsrep,
 
 void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
 {
-  /* filename is derived from wsrep_data_home_dir (which in turn from
-  mysql_real_data_home that has max length limit of FN_REFLEN */
-  char filename[FN_REFLEN + 64]= {0};
-  int len= snprintf(filename, FN_REFLEN + 64, "%s/GRA_%ld_%lld.log",
-                    wsrep_data_home_dir, thd->thread_id,
+  char filename[PATH_MAX]= {0};
+  int len= snprintf(filename, PATH_MAX, "%s/GRA_%u_%lld.log",
+                    wsrep_data_home_dir, thd->thread_id(),
                     (long long)wsrep_thd_trx_seqno(thd));
-  if (len >= (FN_REFLEN + 64))
+  if (len >= PATH_MAX)
   {
     WSREP_ERROR("RBR dump path too long: %d, skipping dump.", len);
     return;
@@ -332,15 +364,13 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
 
 void wsrep_dump_rbr_direct(THD* thd, IO_CACHE* cache)
 {
-  /* filename is derived from wsrep_data_home_dir (which in turn from
-  mysql_real_data_home that has max length limit of FN_REFLEN */
-  char filename[FN_REFLEN + 64]= {0};
-  int len= snprintf(filename, FN_REFLEN + 64, "%s/GRA_%ld_%lld.log",
-                    wsrep_data_home_dir, thd->thread_id,
+  char filename[PATH_MAX]= {0};
+  int len= snprintf(filename, PATH_MAX, "%s/GRA_%u_%lld.log",
+                    wsrep_data_home_dir, thd->thread_id(),
                     (long long)wsrep_thd_trx_seqno(thd));
   size_t bytes_in_cache = 0;
   // check path
-  if (len >= (FN_REFLEN + 64))
+  if (len >= PATH_MAX)
   {
     WSREP_ERROR("RBR dump path too long: %d, skipping dump.", len);
     return ;
