@@ -39,15 +39,13 @@ enum wsrep_trx_status wsrep_run_wsrep_commit(THD *thd, handlerton *hton,
 void wsrep_cleanup_transaction(THD *thd)
 {
   if (!WSREP(thd)) return;
-  DBUG_ASSERT(thd->wsrep_conflict_state != MUST_REPLAY &&
-              thd->wsrep_conflict_state != REPLAYING);
-  if (wsrep_emulate_bin_log) thd_binlog_trx_reset(thd);
+  if (wsrep_emulate_bin_log) wsrep_thd_binlog_trx_reset(thd);
   thd->wsrep_ws_handle.trx_id= WSREP_UNDEFINED_TRX_ID;
   thd->wsrep_trx_meta.gtid= WSREP_GTID_UNDEFINED;
   thd->wsrep_trx_meta.depends_on= WSREP_SEQNO_UNDEFINED;
   thd->wsrep_exec_mode= LOCAL_STATE;
+  thd->wsrep_certify_empty_trx= false;
   thd->wsrep_affected_rows= 0;
-  thd->wsrep_skip_wsrep_GTID= false;
   return;
 }
 
@@ -74,15 +72,6 @@ handlerton *wsrep_hton;
 */
 void wsrep_register_hton(THD* thd, bool all)
 {
-  /* Skip wsrep-hton handler registration.
-  This happens in case like following:
-  - DDL is executed with sql_log_bin = 0 where-in ddl replication is skipped
-    so the exec_mode != TOTAL_ORDER but as thumb rule hton should get registered
-    for DDL/TOI statement.
-  */
-  if (WSREP(thd) && thd->wsrep_skip_wsrep_hton)
-    return;
-
   if (WSREP(thd) && thd->wsrep_exec_mode != TOTAL_ORDER && !thd->wsrep_apply_toi)
   {
     if (thd->wsrep_exec_mode == LOCAL_STATE      &&
@@ -96,18 +85,23 @@ void wsrep_register_hton(THD* thd, bool all)
         return;
     }
 
-    THD_TRANS *trans=all ? &thd->transaction.all : &thd->transaction.stmt;
-    for (Ha_trx_info *i= trans->ha_list; i; i = i->next())
+    Transaction_ctx *trn_ctx= thd->get_transaction();
+    Transaction_ctx::enum_trx_scope trx_scope=
+      all ? Transaction_ctx::SESSION : Transaction_ctx::STMT;
+    Ha_trx_info *ha_info= trn_ctx->ha_trx_info(trx_scope), *ha_info_next;
+
+    for (; ha_info; ha_info= ha_info_next)
     {
-      if (i->ht()->db_type == DB_TYPE_INNODB)
+      handlerton *ht= ha_info->ht();
+      if (ht->db_type == DB_TYPE_INNODB)
       {
-        trans_register_ha(thd, all, wsrep_hton);
+        trans_register_ha(thd, all, wsrep_hton, NULL);
 
         /* follow innodb read/write settting
          * but, as an exception: CTAS with empty result set will not be
          * replicated unless we declare wsrep hton as read/write here
 	 */
-        if (i->is_trx_read_write() ||
+        if (ha_info->is_trx_read_write() ||
             (thd->lex->sql_command == SQLCOM_CREATE_TABLE &&
              thd->wsrep_exec_mode == LOCAL_STATE))
         {
@@ -115,6 +109,7 @@ void wsrep_register_hton(THD* thd, bool all)
         }
         break;
       }
+      ha_info_next= ha_info->next();
     }
   }
 }
@@ -146,28 +141,9 @@ void wsrep_post_commit(THD* thd, bool all)
      /* non-InnoDB statements may have populated events in stmt cache 
 	=> cleanup 
      */
-     if (thd->wsrep_conflict_state != MUST_REPLAY)
-     {
-       WSREP_DEBUG("cleanup transaction for LOCAL_STATE: %s",
-                   WSREP_QUERY(thd));
-     }
-     /*
-       Run post-rollback hook to clean up in the case if
-       some keys were populated for the transaction in provider
-       but during commit time there was no write set to replicate.
-       This may happen when client sets the SAVEPOINT and immediately
-       rolls back to savepoint after first operation. 
-      */
-     if (all && thd->wsrep_conflict_state != MUST_REPLAY &&
-         wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle))
-     {
-         WSREP_WARN("post_rollback fail: %llu %d",
-                    (long long)thd->thread_id, thd->get_stmt_da()->status());
-     }
-     if (thd->wsrep_conflict_state != MUST_REPLAY)
-     {
-       wsrep_cleanup_transaction(thd);
-     }
+     WSREP_DEBUG("cleanup transaction for LOCAL_STATE: %s",
+                 WSREP_QUERY(thd));
+     wsrep_cleanup_transaction(thd);
      break;
    }
   default: break;
@@ -281,7 +257,8 @@ static int wsrep_rollback(handlerton *hton, THD *thd, bool all)
     {
       DBUG_PRINT("wsrep", ("setting rollback fail"));
       WSREP_ERROR("settting rollback fail: thd: %llu, schema: %s, SQL: %s",
-                  (long long)thd->real_id, (thd->db ? thd->db : "(null)"),
+                  (long long)thd->real_id,
+                  ((thd->db().length != 0) ? thd->db().str : "(null)"),
                   WSREP_QUERY(thd));
     }
     wsrep_cleanup_transaction(thd);
@@ -322,7 +299,8 @@ int wsrep_commit(handlerton *hton, THD *thd, bool all)
       {
         DBUG_PRINT("wsrep", ("setting rollback fail"));
         WSREP_ERROR("settting rollback fail: thd: %llu, schema: %s, SQL: %s",
-                    (long long)thd->real_id, (thd->db ? thd->db : "(null)"),
+                    (long long)thd->real_id,
+                    ((thd->db().length != 0) ? thd->db().str : "(null)"),
                     WSREP_QUERY(thd));
       }
       wsrep_cleanup_transaction(thd);
@@ -345,8 +323,9 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   int replay_round= 0;
 
   if (thd->get_stmt_da()->is_error()) {
-    WSREP_DEBUG("commit issue, error: %d %s",
-                thd->get_stmt_da()->sql_errno(), thd->get_stmt_da()->message());
+    WSREP_ERROR("commit issue, error: %d %s",
+                thd->get_stmt_da()->mysql_errno(),
+                thd->get_stmt_da()->message_text());
   }
 
   DBUG_ENTER("wsrep_run_wsrep_commit");
@@ -380,6 +359,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   }
 
   DBUG_PRINT("wsrep", ("replicating commit"));
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_replicating_commit);
 
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
   if (thd->wsrep_conflict_state == MUST_ABORT) {
@@ -395,17 +375,17 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
   while (wsrep_replaying > 0                       &&
          thd->wsrep_conflict_state == NO_CONFLICT  &&
          thd->killed == THD::NOT_KILLED            &&
-         !shutdown_in_progress)
+         !abort_loop)
   {
 
     mysql_mutex_unlock(&LOCK_wsrep_replaying);
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
 
-    mysql_mutex_lock(&thd->mysys_var->mutex);
-    thd_proc_info(thd, "wsrep waiting on replaying");
-    thd->mysys_var->current_mutex= &LOCK_wsrep_replaying;
-    thd->mysys_var->current_cond=  &COND_wsrep_replaying;
-    mysql_mutex_unlock(&thd->mysys_var->mutex);
+    mysql_mutex_lock(&thd->LOCK_current_cond);
+    THD_STAGE_INFO_GUARD_UPDATE(&stage_wsrep_waiting_on_replaying);
+    thd->current_mutex= &LOCK_wsrep_replaying;
+    thd->current_cond= &COND_wsrep_replaying;
+    mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     mysql_mutex_lock(&LOCK_wsrep_replaying);
     // Using timedwait is a hack to avoid deadlock in case if BF victim
@@ -420,22 +400,23 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
 			 &wtime);
 
     if (replay_round++ % 100000 == 0)
-      WSREP_DEBUG("commit waiting for replaying: replayers %d, thd: (%lu) "
+      WSREP_DEBUG("commit waiting for replaying: replayers %d, thd: (%u) "
                   "conflict: %d (round: %d)",
-		  wsrep_replaying, thd->thread_id,
+		  wsrep_replaying, thd->thread_id(),
                   thd->wsrep_conflict_state, replay_round);
 
     mysql_mutex_unlock(&LOCK_wsrep_replaying);
 
-    mysql_mutex_lock(&thd->mysys_var->mutex);
-    thd->mysys_var->current_mutex= 0;
-    thd->mysys_var->current_cond=  0;
-    mysql_mutex_unlock(&thd->mysys_var->mutex);
+    mysql_mutex_lock(&thd->LOCK_current_cond);
+    thd->current_mutex= 0;
+    thd->current_cond= 0;
+    mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     mysql_mutex_lock(&thd->LOCK_wsrep_thd);
     mysql_mutex_lock(&LOCK_wsrep_replaying);
   }
   mysql_mutex_unlock(&LOCK_wsrep_replaying);
+  THD_STAGE_INFO_GUARD_LEAVE();
 
   if (thd->wsrep_conflict_state == MUST_ABORT) {
     DBUG_PRINT("wsrep", ("replicate commit fail"));
@@ -446,11 +427,10 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     DBUG_RETURN(WSREP_TRX_CERT_FAIL);
   }
 
-  wsrep_thd_set_query_state(thd, QUERY_COMMITTING);
+  thd->wsrep_query_state = QUERY_COMMITTING;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
 
-  DEBUG_SYNC(thd, "wsrep_before_certification");
-  cache = get_trans_log(thd);
+  cache = wsrep_get_trans_log(thd);
   rcode = 0;
   if (cache) {
     thd->binlog_flush_pending_rows_event(true);
@@ -474,7 +454,7 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
                   "wsrep status (%d %d %d)",
                   WSREP_QUERY(thd),
                   thd->get_stmt_da()->affected_rows(),
-                  stmt_has_updated_trans_table(thd->transaction.stmt.ha_list),
+                  stmt_has_updated_trans_table(thd->get_transaction()->ha_trx_info(Transaction_ctx::STMT)),
                   thd->variables.sql_log_bin,
                   thd->wsrep_exec_mode, thd->wsrep_query_state,
                   thd->wsrep_conflict_state);
@@ -483,63 +463,73 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     {
       WSREP_DEBUG("empty rbr buffer, query: %s", WSREP_QUERY(thd));
     }
-    wsrep_thd_set_query_state(thd, QUERY_EXEC);
+    thd->wsrep_query_state= QUERY_EXEC;
     DBUG_RETURN(WSREP_TRX_OK);
   }
 
   if (WSREP_UNDEFINED_TRX_ID == thd->wsrep_ws_handle.trx_id)
   {
-    WSREP_WARN("SQL statement was ineffective, THD: %lu, buf: %zu\n"
+    WSREP_WARN("SQL statement was ineffective, THD: %u, buf: %zu\n"
                "schema: %s \n"
 	       "QUERY: %s\n"
 	       " => Skipping replication",
-	       thd->thread_id, data_len,
-               (thd->db ? thd->db : "(null)"),
+	       thd->thread_id(), data_len,
+               ((thd->db().length != 0) ? thd->db().str : "(null)"),
                WSREP_QUERY(thd));
     rcode = WSREP_TRX_FAIL;
   }
   else if (!rcode)
   {
+    THD_STAGE_INFO_GUARD(NULL, NULL);
     if (WSREP_OK == rcode) {
-      if (WSREP(thd))
-        thd_proc_info(thd, "wsrep in pre-commit stage");
+      if (WSREP(thd)) {
+        THD_STAGE_INFO_GUARD_ENTER(thd, &stage_wsrep_pre_commit);
+      }
       rcode = wsrep->pre_commit(wsrep,
-                                (wsrep_conn_id_t)thd->thread_id,
+                                (wsrep_conn_id_t)thd->thread_id(),
                                 &thd->wsrep_ws_handle,
                                 WSREP_FLAG_COMMIT |
                                 ((thd->wsrep_PA_safe) ?
                                  0ULL : WSREP_FLAG_PA_UNSAFE),
                                 &thd->wsrep_trx_meta);
+
+      /* Provide this for some galera test cases that check the thread state */
+      if (WSREP(thd)) {
+        THD_STAGE_INFO_GUARD_LEAVE();
+        DBUG_EXECUTE_IF("thd_proc_info.wsrep_run_wsrep_commit",
+                        { thd_proc_info(thd, "wsrep: in pre-commit stage"); };);
+      }
     }
 
     if (rcode == WSREP_TRX_MISSING) {
-      WSREP_WARN("Transaction missing in provider thd: %ld schema: %s SQL: %s",
-                 thd->thread_id, (thd->db ? thd->db : "(null)"),
+      WSREP_WARN("Transaction missing in provider thd: %u schema: %s SQL: %s",
+                 thd->thread_id(),
+                 ((thd->db().length != 0) ? thd->db().str : "(null)"),
                  WSREP_QUERY(thd));
       rcode = WSREP_TRX_FAIL;
     } else if (rcode == WSREP_BF_ABORT) {
-      WSREP_DEBUG("thd %lu seqno %lld BF aborted by provider, will replay",
-                  thd->thread_id, (long long)thd->wsrep_trx_meta.gtid.seqno);
+      WSREP_DEBUG("thd %u seqno %lld BF aborted by provider, will replay",
+                  thd->thread_id(), (long long)thd->wsrep_trx_meta.gtid.seqno);
       mysql_mutex_lock(&thd->LOCK_wsrep_thd);
       thd->wsrep_conflict_state = MUST_REPLAY;
       DBUG_ASSERT(wsrep_thd_trx_seqno(thd) > 0);
       mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
       mysql_mutex_lock(&LOCK_wsrep_replaying);
       wsrep_replaying++;
-      WSREP_DEBUG("replaying increased: %d, thd: %lu",
-                  wsrep_replaying, thd->thread_id);
+      WSREP_DEBUG("replaying increased: %d, thd: %u",
+                  wsrep_replaying, thd->thread_id());
       mysql_mutex_unlock(&LOCK_wsrep_replaying);
     }
   } else {
     WSREP_ERROR("I/O error reading from thd's binlog iocache: "
-                "errno=%d, io cache code=%d", my_errno, cache->error);
+                "errno=%d, io cache code=%d", my_errno(), cache->error);
     DBUG_ASSERT(0); // failure like this can not normally happen
     DBUG_RETURN(WSREP_TRX_ERROR);
   }
 
-  DEBUG_SYNC(thd, "wsrep_after_replication");
-
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
+  Transaction_ctx *trn_ctx= thd->get_transaction();
+
   switch(rcode) {
   case 0:
     /*
@@ -549,37 +539,36 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
       set to NO_CONFLICT and commit proceeds as usual.
     */
     if (thd->wsrep_conflict_state == MUST_ABORT)
-    {
-      thd->killed= THD::NOT_KILLED;
-      thd->wsrep_conflict_state= NO_CONFLICT;
-    }
+        thd->wsrep_conflict_state= NO_CONFLICT;
 
     if (thd->wsrep_conflict_state != NO_CONFLICT)
     {
-      WSREP_WARN("thd %lu seqno %lld: conflict state %d after post commit",
-                 thd->thread_id,
+      WSREP_WARN("thd %u seqno %lld: conflict state %d after post commit",
+                 thd->thread_id(),
                  (long long)thd->wsrep_trx_meta.gtid.seqno,
                  thd->wsrep_conflict_state);
     }
     thd->wsrep_exec_mode= LOCAL_COMMIT;
     DBUG_ASSERT(thd->wsrep_trx_meta.gtid.seqno != WSREP_SEQNO_UNDEFINED);
     /* Override XID iff it was generated by mysql */
-    if (thd->transaction.xid_state.xid.get_my_xid())
+    if (trn_ctx->xid_state()->get_xid()->get_my_xid())
     {
-      wsrep_xid_init(&thd->transaction.xid_state.xid,
+      wsrep_xid_init(trn_ctx->xid_state()->get_xid(),
                      thd->wsrep_trx_meta.gtid.uuid,
                      thd->wsrep_trx_meta.gtid.seqno);
     }
     DBUG_PRINT("wsrep", ("replicating commit success"));
     break;
+
   case WSREP_BF_ABORT:
     DBUG_ASSERT(thd->wsrep_trx_meta.gtid.seqno != WSREP_SEQNO_UNDEFINED);
+
   case WSREP_TRX_FAIL:
-    WSREP_DEBUG("commit failed for reason: %d %lu %s",
-                rcode, thd->thread_id, WSREP_QUERY(thd));
+    WSREP_DEBUG("commit failed for reason: %d %u %s",
+                rcode, thd->thread_id(), WSREP_QUERY(thd));
     DBUG_PRINT("wsrep", ("replicating commit fail"));
 
-    wsrep_thd_set_query_state(thd, QUERY_EXEC);
+    thd->wsrep_query_state= QUERY_EXEC;
 
     if (thd->wsrep_conflict_state == MUST_ABORT) {
       thd->wsrep_conflict_state= ABORTED;
@@ -601,54 +590,24 @@ wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all)
     WSREP_ERROR("transaction size exceeded");
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
     DBUG_RETURN(WSREP_TRX_SIZE_EXCEEDED);
+
   case WSREP_CONN_FAIL:
     WSREP_ERROR("connection failure");
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
     DBUG_RETURN(WSREP_TRX_ERROR);
+
   default:
     WSREP_ERROR("unknown connection failure");
     mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
     DBUG_RETURN(WSREP_TRX_ERROR);
   }
 
-  wsrep_thd_set_query_state(thd, QUERY_EXEC);
+  thd->wsrep_query_state= QUERY_EXEC;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
 
   DBUG_RETURN(WSREP_TRX_OK);
 }
 
-bool wsrep_replicate_GTID(THD *thd)
-{
-  if (thd->slave_thread)
-  {
-    WSREP_DEBUG("GTID replication");
-    DBUG_ASSERT (WSREP_UNDEFINED_TRX_ID == thd->wsrep_ws_handle.trx_id);
-    (void)wsrep_ws_handle_for_trx(&thd->wsrep_ws_handle, thd->query_id);
-    DBUG_ASSERT (WSREP_UNDEFINED_TRX_ID != thd->wsrep_ws_handle.trx_id);
-    WSREP_DEBUG("slave trx using query ID %llu for replication GTID",
-                (ulonglong) thd->wsrep_ws_handle.trx_id);
-    enum wsrep_trx_status rcode= wsrep_run_wsrep_commit(thd, wsrep_hton, true);
-    if (rcode)
-    {
-      /*
-        TODO: should error here cause stopping of MySQL slave?
-        Slave applying was totally filtered out, and fauílure in replicating
-        GTID event, would cause a hole in GTID history in other cluster nodes
-
-      */
-      WSREP_INFO("GTID replication failed: %d", rcode);
-      wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle);
-      thd->wsrep_replicate_GTID= false;
-      my_message(ER_ERROR_DURING_COMMIT,
-                 "WSREP GTID replication was interrupted", MYF(0));
-      return true;
-    }
-    wsrep_post_commit(thd, true);
-  }
-  thd->wsrep_replicate_GTID= false;
-
-  return false;
-}
 
 static int wsrep_hton_init(void *p)
 {
