@@ -17,12 +17,13 @@
 #define WSREP_MYSQLD_H
 
 #include "mysqld.h"
-#include "wsrep_api.h"
+#include "log.h"
 typedef struct st_mysql_show_var SHOW_VAR;
-#include <sql_priv.h>
+#include "query_options.h"
 #include "rpl_gtid.h"
+#include "../wsrep/wsrep_api.h"
 
-#define WSREP_UNDEFINED_TRX_ID ULONGLONG_MAX
+#define WSREP_UNDEFINED_TRX_ID ULLONG_MAX
 
 class set_var;
 class THD;
@@ -59,14 +60,12 @@ enum wsrep_consistency_check_mode {
     CONSISTENCY_CHECK_RUNNING,
 };
 
-enum enum_wsrep_certification_rules {
-    WSREP_CERTIFICATION_RULES_STRICT,
-    WSREP_CERTIFICATION_RULES_OPTIMIZED
-};
 
 // Global wsrep parameters
 extern wsrep_t*    wsrep;
 
+extern bool        wsrep_replicate_set_stmt;
+
 // MySQL wsrep options
 extern const char* wsrep_provider;
 extern const char* wsrep_provider_options;
@@ -90,7 +89,6 @@ extern ulong       wsrep_max_ws_size;
 extern ulong       wsrep_max_ws_rows;
 extern const char* wsrep_notify_cmd;
 extern my_bool     wsrep_certify_nonPK;
-extern ulong       wsrep_certification_rules;
 extern long        wsrep_max_protocol_version;
 extern long        wsrep_protocol_version;
 extern ulong       wsrep_forced_binlog_format;
@@ -98,14 +96,11 @@ extern my_bool     wsrep_desync;
 extern ulong       wsrep_reject_queries;
 extern my_bool     wsrep_recovery;
 extern my_bool     wsrep_log_conflicts;
-extern ulong       wsrep_mysql_replication_bundle;
 extern my_bool     wsrep_load_data_splitting;
 extern my_bool     wsrep_restart_slave;
 extern my_bool     wsrep_restart_slave_activated;
 extern my_bool     wsrep_slave_FK_checks;
 extern my_bool     wsrep_slave_UK_checks;
-extern ulong       wsrep_running_threads;
-extern ulong       wsrep_RSU_commit_timeout;
 
 enum enum_wsrep_reject_types {
   WSREP_REJECT_NONE,    /* nothing rejected */
@@ -121,16 +116,24 @@ enum enum_wsrep_OSU_method {
 
 enum enum_wsrep_sync_wait {
     WSREP_SYNC_WAIT_NONE = 0x0,
-    // select, begin
+    // show, select, begin
     WSREP_SYNC_WAIT_BEFORE_READ = 0x1,
     WSREP_SYNC_WAIT_BEFORE_UPDATE_DELETE = 0x2,
     WSREP_SYNC_WAIT_BEFORE_INSERT_REPLACE = 0x4,
-    WSREP_SYNC_WAIT_BEFORE_SHOW = 0x8,
-    WSREP_SYNC_WAIT_MAX = 0xF
+    WSREP_SYNC_WAIT_MAX = 0x7
 };
 
+enum enum_pxc_strict_modes {
+    PXC_STRICT_MODE_DISABLED = 0,
+    PXC_STRICT_MODE_PERMISSIVE,
+    PXC_STRICT_MODE_ENFORCING,
+    PXC_STRICT_MODE_MASTER
+};
+extern ulong       pxc_strict_mode;
+
 // MySQL status variables
 extern my_bool     wsrep_connected;
+extern my_bool     wsrep_ready;
 extern const char* wsrep_cluster_state_uuid;
 extern long long   wsrep_cluster_conf_id;
 extern const char* wsrep_cluster_status;
@@ -142,7 +145,6 @@ extern const char* wsrep_provider_version;
 extern const char* wsrep_provider_vendor;
 
 int  wsrep_show_status(THD *thd, SHOW_VAR *var, char *buff);
-int  wsrep_show_ready(THD *thd, SHOW_VAR *var, char *buff);
 void wsrep_free_status(THD *thd);
 
 /* Filters out --wsrep-new-cluster oprtion from argv[]
@@ -158,9 +160,6 @@ bool wsrep_before_SE(); // initialize wsrep before storage
  * @param before wsrep_before_SE() value */
 void wsrep_init_startup(bool before);
 
-/* Setup SST request methods that will be accepted by server */
-bool wsrep_setup_allowed_sst_methods();
-
 
 extern "C" enum wsrep_exec_mode wsrep_thd_exec_mode(THD *thd);
 extern "C" enum wsrep_conflict_state wsrep_thd_conflict_state(THD *thd);
@@ -185,15 +184,14 @@ extern "C" time_t wsrep_thd_query_start(THD *thd);
 extern "C" my_thread_id wsrep_thd_thread_id(THD *thd);
 extern "C" int64_t wsrep_thd_trx_seqno(THD *thd);
 extern "C" query_id_t wsrep_thd_query_id(THD *thd);
-extern "C" char * wsrep_thd_query(THD *thd);
+extern "C" const char * wsrep_thd_query(THD *thd);
 extern "C" query_id_t wsrep_thd_wsrep_last_query_id(THD *thd);
 extern "C" void wsrep_thd_set_wsrep_last_query_id(THD *thd, query_id_t id);
 extern "C" void wsrep_thd_awake(THD *thd, my_bool signal);
 extern "C" int wsrep_thd_retry_counter(THD *thd);
 
-extern "C" void wsrep_handle_fatal_signal(int sig);
 
-extern void wsrep_close_client_connections(my_bool wait_to_end);
+extern void wsrep_close_client_connections(bool wait_to_end, bool server_shutdown);
 extern int  wsrep_wait_committing_connections_close(int wait_time);
 extern void wsrep_close_applier(THD *thd);
 extern void wsrep_wait_appliers_close(THD *thd);
@@ -209,32 +207,19 @@ extern int  wsrep_check_opts (int argc, char* const* argv);
 extern void wsrep_prepend_PATH (const char* path);
 /* some inline functions are defined in wsrep_mysqld_inl.h */
 
-/* Provide a wrapper of the WSREP_ON macro for plugins to use */
-extern "C" bool wsrep_is_wsrep_on(void);
-
-
 /* Other global variables */
 extern wsrep_seqno_t wsrep_locked_seqno;
 
-extern bool wsrep_provider_set;
-
-#define WSREP_ON                          \
-  ((global_system_variables.wsrep_on) &&  \
-   wsrep_provider_set)
-
-/* use xxxxxx_NNULL macros when thd pointer is guaranteed to be non-null to
- * avoid compiler warnings (GCC 6 and later) */
-#define WSREP_NNULL(thd) \
-  (WSREP_ON && wsrep && thd->variables.wsrep_on)
+#define WSREP_ON                         \
+  ((global_system_variables.wsrep_on) && \
+   wsrep_provider                     && \
+   strcmp(wsrep_provider, WSREP_NONE))
 
 #define WSREP(thd) \
-  (thd && WSREP_NNULL(thd))
+  (WSREP_ON && wsrep && (thd && thd->variables.wsrep_on))
 
 #define WSREP_CLIENT(thd) \
-  (WSREP(thd) && thd->wsrep_client_thread)
-
-#define WSREP_EMULATE_BINLOG_NNULL(thd) \
-  (WSREP_NNULL(thd) && wsrep_emulate_bin_log)
+    (WSREP(thd) && thd->wsrep_client_thread)
 
 #define WSREP_EMULATE_BINLOG(thd) \
   (WSREP(thd) && wsrep_emulate_bin_log)
@@ -243,8 +228,8 @@ extern bool wsrep_provider_set;
 // This is a workaround. It also prefixes all messages with "WSREP"
 #define WSREP_LOG(fun, ...)                                       \
     {                                                             \
-        char msg[4096] = {'\0'};                                  \
-        snprintf(msg, sizeof(msg) - 1, ## __VA_ARGS__);           \
+        char msg[1024] = {'\0'};                                  \
+        snprintf(msg, sizeof(msg) - 1, ##__VA_ARGS__);            \
         fun("WSREP: %s", msg);                                    \
     }
 
@@ -257,7 +242,7 @@ extern bool wsrep_provider_set;
 #define WSREP_LOG_CONFLICT_THD(thd, role)                                      \
     WSREP_LOG(sql_print_information, 	                                       \
       "%s: \n "       	                                                       \
-      "  THD: %lu, mode: %s, state: %s, conflict: %s, seqno: %lld\n "          \
+      "  THD: %u, mode: %s, state: %s, conflict: %s, seqno: %lld\n "          \
       "  SQL: %s",							       \
       role, wsrep_thd_thread_id(thd), wsrep_thd_exec_mode_str(thd),            \
       wsrep_thd_query_state_str(thd),                                          \
@@ -275,11 +260,10 @@ extern bool wsrep_provider_set;
     if (victim_thd) WSREP_LOG_CONFLICT_THD(victim_thd, "Victim thread");       \
   }
 
-#define WSREP_QUERY(thd)                                \
-  ((!opt_log_raw) && thd->rewritten_query.length()      \
-   ? thd->rewritten_query.c_ptr_safe() : thd->query())
+#define WSREP_QUERY(thd)					\
+  (((!opt_general_log_raw) && thd->rewritten_query.length())	\
+   ? (thd->rewritten_query.c_ptr_safe()) : (thd->query().str))
 
-extern my_bool wsrep_ready_get();
 extern void wsrep_ready_wait();
 
 enum wsrep_trx_status {
@@ -333,20 +317,24 @@ extern PSI_mutex_key key_LOCK_wsrep_sst;
 extern PSI_cond_key  key_COND_wsrep_sst;
 extern PSI_mutex_key key_LOCK_wsrep_sst_init;
 extern PSI_cond_key  key_COND_wsrep_sst_init;
-extern PSI_mutex_key key_LOCK_wsrep_sst_thread;
-extern PSI_cond_key  key_COND_wsrep_sst_thread;
 extern PSI_mutex_key key_LOCK_wsrep_rollback;
 extern PSI_cond_key  key_COND_wsrep_rollback;
 extern PSI_mutex_key key_LOCK_wsrep_replaying;
 extern PSI_cond_key  key_COND_wsrep_replaying;
 extern PSI_mutex_key key_LOCK_wsrep_slave_threads;
 extern PSI_mutex_key key_LOCK_wsrep_desync;
+
+extern PSI_mutex_key key_LOCK_wsrep_sst_thread;
+extern PSI_cond_key  key_COND_wsrep_sst_thread;
+
+extern PSI_thread_key key_THREAD_wsrep_sst_joiner;
+extern PSI_thread_key key_THREAD_wsrep_sst_donor;
+extern PSI_thread_key key_THREAD_wsrep_applier;
+extern PSI_thread_key key_THREAD_wsrep_rollbacker;
 #endif /* HAVE_PSI_INTERFACE */
 struct TABLE_LIST;
-class Alter_info;
-int wsrep_to_isolation_begin(THD *thd, char *db_, char *table_,
-                             const TABLE_LIST* table_list,
-                             Alter_info* alter_info = NULL);
+int wsrep_to_isolation_begin(THD *thd, const char *db_, const char *table_,
+                             const TABLE_LIST* table_list);
 void wsrep_to_isolation_end(THD *thd);
 void wsrep_cleanup_transaction(THD *thd);
 int wsrep_to_buf_helper(
@@ -361,17 +349,4 @@ bool wsrep_stmt_rollback_is_safe(THD* thd);
 void wsrep_init_sidno(const wsrep_uuid_t&);
 bool wsrep_node_is_donor();
 bool wsrep_node_is_synced();
-bool wsrep_replicate_GTID(THD* thd);
-
-typedef struct wsrep_key_arr
-{
-    wsrep_key_t* keys;
-    size_t       keys_len;
-} wsrep_key_arr_t;
-bool wsrep_prepare_keys_for_isolation(THD*              thd,
-                                      const char*       db,
-                                      const char*       table,
-                                      const TABLE_LIST* table_list,
-                                      wsrep_key_arr_t*  ka);
-void wsrep_keys_free(wsrep_key_arr_t* key_arr);
 #endif /* WSREP_MYSQLD_H */
