@@ -13,90 +13,25 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
-#include "wsrep_sst.h"
-
 #include <mysqld.h>
 #include <sql_class.h>
-#include <set_var.h>
-#include <sql_acl.h>
-#include <sql_reload.h>
 #include <sql_parse.h>
-#include <mysql_version.h>
+#include <sql_reload.h>
+#include "debug_sync.h"
 #include "wsrep_priv.h"
+#include "wsrep_thd.h"
+#include "wsrep_sst.h"
 #include "wsrep_utils.h"
+#include "wsrep_var.h"
+#include "wsrep_binlog.h"
+#include "wsrep_applier.h"
 #include "wsrep_xid.h"
 #include <cstdio>
 #include <cstdlib>
-#include <functional>
-#include <sstream>
-#include <string>
-
-/* We don't need PSI instrumentation for regex mutex, however
-   Regex class has its constructor compiled conditionally.
-   Disable instrumentation.
-   Regex class uses c++11 nullptr. As we are in c++03 world
-   let's define nullptr to satisfy compiler */
-#ifdef HAVE_PSI_INTERFACE
-#undef HAVE_PSI_INTERFACE
-#endif
-#define nullptr 0
-#include <sql_regex.h>
-
-/* Check if the string is valid. In method name we allow only:
-   alpha-num
-   underscore
-   dash
-   dot
-   but we do not allow leading and trailing dash/dot.
-   Method names have to be separated by colons (spaces before/after colon are allowed).
-   Below regex may seem difficult, but in fact it is not:
-
-   1. Any number of leading spaces
-
-     2. Followed by  alpha-num character or underscore
-       3. Followed by any number of underscore/dash/dot
-       4. Followed by alpha-num character or underscore
-       5. 3 - 4 group is optional
-     6. Followed by colon (possibly with leading and/or trailing spaces)
-     7. 2 - 6 group is optional
-
-   8. Followed by alpha-num character or underscore
-     9. Followed by any number of underscore/dash/dot
-     10. Followed by alpha-num character or underscore
-     11. 9 -10 group is optional
-
-   11. Followed by any number of trailing spaces
-
-   It is the same as:
-   ^\s*(\w([-_.]*\w)*(\s*,\s*))*\w([-_.]*\w)*\s*$    */
-
-#define COLON_REGEX "([[:blank:]]*,[[:blank:]]*)"
-#define METHOD_REGEX "[[:alnum:]_]([-_.]*[[:alnum:]_])*"
-#define SST_ALLOWED_METHODS_REGEX_PATTERN \
-    "^[[:blank:]]*(" METHOD_REGEX COLON_REGEX ")*" METHOD_REGEX "[[:blank:]]*$"
-
-static Regex sst_allowed_methods_regex;
-
-/* We allow custom sst scripts, so data can be anything.
-
-   We could create and maintain the list of forbidden
-   characters and the ways they could be used to inject the command to
-   the OS. However this approach seems to be too error prone.
-   Instead of this we will just allow alpha-num + a few special characters
-   (colon, slash, dot, underscore, square brackets, hyphen).
-
-   It is the same as ^[\w:/.[\]-]+$    */
-
-   /* For some reason regex engine is not able to handle the following pattern
-   allowing square brackets:
-   "^[[:alnum:]:/._[\\]-]+$"
-   We will substitute square brackets with lt/gt for validation.
-   Because of another regex engine problem, hyphen literal has to be located
-   at the end of matching set (cannot be escaped when in the middle). */
-static const char *sst_method_allowed_chars_regex_pattern=
-    "^[[:alnum:]:/._<>-]+$";
-static Regex sst_method_allowed_chars_regex;
-
+#include "log_event.h"
+#include <rpl_slave.h>
+#include "sql_base.h"           // TEMP_PREFIX 
+#include "rpl_msr.h"           // channel_map
 
 extern const char wsrep_defaults_file[];
 extern const char wsrep_defaults_group_suffix[];
@@ -109,7 +44,6 @@ extern const char wsrep_defaults_group_suffix[];
 #define WSREP_SST_OPT_CONF_SUFFIX "--defaults-group-suffix"
 #define WSREP_SST_OPT_PARENT   "--parent"
 #define WSREP_SST_OPT_BINLOG   "--binlog"
-#define WSREP_SST_OPT_VERSION  "--mysqld-version"
 
 // mysqldump-specific options
 #define WSREP_SST_OPT_USER     "--user"
@@ -137,15 +71,6 @@ const char* wsrep_sst_receive_address = WSREP_SST_ADDRESS_AUTO;
 const char* wsrep_sst_donor           = "";
       char* wsrep_sst_auth            = NULL;
 
-#define WSREP_SST_ALLOWED_METHODS_DEFAULT WSREP_SST_MYSQLDUMP "," \
-                                          WSREP_SST_RSYNC "," \
-                                          WSREP_SST_SKIP "," \
-                                          WSREP_SST_XTRABACKUP "," \
-                                          WSREP_SST_XTRABACKUP_V2
-
-const char *wsrep_sst_allowed_methods = WSREP_SST_ALLOWED_METHODS_DEFAULT;
-static std::vector<std::string> allowed_sst_methods;
-
 // container for real auth string
 static const char* sst_auth_real      = NULL;
 my_bool wsrep_sst_donor_rejects_queries = FALSE;
@@ -171,6 +96,7 @@ bool wsrep_sst_method_update (sys_var *self, THD* thd, enum_var_type type)
     return 0;
 }
 
+#if 0
 static bool sst_receive_address_check (const char* str)
 {
     if (!strncasecmp(str, "127.0.0.1", strlen("127.0.0.1")) ||
@@ -181,6 +107,7 @@ static bool sst_receive_address_check (const char* str)
 
     return 0;
 }
+#endif
 
 /* Function checks if the new value for sst_recieve_address is valid.
 @return false if no error encountered with check else return true. */
@@ -198,10 +125,12 @@ bool wsrep_sst_receive_address_check (sys_var *self, THD* thd, set_var* var)
          var->save_result.string_value.length);
   addr_buf[var->save_result.string_value.length]= 0;
 
+#if 0
   if (sst_receive_address_check(addr_buf))
   {
     goto err;
   }
+#endif
 
   return false;
 
@@ -237,7 +166,7 @@ static bool sst_auth_real_set (const char* value)
 
   if (value)
   {
-    v= my_strdup(value, MYF(0));
+    v= my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0));
   }
   else                                          // its NULL
   {
@@ -255,7 +184,7 @@ static bool sst_auth_real_set (const char* value)
     if (strlen(sst_auth_real))
     {
       if (wsrep_sst_auth) { my_free((void*) wsrep_sst_auth); }
-      wsrep_sst_auth= my_strdup(WSREP_SST_AUTH_MASK, MYF(0));
+      wsrep_sst_auth= my_strdup(PSI_NOT_INSTRUMENTED, WSREP_SST_AUTH_MASK, MYF(0));
     }
     return 0;
   }
@@ -293,65 +222,6 @@ bool wsrep_before_SE()
           && strcmp (wsrep_sst_method, WSREP_SST_MYSQLDUMP));
 }
 
-/* Helper class for splitting string tokens separated with
-   commas. Spaces are stripped. */
-struct Splitter : public std::unary_function<char, void> {
-  Splitter(std::vector<std::string> &vec) : vec_(vec) {}
-
-  void flush()
-  {
-    if (tmp_.length() > 0)
-    {
-      vec_.push_back(tmp_);
-      tmp_.clear();
-    }
-  }
-  void operator()(char ch)
-  {
-    if (ch == ',')
-    {
-      vec_.push_back(tmp_);
-      tmp_.clear();
-    }
-    else
-    {
-      if (ch != ' ')
-      {
-        tmp_+= ch;
-      }
-    }
-  }
-  std::vector<std::string> &vec_;
-  std::string               tmp_;
-};
-
-bool wsrep_setup_allowed_sst_methods()
-{
-  sst_allowed_methods_regex.compile(SST_ALLOWED_METHODS_REGEX_PATTERN,
-                                    MY_REG_EXTENDED | MY_REG_NOSUB,
-                                    &my_charset_latin1);
-  sst_method_allowed_chars_regex.compile(
-      sst_method_allowed_chars_regex_pattern, MY_REG_EXTENDED | MY_REG_NOSUB,
-      &my_charset_latin1);
-
-  std::string methods(wsrep_sst_allowed_methods);
-
-  if (!sst_allowed_methods_regex.match(methods))
-  {
-    WSREP_ERROR(
-        "Wrong format of --wsrep-sst-allowed-methods parameter value: %s",
-        wsrep_sst_allowed_methods);
-    return true;
-  }
-
-  // split it into tokens and strip
-  Splitter splitter(allowed_sst_methods);
-  std::for_each(methods.begin(), methods.end(), splitter).flush();
-
-  return false;
-}
-
-
 static bool            sst_complete = false;
 static bool            sst_needed   = false;
 
@@ -472,10 +342,7 @@ void wsrep_sst_continue ()
   if (sst_needed)
   {
     WSREP_INFO("Signalling provider to continue.");
-    // local_uuid and local_seqno are global variables and are volatile
-    wsrep_uuid_t  const sst_uuid  = local_uuid;
-    wsrep_seqno_t const sst_seqno = local_seqno;
-    wsrep_sst_received (wsrep, sst_uuid, sst_seqno, NULL, 0);
+    wsrep_sst_received (wsrep, local_uuid, local_seqno, NULL, 0);
   }
 }
 
@@ -485,21 +352,22 @@ struct sst_thread_arg
   char**          env;
   char*           ret_str;
   int             err;
-  mysql_mutex_t   lock;
-  mysql_cond_t    cond;
+  mysql_mutex_t   LOCK_wsrep_sst_thread;
+  mysql_cond_t    COND_wsrep_sst_thread;
 
   sst_thread_arg (const char* c, char** e)
     : cmd(c), env(e), ret_str(0), err(-1)
   {
-    mysql_mutex_init(key_LOCK_wsrep_sst_thread, &lock, MY_MUTEX_INIT_FAST);
-    mysql_cond_init(key_COND_wsrep_sst_thread, &cond, NULL);
+    mysql_mutex_init(key_LOCK_wsrep_sst_thread,
+                     &LOCK_wsrep_sst_thread, MY_MUTEX_INIT_FAST);
+    mysql_cond_init(key_COND_wsrep_sst_thread, &COND_wsrep_sst_thread);
   }
 
   ~sst_thread_arg()
   {
-    mysql_cond_destroy  (&cond);
-    mysql_mutex_unlock  (&lock);
-    mysql_mutex_destroy (&lock);
+    mysql_cond_destroy  (&COND_wsrep_sst_thread);
+    mysql_mutex_unlock  (&LOCK_wsrep_sst_thread);
+    mysql_mutex_destroy (&LOCK_wsrep_sst_thread);
   }
 };
 
@@ -547,14 +415,16 @@ static int generate_binlog_opt_val(char** ret)
 {
   DBUG_ASSERT(ret);
   *ret= NULL;
-  if (opt_bin_log && gtid_mode > 0)
+  if (opt_bin_log && get_gtid_mode(GTID_MODE_LOCK_NONE) > GTID_MODE_OFF)
   {
     assert(opt_bin_logname);
-    *ret= my_strdup(opt_bin_logname, MYF(0));
+    *ret= strcmp(opt_bin_logname, "0") ?
+        my_strdup(PSI_NOT_INSTRUMENTED, opt_bin_logname, MYF(0)) :
+        my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(0));
   }
   else
   {
-    *ret= my_strdup("", MYF(0));
+    *ret= my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(0));
   }
   if (!*ret) return -ENOMEM;
   return 0;
@@ -568,29 +438,18 @@ static wsp::process * sst_process= static_cast<wsp::process*>(NULL);
 
 static bool sst_cancelled= false;
 
-void wsrep_sst_cancel (bool call_wsrep_cb)
+void wsrep_sst_cancel ()
 {
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   if (!sst_cancelled)
   {
     sst_cancelled=true;
-    /*
-      When we launched the SST process, then we need
-      to terminate it before exit from the parent (server)
-      process:
-    */
     if (sst_process)
     {
       sst_process->terminate();
       sst_process = NULL;
     }
-    /*
-      If this is a normal shutdown, then we need to notify
-      the wsrep provider about completion of the SST, to
-      prevent infinite waitng in the wsrep provider after
-      the SST process was canceled:
-    */
-    if (call_wsrep_cb && sst_awaiting_callback)
+    if (sst_awaiting_callback)
     {
       WSREP_INFO("Signalling cancellation of the SST request.");
       wsrep_gtid_t const state_id = {
@@ -604,28 +463,15 @@ void wsrep_sst_cancel (bool call_wsrep_cb)
   mysql_mutex_unlock (&LOCK_wsrep_sst);
 }
 
-/*
-  Handling of fatal signals: SIGABRT, SIGTERM, etc.
-*/
-extern "C" void wsrep_handle_fatal_signal (int sig)
-{
-  wsrep_sst_cancel(false);
-}
-
-/*
-  This callback is invoked in the case of abnormal
-  termination of the wsrep provider:
-*/
-void wsrep_abort_cb (void)
-{
-  wsrep_sst_cancel(false);
-}
-
 static void* sst_joiner_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*) a;
   int err= 1;
 
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_register_thread(key_THREAD_wsrep_sst_joiner);
+#endif /* HAVE_PSI_INTERFACE */
+
   {
     const char magic[] = "ready";
     const size_t magic_len = sizeof(magic) - 1;
@@ -665,8 +511,10 @@ static void* sst_joiner_thread (void* a)
       }
       else
       {
-        // Do not clear the pointer to SST process. It will be useful if 
-        // we decide to interrupt it.
+        // Clear the pointer to SST process:
+        if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
+        sst_process = NULL;
+        mysql_mutex_unlock (&LOCK_wsrep_sst);
         err = 0;
       }
     }
@@ -683,26 +531,20 @@ static void* sst_joiner_thread (void* a)
 
     // signal sst_prepare thread with ret code,
     // it will go on sending SST request
-    mysql_mutex_lock   (&arg->lock);
+    mysql_mutex_lock   (&arg->LOCK_wsrep_sst_thread);
     if (!err)
     {
       arg->ret_str = strdup (out + magic_len + 1);
       if (!arg->ret_str) err = ENOMEM;
     }
     arg->err = -err;
-    mysql_cond_signal  (&arg->cond);
-    mysql_mutex_unlock (&arg->lock); //! @note arg is unusable after that.
+    mysql_cond_signal  (&arg->COND_wsrep_sst_thread);
+    mysql_mutex_unlock (&arg->LOCK_wsrep_sst_thread); //! @note arg is unusable after that.
 
-    if (err)
-    {
-      if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
-      sst_process = NULL;
-      mysql_mutex_unlock (&LOCK_wsrep_sst);
+    if (err) return NULL; /* lp:808417 - return immediately, don't signal
+                           * initializer thread to ensure single thread of
+                           * shutdown. */
 
-      return NULL; /* lp:808417 - return immediately, don't signal
-                    * initializer thread to ensure single thread of
-                    * shutdown. */
-    }
     wsrep_uuid_t  ret_uuid  = WSREP_UUID_UNDEFINED;
     wsrep_seqno_t ret_seqno = WSREP_SEQNO_UNDEFINED;
 
@@ -737,9 +579,9 @@ static void* sst_joiner_thread (void* a)
     wsrep_sst_complete (&ret_uuid, ret_seqno, true);
   }
 
-  if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
-  sst_process = NULL;
-  mysql_mutex_unlock (&LOCK_wsrep_sst);
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_delete_thread();
+#endif /* HAVE_PSI_INTERFACE */
 
   return NULL;
 }
@@ -801,13 +643,11 @@ static ssize_t sst_prepare_other (const char*  method,
                  WSREP_SST_OPT_DATA" '%s' "
                  WSREP_SST_OPT_CONF" '%s' "
                  WSREP_SST_OPT_CONF_SUFFIX" '%s' "
-                 WSREP_SST_OPT_PARENT" '%d' "
-                 WSREP_SST_OPT_VERSION" '%s' "
+                 WSREP_SST_OPT_PARENT" '%d'"
                  " %s '%s' ",
                  method, addr_in, mysql_real_data_home,
                  wsrep_defaults_file, wsrep_defaults_group_suffix,
-                 (int)getpid(), MYSQL_SERVER_VERSION MYSQL_SERVER_SUFFIX_DEF,
-                 binlog_opt, binlog_opt_val);
+                 (int)getpid(), binlog_opt, binlog_opt_val);
   my_free(binlog_opt_val);
 
   if (ret < 0 || ret >= cmd_len)
@@ -831,15 +671,15 @@ static ssize_t sst_prepare_other (const char*  method,
 
   pthread_t tmp;
   sst_thread_arg arg(cmd_str(), env());
-  mysql_mutex_lock (&arg.lock);
-  ret = pthread_create (&tmp, NULL, sst_joiner_thread, &arg);
+  mysql_mutex_lock (&arg.LOCK_wsrep_sst_thread);
+  ret = pthread_create(&tmp, NULL, sst_joiner_thread, &arg);
   if (ret)
   {
     WSREP_ERROR("sst_prepare_other(): pthread_create() failed: %d (%s)",
                 ret, strerror(ret));
     return -ret;
   }
-  mysql_cond_wait (&arg.cond, &arg.lock);
+  mysql_cond_wait (&arg.COND_wsrep_sst_thread, &arg.LOCK_wsrep_sst_thread);
 
   *addr_out= arg.ret_str;
 
@@ -895,7 +735,7 @@ static ssize_t sst_prepare_mysqldump (const char*  addr_in,
   return ret;
 }
 
-static enum Wsrep_SE_init_result SE_initialized= WSREP_SE_INIT_RESULT_NONE;
+static bool SE_initialized = false;
 
 ssize_t wsrep_sst_prepare (void** msg)
 {
@@ -918,19 +758,18 @@ ssize_t wsrep_sst_prepare (void** msg)
 
   // Figure out SST address. Common for all SST methods
   if (wsrep_sst_receive_address &&
-      strcmp (wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
+    strcmp (wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
   {
     addr_in= wsrep_sst_receive_address;
   }
   else if (wsrep_node_address && strlen(wsrep_node_address))
   {
-    size_t const addr_len= strlen(wsrep_node_address);
-    size_t const host_len= wsrep_host_len(wsrep_node_address, addr_len);
-
-    if (host_len < addr_len)
+    const char* const colon= strchr (wsrep_node_address, ':');
+    if (colon)
     {
-      strncpy (ip_buf, wsrep_node_address, host_len);
-      ip_buf[host_len]= '\0';
+      ptrdiff_t const len= colon - wsrep_node_address;
+      strncpy (ip_buf, wsrep_node_address, len);
+      ip_buf[len]= '\0';
       addr_in= ip_buf;
     }
     else
@@ -1072,7 +911,7 @@ static void sst_reject_queries(my_bool close_conn)
 {
     wsrep_ready_set (FALSE); // this will be resotred when donor becomes synced
     WSREP_INFO("Rejecting client queries for the duration of SST.");
-    if (TRUE == close_conn) wsrep_close_client_connections(FALSE);
+    if (TRUE == close_conn) wsrep_close_client_connections(false, false);
 }
 
 static int sst_donate_mysqldump (const char*         addr,
@@ -1082,6 +921,25 @@ static int sst_donate_mysqldump (const char*         addr,
                                  bool                bypass,
                                  char**              env) // carries auth info
 {
+  size_t host_len;
+  const char* port = strchr (addr, ':');
+
+  if (port)
+  {
+    port += 1;
+    host_len = port - addr;
+  }
+  else
+  {
+    port = "";
+    host_len = strlen (addr) + 1;
+  }
+
+  char *host= static_cast<char *>(alloca(host_len));
+
+  strncpy (host, addr, host_len - 1);
+  host[host_len - 1] = '\0';
+
   int const cmd_len= 4096;
   wsp::string  cmd_str(cmd_len);
 
@@ -1096,18 +954,16 @@ static int sst_donate_mysqldump (const char*         addr,
 
   int ret= snprintf (cmd_str(), cmd_len,
                      "wsrep_sst_mysqldump "
-                     WSREP_SST_OPT_ADDR" '%s' "
+                     WSREP_SST_OPT_HOST" '%s' "
+                     WSREP_SST_OPT_PORT" '%s' "
                      WSREP_SST_OPT_LPORT" '%u' "
                      WSREP_SST_OPT_SOCKET" '%s' "
                      WSREP_SST_OPT_CONF" '%s' "
-                     WSREP_SST_OPT_GTID" '%s:%lld' "
-                     WSREP_SST_OPT_VERSION" '%s' "
+                     WSREP_SST_OPT_GTID" '%s:%lld'"
                      "%s",
-                     addr, mysqld_port, mysqld_unix_port,
+                     host, port, mysqld_port, mysqld_unix_port,
                      wsrep_defaults_file, uuid_str,
-                     (long long)seqno,
-                     MYSQL_SERVER_VERSION MYSQL_SERVER_SUFFIX_DEF,
-                     bypass ? " " WSREP_SST_OPT_BYPASS : "");
+                     (long long)seqno, bypass ? " "WSREP_SST_OPT_BYPASS : "");
 
   if (ret < 0 || ret >= cmd_len)
   {
@@ -1123,6 +979,7 @@ static int sst_donate_mysqldump (const char*         addr,
 
   wsrep->sst_sent (wsrep, &state_id, ret);
 
+
   return ret;
 }
 
@@ -1133,18 +990,18 @@ static int run_sql_command(THD *thd, const char *query)
   thd->set_query((char *)query, strlen(query));
 
   Parser_state ps;
-  if (ps.init(thd, thd->query(), thd->query_length()))
+  if (ps.init(thd, thd->query().str, thd->query().length))
   {
     WSREP_ERROR("SST query: %s failed", query);
     return -1;
   }
 
-  mysql_parse(thd, thd->query(), thd->query_length(), &ps, false);
+  mysql_parse(thd, &ps);
   if (thd->is_error())
   {
-    int const err= thd->get_stmt_da()->sql_errno();
+    int const err= thd->get_stmt_da()->mysql_errno();
     WSREP_WARN ("error executing '%s': %d (%s)%s",
-                query, err, thd->get_stmt_da()->message(),
+                query, err, thd->get_stmt_da()->message_text(),
                 err == ER_UNKNOWN_SYSTEM_VARIABLE ?
                 ". Was mysqld built with --with-innodb-disallow-writes ?" : "");
     thd->clear_error();
@@ -1226,6 +1083,10 @@ static void* sst_donor_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*)a;
 
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_register_thread(key_THREAD_wsrep_sst_donor);
+#endif /* HAVE_PSI_INTERFACE */
+
   WSREP_INFO("Running: '%s'", arg->cmd);
 
   int  err= 1;
@@ -1252,10 +1113,10 @@ static void* sst_donor_thread (void* a)
   err= proc.error();
 
 /* Inform server about SST script startup and release TO isolation */
-  mysql_mutex_lock   (&arg->lock);
+  mysql_mutex_lock   (&arg->LOCK_wsrep_sst_thread);
   arg->err = -err;
-  mysql_cond_signal  (&arg->cond);
-  mysql_mutex_unlock (&arg->lock); //! @note arg is unusable after that.
+  mysql_cond_signal  (&arg->COND_wsrep_sst_thread);
+  mysql_mutex_unlock (&arg->LOCK_wsrep_sst_thread); //! @note arg is unusable after that.
 
   if (proc.pipe() && !err)
   {
@@ -1345,6 +1206,10 @@ skip_clear_pointer:
   wsrep->sst_sent (wsrep, &state_id, -err);
   proc.wait();
 
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_delete_thread();
+#endif /* HAVE_PSI_INTERFACE */
+
   return NULL;
 }
 
@@ -1386,16 +1251,14 @@ static int sst_donate_other (const char*   method,
                  WSREP_SST_OPT_DATA" '%s' "
                  WSREP_SST_OPT_CONF" '%s' "
                  WSREP_SST_OPT_CONF_SUFFIX" '%s' "
-                 WSREP_SST_OPT_VERSION" '%s' "
                  " %s '%s' "
-                 WSREP_SST_OPT_GTID" '%s:%lld' "
+                 WSREP_SST_OPT_GTID" '%s:%lld'"
                  "%s",
                  method, addr, mysqld_unix_port, mysql_real_data_home,
                  wsrep_defaults_file, wsrep_defaults_group_suffix,
-                 MYSQL_SERVER_VERSION MYSQL_SERVER_SUFFIX_DEF,
                  binlog_opt, binlog_opt_val,
                  uuid, (long long) seqno,
-                 bypass ? " " WSREP_SST_OPT_BYPASS : "");
+                 bypass ? " "WSREP_SST_OPT_BYPASS : "");
   my_free(binlog_opt_val);
 
   if (ret < 0 || ret >= cmd_len)
@@ -1408,108 +1271,31 @@ static int sst_donate_other (const char*   method,
 
   pthread_t tmp;
   sst_thread_arg arg(cmd_str(), env);
-  mysql_mutex_lock (&arg.lock);
-  ret = pthread_create (&tmp, NULL, sst_donor_thread, &arg);
+  mysql_mutex_lock (&arg.LOCK_wsrep_sst_thread);
+  ret = pthread_create(&tmp, NULL, sst_donor_thread, &arg);
   if (ret)
   {
     WSREP_ERROR("sst_donate_other(): pthread_create() failed: %d (%s)",
                 ret, strerror(ret));
-    return -ret;
+    return ret;
   }
-  mysql_cond_wait (&arg.cond, &arg.lock);
+  mysql_cond_wait (&arg.COND_wsrep_sst_thread, &arg.LOCK_wsrep_sst_thread);
 
   WSREP_INFO("sst_donor_thread signaled with %d", arg.err);
+  pthread_detach(tmp);
   return arg.err;
 }
 
-/*
-  Validate SST request string.
-  The protocol is: method\0data\0
-
-  For xtrabackup-v2 it is: method\0addresspath\0
-  like:
-  "xtrabackup-v2\000127.0.0.1:4444/xtrabackup_sst//1"
-*/
-static bool is_sst_request_valid(const std::string &msg)
-{
-  size_t method_len= strlen(msg.c_str());
-
-  if (method_len == 0)
-  {
-    return false;
-  }
-
-  std::string method= msg.substr(0, method_len);
-
-  // Is this method allowed?
-  std::vector<std::string>::iterator res= std::find(
-      allowed_sst_methods.begin(), allowed_sst_methods.end(), method);
-  if (res == allowed_sst_methods.end())
-  {
-    return false;
-  }
-
-  const char *data_ptr= msg.c_str() + method_len + 1;
-  size_t      data_len= strlen(data_ptr);
-
-  // method + null + data + null
-  if (method_len + 1 + data_len + 1 != msg.length())
-  {
-    // Someone tries to piggyback after 2nd null
-    return false;
-  }
-
-  if (data_len > 0)
-  {
-    std::string data= msg.substr(method_len + 1, data_len);
-
-    // first make sure we don't have lt/gt which are not allowed
-    if (std::string::npos != data.find('<') ||
-        std::string::npos != data.find('>'))
-    {
-      return false;
-    }
-
-    // square brackets workaround
-    std::replace(data.begin(), data.end(), '[', '<');
-    std::replace(data.begin(), data.end(), ']', '>');
-
-    return sst_method_allowed_chars_regex.match(data);
-  }
-  return true;
-}
-
-/* Helper functor class for conversion of std::string
-   containing null characters into printable string */
-struct StringBuilder : public std::unary_function<char, void> {
-  StringBuilder(std::ostringstream &ss) : ss_(ss) {}
-
-  void operator()(char ch)
-  {
-    if (ch != 0)
-      ss_ << ch;
-    else
-      ss_ << "<nullptr>";
-  }
-  std::ostringstream &ss_;
-};
-
 wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
                                        const void* msg, size_t msg_len,
                                        const wsrep_gtid_t* current_gtid,
                                        const char* state, size_t state_len,
                                        bool bypass)
 {
-
-  std::string message(reinterpret_cast<const char *>(msg), msg_len);
-  if (!is_sst_request_valid(message))
-  {
-    std::ostringstream ss;
-    StringBuilder      stringBuilder(ss);
-    std::for_each(message.begin(), message.end(), stringBuilder);
-    WSREP_ERROR("Invalid sst_request: %s", ss.str().c_str());
-    return WSREP_CB_FAILURE;
-  }
+  /* This will be reset when sync callback is called.
+   * Should we set wsrep_ready to FALSE here too? */
+//  wsrep_notify_status(WSREP_MEMBER_DONOR);
+  local_status.set(WSREP_MEMBER_DONOR);
 
   const char* method = (char*)msg;
   size_t method_len  = strlen (method);
@@ -1518,15 +1304,6 @@ wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
   char uuid_str[37];
   wsrep_uuid_print (&current_gtid->uuid, uuid_str, sizeof(uuid_str));
 
-  DBUG_EXECUTE_IF("wsrep_sst_donate_cb_fails", 
-  {
-    return WSREP_CB_FAILURE;
-  });
-
-  /* This will be reset when sync callback is called.
-   * Should we set wsrep_ready to FALSE here too? */
-  local_status.set(WSREP_MEMBER_DONOR);
-
   wsp::env env(NULL);
   if (env.error())
   {
@@ -1552,32 +1329,35 @@ wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
                            current_gtid->seqno, bypass, env());
   }
 
-  /* Above methods should return 0 in case of success and negative value
-   * in case of failure. If we have any positive value here it means that we
-   * handle errors in above functions in the wrong way */
-  DBUG_ASSERT(ret <= 0); 
   return (ret >= 0 ? WSREP_CB_SUCCESS : WSREP_CB_FAILURE);
 }
 
-enum Wsrep_SE_init_result wsrep_SE_init_wait()
+void wsrep_SE_init_grab()
 {
-  mysql_mutex_lock (&LOCK_wsrep_sst_init);
-  while (SE_initialized == WSREP_SE_INIT_RESULT_NONE)
+  if (mysql_mutex_lock (&LOCK_wsrep_sst_init)) abort();
+}
+
+void wsrep_SE_init_wait()
+{
+  while (SE_initialized == false)
   {
     mysql_cond_wait (&COND_wsrep_sst_init, &LOCK_wsrep_sst_init);
   }
-  enum Wsrep_SE_init_result ret= SE_initialized;
   mysql_mutex_unlock (&LOCK_wsrep_sst_init);
-  return ret;
 }
 
-void wsrep_SE_initialized(enum Wsrep_SE_init_result result)
+void wsrep_SE_init_done()
 {
-  mysql_mutex_lock (&LOCK_wsrep_sst_init);
-  if (SE_initialized == WSREP_SE_INIT_RESULT_NONE)
-  {
-    SE_initialized= result;
-  }
   mysql_cond_signal (&COND_wsrep_sst_init);
   mysql_mutex_unlock (&LOCK_wsrep_sst_init);
 }
+
+void wsrep_SE_initialized()
+{
+  SE_initialized = true;
+}
+
+bool wsrep_is_SE_initialized()
+{
+  return SE_initialized;
+}
