@@ -16,7 +16,15 @@
 #ifndef WSREP_SST_H
 #define WSREP_SST_H
 
-#include <mysql.h> // my_bool
+// #include <mysql.h> // my_bool
+
+#include "mysqld.h"
+#include "log.h"
+typedef struct st_mysql_show_var SHOW_VAR;
+#include "query_options.h"
+#include "rpl_gtid.h"
+#include "../wsrep/wsrep_api.h"
+
 
 /* system variables */
 extern const char* wsrep_sst_method;
@@ -24,7 +32,6 @@ extern const char* wsrep_sst_receive_address;
 extern const char* wsrep_sst_donor;
 extern       char* wsrep_sst_auth;
 extern    my_bool  wsrep_sst_donor_rejects_queries;
-extern const char *wsrep_sst_allowed_methods;
 
 /*! Synchronizes applier thread start with init thread */
 extern void wsrep_sst_grab();
@@ -33,30 +40,12 @@ extern bool wsrep_sst_wait();
 /*! Signals wsrep that initialization is complete, writesets can be applied */
 extern void wsrep_sst_continue();
 /*! Cancel the SST script if it is running */
-extern void wsrep_sst_cancel(bool call_wsrep_cb);
+extern void wsrep_sst_cancel();
 
 extern void wsrep_SE_init_grab();   /*! grab init critical section */
+extern void wsrep_SE_init_wait();   /*! wait for SE init to complete */
 extern void wsrep_SE_init_done();   /*! signal that SE init is complte */
-/*! Free SST auth memory allocated strings */
-extern void wsrep_sst_auth_free();
-
-/*!  Result enumeration of the SE initialization. */
-enum Wsrep_SE_init_result {
-    WSREP_SE_INIT_RESULT_NONE,    /*! Initialization not done yet. */
-    WSREP_SE_INIT_RESULT_SUCCESS, /*! Initialization succeeded */
-    WSREP_SE_INIT_RESULT_FAILURE  /*! Initialization failed */
-};
-/*!
- * Wait for SE init to complete.
- *
- * @return WSREP_SE_INIT_RESULT_SUCCESS if SE initialization was successful.
- * @return WSREP_SE_INIT_RESULT_FAILURE if SE initialization failed or
- *         an error was encountered before SE initialization took place.
- */
-extern enum Wsrep_SE_init_result wsrep_SE_init_wait();
-/*!
- * Mark SE initialization complete with result.
- */
-extern void wsrep_SE_initialized(enum Wsrep_SE_init_result result);
+extern void wsrep_SE_initialized(); /*! mark SE initialization complete */
+extern bool wsrep_is_SE_initialized();
 
 #endif /* WSREP_SST_H */
