@@ -19,7 +19,6 @@
 #include "rpl_rli.h"
 #include "log_event.h"
 #include "sql_parse.h"
-#include "global_threads.h" // LOCK_thread_count, etc.
 #include "sql_base.h" // close_thread_tables()
 #include "mysqld.h"   // start_wsrep_THD();
 
@@ -36,8 +35,8 @@ int wsrep_show_bf_aborts (THD *thd, SHOW_VAR *var, char *buff)
 /* must have (&thd->LOCK_wsrep_thd) */
 void wsrep_client_rollback(THD *thd)
 {
-  WSREP_DEBUG("client rollback due to BF abort for (%ld %lld), query: %s",
-              thd->thread_id, thd->query_id, WSREP_QUERY(thd));
+  WSREP_DEBUG("client rollback due to BF abort for (%u), query: %s",
+              thd->thread_id(), WSREP_QUERY(thd));
 
   my_atomic_add64(&wsrep_bf_aborts_counter, 1);
 
@@ -47,22 +46,18 @@ void wsrep_client_rollback(THD *thd)
   /* Check for comments in Relay_log_info::cleanup_context */
   trans_rollback_stmt(thd);
 
-  if (trans_rollback(thd))
-  {
-    WSREP_WARN("client rollback failed for: %lu %lld, conf: %d",
-               thd->thread_id, thd->query_id, thd->wsrep_conflict_state);
-  }
+  trans_rollback(thd);
 
   if (thd->locked_tables_mode && thd->lock)
   {
-    WSREP_DEBUG("unlocking tables for BF abort (%ld)", thd->thread_id);
+    WSREP_DEBUG("unlocking tables for BF abort (%u)", thd->thread_id());
     thd->locked_tables_list.unlock_locked_tables(thd);
     thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
   }
 
   if (thd->global_read_lock.is_acquired())
   {
-    WSREP_DEBUG("unlocking GRL for BF abort (%ld)", thd->thread_id);
+    WSREP_DEBUG("unlocking GRL for BF abort (%u)", thd->thread_id());
     thd->global_read_lock.unlock_global_read_lock(thd);
   }
 
@@ -77,7 +72,7 @@ void wsrep_client_rollback(THD *thd)
 
   if (thd->get_binlog_table_maps())
   {
-    WSREP_DEBUG("clearing binlog table map for BF abort (%ld)", thd->thread_id);
+    WSREP_DEBUG("clearing binlog table map for BF abort (%u)", thd->thread_id());
     thd->clear_binlog_table_maps();
   }
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
@@ -92,11 +87,66 @@ static Relay_log_info* wsrep_relay_log_init(const char* log_fname)
 {
   uint rli_option = INFO_REPOSITORY_DUMMY;
   Relay_log_info *rli= NULL;
-  rli = Rpl_info_factory::create_rli(rli_option, false);
+  rli = Rpl_info_factory::create_rli(rli_option, false, (const char*)"", true);
   rli->set_rli_description_event(
       new Format_description_log_event(BINLOG_VERSION));
 
   return (rli);
+
+#ifdef REMOVED
+  Rpl_info_handler* handler_src= NULL;
+  Rpl_info_handler* handler_dest= NULL;
+  ulong *key_info_idx= NULL;
+  const char *msg= "Failed to allocate memory for the relay log info "
+                   "structure";
+
+  DBUG_ENTER("Rpl_info_factory::create_rli");
+
+  if (!(rli= new Relay_log_info(false
+#ifdef HAVE_PSI_INTERFACE
+                                ,&key_relay_log_info_run_lock,
+                                &key_relay_log_info_data_lock,
+                                &key_relay_log_info_sleep_lock,
+                                &key_relay_log_info_data_cond,
+                                &key_relay_log_info_start_cond,
+                                &key_relay_log_info_stop_cond,
+                                &key_relay_log_info_sleep_cond
+#endif /* HAVE_PSI_INTERFACE */
+                               )))
+    goto err;
+
+  if (!(key_info_idx= new ulong[NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR]))
+     goto err;
+  key_info_idx[0]= server_id;
+  rli->set_idx_info(key_info_idx, NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR);
+
+  if(Rpl_info_factory::init_rli_repositories(rli, rli_option, &handler_src,
+                                             &handler_dest, &msg))
+    goto err;
+
+  if (Rpl_info_factory::decide_repository(rli, rli_option, &handler_src,
+                                          &handler_dest, &msg))
+    goto err;
+
+  DBUG_RETURN(rli);
+err:
+  delete handler_src;
+  delete handler_dest;
+  delete []key_info_idx;
+  if (rli)
+  {
+    /*
+      The handler was previously deleted so we need to remove
+      any reference to it.
+    */
+    rli->set_idx_info(NULL, 0);
+    rli->set_rpl_info_handler(NULL);
+    rli->set_rpl_info_type(INVALID_INFO_REPOSITORY);
+    delete rli;
+  }
+  WSREP_ERROR("Error creating relay log info: %s.", msg);
+  DBUG_RETURN(NULL);
+#endif /* REMOVED */
 }
 
 static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -104,7 +154,7 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
   shadow->options       = thd->variables.option_bits;
   shadow->server_status = thd->server_status;
   shadow->wsrep_exec_mode = thd->wsrep_exec_mode;
-  shadow->vio           = thd->net.vio;
+  shadow->vio           = thd->get_protocol_classic()->get_vio();
 
   // Disable general logging on applier threads
   thd->variables.option_bits |= OPTION_LOG_OFF;
@@ -116,21 +166,26 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
 
   if (!thd->wsrep_rli) thd->wsrep_rli= wsrep_relay_log_init("wsrep_relay");
   thd->wsrep_rli->info_thd = thd;
+  thd->init_for_queries(thd->wsrep_rli);
+
+  thd->wsrep_rli->current_mts_submode= new Mts_submode_database();
 
   thd->wsrep_exec_mode= REPL_RECV;
-  thd->net.vio= 0;
+  thd->get_protocol_classic()->set_vio(NULL);
   thd->clear_error();
 
   shadow->tx_isolation        = thd->variables.tx_isolation;
   thd->variables.tx_isolation = ISO_READ_COMMITTED;
   thd->tx_isolation           = ISO_READ_COMMITTED;
 
-  shadow->db            = thd->db;
-  shadow->db_length     = thd->db_length;
-  thd->reset_db(NULL, 0);
+  // intended shallow copy.
+  shadow->db            = thd->db();
 
+  LEX_CSTRING  dummy;
+  dummy.str= NULL;
+  dummy.length= 0;
+  thd->reset_db(dummy);
   shadow->user_time = thd->user_time;
-  shadow->row_count_func= thd->get_row_count_func();
 }
 
 static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -138,115 +193,10 @@ static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
   thd->variables.option_bits  = shadow->options;
   thd->server_status          = shadow->server_status;
   thd->wsrep_exec_mode        = shadow->wsrep_exec_mode;
-  thd->net.vio                = shadow->vio;
+  thd->get_protocol_classic()->set_vio(shadow->vio);
   thd->variables.tx_isolation = shadow->tx_isolation;
-  thd->reset_db(shadow->db, shadow->db_length);
+  thd->reset_db(shadow->db);
   thd->user_time              = shadow->user_time;
-  thd->set_row_count_func(shadow->row_count_func);
-}
-
-void wsrep_replay_sp_transaction(THD* thd)
-{
-  DBUG_ENTER("wsrep_replay_sp_transaction");
-  mysql_mutex_assert_owner(&thd->LOCK_wsrep_thd);
-  DBUG_ASSERT(thd->wsrep_conflict_state == MUST_REPLAY);
-  DBUG_ASSERT(thd->sp_runtime_ctx);
-  DBUG_ASSERT(wsrep_thd_trx_seqno(thd) > 0);
-
-  close_thread_tables(thd);
-  if (thd->locked_tables_mode && thd->lock)
-  {
-    WSREP_DEBUG("releasing table lock for replaying (%lu)",
-                thd->thread_id);
-    thd->locked_tables_list.unlock_locked_tables(thd);
-    thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
-  }
-  thd->mdl_context.release_transactional_locks();
-
-  THD *replay_thd= new THD(true, true);
-  replay_thd->thread_stack= thd->thread_stack;
-
-  struct wsrep_thd_shadow shadow;
-  wsrep_prepare_bf_thd(replay_thd, &shadow);
-  replay_thd->wsrep_trx_meta= thd->wsrep_trx_meta;
-  replay_thd->wsrep_ws_handle= thd->wsrep_ws_handle;
-  replay_thd->wsrep_ws_handle.trx_id= WSREP_UNDEFINED_TRX_ID;
-  replay_thd->wsrep_conflict_state= REPLAYING;
-
-  replay_thd->variables.option_bits|= OPTION_BEGIN;
-  replay_thd->server_status|= SERVER_STATUS_IN_TRANS;
-
-  mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
-
-  thd->restore_globals();
-  replay_thd->store_globals();
-  wsrep_status_t rcode= wsrep->replay_trx(wsrep,
-                                          &replay_thd->wsrep_ws_handle,
-                                          (void*) replay_thd);
-
-  wsrep_return_from_bf_mode(replay_thd, &shadow);
-  replay_thd->restore_globals();
-  delete replay_thd;
-
-  mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-
-  thd->store_globals();
-
-  switch (rcode)
-  {
-  case WSREP_OK:
-    {
-      thd->wsrep_conflict_state= NO_CONFLICT;
-      thd->killed= THD::NOT_KILLED;
-      wsrep_status_t rcode= wsrep->post_commit(wsrep, &thd->wsrep_ws_handle);
-      if (rcode != WSREP_OK)
-      {
-        WSREP_WARN("Post commit failed for SP replay: thd: %lu error: %d",
-                   thd->thread_id, rcode);
-      }
-      /* As replaying the transaction was successful, an error must not
-         be returned to client, so we need to reset the error state of
-         the diagnostics area */
-      thd->get_stmt_da()->reset_diagnostics_area();
-      break;
-    }
-  case WSREP_TRX_FAIL:
-    {
-      thd->wsrep_conflict_state= ABORTED;
-      wsrep_status_t rcode= wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle);
-      if (rcode != WSREP_OK)
-      {
-        WSREP_WARN("Post rollback failed for SP replay: thd: %lu error: %d",
-                   thd->thread_id, rcode);
-      }
-      if (thd->get_stmt_da()->is_set())
-      {
-        thd->get_stmt_da()->reset_diagnostics_area();
-      }
-      my_error(ER_LOCK_DEADLOCK, MYF(0));
-      break;
-    }
-  default:
-    WSREP_ERROR("trx_replay failed for: %d, schema: %s, query: %s",
-                rcode,
-                (thd->db ? thd->db : "(null)"),
-                WSREP_QUERY(thd));
-    /* we're now in inconsistent state, must abort */
-    mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
-    unireg_abort(1);
-    break;
-  }
-
-  wsrep_cleanup_transaction(thd);
-
-  mysql_mutex_lock(&LOCK_wsrep_replaying);
-  wsrep_replaying--;
-  WSREP_DEBUG("replaying decreased: %d, thd: %lu",
-              wsrep_replaying, thd->thread_id);
-  mysql_cond_broadcast(&COND_wsrep_replaying);
-  mysql_mutex_unlock(&LOCK_wsrep_replaying);
-
-  DBUG_VOID_RETURN;
 }
 
 void wsrep_replay_transaction(THD *thd)
@@ -266,28 +216,9 @@ void wsrep_replay_transaction(THD *thd)
        */
       if (thd->get_reprepare_observer() && wsrep_log_conflicts)
       {
-        WSREP_WARN("dangling observer in replay transaction: (thr %lu %lld %s)",
-                   thd->thread_id, thd->query_id, thd->query());
+        WSREP_WARN("dangling observer in replay transaction: (thr %u %lld %s)",
+                   thd->thread_id(), thd->query_id, thd->query().str);
       }
-
-      struct da_shadow
-      {
-          enum Diagnostics_area::enum_diagnostics_status status;
-          ulonglong affected_rows;
-          ulonglong last_insert_id;
-          char message[MYSQL_ERRMSG_SIZE];
-      };
-      struct da_shadow da_status;
-      da_status.status= thd->get_stmt_da()->status();
-      if (da_status.status == Diagnostics_area::DA_OK)
-      {
-        da_status.affected_rows= thd->get_stmt_da()->affected_rows();
-        da_status.last_insert_id= thd->get_stmt_da()->last_insert_id();
-        strmake(da_status.message,
-                thd->get_stmt_da()->message(),
-                sizeof(da_status.message)-1);
-      }
-
       thd->get_stmt_da()->reset_diagnostics_area();
 
       thd->wsrep_conflict_state= REPLAYING;
@@ -298,8 +229,8 @@ void wsrep_replay_transaction(THD *thd)
       close_thread_tables(thd);
       if (thd->locked_tables_mode && thd->lock)
       {
-        WSREP_DEBUG("releasing table lock for replaying (%ld)",
-                    thd->thread_id);
+        WSREP_DEBUG("releasing table lock for replaying (%u)",
+                    thd->thread_id());
         thd->locked_tables_list.unlock_locked_tables(thd);
         thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
       }
@@ -312,7 +243,7 @@ void wsrep_replay_transaction(THD *thd)
       MYSQL_END_STATEMENT(thd->m_statement_psi, thd->get_stmt_da());
       thd->m_statement_psi= NULL;
       thd->m_digest= NULL;
-      thd_proc_info(thd, "wsrep replaying trx");
+      THD_STAGE_INFO_GUARD(thd, &stage_wsrep_replaying_trx);
       WSREP_DEBUG("replay trx: %s %lld",
                   WSREP_QUERY(thd),
                   (long long)wsrep_thd_trx_seqno(thd));
@@ -338,8 +269,8 @@ void wsrep_replay_transaction(THD *thd)
       case WSREP_OK:
         thd->wsrep_conflict_state= NO_CONFLICT;
         wsrep->post_commit(wsrep, &thd->wsrep_ws_handle);
-        WSREP_DEBUG("trx_replay successful for: %ld %llu",
-                    thd->thread_id, (long long)thd->real_id);
+        WSREP_DEBUG("trx_replay successful for: %u %llu",
+                    thd->thread_id(), (long long)thd->real_id);
         if (thd->get_stmt_da()->is_sent())
         {
           WSREP_WARN("replay ok, thd has reported status");
@@ -354,17 +285,7 @@ void wsrep_replay_transaction(THD *thd)
         }
         else
         {
-          if (da_status.status == Diagnostics_area::DA_OK)
-          {
-            my_ok(thd,
-                  da_status.affected_rows,
-                  da_status.last_insert_id,
-                  da_status.message);
-          }
-          else
-          {
-            my_ok(thd);
-          }
+          my_ok(thd);
         }
         break;
       case WSREP_TRX_FAIL:
@@ -382,7 +303,8 @@ void wsrep_replay_transaction(THD *thd)
         break;
       default:
         WSREP_ERROR("trx_replay failed for: %d, schema: %s, query: %s",
-                    rcode, (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+                    rcode, (thd->db().length ? thd->db().str : "(null)"),
+                    WSREP_QUERY(thd));
         /* we're now in inconsistent state, must abort */
 	mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
         unireg_abort(1);
@@ -393,8 +315,8 @@ void wsrep_replay_transaction(THD *thd)
 
       mysql_mutex_lock(&LOCK_wsrep_replaying);
       wsrep_replaying--;
-      WSREP_DEBUG("replaying decreased: %d, thd: %lu",
-                  wsrep_replaying, thd->thread_id);
+      WSREP_DEBUG("replaying decreased: %d, thd: %u",
+                  wsrep_replaying, thd->thread_id());
       mysql_cond_broadcast(&COND_wsrep_replaying);
       mysql_mutex_unlock(&LOCK_wsrep_replaying);
     }
@@ -402,18 +324,30 @@ void wsrep_replay_transaction(THD *thd)
   DBUG_VOID_RETURN;
 }
 
+/* Applier/Slave Threads. Controlled using wsrep_slave_threads configuration
+variable. These are worker thread capable of applying write-set in parallel.
+Parallelism is use to explore efficency. */
 static void wsrep_replication_process(THD *thd)
 {
   int rcode;
   DBUG_ENTER("wsrep_replication_process");
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_applier_idle);
+
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_register_thread(key_THREAD_wsrep_applier);
+#endif /* HAVE_PSI_INTERFACE */
 
   struct wsrep_thd_shadow shadow;
 
   wsrep_prepare_bf_thd(thd, &shadow);
 
+  /* This should be set as part of trans_begin and not explicitly.
+  Note sure why Galera thought of setting it here but it conflicts
+  as the first BEGIN statement is now intepreted wrongly as statement
+  inside an active transaction. */
   /* From trans_begin() */
-  thd->variables.option_bits|= OPTION_BEGIN;
-  thd->server_status|= SERVER_STATUS_IN_TRANS;
+  //thd->variables.option_bits|= OPTION_BEGIN;
+  //thd->server_status|= SERVER_STATUS_IN_TRANS;
 
   rcode = wsrep->recv(wsrep, (void *)thd);
   DBUG_PRINT("wsrep",("wsrep_repl returned: %d", rcode));
@@ -441,7 +375,6 @@ static void wsrep_replication_process(THD *thd)
     /* these suggests a bug in provider code */
     WSREP_WARN("bad return from recv() call: %d", rcode);
     /* fall through to node shutdown */
-    // fall through
   case WSREP_FATAL:
     /* Cluster connectivity is lost.
      *
@@ -456,43 +389,29 @@ static void wsrep_replication_process(THD *thd)
     break;
   }
 
-  mysql_mutex_lock(&LOCK_thread_count);
+  thd_lock_thread_count(thd);
   wsrep_close_applier(thd);
-  mysql_cond_broadcast(&COND_thread_count);
-  mysql_mutex_unlock(&LOCK_thread_count);
+  thd_unlock_thread_count(thd);
 
   TABLE *tmp;
   while ((tmp = thd->temporary_tables))
   {
-    WSREP_WARN("Applier %lu, has temporary tables at exit: %s.%s",
-                  thd->thread_id, 
+    WSREP_WARN("Applier %u, has temporary tables at exit: %s.%s",
+                  thd->thread_id(), 
                   (tmp->s) ? tmp->s->db.str : "void",
                   (tmp->s) ? tmp->s->table_name.str : "void");
   }
   wsrep_return_from_bf_mode(thd, &shadow);
+
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_delete_thread();
+#endif /* HAVE_PSI_INTERFACE */
+
   DBUG_VOID_RETURN;
 }
 
-static bool create_wsrep_THD(wsrep_thd_processor_fun processor)
-{
-  ulong old_wsrep_running_threads= wsrep_running_threads;
-  pthread_t unused;
-  mysql_mutex_lock(&LOCK_thread_count);
-  bool res= pthread_create(&unused, &connection_attrib, start_wsrep_THD,
-                           (void*)processor);
-  /*
-    if starting a thread on server startup, wait until this thread's THD
-    is fully initialized (otherwise a THD initialization code might
-    try to access a partially initialized server data structure - MDEV-8208).
-    (https://jira.mariadb.org/browse/MDEV-8208)
-  */
-  if (!mysqld_server_initialized)
-    while (old_wsrep_running_threads == wsrep_running_threads)
-      mysql_cond_wait(&COND_thread_count, &LOCK_thread_count);
-  mysql_mutex_unlock(&LOCK_thread_count);
-  return res;
-}
-
+/* Create applier threads based on configuration.
+Applier thread are responsible for applying write-set. */
 void wsrep_create_appliers(long threads)
 {
   if (!wsrep_connected)
@@ -508,8 +427,11 @@ void wsrep_create_appliers(long threads)
   }
 
   long wsrep_threads=0;
+  pthread_t hThread;
   while (wsrep_threads++ < threads) {
-    if (create_wsrep_THD(wsrep_replication_process))
+    if (pthread_create(
+      &hThread, &connection_attrib,
+      start_wsrep_THD, (void*)wsrep_replication_process))
       WSREP_WARN("Can't create thread to manage wsrep replication");
   }
 }
@@ -518,23 +440,32 @@ static void wsrep_rollback_process(THD *thd)
 {
   DBUG_ENTER("wsrep_rollback_process");
 
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_register_thread(key_THREAD_wsrep_rollbacker);
+#endif /* HAVE_PSI_INTERFACE */
+
   mysql_mutex_lock(&LOCK_wsrep_rollback);
   wsrep_aborting_thd= NULL;
 
+  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_in_rollback_thread);
+
   while (thd->killed == THD::NOT_KILLED) {
-    thd_proc_info(thd, "wsrep aborter idle");
-    thd->mysys_var->current_mutex= &LOCK_wsrep_rollback;
-    thd->mysys_var->current_cond=  &COND_wsrep_rollback;
+
+    mysql_mutex_lock(&thd->LOCK_current_cond);
+    THD_STAGE_INFO_GUARD_UPDATE(&stage_wsrep_aborter_idle);
+    thd->current_mutex= &LOCK_wsrep_rollback;
+    thd->current_cond= &COND_wsrep_rollback;
+    mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     mysql_cond_wait(&COND_wsrep_rollback,&LOCK_wsrep_rollback);
 
     WSREP_DEBUG("WSREP rollback thread wakes for signal");
 
-    mysql_mutex_lock(&thd->mysys_var->mutex);
-    thd_proc_info(thd, "wsrep aborter active");
-    thd->mysys_var->current_mutex= 0;
-    thd->mysys_var->current_cond=  0;
-    mysql_mutex_unlock(&thd->mysys_var->mutex);
+    mysql_mutex_lock(&thd->LOCK_current_cond);
+    THD_STAGE_INFO_GUARD_UPDATE(&stage_wsrep_aborter_active);
+    thd->current_mutex= 0;
+    thd->current_cond= 0;
+    mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     /* check for false alarms */
     if (!wsrep_aborting_thd)
@@ -572,20 +503,11 @@ static void wsrep_rollback_process(THD *thd)
       aborting->store_globals();
 
       mysql_mutex_lock(&aborting->LOCK_wsrep_thd);
-
-      /* prepare THD for rollback processing */
-      mysql_reset_thd_for_next_command(aborting);
-      aborting->lex->sql_command= SQLCOM_ROLLBACK;
-
       wsrep_client_rollback(aborting);
-      WSREP_DEBUG("WSREP rollbacker aborted thd: (%lu %llu)",
-                  aborting->thread_id, (long long)aborting->real_id);
-
+      WSREP_DEBUG("WSREP rollbacker aborted thd: (%u %llu)",
+                  aborting->thread_id(), (long long)aborting->real_id);
       mysql_mutex_unlock(&aborting->LOCK_wsrep_thd);
 
-      /* Clear the thread state, since the rollback thread is done with it */
-      aborting->restore_globals();
-
       mysql_mutex_lock(&LOCK_wsrep_rollback);
     }
   }
@@ -593,6 +515,10 @@ static void wsrep_rollback_process(THD *thd)
   mysql_mutex_unlock(&LOCK_wsrep_rollback);
   sql_print_information("WSREP: rollbacker thread exiting");
 
+#ifdef HAVE_PSI_INTERFACE
+  wsrep_pfs_delete_thread();
+#endif /* HAVE_PSI_INTERFACE */
+
   DBUG_PRINT("wsrep",("wsrep rollbacker thread exiting"));
   DBUG_VOID_RETURN;
 }
@@ -601,8 +527,10 @@ void wsrep_create_rollbacker()
 {
   if (wsrep_provider && strcasecmp(wsrep_provider, "none"))
   {
+    pthread_t hThread;
     /* create rollbacker */
-    if (create_wsrep_THD(wsrep_rollback_process))
+    if (pthread_create( &hThread, &connection_attrib,
+                        start_wsrep_THD, (void*)wsrep_rollback_process))
       WSREP_WARN("Can't create thread to manage wsrep rollback");
   }
 }
@@ -637,9 +565,9 @@ my_bool wsrep_thd_is_BF(void *thd_ptr, my_bool sync)
   {
     THD* thd = (THD*)thd_ptr;
     if (sync) mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-
+    
     status = ((thd->wsrep_exec_mode == REPL_RECV)    ||
-              (thd->wsrep_exec_mode == TOTAL_ORDER));
+	      (thd->wsrep_exec_mode == TOTAL_ORDER));
     if (sync) mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
   }
   return status;
@@ -653,10 +581,10 @@ my_bool wsrep_thd_is_BF_or_commit(void *thd_ptr, my_bool sync)
   {
     THD* thd = (THD*)thd_ptr;
     if (sync) mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-
+    
     status = ((thd->wsrep_exec_mode == REPL_RECV)    ||
-              (thd->wsrep_exec_mode == TOTAL_ORDER)  ||
-              (thd->wsrep_exec_mode == LOCAL_COMMIT));
+	      (thd->wsrep_exec_mode == TOTAL_ORDER)  ||
+	      (thd->wsrep_exec_mode == LOCAL_COMMIT));
     if (sync) mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
   }
   return status;
