@@ -34,6 +34,8 @@
 #include <netdb.h>    // getaddrinfo()
 #include <signal.h>
 
+#include "socket_connection.h"           // MY_BIND_ALL_ADDRESSES
+
 #ifdef HAVE_SYS_PRCTL_H
 #include <sys/prctl.h>
 #endif
@@ -94,17 +96,6 @@ wsrep_prepend_PATH (const char* path)
                  "State snapshot transfer may not be working.");
 }
 
-/**
- * This is a wrapper around the WSREP_ON macro.  This allows this check
- * to be called from plugins.
- */
-extern "C" bool
-wsrep_is_wsrep_on(void)
-{
-  return WSREP_ON;
-}
-
-
 namespace wsp
 {
 
@@ -453,7 +444,7 @@ process::process (const char* cmd, const char* type, char** env)
 
     if (prctl(PR_SET_PDEATHSIG, SIGTERM))
     {
-      sql_perror("prctl() failed");
+      sql_print_error("prctl() failed");
       _exit(EXIT_FAILURE);
     }
 #endif
@@ -465,7 +456,7 @@ process::process (const char* cmd, const char* type, char** env)
 
     if (sigprocmask(SIG_SETMASK, &set, NULL))
     {
-      sql_perror("sigprocmask() failed");
+      sql_print_error("sigprocmask() failed");
       _exit(EXIT_FAILURE);
     }
 
@@ -491,7 +482,7 @@ process::process (const char* cmd, const char* type, char** env)
 
     if (setsid() < 0)
     {
-      sql_perror("setsid() failed");
+      sql_print_error("setsid() failed");
       _exit(EXIT_FAILURE);
     }
 
@@ -501,7 +492,7 @@ process::process (const char* cmd, const char* type, char** env)
     /* 
     if (close(close_fd) < 0)
     {
-      sql_perror("close() failed");
+      sql_print_error("close() failed");
       _exit(EXIT_FAILURE);
     }
     */
@@ -510,7 +501,7 @@ process::process (const char* cmd, const char* type, char** env)
 
     if (dup2(pipe_fds[child_end], close_fd) < 0)
     {
-      sql_perror("dup2() failed");
+      sql_print_error("dup2() failed");
       _exit(EXIT_FAILURE);
     }
 
@@ -530,7 +521,7 @@ process::process (const char* cmd, const char* type, char** env)
 
     execvpe(pargv[0], pargv, env);
 
-    sql_perror("execlp() failed");
+    sql_print_error("execlp() failed");
     _exit(EXIT_FAILURE);
 
 #else // _POSIX_SPAWN is defined:
@@ -788,15 +779,7 @@ void process::terminate ()
 {
   if (pid_)
   {
-    /*
-      If we have an appropriated system call, then we try
-      to terminate entire process group:
-    */
-#if _XOPEN_SOURCE >= 500 || _DEFAULT_SOURCE || _BSD_SOURCE
-    if (killpg(pid_, SIGTERM))
-#else
     if (kill(pid_, SIGTERM))
-#endif
     {
       WSREP_WARN("Unable to terminate process: %s: %d (%s)",
                  str_, errno, strerror(errno));
@@ -812,7 +795,7 @@ thd::thd (my_bool won) : init(), ptr(new THD)
     ptr->store_globals();
     ptr->variables.option_bits&= ~OPTION_BIN_LOG; // disable binlog
     ptr->variables.wsrep_on = won;
-    ptr->security_ctx->master_access= ~(ulong)0;
+    ptr->m_security_ctx->set_master_access(~(ulong)0);
     lex_start(ptr);
   }
 }
@@ -822,7 +805,6 @@ thd::~thd ()
   if (ptr)
   {
     delete ptr;
-    my_pthread_setspecific_ptr (THR_THD, 0);
   }
 }
 
@@ -944,18 +926,3 @@ size_t wsrep_guess_ip (char* buf, size_t buf_len)
 
   return 0;
 }
-
-/* returns the length of the host part of the address string */
-size_t wsrep_host_len(const char* const addr, size_t const addr_len)
-{
-  // check for IPv6 notation first
-  const char* const bracket= ('[' == addr[0] ? strchr(addr, ']') : NULL);
-
-  if (bracket) { // IPv6
-    return (bracket - addr + 1);
-  }
-  else { // host part ends at ':' or end of string
-    const char* const colon= strchr(addr, ':');
-    return (colon ? colon - addr : addr_len);
-  }
-}
