@@ -28,7 +28,6 @@
 #include <cstdlib>
 
 const  char* wsrep_provider         = 0;
-bool         wsrep_provider_set     = false;
 const  char* wsrep_provider_options = 0;
 const  char* wsrep_cluster_address  = 0;
 const  char* wsrep_cluster_name     = 0;
@@ -40,15 +39,18 @@ ulong   wsrep_reject_queries;
 
 int wsrep_init_vars()
 {
-  wsrep_provider        = my_strdup(WSREP_NONE, MYF(MY_WME));
-  wsrep_provider_set    = false;
-  wsrep_provider_options= my_strdup("", MYF(MY_WME));
-  wsrep_cluster_address = my_strdup("", MYF(MY_WME));
-  wsrep_cluster_name    = my_strdup(WSREP_CLUSTER_NAME, MYF(MY_WME));
-  wsrep_node_name       = my_strdup("", MYF(MY_WME));
-  wsrep_node_address    = my_strdup("", MYF(MY_WME));
-  wsrep_node_incoming_address= my_strdup(WSREP_NODE_INCOMING_AUTO, MYF(MY_WME));
-  wsrep_start_position  = my_strdup(WSREP_START_POSITION_ZERO, MYF(MY_WME));
+  wsrep_provider        = my_strdup(PSI_NOT_INSTRUMENTED,
+                                    WSREP_NONE, MYF(MY_WME));
+  wsrep_provider_options= my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(MY_WME));
+  wsrep_cluster_address = my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(MY_WME));
+  wsrep_cluster_name    = my_strdup(PSI_NOT_INSTRUMENTED,
+                                    WSREP_CLUSTER_NAME, MYF(MY_WME));
+  wsrep_node_name       = my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(MY_WME));
+  wsrep_node_address    = my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(MY_WME));
+  wsrep_node_incoming_address= my_strdup(PSI_NOT_INSTRUMENTED,
+                                         WSREP_NODE_INCOMING_AUTO, MYF(MY_WME));
+  wsrep_start_position  = my_strdup(PSI_NOT_INSTRUMENTED,
+                                    WSREP_START_POSITION_ZERO, MYF(MY_WME));
 
   global_system_variables.binlog_format=BINLOG_FORMAT_ROW;
   return 0;
@@ -218,7 +220,8 @@ static int get_provider_option_value(const char* opts,
 {
   int ret= 1;
   ulong opt_value_tmp;
-  char *opt_value_str, *s, *opts_copy= my_strdup(opts, MYF(MY_WME));
+  char *opt_value_str, *s, *opts_copy=
+    my_strdup(PSI_NOT_INSTRUMENTED, opts, MYF(MY_WME));
 
   if ((opt_value_str= strstr(opts_copy, opt_name)) == NULL)
     goto end;
@@ -359,14 +362,6 @@ bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
   mysql_mutex_lock(&LOCK_global_system_variables);
   mysql_mutex_lock(&LOCK_wsrep_slave_threads);
 
-#if 0
-  /* provider status variables are allocated in provider library
-     and need to freed here, otherwise a dangling reference to
-     wsrep_status_vars would remain in THD
-  */
-  wsrep_free_status(thd);
-#endif
-
   wsrep_deinit();
 
   char* tmp= strdup(wsrep_provider); // wsrep_init() rewrites provider 
@@ -378,8 +373,6 @@ bool wsrep_provider_update (sys_var *self, THD* thd, enum_var_type type)
   }
   free(tmp);
 
-  wsrep_provider_set = (wsrep_provider != NULL) && strcmp(wsrep_provider, WSREP_NONE) != 0;
-
   // we sure don't want to use old address with new provider
   wsrep_cluster_address_init(NULL);
   wsrep_provider_options_init(NULL);
@@ -404,8 +397,7 @@ void wsrep_provider_init (const char* value)
   }
 
   if (wsrep_provider) my_free((void *)wsrep_provider);
-  wsrep_provider = my_strdup(value, MYF(0));
-  wsrep_provider_set = (wsrep_provider != NULL) && strcmp(wsrep_provider, WSREP_NONE) != 0;
+  wsrep_provider = my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0));
 }
 
 bool wsrep_provider_options_check (sys_var *self, THD* thd, set_var* var)
@@ -435,7 +427,7 @@ void wsrep_provider_options_init(const char* value)
 {
   if (wsrep_provider_options && wsrep_provider_options != value)
     my_free((void *)wsrep_provider_options);
-  wsrep_provider_options = (value) ? my_strdup(value, MYF(0)) : NULL;
+  wsrep_provider_options = (value) ? my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0)) : NULL;
 }
 
 bool wsrep_reject_queries_update(sys_var *self, THD* thd, enum_var_type type)
@@ -448,7 +440,7 @@ bool wsrep_reject_queries_update(sys_var *self, THD* thd, enum_var_type type)
             WSREP_INFO("Rejecting client queries due to manual setting");
             break;
         case WSREP_REJECT_ALL_KILL:
-            wsrep_close_client_connections(FALSE);
+            wsrep_close_client_connections(false, false);
             WSREP_INFO("Rejecting client queries and killing connections due to manual setting");
             break;
         default:
@@ -529,7 +521,7 @@ void wsrep_cluster_address_init (const char* value)
               (value) ? value : "null");
 
   if (wsrep_cluster_address) my_free ((void*)wsrep_cluster_address);
-  wsrep_cluster_address = (value) ? my_strdup(value, MYF(0)) : NULL;
+  wsrep_cluster_address = (value) ? my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0)) : NULL;
 }
 
 /* Function checks if the new value for cluster_name is valid.
@@ -608,18 +600,21 @@ void wsrep_node_address_init (const char* value)
   if (wsrep_node_address && strcmp(wsrep_node_address, value))
     my_free ((void*)wsrep_node_address);
 
-  wsrep_node_address = (value) ? my_strdup(value, MYF(0)) : NULL;
+  wsrep_node_address = (value) ? my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0)) : NULL;
 }
 
-static void wsrep_slave_count_change_update ()
+bool wsrep_slave_threads_check (sys_var *self, THD* thd, set_var* var)
 {
-  // wsrep_running_threads = appliers threads + 1 rollbacker thread
-  wsrep_slave_count_change = (wsrep_slave_threads - wsrep_running_threads + 1);
+  mysql_mutex_lock(&LOCK_wsrep_slave_threads);
+  wsrep_slave_count_change += (var->save_result.ulonglong_value -
+                               wsrep_slave_threads);
+  mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
+
+  return false;
 }
 
 bool wsrep_slave_threads_update (sys_var *self, THD* thd, enum_var_type type)
 {
-  wsrep_slave_count_change_update();
   if (wsrep_slave_count_change > 0)
   {
     WSREP_DEBUG("Creating %d applier threads, total %ld", wsrep_slave_count_change, wsrep_slave_threads);
@@ -634,38 +629,22 @@ bool wsrep_desync_check (sys_var *self, THD* thd, set_var* var)
   bool new_wsrep_desync = var->save_result.ulonglong_value;
   if (wsrep_desync == new_wsrep_desync) {
     if (new_wsrep_desync) {
-      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::SL_WARNING,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already ON.");
     } else {
-      push_warning (thd, Sql_condition::WARN_LEVEL_WARN,
+      push_warning (thd, Sql_condition::SL_WARNING,
                    ER_WRONG_VALUE_FOR_VAR,
                    "'wsrep_desync' is already OFF.");
     }
     return false;
   }
-
-  /* Setting desync to on/off signals participation of node in flow control.
-  It doesn't turn-off recieval of write-sets.
-  If the node is already in paused state due to some previous action like FTWRL
-  then avoid desync as superset action of pausing the node is already active. */
-
-  if (!thd->global_read_lock.provider_resumed())
-  {
-    char message[1024];
-    sprintf(message,
-            "Explictly desync/resync of already desynced/paused node"
-            " is prohibited");
-    my_message(ER_UNKNOWN_ERROR, message, MYF(0));
-    return true;
-  }
-
   wsrep_status_t ret(WSREP_WARNING);
   if (new_wsrep_desync) {
     ret = wsrep->desync (wsrep);
     if (ret != WSREP_OK) {
       WSREP_WARN ("SET desync failed %d for schema: %s, query: %s", ret,
-                  (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+                  (thd->db().length ? thd->db().str : "(null)"), WSREP_QUERY(thd));
       my_error (ER_CANNOT_USER, MYF(0), "'desync'", thd->query());
       return true;
     }
@@ -673,7 +652,7 @@ bool wsrep_desync_check (sys_var *self, THD* thd, set_var* var)
     ret = wsrep->resync (wsrep);
     if (ret != WSREP_OK) {
       WSREP_WARN ("SET resync failed %d for schema: %s, query: %s", ret,
-                  (thd->db ? thd->db : "(null)"), WSREP_QUERY(thd));
+                  (thd->db().length ? thd->db().str : "(null)"), WSREP_QUERY(thd));
       my_error (ER_CANNOT_USER, MYF(0), "'resync'", thd->query());
       return true;
     }
@@ -712,14 +691,17 @@ wsrep_assign_to_mysql (SHOW_VAR* mysql, wsrep_stats_var* wsrep)
   case WSREP_VAR_INT64:
     mysql->value = (char*) &wsrep->value._int64;
     mysql->type  = SHOW_LONGLONG;
+    mysql->scope = SHOW_SCOPE_GLOBAL;
     break;
   case WSREP_VAR_STRING:
     mysql->value = (char*) &wsrep->value._string;
     mysql->type  = SHOW_CHAR_PTR;
+    mysql->scope = SHOW_SCOPE_GLOBAL;
     break;
   case WSREP_VAR_DOUBLE:
     mysql->value = (char*) &wsrep->value._double;
     mysql->type  = SHOW_DOUBLE;
+    mysql->scope = SHOW_SCOPE_GLOBAL;
     break;
   }
 }
@@ -780,6 +762,7 @@ static void export_wsrep_status_to_mysql(THD* thd)
   mysql_status_vars[wsrep_status_len].name  = NullS;
   mysql_status_vars[wsrep_status_len].value = NullS;
   mysql_status_vars[wsrep_status_len].type  = SHOW_LONG;
+  mysql_status_vars[wsrep_status_len].scope = SHOW_SCOPE_GLOBAL;
 }
 
 int wsrep_show_status (THD *thd, SHOW_VAR *var, char *buff)
@@ -798,3 +781,150 @@ void wsrep_free_status (THD* thd)
     thd->wsrep_status_vars = 0;
   }
 }
+
+
+bool wsrep_replicate_myisam_check(sys_var *self, THD* thd, set_var* var)
+{
+  /* Trying to set wsrep_replicate_myisam to off. */
+  if (var->save_result.ulonglong_value == 0)
+    return false;
+
+  /* If pxc-strict-mode >= ENFORCING we don't allow setting
+  wsrep_replicate_myisam to ON. */
+  bool block= false;
+
+  switch(pxc_strict_mode)
+  {
+  case PXC_STRICT_MODE_DISABLED:
+    break;
+  case PXC_STRICT_MODE_PERMISSIVE:
+    WSREP_WARN("Percona-XtraDB-Cluster doesn't recommend use of MyISAM"
+               " table replication feature"
+               " with pxc_strict_mode = PERMISSIVE");
+    push_warning_printf(
+      thd, Sql_condition::SL_WARNING, ER_UNKNOWN_ERROR, 
+      "Percona-XtraDB-Cluster doesn't recommend use of MyISAM"
+      " table replication feature"
+      " with pxc_strict_mode = PERMISSIVE");
+    break;
+  case PXC_STRICT_MODE_ENFORCING:
+  case PXC_STRICT_MODE_MASTER:
+  default:
+    block= true;
+    WSREP_ERROR("Percona-XtraDB-Cluster prohibits use of MyISAM"
+                " table replication feature"
+                " with pxc_strict_mode = ENFORCING or MASTER");
+    char message[1024];
+    sprintf(message,
+            "Percona-XtraDB-Cluster prohibits use of MyISAM"
+            " table replication feature"
+            " with pxc_strict_mode = ENFORCING or MASTER");
+    my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+    break;
+  }
+
+  return block;
+}
+
+static const char* pxc_strict_mode_to_string(ulong value)
+{
+  switch(value)
+  {
+  case PXC_STRICT_MODE_DISABLED:
+    return "DISABLED";
+  case PXC_STRICT_MODE_PERMISSIVE:
+    return "PERMISSIVE";
+  case PXC_STRICT_MODE_ENFORCING:
+    return "ENFORCING";
+  case PXC_STRICT_MODE_MASTER:
+    return "MASTER";
+  default:
+    return "NULL";
+  }
+}
+
+bool pxc_strict_mode_check(sys_var *self, THD* thd, set_var* var)
+{
+  /* pxc-strict-mode can be changed only if node is cluster-node. */
+  if (!(WSREP_ON))
+  {
+    WSREP_ERROR("pxc_strict_mode can be changed only if node is cluster-node");
+
+    char message[1024];
+    sprintf(message,
+            "pxc_strict_mode can be changed only if node is cluster-node");
+    my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+    return true;
+  }
+
+  bool enforcing_strictness= false;
+  bool block= false;
+
+  if (pxc_strict_mode <= PXC_STRICT_MODE_PERMISSIVE &&
+      var->save_result.ulonglong_value >= PXC_STRICT_MODE_ENFORCING)
+    enforcing_strictness= true;
+
+  if (enforcing_strictness)
+  {
+    /* wsrep_replicate_myisam shouldn't be OFF
+    TODO: Mutex ordering ? */
+    bool replicate_myisam;
+    replicate_myisam= (global_system_variables.wsrep_replicate_myisam ||
+                       thd->variables.wsrep_replicate_myisam);
+
+    bool row_binlog_format;
+    row_binlog_format=
+      (global_system_variables.binlog_format == BINLOG_FORMAT_ROW &&
+       thd->variables.binlog_format == BINLOG_FORMAT_ROW);
+
+    bool safe_log_output;
+    safe_log_output=
+       ((log_output_options & LOG_NONE) || (log_output_options & LOG_FILE));
+
+    bool serializable;
+    serializable= (thd->tx_isolation == ISO_SERIALIZABLE ||
+                   global_system_variables.tx_isolation == ISO_SERIALIZABLE);
+
+    /* replicate_myisam = off
+       row_binlog_format = true (row)
+       safe_log_output = true (none/file)
+       serializable = false */
+    block = !(!replicate_myisam &&
+              row_binlog_format &&
+              safe_log_output   &&
+              !serializable);
+
+    if (replicate_myisam)
+      WSREP_ERROR("Can't change pxc_strict_mode with wsrep_replicate_myisam"
+                  " turned ON");
+
+    if (!row_binlog_format)
+      WSREP_ERROR("Can't change pxc_strict_mode while binlog format != ROW");
+
+    if (!safe_log_output)
+      WSREP_ERROR("Can't change pxc_strict_mode while log_output != NONE/FILE");
+
+    if (serializable)
+      WSREP_ERROR("Can't change pxc_strict_mode while isolation level is"
+                  " SERIALIZABLE");
+
+    if (block)
+    {
+      char message[1024];
+      sprintf(message,
+              "Can't change pxc_strict_mode to %s as%s%s%s%s",
+              pxc_strict_mode_to_string(var->save_result.ulonglong_value),
+              (replicate_myisam ? " wsrep_replicate_myisam is ON" : ""), 
+              (!row_binlog_format ? " binlog_format != ROW" : ""),
+              (!safe_log_output ? " log_output != NONE/FILE" : ""),
+              (serializable ? " isolation level is SERIALIZABLE" : ""));
+      my_message(ER_UNKNOWN_ERROR, message, MYF(0));
+    }
+  }
+
+  if (!block)
+    wsrep_replicate_set_stmt= true;
+
+  return(block);
+}
+
