@@ -23,7 +23,8 @@
 
 // MySQL variables funcs
 
-#include "sql_priv.h"
+#include "query_options.h"
+#include "set_var.h"
 class sys_var;
 class set_var;
 class THD;
@@ -85,6 +86,12 @@ extern bool wsrep_slave_threads_update       UPDATE_ARGS;
 extern bool wsrep_desync_check               CHECK_ARGS;
 extern bool wsrep_desync_update              UPDATE_ARGS;
 
+extern bool wsrep_max_ws_size_update         UPDATE_ARGS;
+
 extern bool wsrep_reject_queries_update    UPDATE_ARGS;
 
+extern bool wsrep_replicate_myisam_check     CHECK_ARGS;
+
+extern bool pxc_strict_mode_check            CHECK_ARGS;
+
 #endif /* WSREP_VAR_H */
