@@ -25,9 +25,6 @@ Copyright (c) 2006, 2015, Percona and/or its affiliates. All rights reserved.
 #ident "Copyright (c) 2006, 2015, Percona and/or its affiliates. All rights reserved."
 
 #include "hatoku_hton.h"
-#include "sql_acl.h"
-#include "tokudb_dir_cmd.h"
-#include "sql_parse.h"
 
 namespace tokudb {
 namespace sysvars {
@@ -43,8 +40,6 @@ namespace sysvars {
 #define TOKUDB_VERSION_STR NULL
 #endif
 
-const size_t error_buffer_max_size = 1024;
-
 ulonglong   cache_size = 0;
 uint        cachetable_pool_threads = 0;
 int         cardinality_scale_percent = 0;
@@ -57,10 +52,10 @@ uint        client_pool_threads = 0;
 my_bool     compress_buffers_before_eviction = TRUE;
 char*       data_dir = NULL;
 ulong       debug = 0;
-#if defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
+#if TOKUDB_DEBUG
 // used to control background job manager
 my_bool     debug_pause_background_job_manager = FALSE;
-#endif  // defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
+#endif
 my_bool     directio = FALSE;
 my_bool     enable_partial_eviction = TRUE;
 int         fs_reserve_percent = 0;
@@ -71,16 +66,13 @@ uint        read_status_frequency = 0;
 my_bool     strip_frm_data = FALSE;
 char*       tmp_dir = NULL;
 uint        write_status_frequency = 0;
-my_bool     dir_per_db = FALSE;
 char*       version = (char*) TOKUDB_VERSION_STR;
 
 // file system reserve as a percentage of total disk space
-#if defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) && \
-    TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+#if TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
 char*       gdb_path = NULL;
 my_bool     gdb_on_fatal = FALSE;
-#endif  // defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) &&
-        // TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+#endif
 
 my_bool        check_jemalloc = TRUE;
 
@@ -115,7 +107,7 @@ static MYSQL_SYSVAR_INT(
     "index cardinality scale percentage",
     NULL,
     NULL,
-    50,
+    100,
     0,
     100,
     0);
@@ -142,10 +134,11 @@ static MYSQL_SYSVAR_UINT(
     0);
 
 static void checkpointing_period_update(
-    TOKUDB_UNUSED(THD* thd),
-    TOKUDB_UNUSED(st_mysql_sys_var* sys_var),
+    THD* thd,
+    st_mysql_sys_var* sys_var,
     void* var,
     const void* save) {
+
     uint* cp = (uint*)var;
     *cp = *(const uint*)save;
     int r = db_env->checkpointing_set_period(db_env, *cp);
@@ -164,10 +157,12 @@ static MYSQL_SYSVAR_UINT(
     ~0U,
     0);
 
-static void cleaner_iterations_update(TOKUDB_UNUSED(THD* thd),
-                                      TOKUDB_UNUSED(st_mysql_sys_var* sys_var),
-                                      void* var,
-                                      const void* save) {
+static void cleaner_iterations_update(
+    THD* thd,
+    st_mysql_sys_var* sys_var,
+    void* var,
+    const void* save) {
+
     ulong* ci = (ulong*)var;
     *ci = *(const ulong*)save;
     int r = db_env->cleaner_set_iterations(db_env, *ci);
@@ -186,10 +181,12 @@ static MYSQL_SYSVAR_ULONG(
     ~0UL,
     0);
 
-static void cleaner_period_update(TOKUDB_UNUSED(THD* thd),
-                                  TOKUDB_UNUSED(st_mysql_sys_var* sys_var),
-                                  void* var,
-                                  const void* save) {
+static void cleaner_period_update(
+    THD* thd,
+    st_mysql_sys_var* sys_var,
+    void* var,
+    const void * save) {
+
     ulong* cp = (ulong*)var;
     *cp = *(const ulong*)save;
     int r = db_env->cleaner_set_period(db_env, *cp);
@@ -250,7 +247,7 @@ static MYSQL_SYSVAR_ULONG(
     ~0UL,
     0);
 
-#if defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
+#if TOKUDB_DEBUG
 static MYSQL_SYSVAR_BOOL(
     debug_pause_background_job_manager,
     debug_pause_background_job_manager,
@@ -259,7 +256,7 @@ static MYSQL_SYSVAR_BOOL(
     NULL,
     NULL,
     FALSE);
-#endif  // defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
+#endif // TOKUDB_DEBUG
 
 static MYSQL_SYSVAR_BOOL(
     directio,
@@ -270,10 +267,11 @@ static MYSQL_SYSVAR_BOOL(
     FALSE);
 
 static void enable_partial_eviction_update(
-    TOKUDB_UNUSED(THD* thd),
-    TOKUDB_UNUSED(st_mysql_sys_var* sys_var),
+    THD* thd,
+    st_mysql_sys_var* sys_var,
     void* var,
     const void* save) {
+
     my_bool* epe = (my_bool*)var;
     *epe = *(const my_bool*)save;
     int r = db_env->evictor_set_enable_partial_eviction(db_env, *epe);
@@ -301,10 +299,12 @@ static MYSQL_SYSVAR_INT(
     100,
     0);
 
-static void fsync_log_period_update(TOKUDB_UNUSED(THD* thd),
-                                    TOKUDB_UNUSED(st_mysql_sys_var* sys_var),
-                                    void* var,
-                                    const void* save) {
+static void fsync_log_period_update(
+    THD* thd,
+    st_mysql_sys_var* sys_var,
+    void* var,
+    const void* save) {
+
     uint* flp = (uint*)var;
     *flp = *(const uint*)save;
     db_env->change_fsync_log_period(db_env, *flp);
@@ -394,22 +394,7 @@ static MYSQL_SYSVAR_UINT(
     ~0U,
     0);
 
-static void tokudb_dir_per_db_update(
-    TOKUDB_UNUSED(THD* thd),
-    TOKUDB_UNUSED(struct st_mysql_sys_var* sys_var),
-    void* var,
-    const void* save) {
-    my_bool *value = (my_bool *) var;
-    *value = *(const my_bool *) save;
-    db_env->set_dir_per_db(db_env, *value);
-}
-
-static MYSQL_SYSVAR_BOOL(dir_per_db, dir_per_db,
-    0, "TokuDB store ft files in db directories",
-    NULL, tokudb_dir_per_db_update, FALSE);
-
-#if defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) && \
-    TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+#if TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
 static MYSQL_SYSVAR_STR(
     gdb_path,
     gdb_path,
@@ -427,8 +412,7 @@ static MYSQL_SYSVAR_BOOL(
     NULL,
     NULL,
     true);
-#endif  // defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) &&
-        // TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+#endif
 
 static MYSQL_SYSVAR_BOOL(
     check_jemalloc,
@@ -468,7 +452,7 @@ static MYSQL_THDVAR_BOOL(
     "dispatch ANALYZE TABLE to background job.",
     NULL,
     NULL,
-    false);
+    true);
 
 const char* srv_analyze_mode_names[] = {
     "TOKUDB_ANALYZE_STANDARD",
@@ -524,7 +508,7 @@ static MYSQL_THDVAR_ULONGLONG(
     "auto analyze threshold (percent)",
     NULL,
     NULL,
-    0,
+    30,
     0,
     ~0U,
     1);
@@ -548,10 +532,12 @@ static MYSQL_THDVAR_BOOL(
     NULL,
     true);
 
-static void checkpoint_lock_update(TOKUDB_UNUSED(THD* thd),
-                                   TOKUDB_UNUSED(st_mysql_sys_var* var),
-                                   void* var_ptr,
-                                   const void* save) {
+static void checkpoint_lock_update(
+    THD* thd,
+    st_mysql_sys_var* var,
+    void* var_ptr,
+    const void* save) {
+
     my_bool* val = (my_bool*)var_ptr;
     *val= *(my_bool*)save ? true : false;
     if (*val) {
@@ -662,13 +648,13 @@ static MYSQL_THDVAR_ULONGLONG(
     ~0ULL,
     1);
 
-static MYSQL_THDVAR_STR(last_lock_timeout,
-                        PLUGIN_VAR_MEMALLOC | PLUGIN_VAR_NOCMDOPT |
-                            PLUGIN_VAR_READONLY,
-                        "last lock timeout",
-                        NULL,
-                        NULL,
-                        NULL);
+static MYSQL_THDVAR_STR(
+    last_lock_timeout,
+    PLUGIN_VAR_MEMALLOC,
+    "last lock timeout",
+    NULL,
+    NULL,
+    NULL);
 
 static MYSQL_THDVAR_BOOL(
     load_save_space,
@@ -741,50 +727,6 @@ static MYSQL_THDVAR_ULONGLONG(
     ~0ULL,
     1);
 
-static const char* deprecated_tokudb_pk_insert_mode =
-    "Using tokudb_pk_insert_mode is deprecated and the "
-    "parameter may be removed in future releases.";
-static const char* deprecated_tokudb_pk_insert_mode_zero =
-    "Using tokudb_pk_insert_mode=0 is deprecated and the "
-    "parameter may be removed in future releases. "
-    "Only tokudb_pk_insert_mode=1|2 is allowed."
-    "Resettig the value to 1.";
-
-static void pk_insert_mode_update(
-    THD* thd,
-    st_mysql_sys_var* var,
-    void* var_ptr,
-    const void* save) {
-    const uint* new_pk_insert_mode = static_cast<const uint*>(save);
-    uint* pk_insert_mode = static_cast<uint*>(var_ptr);
-    if (*new_pk_insert_mode == 0) {
-        push_warning(
-            thd,
-            Sql_condition::WARN_LEVEL_WARN,
-            HA_ERR_WRONG_COMMAND,
-            deprecated_tokudb_pk_insert_mode_zero);
-        *pk_insert_mode = 1;
-    } else {
-        push_warning(
-            thd,
-            Sql_condition::WARN_LEVEL_WARN,
-            HA_ERR_WRONG_COMMAND,
-            deprecated_tokudb_pk_insert_mode);
-        *pk_insert_mode = *new_pk_insert_mode;
-    }
-}
-
-static MYSQL_THDVAR_UINT(
-    pk_insert_mode,
-    0,
-    "set the primary key insert mode",
-    NULL,
-    pk_insert_mode_update,
-    1,
-    0,
-    2,
-    1);
-
 static MYSQL_THDVAR_BOOL(
     prelock_empty,
     0,
@@ -845,7 +787,6 @@ static MYSQL_THDVAR_ENUM(
     SRV_ROW_FORMAT_ZLIB,
     &tokudb_row_format_typelib);
 
-#if defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
 static MYSQL_THDVAR_BOOL(
     rpl_check_readonly,
     PLUGIN_VAR_THDLOCAL,
@@ -891,10 +832,10 @@ static MYSQL_THDVAR_ULONGLONG(
     0,
     ~0ULL,
     1);
-#endif // defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
 
+#if TOKU_INCLUDE_UPSERT
 static MYSQL_THDVAR_BOOL(
-    enable_fast_update,
+    disable_slow_update,
     PLUGIN_VAR_THDLOCAL,
     "disable slow update",
     NULL,
@@ -902,94 +843,44 @@ static MYSQL_THDVAR_BOOL(
     false);
 
 static MYSQL_THDVAR_BOOL(
-    enable_fast_upsert,
+    disable_slow_upsert,
     PLUGIN_VAR_THDLOCAL,
     "disable slow upsert",
     NULL,
     NULL,
     false);
+#endif
+
+static const char* deprecated_tokudb_support_xa =
+    "Using tokudb_support_xa is deprecated and the "
+    "parameter may be removed in future releases.";
+static const char* deprecated_tokudb_support_xa_off =
+    "Using tokudb_support_xa is deprecated and the "
+    "parameter may be removed in future releases. "
+    "Only tokudb_support_xa=ON is allowed.";
+
+static void support_xa_update(
+    THD* thd,
+    st_mysql_sys_var* var,
+    void* var_ptr,
+    const void* save) {
+    my_bool tokudb_support_xa = *static_cast<const my_bool*>(save);
+    push_warning(thd,
+                 Sql_condition::SL_WARNING,
+                 HA_ERR_WRONG_COMMAND,
+                 tokudb_support_xa ? deprecated_tokudb_support_xa :
+                 deprecated_tokudb_support_xa_off);
+}
 
-#if TOKU_INCLUDE_XA
 static MYSQL_THDVAR_BOOL(
     support_xa,
     PLUGIN_VAR_OPCMDARG,
     "Enable TokuDB support for the XA two-phase commit",
     NULL,
-    NULL,
+    support_xa_update,
     true);
-#endif
 
-static int dir_cmd_check(THD* thd, struct st_mysql_sys_var* var,
-                         void* save, struct st_mysql_value* value) ;
 
-static MYSQL_THDVAR_INT(dir_cmd_last_error,
-    PLUGIN_VAR_THDLOCAL,
-    "error from the last dir command. 0 is success",
-    NULL, NULL, 0, 0, 0, 1);
-
-static MYSQL_THDVAR_STR(dir_cmd_last_error_string,
-    PLUGIN_VAR_THDLOCAL + PLUGIN_VAR_MEMALLOC,
-    "error string from the last dir command",
-    NULL, NULL, NULL);
-
-static MYSQL_THDVAR_STR(dir_cmd,
-    PLUGIN_VAR_THDLOCAL + PLUGIN_VAR_MEMALLOC,
-    "name of the directory where the backup is stored",
-    dir_cmd_check, NULL, NULL);
-
-static void MY_ATTRIBUTE((format(printf, 3, 4)))
-    dir_cmd_set_error(THD* thd, int error, const char* error_fmt, ...) {
-    char   buff[error_buffer_max_size];
-    va_list varargs;
-
-    assert(thd);
-    assert(error_fmt);
-
-    va_start(varargs, error_fmt);
-    vsnprintf(buff, sizeof(buff), error_fmt, varargs);
-    va_end(varargs);
-
-    THDVAR_SET(thd, dir_cmd_last_error, &error);
-    THDVAR_SET(thd, dir_cmd_last_error_string, buff);
-}
-
-static void dir_cmd_clear_error(THD* thd) {
-    static constexpr int no_error = 0;
-    static const char* empty_error_str = "";
-    THDVAR_SET(thd, dir_cmd_last_error, &no_error);
-    THDVAR_SET(thd, dir_cmd_last_error_string, empty_error_str);
-}
-
-static int dir_cmd_check(THD* thd,
-                         TOKUDB_UNUSED(struct st_mysql_sys_var* var),
-                         void* save,
-                         struct st_mysql_value* value) {
-    int error = 0;
-    dir_cmd_clear_error(thd);
-
-    if (check_global_access(thd, SUPER_ACL)) {
-        return 1;
-    }
-
-    char buff[STRING_BUFFER_USUAL_SIZE];
-    int length = sizeof(buff);
-    const char *str = value->val_str(value, buff, &length);
-    if (str) {
-        str = thd->strmake(str, length);
-        *(const char**)save = str;
-    }
-
-    if (str) {
-        dir_cmd_callbacks callbacks { .set_error = dir_cmd_set_error };
-        process_dir_cmd(thd, str, callbacks);
-
-        error = THDVAR(thd, dir_cmd_last_error);
-    } else {
-        error = EINVAL;
-    }
-
-    return error;
-}
 
 //******************************************************************************
 // all system variables
@@ -1019,13 +910,11 @@ st_mysql_sys_var* system_variables[] = {
     MYSQL_SYSVAR(tmp_dir),
     MYSQL_SYSVAR(version),
     MYSQL_SYSVAR(write_status_frequency),
-    MYSQL_SYSVAR(dir_per_db),
-#if defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) && \
-    TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+
+#if TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
     MYSQL_SYSVAR(gdb_path),
     MYSQL_SYSVAR(gdb_on_fatal),
-#endif  // defined(TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL) &&
-        // TOKU_INCLUDE_HANDLERTON_HANDLE_FATAL_SIGNAL
+#endif
 
     MYSQL_SYSVAR(check_jemalloc),
 
@@ -1057,30 +946,26 @@ st_mysql_sys_var* system_variables[] = {
     MYSQL_SYSVAR(optimize_index_fraction),
     MYSQL_SYSVAR(optimize_index_name),
     MYSQL_SYSVAR(optimize_throttle),
-    MYSQL_SYSVAR(pk_insert_mode),
     MYSQL_SYSVAR(prelock_empty),
     MYSQL_SYSVAR(read_block_size),
     MYSQL_SYSVAR(read_buf_size),
     MYSQL_SYSVAR(row_format),
-#if defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
     MYSQL_SYSVAR(rpl_check_readonly),
     MYSQL_SYSVAR(rpl_lookup_rows),
     MYSQL_SYSVAR(rpl_lookup_rows_delay),
     MYSQL_SYSVAR(rpl_unique_checks),
     MYSQL_SYSVAR(rpl_unique_checks_delay),
-#endif // defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
-    MYSQL_SYSVAR(enable_fast_update),
-    MYSQL_SYSVAR(enable_fast_upsert),
-#if TOKU_INCLUDE_XA
-    MYSQL_SYSVAR(support_xa),
+
+#if TOKU_INCLUDE_UPSERT
+    MYSQL_SYSVAR(disable_slow_update),
+    MYSQL_SYSVAR(disable_slow_upsert),
 #endif
 
-#if defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
-    MYSQL_SYSVAR(debug_pause_background_job_manager),
-#endif  // defined(TOKUDB_DEBUG) && TOKUDB_DEBUG
-    MYSQL_SYSVAR(dir_cmd_last_error),
-    MYSQL_SYSVAR(dir_cmd_last_error_string),
-    MYSQL_SYSVAR(dir_cmd),
+    MYSQL_SYSVAR(support_xa),
+
+#if TOKUDB_DEBUG
+   MYSQL_SYSVAR(debug_pause_background_job_manager),
+#endif // TOKUDB_DEBUG
 
     NULL
 };
@@ -1127,12 +1012,14 @@ my_bool disable_prefetching(THD* thd) {
 my_bool disable_slow_alter(THD* thd) {
     return (THDVAR(thd, disable_slow_alter) != 0);
 }
-my_bool enable_fast_update(THD* thd) {
-    return (THDVAR(thd, enable_fast_update) != 0);
+#if TOKU_INCLUDE_UPSERT
+my_bool disable_slow_update(THD* thd) {
+    return (THDVAR(thd, disable_slow_update) != 0);
 }
-my_bool enable_fast_upsert(THD* thd) {
-    return (THDVAR(thd, enable_fast_upsert) != 0);
+my_bool disable_slow_upsert(THD* thd) {
+    return (THDVAR(thd, disable_slow_upsert) != 0);
 }
+#endif
 empty_scan_mode_t empty_scan(THD* thd) {
     return (empty_scan_mode_t)THDVAR(thd, empty_scan);
 }
@@ -1172,12 +1059,6 @@ const char* optimize_index_name(THD* thd) {
 ulonglong optimize_throttle(THD* thd) {
     return THDVAR(thd, optimize_throttle);
 }
-uint pk_insert_mode(THD* thd) {
-    return THDVAR(thd, pk_insert_mode);
-}
-void set_pk_insert_mode(THD* thd, uint mode) {
-    THDVAR(thd, pk_insert_mode) = mode;
-}
 my_bool prelock_empty(THD* thd) {
     return (THDVAR(thd, prelock_empty) != 0);
 }
@@ -1190,7 +1071,6 @@ uint read_buf_size(THD* thd) {
 row_format_t row_format(THD *thd) {
     return (row_format_t) THDVAR(thd, row_format);
 }
-#if defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
 my_bool rpl_check_readonly(THD* thd) {
     return (THDVAR(thd, rpl_check_readonly) != 0);
 }
@@ -1206,10 +1086,11 @@ my_bool rpl_unique_checks(THD* thd) {
 ulonglong rpl_unique_checks_delay(THD* thd) {
     return THDVAR(thd, rpl_unique_checks_delay);
 }
-#endif // defined(TOKU_INCLUDE_RFR) && TOKU_INCLUDE_RFR
 my_bool support_xa(THD* thd) {
     return (THDVAR(thd, support_xa) != 0);
 }
-
+void set_support_xa(THD* thd, my_bool xa) {
+    THDVAR(thd, support_xa) = xa;
+}
 } // namespace sysvars
 } // namespace tokudb
